/*
 * Copyright 2023 Conor McCarthy for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "vkd3d_shader_private.h"

#define VKD3D_SM6_VERSION_MAJOR(version) (((version) >> 4) & 0xf)
#define VKD3D_SM6_VERSION_MINOR(version) (((version) >> 0) & 0xf)
#define VKD3D_SM6_MAX_METADATA_TABLES 4
#define VKD3D_SHADER_DESCRIPTOR_TYPE_COUNT (VKD3D_SHADER_DESCRIPTOR_TYPE_SAMPLER + 1)
#define VKD3D_SM6_MAX_IR_INSTRUCTIONS_PER_DXIL_INSTRUCTION 4

#define BITCODE_MAGIC VKD3D_MAKE_TAG('B', 'C', 0xc0, 0xde)
#define DXIL_MAX_ALIGNMENT_EXPONENT 29
#define DXIL_OP_MAX_OPERANDS 18

enum bitcode_block_id
{
    BLOCKINFO_BLOCK           =  0,
    MODULE_BLOCK              =  8,
    PARAMATTR_BLOCK           =  9,
    PARAMATTR_GROUP_BLOCK     = 10,
    CONSTANTS_BLOCK           = 11,
    FUNCTION_BLOCK            = 12,
    VALUE_SYMTAB_BLOCK        = 14,
    METADATA_BLOCK            = 15,
    METADATA_ATTACHMENT_BLOCK = 16,
    TYPE_BLOCK                = 17,
    USELIST_BLOCK             = 18,
};

enum bitcode_blockinfo_code
{
    SETBID = 1,
    BLOCKNAME = 2,
    SETRECORDNAME = 3,
};

enum bitcode_block_abbreviation
{
    END_BLOCK = 0,
    ENTER_SUBBLOCK = 1,
    DEFINE_ABBREV = 2,
    UNABBREV_RECORD = 3,
};

enum bitcode_abbrev_type
{
    ABBREV_FIXED = 1,
    ABBREV_VBR   = 2,
    ABBREV_ARRAY = 3,
    ABBREV_CHAR  = 4,
    ABBREV_BLOB  = 5,
};

enum bitcode_address_space
{
    ADDRESS_SPACE_DEFAULT,
    ADDRESS_SPACE_DEVICEMEM,
    ADDRESS_SPACE_CBUFFER,
    ADDRESS_SPACE_GROUPSHARED,
};

enum bitcode_module_code
{
    MODULE_CODE_VERSION     =  1,
    MODULE_CODE_GLOBALVAR   =  7,
    MODULE_CODE_FUNCTION    =  8,
};

enum bitcode_constant_code
{
    CST_CODE_SETTYPE         =  1,
    CST_CODE_NULL            =  2,
    CST_CODE_UNDEF           =  3,
    CST_CODE_INTEGER         =  4,
    CST_CODE_FLOAT           =  6,
    CST_CODE_STRING          =  8,
    CST_CODE_CE_GEP          = 12,
    CST_CODE_CE_INBOUNDS_GEP = 20,
    CST_CODE_DATA            = 22,
};

enum bitcode_function_code
{
    FUNC_CODE_DECLAREBLOCKS    =  1,
    FUNC_CODE_INST_BINOP       =  2,
    FUNC_CODE_INST_CAST        =  3,
    FUNC_CODE_INST_RET         = 10,
    FUNC_CODE_INST_BR          = 11,
    FUNC_CODE_INST_SWITCH      = 12,
    FUNC_CODE_INST_PHI         = 16,
    FUNC_CODE_INST_ALLOCA      = 19,
    FUNC_CODE_INST_LOAD        = 20,
    FUNC_CODE_INST_EXTRACTVAL  = 26,
    FUNC_CODE_INST_CMP2        = 28,
    FUNC_CODE_INST_VSELECT     = 29,
    FUNC_CODE_INST_CALL        = 34,
    FUNC_CODE_INST_ATOMICRMW   = 38,
    FUNC_CODE_INST_LOADATOMIC  = 41,
    FUNC_CODE_INST_GEP         = 43,
    FUNC_CODE_INST_STORE       = 44,
    FUNC_CODE_INST_STOREATOMIC = 45,
    FUNC_CODE_INST_CMPXCHG     = 46,
};

enum bitcode_metadata_code
{
    METADATA_STRING             =  1,
    METADATA_VALUE              =  2,
    METADATA_NODE               =  3,
    METADATA_NAME               =  4,
    METADATA_DISTINCT_NODE      =  5,
    METADATA_KIND               =  6,
    METADATA_LOCATION           =  7,
    METADATA_NAMED_NODE         = 10,
    METADATA_ATTACHMENT         = 11,
};

enum bitcode_type_code
{
    TYPE_CODE_NUMENTRY     =  1,
    TYPE_CODE_VOID         =  2,
    TYPE_CODE_FLOAT        =  3,
    TYPE_CODE_DOUBLE       =  4,
    TYPE_CODE_LABEL        =  5,
    TYPE_CODE_INTEGER      =  7,
    TYPE_CODE_POINTER      =  8,
    TYPE_CODE_HALF         = 10,
    TYPE_CODE_ARRAY        = 11,
    TYPE_CODE_VECTOR       = 12,
    TYPE_CODE_METADATA     = 16,
    TYPE_CODE_STRUCT_ANON  = 18,
    TYPE_CODE_STRUCT_NAME  = 19,
    TYPE_CODE_STRUCT_NAMED = 20,
    TYPE_CODE_FUNCTION     = 21,
};

enum bitcode_value_symtab_code
{
    VST_CODE_ENTRY   = 1,
    VST_CODE_BBENTRY = 2,
};

enum bitcode_linkage
{
    LINKAGE_EXTERNAL  = 0,
    LINKAGE_APPENDING = 2,
    LINKAGE_INTERNAL  = 3,
};

enum dxil_resource_class
{
    RESOURCE_CLASS_SRV = 0,
    RESOURCE_CLASS_UAV = 1,
    RESOURCE_CLASS_CBV = 2,
    RESOURCE_CLASS_SAMPLER = 3,
};

enum dxil_resource_kind
{
    RESOURCE_KIND_INVALID                 =  0,
    RESOURCE_KIND_TEXTURE1D               =  1,
    RESOURCE_KIND_TEXTURE2D               =  2,
    RESOURCE_KIND_TEXTURE2DMS             =  3,
    RESOURCE_KIND_TEXTURE3D               =  4,
    RESOURCE_KIND_TEXTURECUBE             =  5,
    RESOURCE_KIND_TEXTURE1DARRAY          =  6,
    RESOURCE_KIND_TEXTURE2DARRAY          =  7,
    RESOURCE_KIND_TEXTURE2DMSARRAY        =  8,
    RESOURCE_KIND_TEXTURECUBEARRAY        =  9,
    RESOURCE_KIND_TYPEDBUFFER             = 10,
    RESOURCE_KIND_RAWBUFFER               = 11,
    RESOURCE_KIND_STRUCTUREDBUFFER        = 12,
    RESOURCE_KIND_CBUFFER                 = 13,
    RESOURCE_KIND_SAMPLER                 = 14,
    RESOURCE_KIND_TBUFFER                 = 15,
    RESOURCE_KIND_RTACCELERATIONSTRUCTURE = 16,
    RESOURCE_KIND_FEEDBACKTEXTURE2D       = 17,
    RESOURCE_KIND_FEEDBACKTEXTURE2DARRAY  = 18,
};

enum dxil_resource_type
{
    RESOURCE_TYPE_NON_RAW_STRUCTURED = 0,
    RESOURCE_TYPE_RAW_STRUCTURED     = 1,
};

enum dxil_component_type
{
    COMPONENT_TYPE_INVALID     =  0,
    COMPONENT_TYPE_I1          =  1,
    COMPONENT_TYPE_I16         =  2,
    COMPONENT_TYPE_U16         =  3,
    COMPONENT_TYPE_I32         =  4,
    COMPONENT_TYPE_U32         =  5,
    COMPONENT_TYPE_I64         =  6,
    COMPONENT_TYPE_U64         =  7,
    COMPONENT_TYPE_F16         =  8,
    COMPONENT_TYPE_F32         =  9,
    COMPONENT_TYPE_F64         = 10,
    COMPONENT_TYPE_SNORMF16    = 11,
    COMPONENT_TYPE_UNORMF16    = 12,
    COMPONENT_TYPE_SNORMF32    = 13,
    COMPONENT_TYPE_UNORMF32    = 14,
    COMPONENT_TYPE_SNORMF64    = 15,
    COMPONENT_TYPE_UNORMF64    = 16,
    COMPONENT_TYPE_PACKEDS8X32 = 17,
    COMPONENT_TYPE_PACKEDU8X32 = 18,
};

enum dxil_sampler_kind
{
    SAMPLER_KIND_DEFAULT = 0,
    SAMPLER_KIND_COMPARISON = 1,
    SAMPLER_KIND_MONO = 2,
};

enum dxil_semantic_kind
{
    SEMANTIC_KIND_ARBITRARY            =  0,
    SEMANTIC_KIND_VERTEXID             =  1,
    SEMANTIC_KIND_INSTANCEID           =  2,
    SEMANTIC_KIND_POSITION             =  3,
    SEMANTIC_KIND_RTARRAYINDEX         =  4,
    SEMANTIC_KIND_VIEWPORTARRAYINDEX   =  5,
    SEMANTIC_KIND_CLIPDISTANCE         =  6,
    SEMANTIC_KIND_CULLDISTANCE         =  7,
    SEMANTIC_KIND_OUTPUTCONTROLPOINTID =  8,
    SEMANTIC_KIND_DOMAINLOCATION       =  9,
    SEMANTIC_KIND_PRIMITIVEID          = 10,
    SEMANTIC_KIND_GSINSTANCEID         = 11,
    SEMANTIC_KIND_SAMPLEINDEX          = 12,
    SEMANTIC_KIND_ISFRONTFACE          = 13,
    SEMANTIC_KIND_COVERAGE             = 14,
    SEMANTIC_KIND_INNERCOVERAGE        = 15,
    SEMANTIC_KIND_TARGET               = 16,
    SEMANTIC_KIND_DEPTH                = 17,
    SEMANTIC_KIND_DEPTHLESSEQUAL       = 18,
    SEMANTIC_KIND_DEPTHGREATEREQUAL    = 19,
    SEMANTIC_KIND_STENCILREF           = 20,
    SEMANTIC_KIND_DISPATCHTHREADID     = 21,
    SEMANTIC_KIND_GROUPID              = 22,
    SEMANTIC_KIND_GROUPINDEX           = 23,
    SEMANTIC_KIND_GROUPTHREADID        = 24,
    SEMANTIC_KIND_TESSFACTOR           = 25,
    SEMANTIC_KIND_INSIDETESSFACTOR     = 26,
    SEMANTIC_KIND_VIEWID               = 27,
    SEMANTIC_KIND_BARYCENTRICS         = 28,
    SEMANTIC_KIND_SHADINGRATE          = 29,
    SEMANTIC_KIND_CULLPRIMITIVE        = 30,
    SEMANTIC_KIND_COUNT                = 31,
    SEMANTIC_KIND_INVALID              = SEMANTIC_KIND_COUNT,
};

enum dxil_binop_code
{
    BINOP_ADD  =  0,
    BINOP_SUB  =  1,
    BINOP_MUL  =  2,
    BINOP_UDIV =  3,
    BINOP_SDIV =  4,
    BINOP_UREM =  5,
    BINOP_SREM =  6,
    BINOP_SHL  =  7,
    BINOP_LSHR =  8,
    BINOP_ASHR =  9,
    BINOP_AND  = 10,
    BINOP_OR   = 11,
    BINOP_XOR  = 12
};

enum dxil_fast_fp_flags
{
    FP_ALLOW_UNSAFE_ALGEBRA =  0x1,
    FP_NO_NAN               =  0x2,
    FP_NO_INF               =  0x4,
    FP_NO_SIGNED_ZEROS      =  0x8,
    FP_ALLOW_RECIPROCAL     = 0x10,
};

enum dx_intrinsic_opcode
{
    DX_LOAD_INPUT                   =   4,
    DX_STORE_OUTPUT                 =   5,
    DX_FABS                         =   6,
    DX_SATURATE                     =   7,
    DX_ISNAN                        =   8,
    DX_ISINF                        =   9,
    DX_ISFINITE                     =  10,
    DX_COS                          =  12,
    DX_SIN                          =  13,
    DX_TAN                          =  14,
    DX_ACOS                         =  15,
    DX_ASIN                         =  16,
    DX_ATAN                         =  17,
    DX_HCOS                         =  18,
    DX_HSIN                         =  19,
    DX_HTAN                         =  20,
    DX_EXP                          =  21,
    DX_FRC                          =  22,
    DX_LOG                          =  23,
    DX_SQRT                         =  24,
    DX_RSQRT                        =  25,
    DX_ROUND_NE                     =  26,
    DX_ROUND_NI                     =  27,
    DX_ROUND_PI                     =  28,
    DX_ROUND_Z                      =  29,
    DX_BFREV                        =  30,
    DX_COUNT_BITS                   =  31,
    DX_FIRST_BIT_LO                 =  32,
    DX_FIRST_BIT_HI                 =  33,
    DX_FIRST_BIT_SHI                =  34,
    DX_FMAX                         =  35,
    DX_FMIN                         =  36,
    DX_IMAX                         =  37,
    DX_IMIN                         =  38,
    DX_UMAX                         =  39,
    DX_UMIN                         =  40,
    DX_FMAD                         =  46,
    DX_FMA                          =  47,
    DX_IMAD                         =  48,
    DX_UMAD                         =  49,
    DX_DOT2                         =  54,
    DX_DOT3                         =  55,
    DX_DOT4                         =  56,
    DX_CREATE_HANDLE                =  57,
    DX_CBUFFER_LOAD_LEGACY          =  59,
    DX_SAMPLE                       =  60,
    DX_SAMPLE_B                     =  61,
    DX_SAMPLE_LOD                   =  62,
    DX_SAMPLE_GRAD                  =  63,
    DX_SAMPLE_C                     =  64,
    DX_SAMPLE_C_LZ                  =  65,
    DX_TEXTURE_LOAD                 =  66,
    DX_TEXTURE_STORE                =  67,
    DX_BUFFER_LOAD                  =  68,
    DX_BUFFER_STORE                 =  69,
    DX_BUFFER_UPDATE_COUNTER        =  70,
    DX_GET_DIMENSIONS               =  72,
    DX_TEXTURE_GATHER               =  73,
    DX_TEXTURE_GATHER_CMP           =  74,
    DX_TEX2DMS_GET_SAMPLE_POS       =  75,
    DX_RT_GET_SAMPLE_POS            =  76,
    DX_RT_GET_SAMPLE_COUNT          =  77,
    DX_ATOMIC_BINOP                 =  78,
    DX_ATOMIC_CMP_XCHG              =  79,
    DX_BARRIER                      =  80,
    DX_CALCULATE_LOD                =  81,
    DX_DISCARD                      =  82,
    DX_DERIV_COARSEX                =  83,
    DX_DERIV_COARSEY                =  84,
    DX_DERIV_FINEX                  =  85,
    DX_DERIV_FINEY                  =  86,
    DX_EVAL_SAMPLE_INDEX            =  88,
    DX_EVAL_CENTROID                =  89,
    DX_SAMPLE_INDEX                 =  90,
    DX_COVERAGE                     =  91,
    DX_THREAD_ID                    =  93,
    DX_GROUP_ID                     =  94,
    DX_THREAD_ID_IN_GROUP           =  95,
    DX_FLATTENED_THREAD_ID_IN_GROUP =  96,
    DX_EMIT_STREAM                  =  97,
    DX_CUT_STREAM                   =  98,
    DX_EMIT_THEN_CUT_STREAM         =  99,
    DX_MAKE_DOUBLE                  = 101,
    DX_SPLIT_DOUBLE                 = 102,
    DX_LOAD_OUTPUT_CONTROL_POINT    = 103,
    DX_LOAD_PATCH_CONSTANT          = 104,
    DX_DOMAIN_LOCATION              = 105,
    DX_STORE_PATCH_CONSTANT         = 106,
    DX_OUTPUT_CONTROL_POINT_ID      = 107,
    DX_PRIMITIVE_ID                 = 108,
    DX_WAVE_IS_FIRST_LANE           = 110,
    DX_WAVE_GET_LANE_INDEX          = 111,
    DX_WAVE_GET_LANE_COUNT          = 112,
    DX_WAVE_ANY_TRUE                = 113,
    DX_WAVE_ALL_TRUE                = 114,
    DX_WAVE_ACTIVE_ALL_EQUAL        = 115,
    DX_WAVE_ACTIVE_BALLOT           = 116,
    DX_WAVE_READ_LANE_AT            = 117,
    DX_WAVE_READ_LANE_FIRST         = 118,
    DX_WAVE_ACTIVE_OP               = 119,
    DX_WAVE_ACTIVE_BIT              = 120,
    DX_WAVE_QUAD_OP                 = 123,
    DX_LEGACY_F32TOF16              = 130,
    DX_LEGACY_F16TOF32              = 131,
    DX_WAVE_ALL_BIT_COUNT           = 135,
    DX_WAVE_PREFIX_BIT_COUNT        = 136,
    DX_RAW_BUFFER_LOAD              = 139,
    DX_RAW_BUFFER_STORE             = 140,
};

enum dxil_cast_code
{
    CAST_TRUNC    =  0,
    CAST_ZEXT     =  1,
    CAST_SEXT     =  2,
    CAST_FPTOUI   =  3,
    CAST_FPTOSI   =  4,
    CAST_UITOFP   =  5,
    CAST_SITOFP   =  6,
    CAST_FPTRUNC  =  7,
    CAST_FPEXT    =  8,
    CAST_PTRTOINT =  9,
    CAST_INTTOPTR = 10,
    CAST_BITCAST  = 11,
    CAST_ADDRSPACECAST = 12
};

enum dxil_predicate
{
    FCMP_FALSE =  0,
    FCMP_OEQ   =  1,
    FCMP_OGT   =  2,
    FCMP_OGE   =  3,
    FCMP_OLT   =  4,
    FCMP_OLE   =  5,
    FCMP_ONE   =  6,
    FCMP_ORD   =  7,
    FCMP_UNO   =  8,
    FCMP_UEQ   =  9,
    FCMP_UGT   = 10,
    FCMP_UGE   = 11,
    FCMP_ULT   = 12,
    FCMP_ULE   = 13,
    FCMP_UNE   = 14,
    FCMP_TRUE  = 15,
    ICMP_EQ    = 32,
    ICMP_NE    = 33,
    ICMP_UGT   = 34,
    ICMP_UGE   = 35,
    ICMP_ULT   = 36,
    ICMP_ULE   = 37,
    ICMP_SGT   = 38,
    ICMP_SGE   = 39,
    ICMP_SLT   = 40,
    ICMP_SLE   = 41,
};

enum dxil_rmw_code
{
    RMW_XCHG =  0,
    RMW_ADD  =  1,
    RMW_SUB  =  2,
    RMW_AND  =  3,
    RMW_NAND =  4,
    RMW_OR   =  5,
    RMW_XOR  =  6,
    RMW_MAX  =  7,
    RMW_MIN  =  8,
    RMW_UMAX =  9,
    RMW_UMIN = 10,
};

enum dxil_atomic_ordering
{
    ORDERING_NOTATOMIC = 0,
    ORDERING_UNORDERED = 1,
    ORDERING_MONOTONIC = 2,
    ORDERING_ACQUIRE   = 3,
    ORDERING_RELEASE   = 4,
    ORDERING_ACQREL    = 5,
    ORDERING_SEQCST    = 6,
};

enum dxil_atomic_binop_code
{
    ATOMIC_BINOP_ADD,
    ATOMIC_BINOP_AND,
    ATOMIC_BINOP_OR,
    ATOMIC_BINOP_XOR,
    ATOMIC_BINOP_IMIN,
    ATOMIC_BINOP_IMAX,
    ATOMIC_BINOP_UMIN,
    ATOMIC_BINOP_UMAX,
    ATOMIC_BINOP_XCHG,
    ATOMIC_BINOP_INVALID,
};

enum dxil_wave_bit_op_kind
{
    WAVE_BIT_OP_AND    = 0,
    WAVE_BIT_OP_OR     = 1,
    WAVE_BIT_OP_XOR    = 2,
};

enum dxil_wave_op_kind
{
    WAVE_OP_ADD    = 0,
    WAVE_OP_MUL    = 1,
    WAVE_OP_MIN    = 2,
    WAVE_OP_MAX    = 3,
};

enum dxil_wave_quad_op_kind
{
    QUAD_OP_HORIZONTAL = 0,
    QUAD_OP_VERTICAL   = 1,
    QUAD_OP_DIAGONAL   = 2,
};

enum dxil_input_primitive
{
    INPUT_PRIMITIVE_UNDEFINED            =  0,
    INPUT_PRIMITIVE_POINT                =  1,
    INPUT_PRIMITIVE_LINE                 =  2,
    INPUT_PRIMITIVE_TRIANGLE             =  3,
    INPUT_PRIMITIVE_LINEWITHADJACENCY    =  6,
    INPUT_PRIMITIVE_TRIANGLEWITHADJACENY =  7,
    INPUT_PRIMITIVE_PATCH1               =  8,
    INPUT_PRIMITIVE_PATCH32              = 39,
};

enum dxil_shader_properties_tag
{
    SHADER_PROPERTIES_FLAGS    = 0,
    SHADER_PROPERTIES_GEOMETRY = 1,
    SHADER_PROPERTIES_DOMAIN   = 2,
    SHADER_PROPERTIES_HULL     = 3,
    SHADER_PROPERTIES_COMPUTE  = 4,
};

struct sm6_pointer_info
{
    const struct sm6_type *type;
    enum bitcode_address_space addr_space;
};

struct sm6_struct_info
{
    const char *name;
    unsigned int elem_count;
    const struct sm6_type *elem_types[];
};

struct sm6_function_info
{
    const struct sm6_type *ret_type;
    unsigned int param_count;
    const struct sm6_type *param_types[];
};

struct sm6_array_info
{
    unsigned int count;
    const struct sm6_type *elem_type;
};

enum sm6_type_class
{
    TYPE_CLASS_VOID,
    TYPE_CLASS_INTEGER,
    TYPE_CLASS_FLOAT,
    TYPE_CLASS_POINTER,
    TYPE_CLASS_STRUCT,
    TYPE_CLASS_FUNCTION,
    TYPE_CLASS_VECTOR,
    TYPE_CLASS_ARRAY,
    TYPE_CLASS_LABEL,
    TYPE_CLASS_METADATA,
};

struct sm6_type
{
    enum sm6_type_class class;
    union
    {
        unsigned int width;
        struct sm6_pointer_info pointer;
        struct sm6_struct_info *struc;
        struct sm6_function_info *function;
        struct sm6_array_info array;
    } u;
};

enum sm6_value_type
{
    VALUE_TYPE_FUNCTION,
    VALUE_TYPE_REG,
    VALUE_TYPE_HANDLE,
};

struct sm6_function_data
{
    const char *name;
    bool is_prototype;
    unsigned int attribs_id;
};

struct sm6_handle_data
{
    const struct sm6_resource *resource;
    struct vkd3d_shader_register reg;
};

struct sm6_value
{
    const struct sm6_type *type;
    enum sm6_value_type value_type;
    unsigned int structure_stride;
    bool is_undefined;
    union
    {
        struct sm6_function_data function;
        struct vkd3d_shader_register reg;
        struct sm6_handle_data handle;
    } u;
};

struct dxil_record
{
    unsigned int code;
    unsigned int operand_count;
    uint64_t operands[];
};

struct sm6_symbol
{
    unsigned int id;
    const char *name;
};

struct dxil_block
{
    const struct dxil_block *parent;
    enum bitcode_block_id id;
    unsigned int abbrev_len;
    unsigned int start;
    unsigned int length;
    unsigned int level;

    /* The abbrev, block and record structs are not relocatable. */
    struct dxil_abbrev **abbrevs;
    size_t abbrev_capacity;
    size_t abbrev_count;
    unsigned int blockinfo_bid;
    bool has_bid;

    struct dxil_block **child_blocks;
    size_t child_block_capacity;
    size_t child_block_count;

    struct dxil_record **records;
    size_t record_capacity;
    size_t record_count;
};

enum sm6_metadata_type
{
    VKD3D_METADATA_KIND,
    VKD3D_METADATA_NODE,
    VKD3D_METADATA_STRING,
    VKD3D_METADATA_VALUE,
};

struct sm6_metadata_node
{
    unsigned int index;
    bool is_distinct;
    unsigned int operand_count;
    struct sm6_metadata_value *operands[];
};

struct sm6_metadata_kind
{
    uint64_t id;
    char *name;
};

struct sm6_metadata_value
{
    unsigned int index;
    enum sm6_metadata_type type;
    const struct sm6_type *value_type;
    union
    {
        char *string_value;
        const struct sm6_value *value;
        struct sm6_metadata_node *node;
        struct sm6_metadata_kind kind;
    } u;
};

struct sm6_named_metadata
{
    char *name;
    struct sm6_metadata_value value;
};

struct sm6_resource
{
    enum dxil_resource_class class;
    unsigned int id;
    enum vkd3d_shader_resource_type type;
    enum dxil_resource_kind kind;
    struct vkd3d_shader_register *reg;
    enum vkd3d_data_type data_type;
};

struct sm6_parser
{
    const uint32_t *ptr, *start, *end;
    unsigned int bitpos;

    struct dxil_block root_block;
    struct dxil_block *current_block;

    struct dxil_global_abbrev **abbrevs;
    size_t abbrev_capacity;
    size_t abbrev_count;

    struct sm6_type *types;
    unsigned int type_count;
    struct sm6_type *bool_type;
    struct sm6_type *metadata_type;
    struct sm6_type *handle_type;

    struct sm6_symbol *global_symbols;
    unsigned int global_symbol_count;

    const char *entry_point;
    const char *patch_constant_function;

    struct vkd3d_shader_dst_param *output_params;
    struct vkd3d_shader_dst_param *input_params;
    struct vkd3d_shader_dst_param *patch_constant_params;
    uint64_t input_reg_declarations;
    unsigned int input_control_point_count;
    unsigned int output_control_point_count;

    struct sm6_function *functions;
    unsigned int function_count;

    struct sm6_metadata_value *metadata_tables[VKD3D_SM6_MAX_METADATA_TABLES];
    unsigned int metadata_counts[VKD3D_SM6_MAX_METADATA_TABLES];
    struct sm6_named_metadata *named_metadata;
    unsigned int named_metadata_count;

    struct sm6_resource *resources;
    unsigned int resource_count;

    unsigned int indexable_temp_count;

    struct sm6_value *values;
    unsigned int value_count;
    unsigned int value_capacity;
    unsigned int cur_max_value;
    unsigned int ssa_next_id;

    const struct vkd3d_shader_register *coverage_reg;

    struct vkd3d_shader_parser p;
};

struct dxil_abbrev_operand
{
    uint64_t context;
    bool (*read_operand)(struct sm6_parser *sm6, uint64_t context, uint64_t *operand);
};

struct dxil_abbrev
{
    unsigned int count;
    bool is_array;
    struct dxil_abbrev_operand operands[];
};

struct dxil_global_abbrev
{
    unsigned int block_id;
    struct dxil_abbrev abbrev;
};

static struct sm6_parser *sm6_parser(struct vkd3d_shader_parser *parser)
{
    return CONTAINING_RECORD(parser, struct sm6_parser, p);
}

static bool sm6_parser_is_end(struct sm6_parser *sm6)
{
    return sm6->ptr == sm6->end;
}

static uint32_t sm6_parser_read_uint32(struct sm6_parser *sm6)
{
    if (sm6_parser_is_end(sm6))
    {
        sm6->p.failed = true;
        return 0;
    }
    return *sm6->ptr++;
}

static uint32_t sm6_parser_read_bits(struct sm6_parser *sm6, unsigned int length)
{
    unsigned int l, prev_len = 0;
    uint32_t bits;

    if (!length)
        return 0;

    assert(length < 32);

    if (sm6_parser_is_end(sm6))
    {
        sm6->p.failed = true;
        return 0;
    }

    assert(sm6->bitpos < 32);
    bits = *sm6->ptr >> sm6->bitpos;
    l = 32 - sm6->bitpos;
    if (l <= length)
    {
        ++sm6->ptr;
        if (sm6_parser_is_end(sm6) && l < length)
        {
            sm6->p.failed = true;
            return bits;
        }
        sm6->bitpos = 0;
        bits |= *sm6->ptr << l;
        prev_len = l;
    }
    sm6->bitpos += length - prev_len;

    return bits & ((1 << length) - 1);
}

static uint64_t sm6_parser_read_vbr(struct sm6_parser *sm6, unsigned int length)
{
    unsigned int bits, flag, mask, shift = 0;
    uint64_t result = 0;

    if (!length)
        return 0;

    if (sm6_parser_is_end(sm6))
    {
        sm6->p.failed = true;
        return 0;
    }

    flag = 1 << (length - 1);
    mask = flag - 1;
    do
    {
        bits = sm6_parser_read_bits(sm6, length);
        result |= (uint64_t)(bits & mask) << shift;
        shift += length - 1;
    } while ((bits & flag) && !sm6->p.failed && shift < 64);

    sm6->p.failed |= !!(bits & flag);

    return result;
}

static void sm6_parser_align_32(struct sm6_parser *sm6)
{
    if (!sm6->bitpos)
        return;

    if (sm6_parser_is_end(sm6))
    {
        sm6->p.failed = true;
        return;
    }

    ++sm6->ptr;
    sm6->bitpos = 0;
}

static bool dxil_block_handle_blockinfo_record(struct dxil_block *block, struct dxil_record *record)
{
    /* BLOCKINFO blocks must only occur immediately below the module root block. */
    if (block->level > 1)
    {
        WARN("Invalid blockinfo block level %u.\n", block->level);
        return false;
    }

    switch (record->code)
    {
        case SETBID:
            if (!record->operand_count)
            {
                WARN("Missing id operand.\n");
                return false;
            }
            if (record->operands[0] > UINT_MAX)
                WARN("Truncating block id %"PRIu64".\n", record->operands[0]);
            block->blockinfo_bid = record->operands[0];
            block->has_bid = true;
            break;
        case BLOCKNAME:
        case SETRECORDNAME:
            break;
        default:
            FIXME("Unhandled BLOCKINFO record type %u.\n", record->code);
            break;
    }

    return true;
}

static enum vkd3d_result dxil_block_add_record(struct dxil_block *block, struct dxil_record *record)
{
    unsigned int reserve;

    switch (block->id)
    {
        /* Rough initial reserve sizes for small shaders. */
        case CONSTANTS_BLOCK: reserve = 32; break;
        case FUNCTION_BLOCK: reserve = 128; break;
        case METADATA_BLOCK: reserve = 32; break;
        case TYPE_BLOCK: reserve = 32; break;
        default: reserve = 8; break;
    }
    reserve = max(reserve, block->record_count + 1);
    if (!vkd3d_array_reserve((void **)&block->records, &block->record_capacity, reserve, sizeof(*block->records)))
    {
        ERR("Failed to allocate %u records.\n", reserve);
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    if (block->id == BLOCKINFO_BLOCK && !dxil_block_handle_blockinfo_record(block, record))
        return VKD3D_ERROR_INVALID_SHADER;

    block->records[block->record_count++] = record;

    return VKD3D_OK;
}

static enum vkd3d_result sm6_parser_read_unabbrev_record(struct sm6_parser *sm6)
{
    struct dxil_block *block = sm6->current_block;
    enum vkd3d_result ret = VKD3D_OK;
    unsigned int code, count, i;
    struct dxil_record *record;

    code = sm6_parser_read_vbr(sm6, 6);

    count = sm6_parser_read_vbr(sm6, 6);
    if (!(record = vkd3d_malloc(sizeof(*record) + count * sizeof(record->operands[0]))))
    {
        ERR("Failed to allocate record with %u operands.\n", count);
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    record->code = code;
    record->operand_count = count;

    for (i = 0; i < count; ++i)
        record->operands[i] = sm6_parser_read_vbr(sm6, 6);
    if (sm6->p.failed)
        ret = VKD3D_ERROR_INVALID_SHADER;

    if (ret < 0 || (ret = dxil_block_add_record(block, record)) < 0)
        vkd3d_free(record);

    return ret;
}

static bool sm6_parser_read_literal_operand(struct sm6_parser *sm6, uint64_t context, uint64_t *op)
{
    *op = context;
    return !sm6->p.failed;
}

static bool sm6_parser_read_fixed_operand(struct sm6_parser *sm6, uint64_t context, uint64_t *op)
{
    *op = sm6_parser_read_bits(sm6, context);
    return !sm6->p.failed;
}

static bool sm6_parser_read_vbr_operand(struct sm6_parser *sm6, uint64_t context, uint64_t *op)
{
    *op = sm6_parser_read_vbr(sm6, context);
    return !sm6->p.failed;
}

static bool sm6_parser_read_char6_operand(struct sm6_parser *sm6, uint64_t context, uint64_t *op)
{
    *op = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._"[sm6_parser_read_bits(sm6, 6)];
    return !sm6->p.failed;
}

static bool sm6_parser_read_blob_operand(struct sm6_parser *sm6, uint64_t context, uint64_t *op)
{
    int count = sm6_parser_read_vbr(sm6, 6);
    sm6_parser_align_32(sm6);
    for (; count > 0; count -= 4)
        sm6_parser_read_uint32(sm6);
    FIXME("Unhandled blob operand.\n");
    return false;
}

static enum vkd3d_result dxil_abbrev_init(struct dxil_abbrev *abbrev, unsigned int count, struct sm6_parser *sm6)
{
    enum bitcode_abbrev_type prev_type, type;
    unsigned int i;

    abbrev->is_array = false;

    for (i = 0, prev_type = 0; i < count && !sm6->p.failed; ++i)
    {
        if (sm6_parser_read_bits(sm6, 1))
        {
            if (prev_type == ABBREV_ARRAY)
            {
                WARN("Unexpected literal abbreviation after array.\n");
                return VKD3D_ERROR_INVALID_SHADER;
            }
            abbrev->operands[i].context = sm6_parser_read_vbr(sm6, 8);
            abbrev->operands[i].read_operand = sm6_parser_read_literal_operand;
            continue;
        }

        switch (type = sm6_parser_read_bits(sm6, 3))
        {
            case ABBREV_FIXED:
            case ABBREV_VBR:
                abbrev->operands[i].context = sm6_parser_read_vbr(sm6, 5);
                abbrev->operands[i].read_operand = (type == ABBREV_FIXED) ? sm6_parser_read_fixed_operand
                        : sm6_parser_read_vbr_operand;
                break;

            case ABBREV_ARRAY:
                if (prev_type == ABBREV_ARRAY || i != count - 2)
                {
                    WARN("Unexpected array abbreviation.\n");
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                abbrev->is_array = true;
                --i;
                --count;
                break;

            case ABBREV_CHAR:
                abbrev->operands[i].read_operand = sm6_parser_read_char6_operand;
                break;

            case ABBREV_BLOB:
                if (prev_type == ABBREV_ARRAY || i != count - 1)
                {
                    WARN("Unexpected blob abbreviation.\n");
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                abbrev->operands[i].read_operand = sm6_parser_read_blob_operand;
                break;
        }

        prev_type = type;
    }

    abbrev->count = count;

    return sm6->p.failed ? VKD3D_ERROR_INVALID_SHADER : VKD3D_OK;
}

static enum vkd3d_result sm6_parser_add_global_abbrev(struct sm6_parser *sm6)
{
    struct dxil_block *block = sm6->current_block;
    unsigned int count = sm6_parser_read_vbr(sm6, 5);
    struct dxil_global_abbrev *global_abbrev;
    enum vkd3d_result ret;

    assert(block->id == BLOCKINFO_BLOCK);

    if (!vkd3d_array_reserve((void **)&sm6->abbrevs, &sm6->abbrev_capacity, sm6->abbrev_count + 1, sizeof(*sm6->abbrevs))
            || !(global_abbrev = vkd3d_malloc(sizeof(*global_abbrev) + count * sizeof(global_abbrev->abbrev.operands[0]))))
    {
        ERR("Failed to allocate global abbreviation.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    if ((ret = dxil_abbrev_init(&global_abbrev->abbrev, count, sm6)) < 0)
    {
        vkd3d_free(global_abbrev);
        return ret;
    }

    if (!block->has_bid)
    {
        WARN("Missing blockinfo block id.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }
    if (block->blockinfo_bid == MODULE_BLOCK)
    {
        FIXME("Unhandled global abbreviation for module block.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }
    global_abbrev->block_id = block->blockinfo_bid;

    sm6->abbrevs[sm6->abbrev_count++] = global_abbrev;

    return VKD3D_OK;
}

static enum vkd3d_result sm6_parser_add_block_abbrev(struct sm6_parser *sm6)
{
    struct dxil_block *block = sm6->current_block;
    struct dxil_abbrev *abbrev;
    enum vkd3d_result ret;
    unsigned int count;

    if (block->id == BLOCKINFO_BLOCK)
        return sm6_parser_add_global_abbrev(sm6);

    count = sm6_parser_read_vbr(sm6, 5);
    if (!vkd3d_array_reserve((void **)&block->abbrevs, &block->abbrev_capacity, block->abbrev_count + 1, sizeof(*block->abbrevs))
            || !(abbrev = vkd3d_malloc(sizeof(*abbrev) + count * sizeof(abbrev->operands[0]))))
    {
        ERR("Failed to allocate block abbreviation.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    if ((ret = dxil_abbrev_init(abbrev, count, sm6)) < 0)
    {
        vkd3d_free(abbrev);
        return ret;
    }

    block->abbrevs[block->abbrev_count++] = abbrev;

    return VKD3D_OK;
}

static enum vkd3d_result sm6_parser_read_abbrev_record(struct sm6_parser *sm6, unsigned int abbrev_id)
{
    enum vkd3d_result ret = VKD3D_ERROR_INVALID_SHADER;
    struct dxil_block *block = sm6->current_block;
    struct dxil_record *temp, *record;
    unsigned int i, count, array_len;
    struct dxil_abbrev *abbrev;
    uint64_t code;

    if (abbrev_id >= block->abbrev_count)
    {
        WARN("Invalid abbreviation id %u.\n", abbrev_id);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    abbrev = block->abbrevs[abbrev_id];
    if (!(count = abbrev->count))
        return VKD3D_OK;
    if (count == 1 && abbrev->is_array)
        return VKD3D_ERROR_INVALID_SHADER;

    /* First operand is the record code. The array is included in the count, but will be done separately. */
    count -= abbrev->is_array + 1;
    if (!(record = vkd3d_malloc(sizeof(*record) + count * sizeof(record->operands[0]))))
    {
        ERR("Failed to allocate record with %u operands.\n", count);
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    if (!abbrev->operands[0].read_operand(sm6, abbrev->operands[0].context, &code))
        goto fail;
    if (code > UINT_MAX)
        FIXME("Truncating 64-bit record code %#"PRIx64".\n", code);
    record->code = code;

    for (i = 0; i < count; ++i)
        if (!abbrev->operands[i + 1].read_operand(sm6, abbrev->operands[i + 1].context, &record->operands[i]))
            goto fail;
    record->operand_count = count;

    /* An array can occur only as the last operand. */
    if (abbrev->is_array)
    {
        array_len = sm6_parser_read_vbr(sm6, 6);
        if (!(temp = vkd3d_realloc(record, sizeof(*record) + (count + array_len) * sizeof(record->operands[0]))))
        {
            ERR("Failed to allocate record with %u operands.\n", count + array_len);
            ret = VKD3D_ERROR_OUT_OF_MEMORY;
            goto fail;
        }
        record = temp;

        for (i = 0; i < array_len; ++i)
        {
            if (!abbrev->operands[count + 1].read_operand(sm6, abbrev->operands[count + 1].context,
                    &record->operands[count + i]))
            {
                goto fail;
            }
        }
        record->operand_count += array_len;
    }

    if ((ret = dxil_block_add_record(block, record)) < 0)
        goto fail;

    return VKD3D_OK;

fail:
    vkd3d_free(record);
    return ret;
}

static enum vkd3d_result dxil_block_init(struct dxil_block *block, const struct dxil_block *parent,
        struct sm6_parser *sm6);

static enum vkd3d_result dxil_block_read(struct dxil_block *parent, struct sm6_parser *sm6)
{
    unsigned int reserve = (parent->id == MODULE_BLOCK) ? 12 : 2;
    struct dxil_block *block;
    enum vkd3d_result ret;

    sm6->current_block = parent;

    do
    {
        unsigned int abbrev_id = sm6_parser_read_bits(sm6, parent->abbrev_len);

        switch (abbrev_id)
        {
            case END_BLOCK:
                sm6_parser_align_32(sm6);
                return VKD3D_OK;

            case ENTER_SUBBLOCK:
                if (parent->id != MODULE_BLOCK && parent->id != FUNCTION_BLOCK)
                {
                    WARN("Invalid subblock parent id %u.\n", parent->id);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (!vkd3d_array_reserve((void **)&parent->child_blocks, &parent->child_block_capacity,
                        max(reserve, parent->child_block_count + 1), sizeof(*parent->child_blocks))
                        || !(block = vkd3d_calloc(1, sizeof(*block))))
                {
                    ERR("Failed to allocate block.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }

                if ((ret = dxil_block_init(block, parent, sm6)) < 0)
                {
                    vkd3d_free(block);
                    return ret;
                }

                parent->child_blocks[parent->child_block_count++] = block;
                sm6->current_block = parent;
                break;

            case DEFINE_ABBREV:
                if ((ret = sm6_parser_add_block_abbrev(sm6)) < 0)
                    return ret;
                break;

            case UNABBREV_RECORD:
                if ((ret = sm6_parser_read_unabbrev_record(sm6)) < 0)
                {
                    WARN("Failed to read unabbreviated record.\n");
                    return ret;
                }
                break;

            default:
                if ((ret = sm6_parser_read_abbrev_record(sm6, abbrev_id - 4)) < 0)
                {
                    WARN("Failed to read abbreviated record.\n");
                    return ret;
                }
                break;
        }
    } while (!sm6->p.failed);

    return VKD3D_ERROR_INVALID_SHADER;
}

static unsigned int sm6_parser_compute_global_abbrev_count_for_block_id(struct sm6_parser *sm6,
        unsigned int block_id)
{
    unsigned int i, count;

    for (i = 0, count = 0; i < sm6->abbrev_count; ++i)
        count += sm6->abbrevs[i]->block_id == block_id;

    return count;
}

static void dxil_block_destroy(struct dxil_block *block)
{
    unsigned int i;

    for (i = 0; i < block->record_count; ++i)
        vkd3d_free(block->records[i]);
    vkd3d_free(block->records);

    for (i = 0; i < block->child_block_count; ++i)
    {
        dxil_block_destroy(block->child_blocks[i]);
        vkd3d_free(block->child_blocks[i]);
    }
    vkd3d_free(block->child_blocks);

    block->records = NULL;
    block->record_count = 0;
    block->child_blocks = NULL;
    block->child_block_count = 0;
}

static enum vkd3d_result dxil_block_init(struct dxil_block *block, const struct dxil_block *parent,
        struct sm6_parser *sm6)
{
    unsigned int i, abbrev_count = 0;
    enum vkd3d_result ret;

    block->parent = parent;
    block->level = parent ? parent->level + 1 : 0;
    block->id = sm6_parser_read_vbr(sm6, 8);
    block->abbrev_len = sm6_parser_read_vbr(sm6, 4);
    sm6_parser_align_32(sm6);
    block->length = sm6_parser_read_uint32(sm6);
    block->start = sm6->ptr - sm6->start;

    if (sm6->p.failed)
        return VKD3D_ERROR_INVALID_SHADER;

    if ((block->abbrev_count = sm6_parser_compute_global_abbrev_count_for_block_id(sm6, block->id)))
    {
        if (!vkd3d_array_reserve((void **)&block->abbrevs, &block->abbrev_capacity,
                block->abbrev_count, sizeof(*block->abbrevs)))
        {
            ERR("Failed to allocate block abbreviations.\n");
            return VKD3D_ERROR_OUT_OF_MEMORY;
        }

        for (i = 0; i < sm6->abbrev_count; ++i)
            if (sm6->abbrevs[i]->block_id == block->id)
                block->abbrevs[abbrev_count++] = &sm6->abbrevs[i]->abbrev;

        assert(abbrev_count == block->abbrev_count);
    }

    if ((ret = dxil_block_read(block, sm6)) < 0)
        dxil_block_destroy(block);

    for (i = abbrev_count; i < block->abbrev_count; ++i)
        vkd3d_free(block->abbrevs[i]);
    vkd3d_free(block->abbrevs);
    block->abbrevs = NULL;
    block->abbrev_count = 0;

    return ret;
}

static unsigned int dxil_block_compute_function_count(const struct dxil_block *root)
{
    unsigned int i, count;

    for (i = 0, count = 0; i < root->child_block_count; ++i)
        count += root->child_blocks[i]->id == FUNCTION_BLOCK;

    return count;
}

static unsigned int dxil_block_compute_module_decl_count(const struct dxil_block *block)
{
    unsigned int i, count;

    for (i = 0, count = 0; i < block->record_count; ++i)
        count += block->records[i]->code == MODULE_CODE_FUNCTION || block->records[i]->code == MODULE_CODE_GLOBALVAR;
    return count;
}

static unsigned int dxil_block_compute_constants_count(const struct dxil_block *block)
{
    unsigned int i, count;

    for (i = 0, count = 0; i < block->record_count; ++i)
        count += block->records[i]->code != CST_CODE_SETTYPE;
    return count;
}

static void dxil_global_abbrevs_cleanup(struct dxil_global_abbrev **abbrevs, unsigned int count)
{
    unsigned int i;

    for (i = 0; i < count; ++i)
        vkd3d_free(abbrevs[i]);
    vkd3d_free(abbrevs);
}

static const struct dxil_block *sm6_parser_get_level_one_block(const struct sm6_parser *sm6,
        enum bitcode_block_id id, bool *is_unique)
{
    const struct dxil_block *block, *found = NULL;
    unsigned int i;

    for (i = 0, *is_unique = true; i < sm6->root_block.child_block_count; ++i)
    {
        block = sm6->root_block.child_blocks[i];
        if (block->id != id)
            continue;

        if (!found)
            found = block;
        else
            *is_unique = false;
    }

    return found;
}

static char *dxil_record_to_string(const struct dxil_record *record, unsigned int offset)
{
    unsigned int i;
    char *str;

    assert(offset <= record->operand_count);
    if (!(str = vkd3d_malloc(record->operand_count - offset + 1)))
        return NULL;

    for (i = offset; i < record->operand_count; ++i)
        str[i - offset] = record->operands[i];
    str[i - offset] = 0;

    return str;
}

static enum vkd3d_result sm6_parser_type_table_init(struct sm6_parser *sm6)
{
    unsigned int i, j, count, type_count, type_index;
    const struct dxil_record *record;
    const struct dxil_block *block;
    char *struct_name = NULL;
    struct sm6_type *type;
    uint64_t type_id;
    bool is_unique;

    if (!(block = sm6_parser_get_level_one_block(sm6, TYPE_BLOCK, &is_unique)))
    {
        WARN("No type definitions found.\n");
        return VKD3D_OK;
    }
    if (!is_unique)
        FIXME("Ignoring extra type table(s).\n");

    type_count = 0;
    for (i = 0; i < block->record_count; ++i)
        type_count += block->records[i]->code != TYPE_CODE_NUMENTRY && block->records[i]->code != TYPE_CODE_STRUCT_NAME;

    /* The type array must not be relocated. */
    if (!(sm6->types = vkd3d_calloc(type_count, sizeof(*sm6->types))))
    {
        ERR("Failed to allocate type array.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    for (i = 0; i < block->record_count; ++i)
    {
        record = block->records[i];

        type = &sm6->types[sm6->type_count];
        type_index = sm6->type_count;

        switch (record->code)
        {
            case TYPE_CODE_ARRAY:
            case TYPE_CODE_VECTOR:
                if (record->operand_count < 2)
                {
                    WARN("Missing operands for array/vector type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                type->class = record->code == TYPE_CODE_ARRAY ? TYPE_CLASS_ARRAY : TYPE_CLASS_VECTOR;

                if (!(type->u.array.count = record->operands[0]))
                {
                    TRACE("Setting unbounded for type %u.\n", type_index);
                    type->u.array.count = UINT_MAX;
                }

                if ((type_id = record->operands[1]) >= type_count)
                {
                    WARN("Invalid contained type id %"PRIu64" for type %u.\n", type_id, type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                type->u.array.elem_type = &sm6->types[type_id];
                break;

            case TYPE_CODE_DOUBLE:
                type->class = TYPE_CLASS_FLOAT;
                type->u.width = 64;
                break;

            case TYPE_CODE_FLOAT:
                type->class = TYPE_CLASS_FLOAT;
                type->u.width = 32;
                break;

            case TYPE_CODE_FUNCTION:
                if (record->operand_count < 2)
                {
                    WARN("Missing operands for function type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (record->operands[0])
                    FIXME("Unhandled vararg function type %u.\n", type_index);

                type->class = TYPE_CLASS_FUNCTION;

                if ((type_id = record->operands[1]) >= type_count)
                {
                    WARN("Invalid return type id %"PRIu64" for type %u.\n", type_id, type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                count = record->operand_count - 2;
                if (!(type->u.function = vkd3d_malloc(offsetof(struct sm6_function_info, param_types[count]))))
                {
                    ERR("Failed to allocate function parameter types.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }

                type->u.function->ret_type = &sm6->types[type_id];
                type->u.function->param_count = count;
                for (j = 0; j < count; ++j)
                {
                    if ((type_id = record->operands[j + 2]) >= type_count)
                    {
                        WARN("Invalid parameter type id %"PRIu64" for type %u.\n", type_id, type_index);
                        vkd3d_free(type->u.function);
                        return VKD3D_ERROR_INVALID_SHADER;
                    }
                    type->u.function->param_types[j] = &sm6->types[type_id];
                }
                break;

            case TYPE_CODE_HALF:
                type->class = TYPE_CLASS_FLOAT;
                type->u.width = 16;
                break;

            case TYPE_CODE_INTEGER:
            {
                uint64_t width;

                if (!record->operand_count)
                {
                    WARN("Missing integer width for type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                type->class = TYPE_CLASS_INTEGER;

                if ((width = record->operands[0]) == 1)
                {
                    sm6->bool_type = type;
                }
                else if (width > 64 || width < 8 || vkd3d_popcount(width) != 1)
                {
                    WARN("Invalid integer width %"PRIu64" for type %u.\n", width, type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                type->u.width = width;
                break;
            }

            case TYPE_CODE_LABEL:
                type->class = TYPE_CLASS_LABEL;
                break;

            case TYPE_CODE_METADATA:
                type->class = TYPE_CLASS_METADATA;
                sm6->metadata_type = type;
                break;

            case TYPE_CODE_NUMENTRY:
                continue;

            case TYPE_CODE_POINTER:
                if (!record->operand_count)
                {
                    WARN("Missing pointee type for pointer type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                type->class = TYPE_CLASS_POINTER;

                if ((type_id = record->operands[0]) >= type_count)
                {
                    WARN("Invalid pointee type id %"PRIu64" for type %u.\n", type_id, type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                type->u.pointer.type = &sm6->types[type_id];
                type->u.pointer.addr_space = (record->operand_count > 1) ? record->operands[1] : ADDRESS_SPACE_DEFAULT;
                break;

            case TYPE_CODE_STRUCT_ANON:
            case TYPE_CODE_STRUCT_NAMED:
                if (record->operand_count < 2)
                {
                    WARN("Missing operands for struct type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (record->code == TYPE_CODE_STRUCT_NAMED && !struct_name)
                {
                    WARN("Missing struct name before struct type %u.\n", type_index);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                type->class = TYPE_CLASS_STRUCT;

                count = record->operand_count - 1;
                if (!(type->u.struc = vkd3d_malloc(offsetof(struct sm6_struct_info, elem_types[count]))))
                {
                    ERR("Failed to allocate struct element types.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }

                if (record->operands[0])
                    FIXME("Ignoring struct packed attribute.\n");

                type->u.struc->elem_count = count;
                for (j = 0; j < count; ++j)
                {
                    if ((type_id = record->operands[j + 1]) >= type_count)
                    {
                        WARN("Invalid contained type id %"PRIu64" for type %u.\n", type_id, type_index);
                        vkd3d_free(type->u.struc);
                        return VKD3D_ERROR_INVALID_SHADER;
                    }
                    type->u.struc->elem_types[j] = &sm6->types[type_id];
                }

                if (record->code == TYPE_CODE_STRUCT_ANON)
                {
                    type->u.struc->name = NULL;
                    break;
                }

                if (!strcmp(struct_name, "dx.types.Handle"))
                    sm6->handle_type = type;

                type->u.struc->name = struct_name;
                struct_name = NULL;
                break;

            case TYPE_CODE_STRUCT_NAME:
                if (!(struct_name = dxil_record_to_string(record, 0)))
                {
                    ERR("Failed to allocate struct name.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }
                if (!struct_name[0])
                    WARN("Struct name is empty for type %u.\n", type_index);
                continue;

            case TYPE_CODE_VOID:
                type->class = TYPE_CLASS_VOID;
                break;

            default:
                FIXME("Unhandled type %u at index %u.\n", record->code, type_index);
                return VKD3D_ERROR_INVALID_SHADER;
        }
        ++sm6->type_count;
    }

    assert(sm6->type_count == type_count);

    if (struct_name)
    {
        WARN("Unused struct name %s.\n", struct_name);
        vkd3d_free(struct_name);
    }

    return VKD3D_OK;
}

static inline bool sm6_type_is_void(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_VOID;
}

static inline bool sm6_type_is_integer(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER;
}

static inline bool sm6_type_is_1wil(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER && (type->u.width == 1 || type->u.width >= 16);
}

static inline bool sm6_type_is_bool(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER && type->u.width == 1;
}

static inline bool sm6_type_is_i8(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER && type->u.width == 8;
}

static inline bool sm6_type_is_i32(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER && type->u.width == 32;
}

static inline bool sm6_type_is_wil(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER && type->u.width >= 16;
}

static inline bool sm6_type_is_float(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_FLOAT && type->u.width == 32;
}

static inline bool sm6_type_is_double(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_FLOAT && type->u.width == 64;
}

static inline bool sm6_type_is_floating_point(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_FLOAT;
}

static inline bool sm6_type_is_scalar(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER || type->class == TYPE_CLASS_FLOAT || type->class == TYPE_CLASS_POINTER;
}

static inline bool sm6_type_is_numeric(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_INTEGER || type->class == TYPE_CLASS_FLOAT;
}

static inline bool sm6_type_is_pointer(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_POINTER;
}

static inline bool sm6_type_is_aggregate(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_STRUCT || type->class == TYPE_CLASS_VECTOR || type->class == TYPE_CLASS_ARRAY;
}

static inline bool sm6_type_is_numeric_aggregate(const struct sm6_type *type)
{
    unsigned int i;

    switch (type->class)
    {
        case TYPE_CLASS_ARRAY:
        case TYPE_CLASS_VECTOR:
            return sm6_type_is_numeric(type->u.array.elem_type);

        case TYPE_CLASS_STRUCT:
            for (i = 0; i < type->u.struc->elem_count; ++i)
                if (!sm6_type_is_numeric(type->u.struc->elem_types[i]))
                    return false;
            return true;

        default:
            return false;
    }
}

static inline bool sm6_type_is_array(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_ARRAY;
}

static inline bool sm6_type_is_vector(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_VECTOR;
}

static inline bool sm6_type_is_struct(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_STRUCT;
}

static inline bool sm6_type_is_function(const struct sm6_type *type)
{
    return type->class == TYPE_CLASS_FUNCTION;
}

static inline bool sm6_type_is_function_pointer(const struct sm6_type *type)
{
    return sm6_type_is_pointer(type) && sm6_type_is_function(type->u.pointer.type);
}

static inline bool sm6_type_is_handle(const struct sm6_type *type)
{
    return sm6_type_is_struct(type) && !strcmp(type->u.struc->name, "dx.types.Handle");
}

static inline const struct sm6_type *sm6_type_get_element_type(const struct sm6_type *type)
{
    return (type->class == TYPE_CLASS_ARRAY || type->class == TYPE_CLASS_VECTOR) ? type->u.array.elem_type
            : type;
}

static const struct sm6_type *sm6_type_get_pointer_to_type(const struct sm6_type *type,
        enum bitcode_address_space addr_space, struct sm6_parser *sm6)
{
    unsigned int i, start = type - sm6->types;
    const struct sm6_type *pointer_type;

    /* DXC seems usually to place the pointer type immediately after its pointee. */
    for (i = (start + 1) % sm6->type_count; i != start; i = (i + 1) % sm6->type_count)
    {
        pointer_type = &sm6->types[i];
        if (sm6_type_is_pointer(pointer_type) && pointer_type->u.pointer.type == type
                && pointer_type->u.pointer.addr_space == addr_space)
            return pointer_type;
    }

    return NULL;
}

static const struct sm6_type *sm6_type_get_cmpxchg_result_struct(struct sm6_parser *sm6)
{
    const struct sm6_type *type;
    unsigned int i;

    for (i = 0; i < sm6->type_count; ++i)
    {
        type = &sm6->types[i];
        if (sm6_type_is_struct(type) && type->u.struc->elem_count == 2
                && sm6_type_is_i32(type->u.struc->elem_types[0])
                && sm6_type_is_bool(type->u.struc->elem_types[1]))
        {
            return type;
        }
    }

    return NULL;
}

static const struct sm6_type *sm6_type_get_element_type_at_index(const struct sm6_type *type, unsigned int elem_idx)
{
    switch (type->class)
    {
        case TYPE_CLASS_ARRAY:
        case TYPE_CLASS_VECTOR:
            if (elem_idx != UINT_MAX && elem_idx >= type->u.array.count)
                return NULL;
            return type->u.array.elem_type;

        case TYPE_CLASS_POINTER:
            return type->u.pointer.type;

        case TYPE_CLASS_STRUCT:
            if (elem_idx >= type->u.struc->elem_count)
                return NULL;
            return type->u.struc->elem_types[elem_idx];

        default:
            return NULL;
    }
}

static const struct sm6_type *sm6_type_get_scalar_type(const struct sm6_type *type, unsigned int elem_idx)
{
    switch (type->class)
    {
        case TYPE_CLASS_ARRAY:
        case TYPE_CLASS_VECTOR:
            if (elem_idx >= type->u.array.count)
                return NULL;
            return sm6_type_get_scalar_type(type->u.array.elem_type, 0);

        case TYPE_CLASS_POINTER:
            return sm6_type_get_scalar_type(type->u.pointer.type, 0);

        case TYPE_CLASS_STRUCT:
            if (elem_idx >= type->u.struc->elem_count)
                return NULL;
            return sm6_type_get_scalar_type(type->u.struc->elem_types[elem_idx], 0);

        default:
            return type;
    }
}

static inline unsigned int sm6_type_max_vector_size(const struct sm6_type *type)
{
    return min(128u / type->u.width, VKD3D_VEC4_SIZE);
}

/* Extract the element type if type is one of the structs DXC uses to represent vector return values. */
static const struct sm6_type *sm6_type_get_pseudo_vec4_element_type(const struct sm6_type *type)
{
    const struct sm6_type *elem_type;
    unsigned int i, vec_size;

    /* These could also be identified by checking the name against all known dx struct names
     * which hold a vec4. It's unclear which would be more robust. */
    if (type->class != TYPE_CLASS_STRUCT || type->u.struc->elem_count < VKD3D_DVEC2_SIZE
            || type->u.struc->elem_count > VKD3D_VEC4_SIZE + 1)
        return NULL;

    elem_type = type->u.struc->elem_types[0];
    if (!sm6_type_is_numeric(elem_type))
        return NULL;

    vec_size = sm6_type_max_vector_size(elem_type);
    if (type->u.struc->elem_count < vec_size)
        return NULL;

    for (i = 1; i < vec_size; ++i)
        if (type->u.struc->elem_types[i] != elem_type)
            return NULL;

    return elem_type;
}

static const struct sm6_type *sm6_parser_get_type(const struct sm6_parser *sm6, const char *function, uint64_t type_id)
{
    if (type_id >= sm6->type_count)
    {
        vkd3d_dbg_printf(VKD3D_DBG_LEVEL_WARN, function, "Invalid type index %"PRIu64" at %u.\n",
                type_id, sm6->value_count);
        return NULL;
    }
    return &sm6->types[type_id];
}

#define sm6_parser_get_type(sm6, id) sm6_parser_get_type(sm6, __FUNCTION__, id)

static enum vkd3d_result sm6_parser_symtab_init(struct sm6_parser *sm6)
{
    const struct dxil_record *record;
    const struct dxil_block *block;
    struct sm6_symbol *symbol;
    unsigned int i, count;
    bool is_unique;

    if (!(block = sm6_parser_get_level_one_block(sm6, VALUE_SYMTAB_BLOCK, &is_unique)))
    {
        /* There should always be at least one symbol: the name of the entry point function. */
        WARN("No value symtab block found.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }
    if (!is_unique)
        FIXME("Ignoring extra value symtab block(s).\n");

    for (i = 0, count = 0; i < block->record_count; ++i)
        count += block->records[i]->code == VST_CODE_ENTRY;

    if (!(sm6->global_symbols = vkd3d_calloc(count, sizeof(*sm6->global_symbols))))
    {
        ERR("Failed to allocate global symbols.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    for (i = 0; i < block->record_count; ++i)
    {
        record = block->records[i];

        if (record->code != VST_CODE_ENTRY)
        {
            WARN("Skipping code %u.\n", record->code);
            continue;
        }
        if (record->operand_count < 2)
        {
            WARN("Missing operands for function symbol %u.\n", sm6->global_symbol_count);
            continue;
        }

        symbol = &sm6->global_symbols[sm6->global_symbol_count];
        symbol->id = record->operands[0];
        if (!(symbol->name = dxil_record_to_string(record, 1)))
        {
            ERR("Failed to allocate symbol name.\n");
            return VKD3D_ERROR_OUT_OF_MEMORY;
        }
        ++sm6->global_symbol_count;
    }

    return VKD3D_OK;
}

static const char *sm6_parser_get_global_symbol_name(const struct sm6_parser *sm6, unsigned int id)
{
    unsigned int i;

    for (i = 0; i < sm6->global_symbol_count; ++i)
    {
        if (sm6->global_symbols[i].id == id)
            return sm6->global_symbols[i].name;
    }
    return NULL;
}

static bool register_is_constant(const struct vkd3d_shader_register *reg)
{
    return reg->type == VKD3DSPR_IMMCONST || reg->type == VKD3DSPR_IMMCONST64;
}

static unsigned int register_get_uint_value(const struct vkd3d_shader_register *reg)
{
    if (!register_is_constant(reg))
        return UINT_MAX;

    if (reg->immconst_type == VKD3D_IMMCONST_VEC4)
        WARN("Returning vec4.x.\n");

    if (reg->type == VKD3DSPR_IMMCONST64)
    {
        if (reg->u.immconst_uint64[0] > UINT_MAX)
            FIXME("Truncating 64-bit value.\n");
        return reg->u.immconst_uint64[0];
    }

    return reg->u.immconst_uint[0];
}

static uint64_t register_get_uint64_value(const struct vkd3d_shader_register *reg)
{
    if (reg->type != VKD3DSPR_IMMCONST && reg->type != VKD3DSPR_IMMCONST64)
        return UINT64_MAX;

    if (reg->type == VKD3DSPR_IMMCONST64)
        return reg->u.immconst_uint64[0];

    return reg->u.immconst_uint[0];
}

static inline bool sm6_value_is_function_dcl(const struct sm6_value *value)
{
    return value->value_type == VALUE_TYPE_FUNCTION;
}

static inline bool sm6_value_is_dx_intrinsic_dcl(const struct sm6_value *fn)
{
    assert(sm6_value_is_function_dcl(fn));
    return fn->u.function.is_prototype && !strncmp(fn->u.function.name, "dx.op.", 6);
}

static inline struct sm6_value *sm6_parser_get_current_value(const struct sm6_parser *sm6)
{
    return &sm6->values[sm6->value_count];
}

static inline bool sm6_value_is_register(const struct sm6_value *value)
{
    return value->value_type == VALUE_TYPE_REG;
}

static inline bool sm6_value_is_handle(const struct sm6_value *value)
{
    return value->value_type == VALUE_TYPE_HANDLE;
}

static inline bool sm6_value_is_constant(const struct sm6_value *value)
{
    return sm6_value_is_register(value) && register_is_constant(&value->u.reg);
}

static inline bool sm6_value_is_undef(const struct sm6_value *value)
{
    return sm6_value_is_register(value) && value->u.reg.type == VKD3DSPR_UNDEF;
}

static inline bool sm6_value_vector_is_constant(const struct sm6_value **values, unsigned int count)
{
    unsigned int i;
    for (i = 0; i < count; ++i)
        if (!sm6_value_is_constant(values[i]) && !sm6_value_is_undef(values[i]))
            return false;
    return true;
}

static inline bool sm6_value_is_constant_array(const struct sm6_value *value)
{
    return sm6_value_is_register(value) && value->u.reg.type == VKD3DSPR_IMMCONSTARRAY;
}

static inline bool sm6_value_is_ssa(const struct sm6_value *value)
{
    return sm6_value_is_register(value) && register_is_ssa(&value->u.reg);
}

static inline unsigned int sm6_value_get_constant_uint(const struct sm6_value *value)
{
    if (!sm6_value_is_constant(value))
        return UINT_MAX;
    return register_get_uint_value(&value->u.reg);
}

static inline unsigned int sm6_parser_alloc_ssa_id(struct sm6_parser *sm6)
{
    return sm6->ssa_next_id++;
}

static void instruction_init_with_resource(struct vkd3d_shader_instruction *ins,
        enum vkd3d_shader_opcode handler_idx, const struct sm6_value *resource)
{
    shader_instruction_init(ins, handler_idx);
    ins->resource_type = resource->u.handle.resource->type;
    ins->raw = resource->u.handle.resource->kind == RESOURCE_KIND_RAWBUFFER;
    ins->structured = resource->u.handle.resource->kind == RESOURCE_KIND_STRUCTUREDBUFFER;
}

static struct vkd3d_shader_src_param *instruction_src_params_alloc(struct vkd3d_shader_instruction *ins,
        unsigned int count, struct sm6_parser *sm6)
{
    struct vkd3d_shader_src_param *params = shader_src_param_allocator_get(&sm6->p.instructions.src_params, count);
    ins->src = params;
    ins->src_count = count;
    return params;
}

static struct vkd3d_shader_dst_param *instruction_dst_params_alloc(struct vkd3d_shader_instruction *ins,
        unsigned int count, struct sm6_parser *sm6)
{
    struct vkd3d_shader_dst_param *params = shader_dst_param_allocator_get(&sm6->p.instructions.dst_params, count);
    ins->dst = params;
    ins->dst_count = count;
    return params;
}

static void register_init_with_id(struct vkd3d_shader_register *reg,
        enum vkd3d_shader_register_type reg_type, enum vkd3d_data_type data_type, unsigned int index)
{
    reg->type = reg_type;
    reg->precision = VKD3D_SHADER_REGISTER_PRECISION_DEFAULT;
    reg->non_uniform = false;
    reg->data_type = data_type;
    reg->idx[0].offset = index;
    reg->idx[0].rel_addr = NULL;
    reg->idx[1].offset = UINT_MAX;
    reg->idx[1].rel_addr = NULL;
    reg->idx[2].offset = UINT_MAX;
    reg->idx[2].rel_addr = NULL;
    reg->idx_count = 1;
    reg->immconst_type = VKD3D_IMMCONST_SCALAR;
}

static void register_init_ssa_vector(struct vkd3d_shader_register *reg, enum vkd3d_data_type data_type,
        unsigned int component_count, struct sm6_value *value, struct sm6_parser *sm6)
{
    unsigned int id;

    if (value && register_is_ssa(&value->u.reg) && value->u.reg.idx[0].offset)
    {
        id = value->u.reg.idx[0].offset;
        TRACE("Using forward-allocated id %u.\n", id);
    }
    else
    {
        id = sm6_parser_alloc_ssa_id(sm6);
    }
    register_init_with_id(reg, VKD3DSPR_SSA, data_type, id);
    reg->idx[1].offset = data_type;
    reg->idx_count = 2;
    reg->immconst_type = component_count > 1 ? VKD3D_IMMCONST_VEC4 : VKD3D_IMMCONST_SCALAR;
}

void shader_register_make_scalar_ssa(struct vkd3d_shader_register *reg, enum vkd3d_data_type data_type, void *parser)
{
    register_init_ssa_vector(reg, data_type, 1, NULL, parser);
}

static enum vkd3d_data_type vkd3d_data_type_from_sm6_type(const struct sm6_type *type)
{
    if (type->class == TYPE_CLASS_INTEGER)
    {
        switch (type->u.width)
        {
            case 1:
                return VKD3D_DATA_BOOL;
            case 8:
                return VKD3D_DATA_UINT8;
            case 16:
                return VKD3D_DATA_UINT16;
            case 32:
                return VKD3D_DATA_UINT;
            case 64:
                return VKD3D_DATA_UINT64;
            default:
                FIXME("Unhandled width %u.\n", type->u.width);
                return VKD3D_DATA_UINT;
        }
    }
    else if (type->class == TYPE_CLASS_FLOAT)
    {
        switch (type->u.width)
        {
            case 16:
                return VKD3D_DATA_HALF;
            case 32:
                return VKD3D_DATA_FLOAT;
            case 64:
                return VKD3D_DATA_DOUBLE;
            default:
                FIXME("Unhandled width %u.\n", type->u.width);
                return VKD3D_DATA_FLOAT;
        }
    }

    FIXME("Unhandled type %u.\n", type->class);
    return VKD3D_DATA_UINT;
}

static inline void register_init_ssa_scalar(struct vkd3d_shader_register *reg, const struct sm6_type *type,
        struct sm6_value *value, struct sm6_parser *sm6)
{
    register_init_ssa_vector(reg, vkd3d_data_type_from_sm6_type(sm6_type_get_scalar_type(type, 0)),
            1, value, sm6);
}

unsigned int sm6_get_ssa_register_count(const void *parser)
{
    const struct sm6_parser *sm6 = parser;
    return sm6->ssa_next_id;
}

void register_make_constant_bool(struct vkd3d_shader_register *reg, bool value)
{
    shader_register_init(reg, VKD3DSPR_IMMCONST, VKD3D_DATA_BOOL, 0);
    reg->u.immconst_uint[0] = value;
}

void register_make_constant_uint(struct vkd3d_shader_register *reg, unsigned int value)
{
    shader_register_init(reg, VKD3DSPR_IMMCONST, VKD3D_DATA_UINT, 0);
    reg->u.immconst_uint[0] = value;
}

void register_make_undef(struct vkd3d_shader_register *reg, enum vkd3d_data_type data_type)
{
    shader_register_init(reg, VKD3DSPR_UNDEF, data_type, 0);
}

static inline void dst_param_init(struct vkd3d_shader_dst_param *param)
{
    param->write_mask = VKD3DSP_WRITEMASK_0;
    param->modifiers = 0;
    param->shift = 0;
}

static inline void dst_param_init_with_mask(struct vkd3d_shader_dst_param *param, unsigned int mask)
{
    param->write_mask = mask;
    param->modifiers = 0;
    param->shift = 0;
}

static inline void dst_param_init_scalar(struct vkd3d_shader_dst_param *param, unsigned int component_idx)
{
    param->write_mask = 1u << component_idx;
    param->modifiers = 0;
    param->shift = 0;
}

static inline void dst_param_vector_init(struct vkd3d_shader_dst_param *param, unsigned int component_count)
{
    param->write_mask = (1u << component_count) - 1;
    param->modifiers = 0;
    param->shift = 0;
}

static inline void dst_param_init_ssa_scalar(struct vkd3d_shader_dst_param *param, const struct sm6_type *type,
        struct sm6_value *value, struct sm6_parser *sm6)
{
    dst_param_init(param);
    register_init_ssa_scalar(&param->reg, type, value, sm6);
}

static inline void src_param_init(struct vkd3d_shader_src_param *param)
{
    param->swizzle = VKD3D_SHADER_SWIZZLE(X, X, X, X);
    param->modifiers = VKD3DSPSM_NONE;
}

static inline void src_param_init_scalar(struct vkd3d_shader_src_param *param, unsigned int component_idx)
{
    param->swizzle = vkd3d_shader_create_swizzle(component_idx, component_idx, component_idx, component_idx);
    param->modifiers = VKD3DSPSM_NONE;
}

static inline void src_param_init_vector(struct vkd3d_shader_src_param *param, unsigned int component_count)
{
    param->swizzle = VKD3D_SHADER_NO_SWIZZLE & ((1ull << VKD3D_SHADER_SWIZZLE_SHIFT(component_count)) - 1);
    param->modifiers = VKD3DSPSM_NONE;
}

static void src_param_init_from_value(struct vkd3d_shader_src_param *param, const struct sm6_value *src)
{
    src_param_init(param);
    param->reg = src->u.reg;
}

static inline void src_param_init_vector_from_reg(struct vkd3d_shader_src_param *param,
        const struct vkd3d_shader_register *reg)
{
    param->swizzle = VKD3D_SHADER_NO_SWIZZLE;
    param->modifiers = VKD3DSPSM_NONE;
    param->reg = *reg;
}

static void src_param_make_constant_uint(struct vkd3d_shader_src_param *param, unsigned int value)
{
    src_param_init(param);
    shader_register_init(&param->reg, VKD3DSPR_IMMCONST, VKD3D_DATA_UINT, 0);
    param->reg.u.immconst_uint[0] = value;
}

static void register_address_init(struct vkd3d_shader_register *reg, const struct sm6_value *address,
        unsigned int idx, struct sm6_parser *sm6)
{
    assert(idx < ARRAY_SIZE(reg->idx));
    if (sm6_value_is_constant(address))
    {
        reg->idx[idx].offset = sm6_value_get_constant_uint(address);
    }
    else if (sm6_value_is_undef(address))
    {
        reg->idx[idx].offset = 0;
    }
    else
    {
        struct vkd3d_shader_src_param *rel_addr = shader_src_param_allocator_get(&sm6->p.instructions.src_params, 1);
        src_param_init_from_value(rel_addr, address);
        reg->idx[idx].offset = 0;
        reg->idx[idx].rel_addr = rel_addr;
    }
}

static void instruction_dst_param_init_ssa_scalar_component(struct vkd3d_shader_instruction *ins,
        unsigned int component_idx, struct sm6_parser *sm6)
{
    struct vkd3d_shader_dst_param *param = instruction_dst_params_alloc(ins, 1, sm6);
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);

    dst_param_init_ssa_scalar(param, dst->type, dst, sm6);
    param->write_mask = VKD3DSP_WRITEMASK_0 << component_idx;
    dst->u.reg = param->reg;
}

static inline void instruction_dst_param_init_ssa_scalar(struct vkd3d_shader_instruction *ins, struct sm6_parser *sm6)
{
    return instruction_dst_param_init_ssa_scalar_component(ins, 0, sm6);
}

static void instruction_dst_param_init_ssa_scalar_with_modifiers(struct vkd3d_shader_instruction *ins,
        enum vkd3d_shader_dst_modifier modifiers, struct sm6_parser *sm6)
{
    struct vkd3d_shader_dst_param *param = instruction_dst_params_alloc(ins, 1, sm6);
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);

    dst_param_init(param);
    param->modifiers = modifiers;
    register_init_ssa_scalar(&param->reg, dst->type, dst, sm6);
    dst->u.reg = param->reg;
}

static void instruction_dst_param_vector_init_ssa(struct sm6_parser *sm6, struct vkd3d_shader_instruction *ins,
        unsigned int component_count)
{
    struct vkd3d_shader_dst_param *param = instruction_dst_params_alloc(ins, 1, sm6);
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);

    dst_param_vector_init(param, component_count);
    register_init_ssa_vector(&param->reg, vkd3d_data_type_from_sm6_type(sm6_type_get_scalar_type(
            dst->type, 0)), component_count, dst, sm6);
    dst->u.reg = param->reg;
}

/* Does not emit a value. Used when multiple IR instructions are required for a single DXIL instruction. */
static const struct vkd3d_shader_register *instruction_dst_param_vector_init_helper_ssa(struct sm6_parser *sm6,
        struct vkd3d_shader_instruction *ins, enum vkd3d_data_type data_type, unsigned int component_count)
{
    struct vkd3d_shader_dst_param *param = instruction_dst_params_alloc(ins, 1, sm6);

    dst_param_vector_init(param, component_count);
    register_init_ssa_vector(&param->reg, data_type, component_count, NULL, sm6);

    return &param->reg;
}

static unsigned int sm6_parser_compute_max_value_count(struct sm6_parser *sm6,
        const struct dxil_block *block, unsigned int value_count)
{
    unsigned int i, old_value_count = value_count;

    if (block->id == MODULE_BLOCK)
        value_count += dxil_block_compute_module_decl_count(block);

    for (i = 0; i < block->child_block_count; ++i)
        value_count = sm6_parser_compute_max_value_count(sm6, block->child_blocks[i], value_count);

    switch (block->id)
    {
        case CONSTANTS_BLOCK:
            /* Function local constants are contained in a child block of the function block. */
            value_count += dxil_block_compute_constants_count(block);
            break;
        case FUNCTION_BLOCK:
            /* A function must start with a block count, which emits no value. This formula is likely to
             * overestimate the value count somewhat, but this should be no problem. */
            value_count += max(block->record_count, 1u) - 1;
            sm6->value_capacity = max(sm6->value_capacity, value_count);
            sm6->functions[sm6->function_count++].value_count = value_count;
            /* The value count returns to its previous value after handling a function. */
            value_count = old_value_count;
            break;
        default:
            break;
    }

    return value_count;
}

static unsigned int sm6_parser_get_value_index(const struct sm6_parser *sm6, const char *function, uint64_t id)
{
    uint32_t i;

    if ((i = sm6->value_count - id) >= sm6->cur_max_value)
    {
        vkd3d_dbg_printf(VKD3D_DBG_LEVEL_WARN, function, "Invalid value index %"PRIu64" at %u.\n",
                id, sm6->value_count);
        return UINT_MAX;
    }
    if (i == sm6->value_count)
    {
        vkd3d_dbg_printf(VKD3D_DBG_LEVEL_WARN, function, "Invalid value self-reference at %u.\n", sm6->value_count);
        return UINT_MAX;
    }

    return i;
}

static inline const struct sm6_value *sm6_parser_get_value(const struct sm6_parser *sm6, unsigned int id)
{
    return &sm6->values[id];
}

static inline const struct sm6_value *sm6_parser_get_value_unsafe(const struct sm6_parser *sm6, uint64_t id)
{
    return (id >= sm6->value_count) ? NULL : &sm6->values[id];
}

#define DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(min_count) do {\
    if (record->operand_count < (min_count))\
    {\
        WARN("Invalid operand count %u.\n", record->operand_count);\
        return;\
    } } while (false)

#define DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(max_count) do {\
    if (record->operand_count > (max_count))\
        WARN("Ignoring %u extra operands.\n", record->operand_count - (max_count)); } while (false)

#define DXIL_RECORD_VALIDATE_OPERAND_COUNT(min_count, max_count) do {\
    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(min_count);\
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(max_count); } while (false)

#define SM6_VALUE_VALIDATE_IS_HANDLE(v) do {\
    if (!sm6_value_is_handle(v))\
    {\
        WARN("Handle parameter is not a handle.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_REGISTER(v) do {\
    if (!sm6_value_is_register(v))\
    {\
        WARN("Operand is not a register.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_POINTER(ptr) do {\
    if (!sm6_type_is_pointer((ptr)->type))\
    {\
        WARN("Operand is not a pointer.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_POINTER_TO_I32(ptr) do {\
    if (!sm6_type_is_pointer((ptr)->type) || !sm6_type_is_i32((ptr)->type->u.pointer.type))\
    {\
        WARN("Operand is not a pointer to i32.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_NUMERIC(v) do {\
    if (!sm6_type_is_numeric((v)->type))\
    {\
        WARN("Operand is not of scalar type.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_I32(v) do {\
    if (!sm6_type_is_i32((v)->type))\
    {\
        WARN("Operand is not of type i32.\n");\
        return;\
    } } while (false)

#define SM6_VALUE_VALIDATE_IS_BOOL(v) do {\
    if (!sm6_type_is_bool((v)->type))\
    {\
        WARN("Operand is not of boolean type.\n");\
        return;\
    } } while (false)

static unsigned int sm6_parser_get_value_idx_by_ref(struct sm6_parser *sm6, const struct dxil_record *record,
        const struct sm6_type *type, unsigned int *rec_idx)
{
    const struct sm6_type *fwd_type = NULL;
    unsigned int operand, idx;
    struct sm6_value *value;
    uint64_t val_ref;

    idx = *rec_idx;
    val_ref = record->operands[idx++];
    operand = sm6_parser_get_value_index(sm6, __FUNCTION__, val_ref);

    if (operand == UINT_MAX)
        return UINT_MAX;

    if (operand >= sm6->value_count)
    {
        if (!type && record->operand_count <= idx)
        {
            WARN("Invalid operand count %u.\n", record->operand_count);
            return UINT_MAX;
        }
        /* Forward references are followed by a type id unless an earlier operand set the type,
         * or it is contained in a function declaration. */
        fwd_type = type ? type : sm6_parser_get_type(sm6, record->operands[idx++]);
        if (!fwd_type)
            WARN("Failed to get type.\n");
    }
    *rec_idx = idx;

    if (fwd_type)
    {
        value = &sm6->values[operand];
        if (value->type)
        {
            if (value->type != fwd_type)
                WARN("Value already has a mismatching type.\n");
        }
        else
        {
            value->type = fwd_type;
            value->value_type = VALUE_TYPE_REG;
            register_init_with_id(&value->u.reg, VKD3DSPR_SSA, vkd3d_data_type_from_sm6_type(
                    sm6_type_get_scalar_type(fwd_type, 0)), sm6_parser_alloc_ssa_id(sm6));
            value->u.reg.idx[1].offset = value->u.reg.data_type;
            value->u.reg.idx_count = 2;
            value->u.reg.immconst_type = sm6_type_is_scalar(fwd_type) ? VKD3D_IMMCONST_SCALAR
                    : VKD3D_IMMCONST_VEC4;
        }
    }

    return operand;
}

static const struct sm6_value *sm6_parser_get_value_by_ref(struct sm6_parser *sm6,
        const struct dxil_record *record, const struct sm6_type *type, unsigned int *rec_idx)
{
    unsigned int operand = sm6_parser_get_value_idx_by_ref(sm6, record, type, rec_idx);
    return operand == UINT_MAX ? NULL : &sm6->values[operand];
}

static bool sm6_parser_declare_function(struct sm6_parser *sm6, const struct dxil_record *record)
{
    const struct sm6_type *ret_type;
    struct sm6_value *fn;
    unsigned int i, j;

    if (record->operand_count < 5)
    {
        WARN("Missing operands for function %u.\n", sm6->value_count);
        return false;
    }

    fn = sm6_parser_get_current_value(sm6);
    fn->value_type = VALUE_TYPE_FUNCTION;
    if (!(fn->u.function.name = sm6_parser_get_global_symbol_name(sm6, sm6->value_count)))
    {
        WARN("Missing symbol name for function %u.\n", sm6->value_count);
        fn->u.function.name = "";
    }

    if (!(fn->type = sm6_parser_get_type(sm6, record->operands[0])))
        return false;
    if (!sm6_type_is_function(fn->type))
    {
        WARN("Type is not a function.\n");
        return false;
    }
    ret_type = fn->type->u.function->ret_type;

    if (!(fn->type = sm6_type_get_pointer_to_type(fn->type, ADDRESS_SPACE_DEFAULT, sm6)))
    {
        WARN("Failed to get pointer type for type %u.\n", fn->type->class);
        return false;
    }

    if (record->operands[1])
        WARN("Ignoring calling convention %#"PRIx64".\n", record->operands[1]);

    fn->u.function.is_prototype = !!record->operands[2];

    if (record->operands[3])
        WARN("Ignoring linkage %#"PRIx64".\n", record->operands[3]);

    if (record->operands[4] > UINT_MAX)
        WARN("Invalid attributes id %#"PRIx64".\n", record->operands[4]);
    /* 1-based index. */
    if ((fn->u.function.attribs_id = record->operands[4]))
        TRACE("Ignoring function attributes.\n");

    for (i = 5, j = 0; i < record->operand_count; ++i)
        j += !!record->operands[i];
    if (j)
        WARN("Ignoring %u operands.\n", j);

    if (sm6_value_is_dx_intrinsic_dcl(fn) && !sm6_type_is_void(ret_type) && !sm6_type_is_numeric(ret_type)
            && !sm6_type_is_numeric_aggregate(ret_type) && !sm6_type_is_handle(ret_type))
    {
        WARN("Unexpected return type for dx intrinsic function '%s'.\n", fn->u.function.name);
    }

    ++sm6->value_count;

    return true;
}

static inline uint64_t decode_rotated_signed_value(uint64_t value)
{
    if (value != 1)
    {
        bool neg = value & 1;
        value >>= 1;
        return neg ? -value : value;
    }
    return value << 63;
}

static inline float bitcast_uint64_to_float(uint64_t value)
{
    union
    {
        uint32_t uint32_value;
        float float_value;
    } u;

    u.uint32_value = value;
    return u.float_value;
}

static inline double bitcast_uint64_to_double(uint64_t value)
{
    union
    {
        uint64_t uint64_value;
        double double_value;
    } u;

    u.uint64_value = value;
    return u.double_value;
}

static enum vkd3d_result register_allocate_constant_array(struct vkd3d_shader_register *reg, const struct sm6_type *type,
        unsigned int value_idx, unsigned int *elem_size, struct sm6_parser *sm6)
{
    const struct sm6_type *elem_type;
    unsigned int size;

    elem_type = type->u.array.elem_type;
    /* Multidimensional arrays are emitted in flattened form. */
    if (elem_type->class != TYPE_CLASS_INTEGER && elem_type->class != TYPE_CLASS_FLOAT)
    {
        FIXME("Unhandled element type %u for data array constant id %u.\n", elem_type->class, value_idx);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    /* Arrays of bool are not used in DXIL. dxc will emit an array of int32 instead if necessary. */
    if (!(size = elem_type->u.width / 8u))
    {
        WARN("Invalid data type width %u.\n", elem_type->u.width);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    if (!(reg->u.immconst_aggregate.u.pvoid = vkd3d_calloc(type->u.array.count, size)))
    {
        ERR("Failed to allocate constant array.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }
    shader_instruction_array_add_opaque_param(&sm6->p.instructions, reg->u.immconst_aggregate.u.pvoid);
    reg->u.immconst_aggregate.count = type->u.array.count;

    *elem_size = size;
    return VKD3D_OK;
}

static enum vkd3d_result sm6_parser_constants_init(struct sm6_parser *sm6, const struct dxil_block *block)
{
    const struct sm6_type *type, *elem_type, *pointee_type, *gep_type, *ptr_type;
    enum vkd3d_shader_register_type reg_type = VKD3DSPR_INVALID;
    struct vkd3d_shader_aggregate_data data;
    unsigned int i, j, k, value_idx, size;
    const struct sm6_value *operands[3];
    enum vkd3d_data_type reg_data_type;
    const struct dxil_record *record;
    struct sm6_value *elem, *dst;
    uint64_t type_id, value;
    enum vkd3d_result ret;

    for (i = 0, type = NULL; i < block->record_count; ++i)
    {
        record = block->records[i];
        value_idx = sm6->value_count;

        if (record->code == CST_CODE_SETTYPE)
        {
            if (!record->operand_count)
            {
                WARN("Missing type id for constant idx %u.\n", value_idx);
                return VKD3D_ERROR_INVALID_SHADER;
            }
            if ((type_id = record->operands[0]) >= sm6->type_count)
            {
                WARN("Invalid type id %"PRIu64".\n", type_id);
                return VKD3D_ERROR_INVALID_SHADER;
            }

            if (!(type = sm6_parser_get_type(sm6, type_id)))
                return VKD3D_ERROR_INVALID_SHADER;
            elem_type = sm6_type_get_element_type(type);
            if (sm6_type_is_numeric(elem_type))
            {
                reg_data_type = vkd3d_data_type_from_sm6_type(elem_type);
                reg_type = elem_type->u.width > 32 ? VKD3DSPR_IMMCONST64 : VKD3DSPR_IMMCONST;
            }
            else
            {
                reg_data_type = VKD3D_DATA_UNUSED;
                reg_type = VKD3DSPR_IMMCONST;
            }

            if (i == block->record_count - 1)
                WARN("Unused SETTYPE record.\n");

            continue;
        }

        if (!type)
        {
            WARN("Constant record %u has no type.\n", value_idx);
            return VKD3D_ERROR_INVALID_SHADER;
        }

        dst = sm6_parser_get_current_value(sm6);
        dst->type = type;
        dst->value_type = VALUE_TYPE_REG;
        dst->u.reg.type = reg_type;
        dst->u.reg.immconst_type = VKD3D_IMMCONST_SCALAR;
        dst->u.reg.data_type = reg_data_type;

        switch (record->code)
        {
            case CST_CODE_NULL:
                /* Register constant data is already zero-filled. */
                if (sm6_type_is_array(type))
                {
                    dst->u.reg.type = VKD3DSPR_IMMCONSTARRAY;
                    dst->u.reg.u.immconst_aggregate.count = type->u.array.count;
                }
                break;

            case CST_CODE_INTEGER:
                if (!record->operand_count)
                {
                    WARN("Missing integer value for constant idx %u.\n", value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (!sm6_type_is_integer(type))
                {
                    WARN("Invalid integer of non-integer type %u at constant idx %u.\n", type->class, value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                value = decode_rotated_signed_value(record->operands[0]);
                if (type->u.width <= 32)
                    dst->u.reg.u.immconst_uint[0] = value & ((1ull << type->u.width) - 1);
                else
                    dst->u.reg.u.immconst_uint64[0] = value;

                break;

            case CST_CODE_FLOAT:
                if (!record->operand_count)
                {
                    WARN("Missing floating point value for constant idx %u.\n", value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (!sm6_type_is_floating_point(type))
                {
                    WARN("Invalid float of non-fp type %u at constant idx %u.\n", type->class, value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (type->u.width == 16)
                    dst->u.reg.u.immconst_uint[0] = record->operands[0];
                else if (type->u.width == 32)
                    dst->u.reg.u.immconst_float[0] = bitcast_uint64_to_float(record->operands[0]);
                else if (type->u.width == 64)
                    dst->u.reg.u.immconst_double[0] = bitcast_uint64_to_double(record->operands[0]);
                else
                    vkd3d_unreachable();

                break;

            case CST_CODE_DATA:
                if (record->operand_count < type->u.array.count)
                {
                    WARN("Missing operands for array of size %u.\n", type->u.array.count);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (!sm6_type_is_array(type))
                {
                    WARN("Invalid type %u for data constant idx %u.\n", type->class, value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if ((ret = register_allocate_constant_array(&dst->u.reg, type, value_idx, &size, sm6)) < 0)
                    return ret;

                dst->u.reg.type = VKD3DSPR_IMMCONSTARRAY;

                data = dst->u.reg.u.immconst_aggregate;
                if (size == 1)
                    for (j = 0; j < record->operand_count; ++j)
                        data.u.p8[j] = record->operands[j];
                else if (size == 2)
                    for (j = 0; j < record->operand_count; ++j)
                        data.u.p16[j] = record->operands[j];
                else if (size == 4)
                    for (j = 0; j < record->operand_count; ++j)
                        data.u.p32[j] = record->operands[j];
                else
                    for (j = 0; j < record->operand_count; ++j)
                        data.u.p64[j] = record->operands[j];
                break;

            case CST_CODE_CE_GEP:
            case CST_CODE_CE_INBOUNDS_GEP:
                j = 0;
                pointee_type = NULL;
                if (record->operand_count & 1)
                    pointee_type = sm6_parser_get_type(sm6, record->operands[j++]);
                if (record->operand_count < j + 2)
                {
                    WARN("Missing operands for constexpr gep at idx %u.\n", value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if ((k = record->operand_count / 2u) > ARRAY_SIZE(operands))
                {
                    WARN("Insufficient space for %u operands.\n", k);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                for (k = 0; j < record->operand_count; j += 2)
                {
                    if (!(elem_type = sm6_parser_get_type(sm6, record->operands[j])))
                        return VKD3D_ERROR_INVALID_SHADER;
                    if ((value = record->operands[j + 1]) >= sm6->cur_max_value)
                    {
                        WARN("Invalid value index %"PRIu64".\n", value);
                        return VKD3D_ERROR_INVALID_SHADER;
                    }

                    elem = &sm6->values[value];
                    if (value > sm6->value_count)
                        elem->type = elem_type;
                    else if (value == sm6->value_count)
                        WARN("Value for constexpr constant idx %u is self-referencing.\n", value_idx);
                    else if (elem->type != elem_type)
                        WARN("Type mismatch.\n");

                    if (!k)
                    {
                        gep_type = elem_type;
                        dst->structure_stride = elem->structure_stride;
                    }
                    else if (!sm6_value_is_constant(elem) || !(gep_type = sm6_type_get_element_type_at_index(
                            gep_type, sm6_value_get_constant_uint(elem))))
                    {
                        WARN("Failed to get element type.\n");
                        break;
                    }

                    operands[k++] = elem;
                }

                ptr_type = operands[0]->type;
                if (!sm6_type_is_pointer(ptr_type))
                {
                    WARN("Constexpr GEP base value at constant idx %u is not a pointer.\n", value_idx);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (pointee_type && pointee_type != (elem_type = ptr_type->u.pointer.type))
                    WARN("Type mismatch.\n");

                if (!(dst->type = sm6_type_get_pointer_to_type(gep_type, ptr_type->u.pointer.addr_space, sm6)))
                {
                    WARN("Failed to get pointer type for type %u.\n", gep_type->class);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                dst->u.reg = operands[0]->u.reg;
                dst->u.reg.idx[1].offset = sm6_value_get_constant_uint(operands[2]);
                dst->u.reg.idx_count = 2;
                break;

            case CST_CODE_UNDEF:
                if (record->operand_count)
                    WARN("Unexpected value for undefined constant at idx %u.\n", value_idx);
                dst->u.reg.type = VKD3DSPR_UNDEF;
                /* Mark as explicitly undefined. */
                dst->is_undefined = true;
                break;

            default:
                FIXME("Unhandled constant code %u.\n", record->code);
                dst->u.reg.type = VKD3DSPR_UNDEF;
                break;
        }

        ++sm6->value_count;
    }

    return VKD3D_OK;
}

static bool bitcode_parse_alignment(uint64_t encoded_alignment, unsigned int *alignment)
{
    if (encoded_alignment > DXIL_MAX_ALIGNMENT_EXPONENT + 1)
    {
        *alignment = 0;
        return false;
    }
    *alignment = (1u << encoded_alignment) >> 1;
    return true;
}

static struct vkd3d_shader_instruction *sm6_parser_require_space(struct sm6_parser *sm6, unsigned int extra)
{
    if (!shader_instruction_array_reserve(&sm6->p.instructions, sm6->p.instructions.count + extra))
    {
        ERR("Failed to allocate instruction.\n");
        return NULL;
    }
    return &sm6->p.instructions.elements[sm6->p.instructions.count];
}

/* Space must be reserved before calling this. */
static struct vkd3d_shader_instruction *sm6_parser_add_instruction(struct sm6_parser *sm6,
        enum vkd3d_shader_opcode handler_idx)
{
    struct vkd3d_shader_instruction *ins = sm6_parser_require_space(sm6, 1);
    assert(ins);
    shader_instruction_init(ins, handler_idx);
    ++sm6->p.instructions.count;
    return ins;
}

static void sm6_parser_declare_indexable_temp(struct sm6_parser *sm6, const struct sm6_type *elem_type,
        unsigned int count, struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    enum vkd3d_data_type data_type = vkd3d_data_type_from_sm6_type(elem_type);

    shader_instruction_init(ins, VKD3DSIH_DCL_INDEXABLE_TEMP);
    ins->declaration.indexable_temp.register_idx = sm6->indexable_temp_count++;
    ins->declaration.indexable_temp.register_size = count;
    ins->declaration.indexable_temp.data_type = data_type;
    ins->declaration.indexable_temp.component_count = 1;
    ins->declaration.indexable_temp.initialiser.type = VKD3DSPR_INVALID;

    register_init_with_id(&dst->u.reg, VKD3DSPR_IDXTEMP, data_type, ins->declaration.indexable_temp.register_idx);
}

static bool sm6_type_alignment_is_default(const struct sm6_type *type, unsigned int alignment)
{
    if (!alignment)
        return true;

    type = sm6_type_get_scalar_type(type, 0);
    if (!sm6_type_is_integer(type) && !sm6_type_is_floating_point(type))
        return false;
    /* TODO: for native 16-bit return alignment == type->u.width / 8 */
    return alignment == max(type->u.width / 8, 4);
}

static bool sm6_parser_declare_global(struct sm6_parser *sm6, const struct dxil_record *record)
{
    const struct sm6_type *type, *scalar_type;
    unsigned int alignment, count, stride = 0;
    struct vkd3d_shader_register *initialiser;
    struct vkd3d_shader_instruction *ins;
    bool is_external, is_constant;
    uint64_t address_space;
    struct sm6_value *dst;
    unsigned int init = 0;
    const char *name;

    if (record->operand_count < 6)
    {
        WARN("Invalid operand count %u.\n", record->operand_count);
        return false;
    }

    if (!(type = sm6_parser_get_type(sm6, record->operands[0])))
        return false;
    if (sm6_type_is_array(type))
    {
        count = type->u.array.count;
        scalar_type = type->u.array.elem_type;
        if (sm6_type_is_numeric(scalar_type))
            stride = scalar_type->u.width / 8u;
    }
    else
    {
        count = 1;
        scalar_type = type;
        if (sm6_type_is_numeric(type))
            stride = type->u.width / 8u;
    }

    if (record->operands[1] & 2)
    {
        enum sm6_type_class class = type->class;

        if ((address_space = record->operands[1] >> 2) > ADDRESS_SPACE_GROUPSHARED)
            WARN("Unknown address space %"PRIu64".\n", address_space);

        /* explicit type */
        if (!(type = sm6_type_get_pointer_to_type(type, address_space, sm6)))
        {
            WARN("Failed to get pointer type for type %u.\n", class);
            return false;
        }
    }
    else
    {
        if (!sm6_type_is_pointer(type))
        {
            WARN("Type is not a pointer.\n");
            return false;
        }
        address_space = type->u.pointer.addr_space;
    }

    if (record->operands[2])
    {
        init = record->operands[2];
        if (init - 1 >= sm6->value_capacity)
        {
            WARN("Invalid value index for initialiser.");
        }
    }

    is_external = record->operands[3] == LINKAGE_EXTERNAL;
    if (!is_external && record->operands[3] != LINKAGE_INTERNAL)
        WARN("Ignoring linkage %"PRIu64".\n", record->operands[3]);

    if ((is_constant = record->operands[1] & 1))
        WARN("Ignoring constant attribute.\n");

    if (!bitcode_parse_alignment(record->operands[4], &alignment))
        WARN("Invalid alignment %"PRIu64".\n", record->operands[4]);
    if (!sm6_type_alignment_is_default(type, alignment))
        WARN("Ignoring alignment %u.\n", alignment);

    if (record->operands[5])
        WARN("Ignoring section code %"PRIu64".\n", record->operands[5]);

    if (!(name = sm6_parser_get_global_symbol_name(sm6, sm6->value_count)))
        WARN("Missing symbol name for global var %u.\n", sm6->value_count);

    if (record->operand_count > 6 && record->operands[6])
        WARN("Ignoring visibility %"PRIu64".\n", record->operands[6]);
    if (record->operand_count > 7 && record->operands[7])
        WARN("Ignoring thread local mode %"PRIu64".\n", record->operands[7]);
    /* record->operands[8] contains unnamed_addr, a flag indicating the address is not important,
     * only the content is. We don't need to use this info for SPIR-V generation. */
    if (record->operand_count > 9 && record->operands[9])
        WARN("Ignoring external_init %"PRIu64".\n", record->operands[9]);
    if (record->operand_count > 10 && record->operands[10])
        WARN("Ignoring dll storage class %"PRIu64".\n", record->operands[10]);
    if (record->operand_count > 11 && record->operands[11])
        WARN("Ignoring comdat %"PRIu64".\n", record->operands[11]);

    dst = sm6_parser_get_current_value(sm6);
    dst->type = type;
    dst->value_type = VALUE_TYPE_REG;

    if (address_space != ADDRESS_SPACE_GROUPSHARED)
    {
        ins = sm6_parser_add_instruction(sm6, 0);
        sm6_parser_declare_indexable_temp(sm6, scalar_type, count, ins, dst);
        dst->structure_stride = 0;
        initialiser = &ins->declaration.indexable_temp.initialiser;
    }
    else if (count == 1)
    {
        if (is_constant)
            WARN("TGSM declared as constant.\n");
        ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TGSM_RAW);
        dst_param_init(&ins->declaration.tgsm_raw.reg);
        register_init_with_id(&ins->declaration.tgsm_raw.reg.reg, VKD3DSPR_GROUPSHAREDMEM,
                vkd3d_data_type_from_sm6_type(scalar_type), sm6->value_count);
        dst->u.reg = ins->declaration.tgsm_raw.reg.reg;
        dst->structure_stride = 0;
        ins->declaration.tgsm_raw.byte_count = stride;
        initialiser = &ins->declaration.tgsm_raw.initialiser;
    }
    else
    {
        if (!stride)
        {
            WARN("Failed to get byte stride for tgsm declaration.\n");
            stride = 4;
        }
        ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TGSM_STRUCTURED);
        dst_param_init(&ins->declaration.tgsm_structured.reg);
        register_init_with_id(&ins->declaration.tgsm_structured.reg.reg, VKD3DSPR_GROUPSHAREDMEM,
                vkd3d_data_type_from_sm6_type(scalar_type), sm6->value_count);
        dst->u.reg = ins->declaration.tgsm_structured.reg.reg;
        dst->structure_stride = stride;
        ins->declaration.tgsm_structured.byte_stride = stride;
        ins->declaration.tgsm_structured.structure_count = count;
        initialiser = &ins->declaration.tgsm_structured.initialiser;
    }

    if (init)
    {
        /* Forward references will be resolved later. */
        initialiser->type = VKD3DSPR_IMMCONSTARRAY;
        initialiser->idx[0].offset = init - 1;
        initialiser->idx_count = 1;
    }
    else
    {
        if (is_constant)
            WARN("Global declared as constant has no initialiser.\n");
        initialiser->type = VKD3DSPR_INVALID;
    }

    ++sm6->value_count;

    return true;
}

static void register_resolve_forward_initialiser(struct vkd3d_shader_register *initialiser,
        struct sm6_parser *sm6)
{
    const struct sm6_value *value;

    if (initialiser->type == VKD3DSPR_IMMCONSTARRAY)
    {
        unsigned int index = initialiser->idx[0].offset;
        if (index >= sm6->value_count || !sm6_value_is_register(value = sm6_parser_get_value(sm6, index)))
        {
            WARN("Ignoring invalid initialiser index %u.\n", index);
            initialiser->type = VKD3DSPR_INVALID;
        }
        else if (value->u.reg.type != VKD3DSPR_UNDEF)
        {
            *initialiser = value->u.reg;
        }
        else
        {
            /* Undefined initialisers are a no-op and are invalid in SPIR-V. */
            initialiser->type = VKD3DSPR_INVALID;
        }
    }
}

static enum vkd3d_result sm6_parser_globals_init(struct sm6_parser *sm6)
{
    const struct dxil_block *block = &sm6->root_block;
    struct vkd3d_shader_instruction *ins;
    const struct dxil_record *record;
    unsigned int i, count;
    enum vkd3d_result ret;
    uint64_t version;

    for (i = 0, count = 0; i < block->record_count; ++i)
        count += block->records[i]->code == MODULE_CODE_GLOBALVAR;
    sm6_parser_require_space(sm6, count);

    for (i = 0; i < block->record_count; ++i)
    {
        record = block->records[i];
        switch (record->code)
        {
            case MODULE_CODE_FUNCTION:
                if (!sm6_parser_declare_function(sm6, record))
                {
                    vkd3d_shader_parser_error(&sm6->p, VKD3D_SHADER_ERROR_DXIL_INVALID_FUNCTION_DCL,
                            "A DXIL function declaration is invalid.");
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                break;
            case MODULE_CODE_GLOBALVAR:
                if (!sm6_parser_declare_global(sm6, record))
                    return VKD3D_ERROR_INVALID_SHADER;
                break;
            case MODULE_CODE_VERSION:
                if (!record->operand_count)
                {
                    WARN("Missing version operand.\n");
                }
                else if ((version = record->operands[0]) != 1)
                {
                    FIXME("Unsupported format version %#"PRIx64".\n", version);
                    vkd3d_shader_parser_error(&sm6->p, VKD3D_SHADER_ERROR_DXIL_UNSUPPORTED_BITCODE_FORMAT,
                            "Bitcode format version %#"PRIx64" is unsupported.", version);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
            default:
                break;
        }
    }

    for (i = 0; i < block->child_block_count; ++i)
    {
        if (block->child_blocks[i]->id == CONSTANTS_BLOCK && (ret = sm6_parser_constants_init(sm6,
                block->child_blocks[i])) < 0)
            return ret;
    }

    /* Resolve initialiser forward references. */
    for (i = 0; i < sm6->p.instructions.count; ++i)
    {
        ins = &sm6->p.instructions.elements[i];
        if (ins->handler_idx == VKD3DSIH_DCL_INDEXABLE_TEMP)
            register_resolve_forward_initialiser(&ins->declaration.indexable_temp.initialiser, sm6);
        else if (ins->handler_idx == VKD3DSIH_DCL_TGSM_RAW)
            register_resolve_forward_initialiser(&ins->declaration.tgsm_raw.initialiser, sm6);
        else if (ins->handler_idx == VKD3DSIH_DCL_TGSM_STRUCTURED)
            register_resolve_forward_initialiser(&ins->declaration.tgsm_structured.initialiser, sm6);
    }

    return VKD3D_OK;
}

static inline void dst_param_io_init(struct vkd3d_shader_dst_param *param,
        const struct signature_element *e, enum vkd3d_shader_register_type reg_type)
{
    enum vkd3d_shader_component_type component_type;

    param->write_mask = e->mask;
    param->modifiers = 0;
    param->shift = 0;
    component_type = e->component_type == VKD3D_SHADER_COMPONENT_INT ? VKD3D_SHADER_COMPONENT_UINT : e->component_type;
    shader_register_init(&param->reg, reg_type, vkd3d_data_type_from_component_type(component_type), 0);
}

static void sm6_parser_init_signature(struct sm6_parser *sm6, const struct shader_signature *s,
        bool is_input, enum vkd3d_shader_register_type reg_type, struct vkd3d_shader_dst_param *params)
{
    enum vkd3d_shader_type shader_type = sm6->p.shader_version.type;
    bool is_patch_constant, is_control_point;
    struct vkd3d_shader_dst_param *param;
    const struct signature_element *e;
    unsigned int i, j;

    is_patch_constant = reg_type == VKD3DSPR_PATCHCONST;
    is_control_point = !is_patch_constant && ((is_input && (shader_type == VKD3D_SHADER_TYPE_HULL
            || shader_type == VKD3D_SHADER_TYPE_DOMAIN || shader_type == VKD3D_SHADER_TYPE_GEOMETRY))
            || (!is_input && !is_patch_constant && shader_type == VKD3D_SHADER_TYPE_HULL));

    for (i = 0; i < s->element_count; ++i)
    {
        e = &s->elements[i];

        if (e->register_index >= MAX_REG_OUTPUT && e->register_index != UINT_MAX)
            reg_type = e->register_index - MAX_REG_OUTPUT;

        param = &params[i];
        dst_param_io_init(param, e, reg_type);
        j = 0;
        if (e->register_count > 1 || (is_patch_constant && sysval_semantic_is_tess_factor(e->sysval_semantic)))
            param->reg.idx[j++].offset = 0;
        if (is_control_point)
        {
            if (reg_type == VKD3DSPR_OUTPUT)
                param->reg.idx[j].rel_addr = instruction_array_create_outpointid_param(&sm6->p.instructions);
            param->reg.idx[j++].offset = 0;
        }
        assert(j < ARRAY_SIZE(param->reg.idx));
        param->reg.idx[j].offset = i;
        param->reg.idx_count = j + 1;
    }
}

static void sm6_parser_emit_signature(struct sm6_parser *sm6, const struct shader_signature *s,
        enum vkd3d_shader_opcode handler_idx, enum vkd3d_shader_opcode siv_handler_idx,
        struct vkd3d_shader_dst_param *params)
{
    enum vkd3d_shader_type shader_type = sm6->p.shader_version.type;
    bool is_patch_constant, is_input, strip_sysval;
    unsigned int i, j, control_point_count = 0;
    enum vkd3d_shader_register_type reg_type;
    struct vkd3d_shader_instruction *ins;
    struct vkd3d_shader_dst_param *param;
    const struct signature_element *e;

    if (!s->element_count)
        return;

    reg_type = params[0].reg.type;
    is_patch_constant = reg_type == VKD3DSPR_PATCHCONST;
    is_input = handler_idx == VKD3DSIH_DCL_INPUT;
    strip_sysval = !is_patch_constant && ((!is_input && shader_type == VKD3D_SHADER_TYPE_HULL)
            || reg_type == VKD3DSPR_INCONTROLPOINT);

    if (is_input && !is_patch_constant && (shader_type == VKD3D_SHADER_TYPE_HULL
            || shader_type == VKD3D_SHADER_TYPE_DOMAIN || shader_type == VKD3D_SHADER_TYPE_GEOMETRY))
    {
        control_point_count = sm6->input_control_point_count;
    }
    else if (reg_type == VKD3DSPR_OUTPUT && shader_type == VKD3D_SHADER_TYPE_HULL)
    {
        control_point_count = sm6->output_control_point_count;
        assert(control_point_count);
    }

    for (i = 0; i < s->element_count; ++i)
    {
        e = &s->elements[i];

        /* Do not check e->used_mask because in some cases it is zero for used elements.
         * TODO: scan ahead for used I/O elements. */

        if (e->sysval_semantic != VKD3D_SHADER_SV_NONE && e->sysval_semantic != VKD3D_SHADER_SV_TARGET && !strip_sysval)
        {
            if (e->register_index == UINT_MAX)
                siv_handler_idx = (siv_handler_idx == VKD3DSIH_DCL_INPUT_PS_SIV) ? VKD3DSIH_DCL_INPUT_PS_SGV
                        : VKD3DSIH_DCL_INPUT_SGV;
            ins = sm6_parser_add_instruction(sm6, siv_handler_idx);
            param = &ins->declaration.register_semantic.reg;
            ins->declaration.register_semantic.sysval_semantic = vkd3d_siv_from_sysval(e->sysval_semantic);
        }
        else
        {
            ins = sm6_parser_add_instruction(sm6, handler_idx);
            param = &ins->declaration.dst;
        }

        ins->flags = e->interpolation_mode;
        *param = params[i];

        j = 0;
        if (e->register_count > 1 || (is_patch_constant && sysval_semantic_is_tess_factor(e->sysval_semantic)))
            param->reg.idx[j++].offset = e->register_count;
        if (control_point_count)
        {
            param->reg.idx[j].rel_addr = NULL;
            param->reg.idx[j++].offset = control_point_count;
        }
    }
}

static void sm6_parser_init_output_signature(struct sm6_parser *sm6, const struct shader_signature *output_signature)
{
    sm6_parser_init_signature(sm6, output_signature, false,
            (sm6->p.shader_version.type == VKD3D_SHADER_TYPE_PIXEL) ? VKD3DSPR_COLOROUT : VKD3DSPR_OUTPUT,
            sm6->output_params);
}

static void sm6_parser_init_input_signature(struct sm6_parser *sm6, const struct shader_signature *input_signature)
{
    sm6_parser_init_signature(sm6, input_signature, true,
            sm6->p.shader_version.type == VKD3D_SHADER_TYPE_DOMAIN ? VKD3DSPR_INCONTROLPOINT : VKD3DSPR_INPUT,
            sm6->input_params);
}

static void sm6_parser_init_patch_constant_signature(struct sm6_parser *sm6,
        const struct shader_signature *patch_constant_signature)
{
    bool is_input = sm6->p.shader_version.type == VKD3D_SHADER_TYPE_DOMAIN;

    sm6_parser_init_signature(sm6, patch_constant_signature, is_input, VKD3DSPR_PATCHCONST,
            sm6->patch_constant_params);
}

static void sm6_parser_emit_output_signature(struct sm6_parser *sm6, const struct shader_signature *output_signature)
{
    sm6_parser_emit_signature(sm6, output_signature, VKD3DSIH_DCL_OUTPUT, VKD3DSIH_DCL_OUTPUT_SIV, sm6->output_params);
}

static void sm6_parser_emit_input_signature(struct sm6_parser *sm6, const struct shader_signature *input_signature)
{
    sm6_parser_emit_signature(sm6, input_signature,
            (sm6->p.shader_version.type == VKD3D_SHADER_TYPE_PIXEL) ? VKD3DSIH_DCL_INPUT_PS : VKD3DSIH_DCL_INPUT,
            (sm6->p.shader_version.type == VKD3D_SHADER_TYPE_PIXEL) ? VKD3DSIH_DCL_INPUT_PS_SIV : VKD3DSIH_DCL_INPUT_SIV,
            sm6->input_params);
}

static void sm6_parser_emit_patch_constant_signature(struct sm6_parser *sm6,
        const struct shader_signature *patch_constant_signature)
{
    bool is_input = sm6->p.shader_version.type == VKD3D_SHADER_TYPE_DOMAIN;

    sm6_parser_emit_signature(sm6, patch_constant_signature, is_input ? VKD3DSIH_DCL_INPUT : VKD3DSIH_DCL_OUTPUT,
            is_input ? VKD3DSIH_DCL_INPUT_SIV : VKD3DSIH_DCL_OUTPUT_SIV, sm6->patch_constant_params);
}

static struct vkd3d_shader_register *sm6_parser_emit_input_register(struct sm6_parser *sm6,
        enum vkd3d_shader_register_type reg_type, enum vkd3d_data_type data_type)
{
    struct vkd3d_shader_instruction *ins;
    struct vkd3d_shader_dst_param *param;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_INPUT);

    param = &ins->declaration.dst;
    dst_param_init(param);
    shader_register_init(&param->reg, reg_type, data_type, 0);

    return &param->reg;
}

static unsigned int sm6_signature_find_element_by_sysval(const struct shader_signature *signature,
        enum vkd3d_shader_sysval_semantic sysval_semantic)
{
    unsigned int i;

    for (i = 0; i < signature->element_count; ++i)
        if (signature->elements[i].sysval_semantic == sysval_semantic)
            return i;
    return UINT_MAX;
}

static inline void sm6_src_operands_init(struct vkd3d_shader_src_param *src_params,
        const struct sm6_value **operands, unsigned int count)
{
    unsigned int i;

    for (i = 0; i < count; ++i)
        src_param_init_from_value(&src_params[i], operands[i]);
}

static const struct sm6_value *sm6_parser_next_function_definition(struct sm6_parser *sm6)
{
    unsigned int i, count = sm6->function_count;

    for (i = 0; i < sm6->value_count; ++i)
    {
        if (sm6_type_is_function_pointer(sm6->values[i].type) && !sm6->values[i].u.function.is_prototype && !count--)
            break;
    }
    if (i == sm6->value_count)
        return NULL;

    ++sm6->function_count;
    return &sm6->values[i];
}

static struct sm6_block *sm6_block_create()
{
    struct sm6_block *block = vkd3d_calloc(1, sizeof(*block));
    return block;
}

bool sm6_block_set_name(struct sm6_block *block, const char *fmt, ...)
{
    va_list args;
    int len;

    va_start(args, fmt);
    len = vsnprintf(NULL, 0, fmt, args);
    va_end(args);
    if (!(block->name = vkd3d_realloc(block->name, len + 1)))
        return false;
    va_start(args, fmt);
    vsprintf(block->name, fmt, args);
    va_end(args);

    return true;
}

void sm6_block_array_init(struct sm6_block_array *array)
{
    memset(array, 0, sizeof(*array));
}

void sm6_block_array_destroy(struct sm6_block_array *array)
{
   vkd3d_free(array->blocks);
}

bool sm6_block_array_append(struct sm6_block_array *array, struct sm6_block *block)
{
    if (!sm6_block_array_reserve(array, array->count + 1))
        return false;
    array->blocks[array->count++] = block;
    return true;
};

bool sm6_block_array_append_unique(struct sm6_block_array *array, struct sm6_block *block)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (array->blocks[i] == block)
            return true;
    return sm6_block_array_append(array, block);
}

unsigned int sm6_block_array_find(const struct sm6_block_array *array,
        const struct sm6_block *block)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (array->blocks[i] == block)
            break;
    return i;
};

struct sm6_phi *sm6_block_phi_new(struct sm6_block *block, const struct vkd3d_shader_register *reg)
{
    struct sm6_phi *phi;

    if (!vkd3d_array_reserve((void **)&block->phi, &block->phi_capacity, block->phi_count + 1, sizeof(*block->phi)))
    {
        ERR("Failed to allocate phi array.\n");
        return NULL;
    }
    phi = &block->phi[block->phi_count++];

    phi->name = NULL;
    if (reg)
        phi->reg = *reg;
    phi->incoming = NULL;
    phi->incoming_capacity = 0;
    phi->incoming_count = 0;

    return phi;
}

struct incoming_value *sm6_phi_incoming_new(struct sm6_phi *phi)
{
    if (!vkd3d_array_reserve((void **)&phi->incoming, &phi->incoming_capacity,
            phi->incoming_count + 1, sizeof(*phi->incoming)))
    {
        ERR("Failed to allocate phi incoming value array.\n");
        abort();
    }
    return &phi->incoming[phi->incoming_count++];
}

static const struct vkd3d_shader_register *sm6_parser_emit_composite_construct(struct sm6_parser *sm6,
        const struct sm6_value **operands, unsigned int component_count, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *reg;

    if (component_count == 1)
        return &operands[0]->u.reg;

    shader_instruction_init(ins, VKD3DSIH_COMPOSITE_CONSTRUCT);

    src_params = instruction_src_params_alloc(ins, component_count, sm6);
    sm6_src_operands_init(src_params, operands, component_count);

    reg = instruction_dst_param_vector_init_helper_ssa(sm6, ins, vkd3d_data_type_from_sm6_type(operands[0]->type),
            component_count);

    return reg;
}

static const struct vkd3d_shader_register *sm6_parser_emit_reg_composite_construct(struct sm6_parser *sm6,
        const struct vkd3d_shader_register *operands, unsigned int component_count,
        struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *reg;
    unsigned int i;

    shader_instruction_init(ins, VKD3DSIH_COMPOSITE_CONSTRUCT);

    src_params = instruction_src_params_alloc(ins, component_count, sm6);

    for (i = 0; i < component_count; ++i)
    {
        src_param_init_scalar(&src_params[i], 0);
        src_params[i].reg = operands[i];
    }

    reg = instruction_dst_param_vector_init_helper_ssa(sm6, ins, operands[0].data_type, component_count);

    return reg;
}

static void sm6_parser_emit_alloca(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    const struct sm6_type *type[2];
    const struct sm6_value *src;
    uint64_t alignment;
    unsigned int i;

    DXIL_RECORD_VALIDATE_OPERAND_COUNT(4, 4);

    for (i = 0; i < 2; ++i)
    {
        if (!(type[i] = sm6_parser_get_type(sm6, record->operands[i])))
            return;
    }

    if (!(dst->type = sm6_type_get_pointer_to_type(type[0], ADDRESS_SPACE_DEFAULT, sm6)))
    {
        FIXME("Failed to get pointer type for type %u.\n", type[0]->class);
        return;
    }

    if (record->operands[2] >= sm6->value_count)
    {
        FIXME("Unhandled forward ref %"PRIu64".\n", record->operands[2]);
        return;
    }

    src = sm6_parser_get_value(sm6, record->operands[2]);
    if (sm6_value_get_constant_uint(src) != 1)
    {
        FIXME("Array dimension is not 1.\n");
        return;
    }

    alignment = record->operands[3];
    if (alignment & (1u << 5))
        WARN("Ignoring in_alloca flag.\n");
    if (!(alignment & (1u << 6)))
        WARN("Unhandled implicit type.\n");

    alignment &= ~(3u << 5);
    if (!bitcode_parse_alignment(alignment, &i))
        WARN("Invalid alignment %"PRIu64".\n", alignment);
    if (!sm6_type_alignment_is_default(type[0], i))
        WARN("Ignoring alignment %u.\n", i);

    if (!sm6_type_is_array(type[0]) || !sm6_type_is_numeric(type[0]->u.array.elem_type))
    {
        FIXME("Type is not an array.\n");
        return;
    }

    sm6_parser_declare_indexable_temp(sm6, type[0]->u.array.elem_type, type[0]->u.array.count, ins, dst);
}

static enum vkd3d_shader_opcode sm6_map_atomicrmw_op(uint64_t code)
{
    switch (code)
    {
        case RMW_ADD:
            return VKD3DSIH_IMM_ATOMIC_IADD;
        case RMW_AND:
            return VKD3DSIH_IMM_ATOMIC_AND;
        case RMW_MAX:
            return VKD3DSIH_IMM_ATOMIC_IMAX;
        case RMW_MIN:
            return VKD3DSIH_IMM_ATOMIC_IMIN;
        case RMW_OR:
            return VKD3DSIH_IMM_ATOMIC_OR;
        case RMW_UMAX:
            return VKD3DSIH_IMM_ATOMIC_UMAX;
        case RMW_UMIN:
            return VKD3DSIH_IMM_ATOMIC_UMIN;
        case RMW_XCHG:
            return VKD3DSIH_IMM_ATOMIC_EXCH;
        case RMW_XOR:
            return VKD3DSIH_IMM_ATOMIC_XOR;
        default:
            /* DXIL currently doesn't use SUB and NAND. */
            return VKD3DSIH_INVALID;
    }
}

static void sm6_parser_emit_atomicrmw(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_block *code_block, struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_dst_param *dst_params;
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *reg;
    struct vkd3d_shader_register regs[2];
    const struct sm6_value *ptr, *src;
    enum vkd3d_shader_opcode op;
    unsigned int idx = 0;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(6);

    if (!(ptr = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_POINTER_TO_I32(ptr);
    if (ptr->u.reg.type != VKD3DSPR_GROUPSHAREDMEM)
    {
        FIXME("Only groupshared variables are supported.\n");
        return;
    }

    dst->type = ptr->type->u.pointer.type;

    if (!(src = sm6_parser_get_value_by_ref(sm6, record, dst->type, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_I32(src);

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 4);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 4);

    if ((op = sm6_map_atomicrmw_op(record->operands[idx])) == VKD3DSIH_INVALID)
    {
        FIXME("Unhandled atomicrmw op %"PRIu64".\n", record->operands[idx]);
        return;
    }
    if (record->operands[++idx])
        TRACE("Ignoring volatile modifier.\n");

    /* It's currently not possible to specify an atomic ordering in HLSL, and it defaults to seq_cst. */
    if (record->operands[++idx] != ORDERING_SEQCST)
        FIXME("Unhandled atomic ordering %"PRIu64".\n", record->operands[idx]);

    if (record->operands[++idx] != 1)
        WARN("Ignoring synchronisation scope %"PRIu64".\n", record->operands[idx]);

    if (ptr->structure_stride)
    {
        regs[0] = ptr->u.reg.idx[1].rel_addr->reg;
        shader_register_init(&regs[1], VKD3DSPR_IMMCONST, VKD3D_DATA_UINT, 0);
        regs[1].u.immconst_uint[0] = 0;
        reg = sm6_parser_emit_reg_composite_construct(sm6, regs, 2, ins++);
        ++code_block->instruction_count;
    }

    shader_instruction_init(ins, op);
    src_params = instruction_src_params_alloc(ins, 2, sm6);
    if (ptr->structure_stride)
        src_param_init_vector_from_reg(&src_params[0], reg);
    else
        src_param_make_constant_uint(&src_params[0], 0);
    src_param_init_from_value(&src_params[1], src);

    dst_params = instruction_dst_params_alloc(ins, 2, sm6);
    dst_param_init(&dst_params[0]);
    register_init_ssa_scalar(&dst_params[0].reg, dst->type, dst, sm6);
    dst_param_init(&dst_params[1]);
    dst_params[1].reg = ptr->u.reg;
    dst_params[1].reg.idx[1].offset = 0;
    dst_params[1].reg.idx_count = 2;

    dst->u.reg = dst_params[0].reg;
}

static enum vkd3d_shader_opcode sm6_parser_decode_binary_op(const struct sm6_parser *sm6,
        uint64_t code, const struct sm6_type *type_a, const struct sm6_type *type_b)
{
    bool is_bool = sm6_type_is_bool(type_a);
    bool is_int = sm6_type_is_1wil(type_a);
    enum vkd3d_shader_opcode op;
    bool is_valid;

    if (!is_int && !sm6_type_is_floating_point(type_a))
    {
        FIXME("Argument type %u is not bool, int16/32/64 or floating point.\n", type_a->class);
        return VKD3DSIH_INVALID;
    }
    if (type_a != type_b)
    {
        FIXME("Type mismatch, type %u width %u vs type %u width %u.\n", type_a->class,
                type_a->u.width, type_b->class, type_b->u.width);
    }

    switch (code)
    {
        case BINOP_ADD:
        case BINOP_SUB:
            op = is_int ? VKD3DSIH_IADD : VKD3DSIH_ADD;
            is_valid = !is_bool;
            break;
        case BINOP_AND:
            op = VKD3DSIH_AND;
            is_valid = is_int;
            break;
        case BINOP_ASHR:
            op = VKD3DSIH_ISHR;
            is_valid = is_int && !is_bool;
            break;
        case BINOP_LSHR:
            op = VKD3DSIH_USHR;
            is_valid = is_int && !is_bool;
            break;
        case BINOP_MUL:
            op = is_int ? VKD3DSIH_UMUL : VKD3DSIH_MUL;
            is_valid = !is_bool;
            break;
        case BINOP_OR:
            op = VKD3DSIH_OR;
            is_valid = is_int;
            break;
        case BINOP_SDIV:
            op = is_int ? VKD3DSIH_IDIV : VKD3DSIH_DIV;
            is_valid = !is_bool;
            break;
        case BINOP_SREM:
            op = is_int ? VKD3DSIH_IDIV : VKD3DSIH_FREM;
            is_valid = !is_bool;
            break;
        case BINOP_SHL:
            op = VKD3DSIH_ISHL;
            is_valid = is_int && !is_bool;
            break;
        case BINOP_UDIV:
        case BINOP_UREM:
            op = VKD3DSIH_UDIV;
            is_valid = is_int && !is_bool;
            break;
        case BINOP_XOR:
            op = VKD3DSIH_XOR;
            is_valid = is_int;
            break;
        default:
            FIXME("Unhandled binary op %"PRIu64".\n", code);
            return VKD3DSIH_INVALID;
    }

    if (!is_valid)
        WARN("Invalid operation %u for type %u width %u.\n", op, type_a->class, type_a->u.width);

    return op;
}

static void sm6_parser_emit_binop(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_src_param *src_params;
    enum vkd3d_shader_opcode handler_idx;
    const struct sm6_value *a, *b;
    enum dxil_binop_code code;
    unsigned int idx = 0;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    a = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx);
    b = sm6_parser_get_value_by_ref(sm6, record, a->type, &idx);
    if (!a || !b)
        return;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 1);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 2);

    code = record->operands[idx];
    handler_idx = sm6_parser_decode_binary_op(sm6, record->operands[idx++], a->type, b->type);
    if (handler_idx == VKD3DSIH_INVALID)
        return;

    shader_instruction_init(ins, handler_idx);

    if (record->operand_count > idx && record->operands[idx])
    {
        uint64_t flags = record->operands[idx];
        if (handler_idx == VKD3DSIH_ADD || handler_idx == VKD3DSIH_MUL || handler_idx == VKD3DSIH_DIV
                || handler_idx == VKD3DSIH_FREM)
        {
            if (!(flags & FP_ALLOW_UNSAFE_ALGEBRA))
                ins->flags |= VKD3DSI_PRECISE_XYZW;
            flags &= ~FP_ALLOW_UNSAFE_ALGEBRA;
            /* SPIR-V FPFastMathMode is only available in the Kernel executon model. */
            if (flags & ~(FP_NO_NAN | FP_NO_INF | FP_NO_SIGNED_ZEROS | FP_ALLOW_RECIPROCAL))
                WARN("Ignoring flags %#"PRIx64".\n", flags);
            else
                TRACE("Ignoring flags %#"PRIx64".\n", flags);
        }
        else
        {
            TRACE("Ignoring flags %#"PRIx64".\n", flags);
        }
    }

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_from_value(&src_params[0], a);
    src_param_init_from_value(&src_params[1], b);
    if (code == BINOP_SUB)
        src_params[1].modifiers = VKD3DSPSM_NEG;

    dst->type = a->type;

    if (handler_idx == VKD3DSIH_UMUL || handler_idx == VKD3DSIH_UDIV || handler_idx == VKD3DSIH_IDIV)
    {
        struct vkd3d_shader_dst_param *dst_params = instruction_dst_params_alloc(ins, 2, sm6);
        unsigned int index = code != BINOP_UDIV && code != BINOP_SDIV;

        dst_param_init(&dst_params[0]);
        dst_param_init(&dst_params[1]);
        register_init_ssa_scalar(&dst_params[index].reg, a->type, dst, sm6);
        dst_params[index ^ 1].reg.type = VKD3DSPR_NULL;
        dst->u.reg = dst_params[index].reg;
    }
    else
    {
        instruction_dst_param_init_ssa_scalar(ins, sm6);
    }
}

static struct sm6_block *sm6_function_get_block(struct sm6_function *function, uint64_t index)
{
    if (index >= function->block_count)
    {
        WARN("Invalid code block index %#"PRIx64".\n", index);
        return NULL;
    }
    return function->blocks[index];
}

bool sm6_function_declare_temp(struct sm6_function *function, struct vkd3d_shader_register *reg,
        enum vkd3d_data_type data_type, struct sm6_block *block, unsigned int init_ssa_id)
{
    if (!vkd3d_array_reserve((void **)&function->temps, &function->temp_capacity, function->temp_count + 1,
            sizeof(*function->temps)))
    {
        ERR("Failed to allocate temporary variable.\n");
        return false;
    }
    function->temps[function->temp_count].init_block = block;
    function->temps[function->temp_count].init_ssa_id = init_ssa_id;
    function->temps[function->temp_count].data_type = data_type;
    register_init_with_id(reg, VKD3DSPR_TEMP, data_type, function->temp_count++);
    return true;
}

static void sm6_parser_emit_temps(struct sm6_parser *sm6, struct sm6_function *function)
{
    struct vkd3d_shader_src_param *src_param;
    struct vkd3d_shader_dst_param *dst_param;
    struct vkd3d_shader_instruction *ins;
    struct sm6_block *block;
    unsigned int i;

    if (!function->temp_count)
        return;

    sm6_parser_require_space(sm6, 1);
    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TEMPS);
    ins->declaration.count = function->temp_count;

    for (i = 0; i < function->temp_count; ++i)
    {
        block = function->temps[i].init_block;
        if (!vkd3d_array_reserve((void **)&block->instructions, &block->instruction_capacity,
                block->instruction_count + 1, sizeof(*block->instructions)))
        {
            ERR("Failed to allocate instructions.\n");
            return;
        }
        ins = &block->instructions[block->instruction_count++];
        shader_instruction_init(ins, VKD3DSIH_MOV);

        src_param = instruction_src_params_alloc(ins, 1, sm6);
        dst_param = instruction_dst_params_alloc(ins, 1, sm6);
        src_param_init_scalar(src_param, 0);
        register_init_with_id(&src_param->reg, VKD3DSPR_SSA, VKD3D_DATA_FLOAT, function->temps[i].init_ssa_id);
        src_param->reg.idx[1].offset = function->temps[i].data_type;
        src_param->reg.idx_count = 2;

        dst_param_init(dst_param);
        register_init_with_id(&dst_param->reg, VKD3DSPR_TEMP, VKD3D_DATA_FLOAT, i);
    }
}

static bool sm6_block_init_succ(struct sm6_block *block, struct sm6_block *succ)
{
    return sm6_block_array_append_unique(&succ->pred, block) && sm6_block_array_append_unique(&block->succ, succ);
}

static void sm6_parser_emit_br(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_function *function, struct sm6_block *code_block, struct vkd3d_shader_instruction *ins)
{
    const struct sm6_value *value;
    unsigned int idx = 2;
    uint64_t index;

    if (record->operand_count != 1 && record->operand_count != 3)
    {
        FIXME("Invalid operand count %u.\n", record->operand_count);
        return;
    }

    if (record->operand_count == 1)
    {
        code_block->terminator.type = TERMINATOR_UNCOND_BR;
        code_block->terminator.direct_block = sm6_function_get_block(function, record->operands[0]);
        sm6_block_init_succ(code_block, code_block->terminator.direct_block);
    }
    else
    {
        if ((index = record->operands[0]) >= function->block_count
                || (index = record->operands[1]) >= function->block_count)
        {
            FIXME("Invalid block number %"PRIu64".\n", index);
            return;
        }

        if (!sm6->bool_type)
            FIXME("Bool type not found.\n");
        if (!(value = sm6_parser_get_value_by_ref(sm6, record, sm6->bool_type, &idx)))
            return;
        SM6_VALUE_VALIDATE_IS_BOOL(value);

        code_block->terminator.type = TERMINATOR_COND_BR;
        code_block->terminator.conditional_reg = value->u.reg;
        code_block->terminator.true_block = sm6_function_get_block(function, record->operands[0]);
        code_block->terminator.false_block = sm6_function_get_block(function, record->operands[1]);
        sm6_block_init_succ(code_block, code_block->terminator.true_block);
        sm6_block_init_succ(code_block, code_block->terminator.false_block);
    }

    ins->handler_idx = VKD3DSIH_NOP;
}

static const struct vkd3d_shader_register *sm6_parser_emit_coordinate_construct(struct sm6_parser *sm6,
        const struct sm6_value **operands, unsigned int max_operands,
        const struct sm6_value *z_operand, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *reg;
    unsigned int component_count, src_count;

    if (z_operand)
    {
        component_count = 3;
    }
    else
    {
        for (component_count = 0; component_count < max_operands; ++component_count)
            if (operands[component_count]->is_undefined)
                break;
    }
    if (component_count < 2)
        return NULL;

    shader_instruction_init(ins, VKD3DSIH_COMPOSITE_CONSTRUCT);

    src_count = component_count + !!z_operand;
    src_params = instruction_src_params_alloc(ins, src_count, sm6);
    sm6_src_operands_init(src_params, operands, component_count);
    if (z_operand)
        src_param_init_from_value(&src_params[src_count - 1], z_operand);

    reg = instruction_dst_param_vector_init_helper_ssa(sm6, ins, vkd3d_data_type_from_sm6_type(operands[0]->type),
            src_count);

    return reg;
}

static enum vkd3d_shader_opcode sm6_dx_map_void_op(enum dx_intrinsic_opcode op)
{
    switch (op)
    {
        case DX_WAVE_IS_FIRST_LANE:
            return VKD3DSIH_WAVE_IS_FIRST_LANE;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_void(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    shader_instruction_init(ins, sm6_dx_map_void_op(op));
    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static enum vkd3d_shader_opcode sm6_dx_map_unary_op(enum dx_intrinsic_opcode op)
{
    switch (op)
    {
        case DX_ISNAN:
            return VKD3DSIH_ISNAN;
        case DX_ISINF:
            return VKD3DSIH_ISINF;
        case DX_ISFINITE:
            return VKD3DSIH_ISFINITE;
        case DX_TAN:
            return VKD3DSIH_TAN;
        case DX_ACOS:
            return VKD3DSIH_ACOS;
        case DX_ASIN:
            return VKD3DSIH_ASIN;
        case DX_ATAN:
            return VKD3DSIH_ATAN;
        case DX_HCOS:
            return VKD3DSIH_HCOS;
        case DX_HSIN:
            return VKD3DSIH_HSIN;
        case DX_HTAN:
            return VKD3DSIH_HTAN;
        case DX_EXP:
            return VKD3DSIH_EXP;
        case DX_FRC:
            return VKD3DSIH_FRC;
        case DX_LOG:
            return VKD3DSIH_LOG;
        case DX_SQRT:
            return VKD3DSIH_SQRT;
        case DX_RSQRT:
            return VKD3DSIH_RSQ;
        case DX_ROUND_NE:
            return VKD3DSIH_ROUND_NE;
        case DX_ROUND_NI:
            return VKD3DSIH_ROUND_NI;
        case DX_ROUND_PI:
            return VKD3DSIH_ROUND_PI;
        case DX_ROUND_Z:
            return VKD3DSIH_ROUND_Z;
        case DX_BFREV:
            return VKD3DSIH_BFREV;
        case DX_COUNT_BITS:
            return VKD3DSIH_COUNTBITS;
        case DX_FIRST_BIT_LO:
            return VKD3DSIH_FIRSTBIT_LO;
        case DX_FIRST_BIT_HI:
            return VKD3DSIH_FIRSTBIT_HI;
        case DX_FIRST_BIT_SHI:
            return VKD3DSIH_FIRSTBIT_SHI;
        case DX_DERIV_COARSEX:
            return VKD3DSIH_DSX_COARSE;
        case DX_DERIV_COARSEY:
            return VKD3DSIH_DSY_COARSE;
        case DX_DERIV_FINEX:
            return VKD3DSIH_DSX_FINE;
        case DX_DERIV_FINEY:
            return VKD3DSIH_DSY_FINE;
        case DX_LEGACY_F32TOF16:
            return VKD3DSIH_F32TOF16;
        case DX_LEGACY_F16TOF32:
            return VKD3DSIH_F16TOF32;
        case DX_WAVE_ACTIVE_ALL_EQUAL:
            return VKD3DSIH_WAVE_ACTIVE_ALL_EQUAL;
        case DX_WAVE_ALL_BIT_COUNT:
            return VKD3DSIH_WAVE_ALL_BIT_COUNT;
        case DX_WAVE_ALL_TRUE:
            return VKD3DSIH_WAVE_ALL_TRUE;
        case DX_WAVE_ANY_TRUE:
            return VKD3DSIH_WAVE_ANY_TRUE;
        case DX_WAVE_PREFIX_BIT_COUNT:
            return VKD3DSIH_WAVE_PREFIX_BIT_COUNT;
        case DX_WAVE_READ_LANE_FIRST:
            return VKD3DSIH_WAVE_READ_LANE_FIRST;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_unary(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, sm6_dx_map_unary_op(op));
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static enum vkd3d_shader_opcode sm6_dx_map_binary_op(enum dx_intrinsic_opcode op, const struct sm6_type *type)
{
    switch (op)
    {
        case DX_FMAX:
            return type->u.width == 64 ? VKD3DSIH_DMAX : VKD3DSIH_MAX;
        case DX_FMIN:
            return type->u.width == 64 ? VKD3DSIH_DMIN : VKD3DSIH_MIN;
        case DX_IMAX:
            return VKD3DSIH_IMAX;
        case DX_IMIN:
            return VKD3DSIH_IMIN;
        case DX_UMAX:
            return VKD3DSIH_UMAX;
        case DX_UMIN:
            return VKD3DSIH_UMIN;
        case DX_WAVE_READ_LANE_AT:
            return VKD3DSIH_WAVE_READ_LANE_AT;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_binary(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;

    shader_instruction_init(ins, sm6_dx_map_binary_op(op, operands[2]->type));
    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_from_value(&src_params[0], operands[2]);
    src_param_init_from_value(&src_params[1], operands[3]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static enum vkd3d_shader_opcode sm6_map_atomic_binop(const struct sm6_value *operand)
{
    uint64_t code = sm6_value_get_constant_uint(operand);

    switch (code)
    {
        case ATOMIC_BINOP_ADD:
            return VKD3DSIH_IMM_ATOMIC_IADD;
        case ATOMIC_BINOP_AND:
            return VKD3DSIH_IMM_ATOMIC_AND;
        case ATOMIC_BINOP_IMAX:
            return VKD3DSIH_IMM_ATOMIC_IMAX;
        case ATOMIC_BINOP_IMIN:
            return VKD3DSIH_IMM_ATOMIC_IMIN;
        case ATOMIC_BINOP_OR:
            return VKD3DSIH_IMM_ATOMIC_OR;
        case ATOMIC_BINOP_UMAX:
            return VKD3DSIH_IMM_ATOMIC_UMAX;
        case ATOMIC_BINOP_UMIN:
            return VKD3DSIH_IMM_ATOMIC_UMIN;
        case ATOMIC_BINOP_XCHG:
            return VKD3DSIH_IMM_ATOMIC_EXCH;
        case ATOMIC_BINOP_XOR:
            return VKD3DSIH_IMM_ATOMIC_XOR;
        /* DXIL currently doesn't use SUB and NAND. */
        default:
            FIXME("Unhandled atomic binop %"PRIu64".\n", code);
            return VKD3DSIH_INVALID;
    }
}

static void sm6_parser_emit_dx_atomic_binop(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);
    const struct vkd3d_shader_register *reg = NULL;
    bool is_cmp_xchg = op == DX_ATOMIC_CMP_XCHG;
    struct vkd3d_shader_dst_param *dst_params;
    struct vkd3d_shader_src_param *src_params;
    enum vkd3d_shader_opcode handler_idx;
    const struct sm6_value *resource;
    unsigned int coord_idx;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);

    if (is_cmp_xchg)
        handler_idx = VKD3DSIH_IMM_ATOMIC_CMP_EXCH;
    else if ((handler_idx = sm6_map_atomic_binop(operands[3])) == VKD3DSIH_INVALID)
        return;

    coord_idx = 4 - is_cmp_xchg;
    if (resource->u.handle.resource->type != VKD3D_SHADER_RESOURCE_BUFFER)
    {
        if ((reg = sm6_parser_emit_coordinate_construct(sm6, &operands[coord_idx], 3, NULL, ins)))
        {
            ++ins;
            ++code_block->instruction_count;
        }
    }
    else if (resource->u.handle.resource->kind == RESOURCE_KIND_STRUCTUREDBUFFER)
    {
        reg = sm6_parser_emit_coordinate_construct(sm6, &operands[coord_idx], 2, NULL, ins++);
        ++code_block->instruction_count;
    }

    shader_instruction_init(ins, handler_idx);
    src_params = instruction_src_params_alloc(ins, 2 + is_cmp_xchg, sm6);
    if (reg)
    {
        src_param_init_vector_from_reg(&src_params[0], reg);
    }
    else
    {
        src_param_init_from_value(&src_params[0], operands[coord_idx]);
        assert(operands[5]->is_undefined && (is_cmp_xchg || operands[6]->is_undefined));
    }
    if (is_cmp_xchg)
        src_param_init_from_value(&src_params[1], operands[6]);
    src_param_init_from_value(&src_params[1 + is_cmp_xchg], operands[7]);

    dst_params = instruction_dst_params_alloc(ins, 2, sm6);
    dst_param_init(&dst_params[0]);
    register_init_ssa_scalar(&dst_params[0].reg, dst->type, dst, sm6);
    dst_param_init(&dst_params[1]);
    dst_params[1].reg = resource->u.handle.reg;

    dst->u.reg = dst_params[0].reg;
}

static void sm6_parser_emit_dx_barrier(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    unsigned int flags;

    shader_instruction_init(ins, VKD3DSIH_SYNC);
    flags = sm6_value_get_constant_uint(operands[2]);
    ins->flags = flags & ~(VKD3DSSF_GROUP_SHARED_MEMORY | VKD3DSSF_GLOBAL_UAV);
    /* These are swapped in DXIL vs TPF. */
    if (flags & VKD3DSSF_GROUP_SHARED_MEMORY)
        ins->flags |= VKD3DSSF_GLOBAL_UAV;
    if (flags & VKD3DSSF_GLOBAL_UAV)
        ins->flags |= VKD3DSSF_GROUP_SHARED_MEMORY;
}

static void sm6_parser_emit_dx_buffer_update_counter(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *resource;
    unsigned int i;
    int8_t inc;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);

    if ((i = sm6_value_get_constant_uint(operands[3])) == UINT_MAX)
        WARN("Increment/decrement parameter is not a constant.\n");
    inc = i;

    shader_instruction_init(ins, inc < 0 ? VKD3DSIH_IMM_ATOMIC_CONSUME : VKD3DSIH_IMM_ATOMIC_ALLOC);
    src_params = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector_from_reg(&src_params[0], &resource->u.handle.reg);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_calculate_lod(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct sm6_value *resource, *sampler;
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *coord;
    unsigned int clamp;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    sampler = operands[3];
    SM6_VALUE_VALIDATE_IS_HANDLE(sampler);

    if ((coord = sm6_parser_emit_coordinate_construct(sm6, &operands[4], 3, NULL, ins)))
    {
        ++ins;
        ++code_block->instruction_count;
    }

    if ((clamp = sm6_value_get_constant_uint(operands[7])) == UINT_MAX)
        WARN("Clamp is not a constant uint.\n");

    shader_instruction_init(ins, VKD3DSIH_LOD);
    src_params = instruction_src_params_alloc(ins, 3, sm6);
    src_param_init_vector_from_reg(&src_params[0], coord ? coord : &operands[4]->u.reg);
    src_param_init_vector_from_reg(&src_params[1], &resource->u.handle.reg);
    src_param_init_vector_from_reg(&src_params[2], &sampler->u.handle.reg);

    instruction_dst_param_init_ssa_scalar_component(ins, !clamp, sm6);
}

static void sm6_parser_emit_dx_cbuffer_load(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);
    unsigned int component_count = VKD3D_VEC4_SIZE;
    struct vkd3d_shader_src_param *src_param;
    const struct sm6_value *buffer;
    const struct sm6_type *type;

    buffer = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(buffer);

    shader_instruction_init(ins, VKD3DSIH_MOV);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector_from_reg(src_param, &buffer->u.handle.reg);
    if (!(type = sm6_type_get_pseudo_vec4_element_type(dst->type)))
    {
        WARN("Failed to get return type.\n");
    }
    else
    {
        src_param->reg.data_type = vkd3d_data_type_from_sm6_type(type);
        component_count = sm6_type_max_vector_size(type);
    }
    register_address_init(&src_param->reg, operands[3], 2, sm6);
    src_param->reg.idx_count = 3;

    instruction_dst_param_vector_init_ssa(sm6, ins, component_count);
}

static void sm6_parser_emit_dx_coverage(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_MOV);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_scalar(src_param, 0);
    if (!sm6->coverage_reg)
        sm6->coverage_reg = sm6_parser_emit_input_register(sm6, VKD3DSPR_COVERAGE, VKD3D_DATA_UINT);
    src_param->reg = *sm6->coverage_reg;

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static unsigned int sm6_parser_get_resource(struct sm6_parser *sm6, enum dxil_resource_class class, unsigned int id)
{
    unsigned int i;

    for (i = 0; i < sm6->resource_count; ++i)
        if (sm6->resources[i].class == class && sm6->resources[i].id == id)
            return i;

    return UINT_MAX;
}

static void sm6_parser_emit_dx_create_handle(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_register *reg;
    enum dxil_resource_class class;
    struct sm6_value *dst;
    unsigned int i, id;

    class = sm6_value_get_constant_uint(operands[2]);
    id = sm6_value_get_constant_uint(operands[3]);
    if ((i = sm6_parser_get_resource(sm6, class, id)) >= sm6->resource_count)
    {
        WARN("Failed to find resource.\n");
        return;
    }

    dst = sm6_parser_get_current_value(sm6);
    dst->value_type = VALUE_TYPE_HANDLE;
    dst->u.handle.resource = &sm6->resources[i];
    dst->u.handle.reg = *sm6->resources[i].reg;

    reg = &dst->u.handle.reg;
    reg->idx[2].offset = UINT_MAX;
    register_address_init(reg, operands[4], 1, sm6);
    reg->idx_count = 2;
    if(!sm6_value_is_constant(operands[5]))
        WARN("Non-uniform attribute is not constant.\n");
    else
        reg->non_uniform = !!sm6_value_get_constant_uint(operands[5]);

    /* NOP is used to flag no instruction emitted. */
    ins->handler_idx = VKD3DSIH_NOP;
}

static void sm6_parser_emit_dx_stream(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, op == DX_CUT_STREAM ? VKD3DSIH_CUT_STREAM : VKD3DSIH_EMIT_STREAM);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    if (op == DX_EMIT_THEN_CUT_STREAM)
    {
        sm6_parser_emit_dx_stream(sm6, code_block, DX_CUT_STREAM, operands, ++ins);
        ++code_block->instruction_count;
    }
}

static void sm6_parser_emit_dx_discard(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_TEXKILL);
    ins->flags = VKD3D_SHADER_CONDITIONAL_OP_PARAM;

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);
}

static void sm6_parser_dcl_register_builtin(struct sm6_parser *sm6,
        enum vkd3d_shader_register_type reg_type, enum vkd3d_data_type data_type, unsigned int component_count)
{
    struct vkd3d_shader_dst_param *dst_param;
    struct vkd3d_shader_instruction *ins;
    uint64_t reg_flag;

    assert(reg_type < 64u);
    reg_flag = 1ull << reg_type;
    if (!(sm6->input_reg_declarations & reg_flag))
    {
        sm6->input_reg_declarations |= reg_flag;
        ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_INPUT);
        dst_param = &ins->declaration.dst;
        dst_param->write_mask = vkd3d_write_mask_from_component_count(component_count);
        dst_param->modifiers = 0;
        dst_param->shift = 0;
        shader_register_init(&dst_param->reg, reg_type, data_type, 0);
    }
}

static void sm6_parser_emit_dx_domain_location(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    sm6_parser_dcl_register_builtin(sm6, VKD3DSPR_TESSCOORD, VKD3D_DATA_FLOAT, 3);

    shader_instruction_init(ins, VKD3DSIH_MOV);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector(src_param, 3);
    shader_register_init(&src_param->reg, VKD3DSPR_TESSCOORD, VKD3D_DATA_FLOAT, 0);

    instruction_dst_param_init_ssa_scalar_component(ins, sm6_value_get_constant_uint(operands[2]), sm6);
}

static void sm6_parser_emit_dx_dot(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct vkd3d_shader_register *regs[2];
    struct vkd3d_shader_src_param *src_params;
    enum vkd3d_shader_opcode handler_idx;
    unsigned int component_count;

    switch (op)
    {
        case DX_DOT2:
            handler_idx = VKD3DSIH_DP2;
            component_count = 2;
            break;
        case DX_DOT3:
            handler_idx = VKD3DSIH_DP3;
            component_count = 3;
            break;
        case DX_DOT4:
            handler_idx = VKD3DSIH_DP4;
            component_count = 4;
            break;
        default:
            vkd3d_unreachable();
    }

    regs[0] = sm6_parser_emit_composite_construct(sm6, &operands[2], component_count, ins++);
    regs[1] = sm6_parser_emit_composite_construct(sm6, &operands[2 + component_count], component_count, ins++);
    code_block->instruction_count += 2;

    shader_instruction_init(ins, handler_idx);
    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_vector_from_reg(&src_params[0], regs[0]);
    src_param_init_vector_from_reg(&src_params[1], regs[1]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_eval_attrib(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct shader_signature *signature;
    unsigned int element_id, column_index;
    const struct signature_element *e;

    element_id = sm6_value_get_constant_uint(operands[2]);
    column_index = sm6_value_get_constant_uint(operands[4]);

    signature = &sm6->p.shader_desc.input_signature;
    if (element_id >= signature->element_count)
    {
        WARN("Invalid signature element id %u.\n", element_id);
        return;
    }
    if (column_index >= VKD3D_VEC4_SIZE)
    {
        WARN("Invalid column index %u.\n", column_index);
        return;
    }

    shader_instruction_init(ins, op == DX_EVAL_CENTROID ? VKD3DSIH_EVAL_CENTROID : VKD3DSIH_EVAL_SAMPLE_INDEX);

    e = &signature->elements[element_id];

    src_params = instruction_src_params_alloc(ins, 1 + (op == DX_EVAL_SAMPLE_INDEX), sm6);
    src_params[0].reg = sm6->input_params[element_id].reg;
    src_param_init_scalar(&src_params[0], column_index);
    if (e->register_count > 1)
        register_address_init(&src_params[0].reg, operands[3], 0, sm6);

    if (op == DX_EVAL_SAMPLE_INDEX)
        src_param_init_from_value(&src_params[1], operands[5]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_fabs(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_MOV);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);
    src_param->modifiers = VKD3DSPSM_ABS;

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_compute_builtin(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    unsigned int component_count = 3, component_idx = 0;
    struct vkd3d_shader_src_param *src_param;
    enum vkd3d_shader_register_type reg_type;

    switch (op)
    {
        case DX_THREAD_ID:
            reg_type = VKD3DSPR_THREADID;
            break;
        case DX_GROUP_ID:
            reg_type = VKD3DSPR_THREADGROUPID;
            break;
        case DX_THREAD_ID_IN_GROUP:
            reg_type = VKD3DSPR_LOCALTHREADID;
            break;
        case DX_FLATTENED_THREAD_ID_IN_GROUP:
            reg_type = VKD3DSPR_LOCALTHREADINDEX;
            component_count = 1;
            break;
        default:
            vkd3d_unreachable();
    }

    sm6_parser_dcl_register_builtin(sm6, reg_type, VKD3D_DATA_UINT, component_count);
    shader_instruction_init(ins, VKD3DSIH_MOV);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector(src_param, component_count);
    shader_register_init(&src_param->reg, reg_type, VKD3D_DATA_UINT, 0);

    if (component_count > 1)
    {
        if (!sm6_value_is_constant(operands[2]))
            WARN("Unsupported non-constant component index.\n");
        else
            component_idx = sm6_value_get_constant_uint(operands[2]);
    }
    instruction_dst_param_init_ssa_scalar_component(ins, component_idx, sm6);
}

static enum vkd3d_shader_opcode sm6_dx_map_ma_op(enum dx_intrinsic_opcode op, const struct sm6_type *type)
{
    switch (op)
    {
        case DX_FMA:
        case DX_FMAD:
            return (type->u.width == 64) ? VKD3DSIH_DFMA : VKD3DSIH_MAD;
        case DX_IMAD:
        case DX_UMAD:
            if (type->u.width == 64)
                FIXME("64-bit integers are not supported yet.\n");
            return VKD3DSIH_IMAD;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_ma(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    unsigned int i;

    shader_instruction_init(ins, sm6_dx_map_ma_op(op, operands[2]->type));
    src_params = instruction_src_params_alloc(ins, 3, sm6);
    for (i = 0; i < 3; ++i)
        src_param_init_from_value(&src_params[i], operands[2 + i]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_get_dimensions(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    unsigned int is_texture, component_count;
    const struct sm6_value *resource;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    is_texture = resource->u.handle.resource->type != VKD3D_SHADER_RESOURCE_BUFFER;

    instruction_init_with_resource(ins, is_texture ? VKD3DSIH_RESINFO : VKD3DSIH_BUFINFO, resource);

    src_params = instruction_src_params_alloc(ins, 1 + is_texture, sm6);
    src_param_init_vector_from_reg(&src_params[is_texture], &resource->u.handle.reg);
    if (is_texture)
    {
        ins->flags = VKD3DSI_RESINFO_UINT4;
        src_param_init_from_value(&src_params[0], operands[3]);
        component_count = VKD3D_VEC4_SIZE;
    }
    else
    {
        component_count = 1 + (resource->u.handle.resource->kind == RESOURCE_KIND_STRUCTUREDBUFFER);
    }

    instruction_dst_param_vector_init_ssa(sm6, ins, component_count);
}

static void sm6_parser_emit_dx_load_input(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    bool is_control_point = op == DX_LOAD_OUTPUT_CONTROL_POINT;
    bool is_patch_constant = op == DX_LOAD_PATCH_CONSTANT;
    struct vkd3d_shader_src_param *src_param;
    unsigned int i, element_id, column_index;
    const struct shader_signature *signature;
    const struct signature_element *e;

    element_id = sm6_value_get_constant_uint(operands[2]);
    column_index = sm6_value_get_constant_uint(operands[4]);
    assert(!is_control_point || !operands[5]->is_undefined);

    shader_instruction_init(ins, VKD3DSIH_MOV);

    signature = is_patch_constant ? &sm6->p.shader_desc.patch_constant_signature : &sm6->p.shader_desc.input_signature;
    e = &signature->elements[element_id];

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param->reg = is_patch_constant ? sm6->patch_constant_params[element_id].reg : sm6->input_params[element_id].reg;
    src_param_init_scalar(src_param, column_index);
    i = 0;
    if (e->register_count > 1)
        register_address_init(&src_param->reg, operands[3], i++, sm6);
    if (!is_patch_constant && !operands[5]->is_undefined)
    {
        assert(src_param->reg.idx[i].offset != UINT_MAX);
        register_address_init(&src_param->reg, operands[5], i, sm6);
    }

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_make_double(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *reg;

    reg = sm6_parser_emit_composite_construct(sm6, &operands[2], 2, ins++);
    ++code_block->instruction_count;

    shader_instruction_init(ins, VKD3DSIH_BITCAST);
    src_params = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector(&src_params[0], 2);
    src_params[0].reg = *reg;

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_input_register_mov(struct sm6_parser *sm6,
        struct vkd3d_shader_instruction *ins, enum vkd3d_shader_register_type reg_type, enum vkd3d_data_type data_type)
{
    struct vkd3d_shader_src_param *src_param;

    sm6_parser_dcl_register_builtin(sm6, reg_type, data_type, 1);

    shader_instruction_init(ins, VKD3DSIH_MOV);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init(src_param);
    shader_register_init(&src_param->reg, reg_type, data_type, 0);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_output_control_point_id(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    sm6_parser_emit_dx_input_register_mov(sm6, ins, VKD3DSPR_OUTPOINTID, VKD3D_DATA_UINT);
}

static void sm6_parser_emit_dx_primitive_id(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    sm6_parser_emit_dx_input_register_mov(sm6, ins, VKD3DSPR_PRIMID, VKD3D_DATA_UINT);
}

static void sm6_parser_emit_dx_raw_buffer_load(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    unsigned int operand_count, write_mask, component_count = VKD3D_VEC4_SIZE;
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *resource;
    bool raw;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    raw = resource->u.handle.resource->kind == RESOURCE_KIND_RAWBUFFER;

    instruction_init_with_resource(ins, raw ? VKD3DSIH_LD_RAW : VKD3DSIH_LD_STRUCTURED, resource);
    operand_count = 2 + !raw;
    src_params = instruction_src_params_alloc(ins, operand_count, sm6);
    sm6_src_operands_init(src_params, &operands[3], operand_count - 1);
    src_param_init_vector_from_reg(&src_params[operand_count - 1], &resource->u.handle.reg);

    if (op == DX_RAW_BUFFER_LOAD)
    {
        write_mask = sm6_value_get_constant_uint(operands[5]);
        if (!write_mask || write_mask > VKD3DSP_WRITEMASK_ALL)
            WARN("Invalid write mask %#x.\n", write_mask);
        component_count = vkd3d_write_mask_component_count(write_mask);
        if (write_mask >= (1u << component_count))
            FIXME("Unhandled nonzero base component index.\n");
    }
    instruction_dst_param_vector_init_ssa(sm6, ins, component_count);
}

static void sm6_parser_emit_dx_raw_buffer_store(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    unsigned int write_mask, component_count, operand_count;
    struct vkd3d_shader_src_param *src_params;
    struct vkd3d_shader_dst_param *dst_param;
    const struct vkd3d_shader_register *data;
    const struct sm6_value *resource;
    bool raw;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    raw = resource->u.handle.resource->kind == RESOURCE_KIND_RAWBUFFER;

    write_mask = sm6_value_get_constant_uint(operands[9]);
    if (!write_mask || write_mask > VKD3DSP_WRITEMASK_ALL)
        WARN("Invalid write mask %#x.\n", write_mask);
    component_count = vkd3d_write_mask_component_count(write_mask);
    if (write_mask >= (1u << component_count))
        FIXME("Unhandled nonzero base component index.\n");
    data = &operands[5]->u.reg;
    if (component_count > 1)
    {
        data = sm6_parser_emit_composite_construct(sm6, &operands[5], component_count, ins);
        ++ins;
        ++code_block->instruction_count;
    }

    shader_instruction_init(ins, raw ? VKD3DSIH_STORE_RAW : VKD3DSIH_STORE_STRUCTURED);
    operand_count = 2 + !raw;

    src_params = instruction_src_params_alloc(ins, operand_count, sm6);
    sm6_src_operands_init(src_params, &operands[3], operand_count - 1);
    src_param_init_vector_from_reg(&src_params[operand_count - 1], data);

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init_with_mask(dst_param, write_mask);
    dst_param->reg = resource->u.handle.reg;
}

static void sm6_parser_emit_dx_buffer_load(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *resource;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);

    if (resource->u.handle.resource->kind == RESOURCE_KIND_RAWBUFFER
            || resource->u.handle.resource->kind == RESOURCE_KIND_STRUCTUREDBUFFER)
        return sm6_parser_emit_dx_raw_buffer_load(sm6, code_block, op, operands, ins);

    if (resource->u.handle.resource->kind != RESOURCE_KIND_TYPEDBUFFER)
        WARN("Resource is not a typed buffer.\n");

    instruction_init_with_resource(ins, resource->u.handle.resource->class == RESOURCE_CLASS_UAV
            ? VKD3DSIH_LD_UAV_TYPED : VKD3DSIH_LD, resource);

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_from_value(&src_params[0], operands[3]);
    if (!sm6_value_is_undef(operands[4]))
        WARN("Ignoring structure offset.\n");
    src_param_init_vector_from_reg(&src_params[1], &resource->u.handle.reg);

    instruction_dst_param_vector_init_ssa(sm6, ins, VKD3D_VEC4_SIZE);
}

static void sm6_parser_emit_dx_buffer_store(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct vkd3d_shader_register *texel;
    struct vkd3d_shader_src_param *src_params;
    struct vkd3d_shader_dst_param *dst_param;
    unsigned int write_mask, component_count;
    const struct sm6_value *resource;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    if (resource->u.handle.resource->kind == RESOURCE_KIND_RAWBUFFER
            || resource->u.handle.resource->kind == RESOURCE_KIND_STRUCTUREDBUFFER)
        return sm6_parser_emit_dx_raw_buffer_store(sm6, code_block, op, operands, ins);

    write_mask = sm6_value_get_constant_uint(operands[9]);
    component_count = vkd3d_write_mask_component_count(write_mask);
    if (write_mask > VKD3DSP_WRITEMASK_ALL)
        WARN("Invalid write mask %#x.\n", write_mask);
    texel = &operands[5]->u.reg;
    if (component_count > 1)
    {
        texel = sm6_parser_emit_composite_construct(sm6, &operands[5], component_count, ins);
        ++ins;
        ++code_block->instruction_count;
    }

    shader_instruction_init(ins, VKD3DSIH_STORE_UAV_TYPED);

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_from_value(&src_params[0], operands[3]);
    if (!sm6_value_is_undef(operands[4]))
        WARN("Ignoring structure offset.\n");
    src_param_init_vector_from_reg(&src_params[1], texel);

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init_with_mask(dst_param, write_mask);
    dst_param->reg = resource->u.handle.reg;
}

static void sm6_parser_emit_dx_get_sample_pos(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *resource;

    shader_instruction_init(ins, VKD3DSIH_SAMPLE_POS);

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    if (op == DX_RT_GET_SAMPLE_POS)
    {
        src_param_init_vector(&src_params[0], 2);
        shader_register_init(&src_params[0].reg, VKD3DSPR_RASTERIZER, VKD3D_DATA_FLOAT, 0);
        src_param_init_from_value(&src_params[1], operands[2]);
    }
    else
    {
        resource = operands[2];
        SM6_VALUE_VALIDATE_IS_HANDLE(resource);
        src_param_init_vector_from_reg(&src_params[0], &resource->u.handle.reg);
        src_param_init_from_value(&src_params[1], operands[3]);
    }

    instruction_dst_param_vector_init_ssa(sm6, ins, 2);
}

static void sm6_parser_emit_dx_get_sample_count(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_params;

    shader_instruction_init(ins, VKD3DSIH_SAMPLE_INFO);
    ins->flags = VKD3DSI_SAMPLE_INFO_UINT;

    src_params = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_vector(&src_params[0], 2);
    shader_register_init(&src_params[0].reg, VKD3DSPR_RASTERIZER, VKD3D_DATA_FLOAT, 0);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void shader_instruction_set_texel_offset(struct vkd3d_shader_instruction *ins,
        const struct sm6_value **operands)
{
    unsigned int i, offsets[3];

    for (i = 0; i < ARRAY_SIZE(offsets); ++i)
        offsets[i] = sm6_value_is_undef(operands[i]) ? 0 : sm6_value_get_constant_uint(operands[i]);

    ins->texel_offset.u = offsets[0];
    ins->texel_offset.v = offsets[1];
    ins->texel_offset.w = offsets[2];
}

static void sm6_parser_emit_dx_sample(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    unsigned int clamp_idx = 0, component_count = VKD3D_VEC4_SIZE;
    const struct vkd3d_shader_register *coord, *ddx, *ddy;
    const struct sm6_value *resource, *sampler;
    struct vkd3d_shader_src_param *src_params;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    sampler = operands[3];
    SM6_VALUE_VALIDATE_IS_HANDLE(sampler);

    if ((coord = sm6_parser_emit_coordinate_construct(sm6, &operands[4], VKD3D_VEC4_SIZE, NULL, ins)))
    {
        ++ins;
        ++code_block->instruction_count;
    }

    if (op == DX_SAMPLE_GRAD)
    {
        ddx = sm6_parser_emit_coordinate_construct(sm6, &operands[11], VKD3D_VEC4_SIZE, NULL, ins++);
        ddy = sm6_parser_emit_coordinate_construct(sm6, &operands[14], VKD3D_VEC4_SIZE, NULL, ins++);
        if (!ddx || !ddy)
        {
            FIXME("Unexpected undefined coordinates.\n");
            return;
        }
        code_block->instruction_count += 2;
    }

    switch (op)
    {
        case DX_SAMPLE:
            instruction_init_with_resource(ins, VKD3DSIH_SAMPLE, resource);
            src_params = instruction_src_params_alloc(ins, 3, sm6);
            clamp_idx = 11;
            break;
        case DX_SAMPLE_B:
            clamp_idx = 12;
            /* fall through */
        case DX_SAMPLE_LOD:
            instruction_init_with_resource(ins, (op == DX_SAMPLE_B) ? VKD3DSIH_SAMPLE_B : VKD3DSIH_SAMPLE_LOD,
                    resource);
            src_params = instruction_src_params_alloc(ins, 4, sm6);
            src_param_init_from_value(&src_params[3], operands[11]);
            break;
        case DX_SAMPLE_C:
            clamp_idx = 12;
            component_count = 1;
            /* fall through */
        case DX_SAMPLE_C_LZ:
            instruction_init_with_resource(ins, (op == DX_SAMPLE_C_LZ) ? VKD3DSIH_SAMPLE_C_LZ
                    : VKD3DSIH_SAMPLE_C, resource);
            src_params = instruction_src_params_alloc(ins, 4, sm6);
            src_param_init_from_value(&src_params[3], operands[11]);
            component_count = 1;
            break;
        case DX_SAMPLE_GRAD:
            instruction_init_with_resource(ins, VKD3DSIH_SAMPLE_GRAD, resource);
            src_params = instruction_src_params_alloc(ins, 5, sm6);
            src_param_init_vector_from_reg(&src_params[3], ddx);
            src_param_init_vector_from_reg(&src_params[4], ddy);
            clamp_idx = 17;
            break;
        default:
            vkd3d_unreachable();
    }

    if (clamp_idx && !sm6_value_is_undef(operands[clamp_idx]))
        FIXME("Ignoring LOD clamp value.\n");

    src_param_init_vector_from_reg(&src_params[0], coord ? coord : &operands[4]->u.reg);
    src_param_init_vector_from_reg(&src_params[1], &resource->u.handle.reg);
    src_param_init_vector_from_reg(&src_params[2], &sampler->u.handle.reg);
    shader_instruction_set_texel_offset(ins, &operands[8]);

    instruction_dst_param_vector_init_ssa(sm6, ins, component_count);
}

static void sm6_parser_emit_dx_sample_index(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct shader_signature *signature = &sm6->p.shader_desc.input_signature;
    struct vkd3d_shader_src_param *src_param;
    unsigned int element_idx;

    shader_instruction_init(ins, VKD3DSIH_MOV);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    if ((element_idx = sm6_signature_find_element_by_sysval(signature, VKD3D_SHADER_SV_SAMPLE_INDEX)) == UINT_MAX)
    {
        WARN("Sample index is not in the signature.\n");
        register_make_undef(&src_param->reg, VKD3D_DATA_UINT);
    }
    else
    {
        src_param->reg = sm6->input_params[element_idx].reg;
    }
    src_param_init_scalar(src_param, 0);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_saturate(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_MOV);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_init_ssa_scalar_with_modifiers(ins, VKD3DSPDM_SATURATE, sm6);
}

static void sm6_parser_emit_dx_sincos(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct sm6_value *dst = sm6_parser_get_current_value(sm6);
    struct vkd3d_shader_dst_param *dst_params;
    struct vkd3d_shader_src_param *src_param;
    unsigned int index;

    shader_instruction_init(ins, VKD3DSIH_SINCOS);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    index = op == DX_COS;
    dst_params = instruction_dst_params_alloc(ins, 2, sm6);
    dst_param_init(&dst_params[0]);
    dst_param_init(&dst_params[1]);
    register_init_ssa_scalar(&dst_params[index].reg, dst->type, dst, sm6);
    dst_params[index ^ 1].reg.type = VKD3DSPR_NULL;
    dst->u.reg = dst_params[index].reg;
}

static void sm6_parser_emit_dx_split_double(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_BITCAST);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_vector_init_ssa(sm6, ins, 2);
}

static void sm6_parser_emit_dx_store_output(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    bool is_patch_constant = op == DX_STORE_PATCH_CONSTANT;
    struct vkd3d_shader_src_param *src_param;
    struct vkd3d_shader_dst_param *dst_param;
    const struct shader_signature *signature;
    unsigned int element_id, column_index;
    const struct signature_element *e;
    const struct sm6_value *value;

    element_id = sm6_value_get_constant_uint(operands[2]);
    column_index = sm6_value_get_constant_uint(operands[4]);

    signature = is_patch_constant ? &sm6->p.shader_desc.patch_constant_signature : &sm6->p.shader_desc.output_signature;
    if (element_id >= signature->element_count)
    {
        WARN("Invalid element id %u.\n", element_id);
        return;
    }
    e = &signature->elements[element_id];

    value = operands[5];
    if (!sm6_value_is_register(value))
    {
        WARN("Source value is not a register.\n");
        return;
    }

    shader_instruction_init(ins, VKD3DSIH_MOV);

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init_scalar(dst_param, column_index);
    dst_param->reg = is_patch_constant ? sm6->patch_constant_params[element_id].reg : sm6->output_params[element_id].reg;
    if (e->register_count > 1)
        register_address_init(&dst_param->reg, operands[3], 0, sm6);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, value);
}

static void sm6_parser_emit_dx_texture_gather(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct vkd3d_shader_register *coord, *offset;
    const struct sm6_value *resource, *sampler;
    struct vkd3d_shader_src_param *src_params;
    unsigned int extended_offset;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    sampler = operands[3];
    SM6_VALUE_VALIDATE_IS_HANDLE(sampler);

    if ((coord = sm6_parser_emit_coordinate_construct(sm6, &operands[4], VKD3D_VEC4_SIZE, NULL, ins)))
    {
        ++ins;
        ++code_block->instruction_count;
    }

    if ((extended_offset = !sm6_value_vector_is_constant(&operands[8], 2)))
    {
        offset = sm6_parser_emit_coordinate_construct(sm6, &operands[8], 2, NULL, ins++);
        ++code_block->instruction_count;
    }

    if (op == DX_TEXTURE_GATHER)
    {
        instruction_init_with_resource(ins, extended_offset ? VKD3DSIH_GATHER4_PO : VKD3DSIH_GATHER4, resource);
        src_params = instruction_src_params_alloc(ins, 3 + extended_offset, sm6);
    }
    else
    {
        instruction_init_with_resource(ins, extended_offset ? VKD3DSIH_GATHER4_PO_C : VKD3DSIH_GATHER4_C, resource);
        src_params = instruction_src_params_alloc(ins, 4 + extended_offset, sm6);
        src_param_init_from_value(&src_params[3 + extended_offset], operands[11]);
    }

    src_param_init_vector_from_reg(&src_params[0], coord ? coord : &operands[4]->u.reg);
    if (extended_offset)
        src_param_init_vector_from_reg(&src_params[1], offset);
    else
        shader_instruction_set_texel_offset(ins, &operands[8]);
    src_param_init_vector_from_reg(&src_params[1 + extended_offset], &resource->u.handle.reg);
    src_param_init_vector_from_reg(&src_params[2 + extended_offset], &sampler->u.handle.reg);
    src_params[2 + extended_offset].swizzle = sm6_value_get_constant_uint(operands[10]);

    instruction_dst_param_vector_init_ssa(sm6, ins, VKD3D_VEC4_SIZE);
}

static void sm6_parser_emit_dx_texture_load(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct sm6_value *resource, *mip_level_or_sample_count;
    enum vkd3d_shader_resource_type resource_type;
    struct vkd3d_shader_src_param *src_params;
    const struct vkd3d_shader_register *coord;
    bool is_multisample, is_uav;
    unsigned int i;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);
    resource_type = resource->u.handle.resource->type;
    is_multisample = resource_type == VKD3D_SHADER_RESOURCE_TEXTURE_2DMS
            || resource_type == VKD3D_SHADER_RESOURCE_TEXTURE_2DMSARRAY;
    is_uav = resource->u.handle.resource->class == RESOURCE_CLASS_UAV;

    mip_level_or_sample_count = (resource_type != VKD3D_SHADER_RESOURCE_BUFFER) ? operands[3] : NULL;
    if ((coord = sm6_parser_emit_coordinate_construct(sm6, &operands[4], 3,
            is_multisample ? NULL : mip_level_or_sample_count, ins)))
    {
        ++ins;
        ++code_block->instruction_count;
    }

    instruction_init_with_resource(ins, is_uav ? VKD3DSIH_LD_UAV_TYPED
            : is_multisample ? VKD3DSIH_LD2DMS : VKD3DSIH_LD, resource);
    shader_instruction_set_texel_offset(ins, &operands[7]);
    for (i = 0; i < VKD3D_VEC4_SIZE; ++i)
        ins->resource_data_type[i] = resource->u.handle.resource->data_type;

    src_params = instruction_src_params_alloc(ins, 2 + is_multisample, sm6);
    src_param_init_vector_from_reg(&src_params[0], coord ? coord : &operands[4]->u.reg);
    src_param_init_vector_from_reg(&src_params[1], &resource->u.handle.reg);
    if (is_multisample)
        src_param_init_from_value(&src_params[2], mip_level_or_sample_count);

    instruction_dst_param_vector_init_ssa(sm6, ins, VKD3D_VEC4_SIZE);
}

static void sm6_parser_emit_dx_texture_store(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    const struct vkd3d_shader_register *coord, *texel;
    struct vkd3d_shader_src_param *src_params;
    struct vkd3d_shader_dst_param *dst_param;
    unsigned int write_mask, component_count;
    const struct sm6_value *resource;

    resource = operands[2];
    SM6_VALUE_VALIDATE_IS_HANDLE(resource);

    if ((coord = sm6_parser_emit_coordinate_construct(sm6, &operands[3], 3, NULL, ins)))
    {
        ++ins;
        ++code_block->instruction_count;
    }

    write_mask = sm6_value_get_constant_uint(operands[10]);
    component_count = vkd3d_write_mask_component_count(write_mask);
    if (write_mask > VKD3DSP_WRITEMASK_ALL)
        WARN("Invalid write mask %#x.\n", write_mask);
    texel = &operands[6]->u.reg;
    if (component_count > 1)
    {
        texel = sm6_parser_emit_composite_construct(sm6, &operands[6], component_count, ins);
        ++ins;
        ++code_block->instruction_count;
    }

    shader_instruction_init(ins, VKD3DSIH_STORE_UAV_TYPED);

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    src_param_init_vector_from_reg(&src_params[0], coord ? coord : &operands[3]->u.reg);
    src_param_init_vector_from_reg(&src_params[1], texel);

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init_with_mask(dst_param, write_mask);
    dst_param->reg = resource->u.handle.reg;
}

static enum vkd3d_shader_opcode sm6_dx_map_wave_bit_op(enum dxil_wave_bit_op_kind op)
{
    switch (op)
    {
        case WAVE_BIT_OP_AND:
            return VKD3DSIH_WAVE_ACTIVE_BIT_AND;
        case WAVE_BIT_OP_OR:
            return VKD3DSIH_WAVE_ACTIVE_BIT_OR;
        case WAVE_BIT_OP_XOR:
            return VKD3DSIH_WAVE_ACTIVE_BIT_XOR;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_wave_active_ballot(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;

    shader_instruction_init(ins, VKD3DSIH_WAVE_ACTIVE_BALLOT);
    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_vector_init_ssa(sm6, ins, VKD3D_VEC4_SIZE);
}

static void sm6_parser_emit_dx_wave_active_bit(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;
    enum dxil_wave_bit_op_kind wave_op;

    wave_op = sm6_value_get_constant_uint(operands[3]);
    shader_instruction_init(ins, sm6_dx_map_wave_bit_op(wave_op));

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static enum vkd3d_shader_opcode sm6_dx_map_wave_op(enum dxil_wave_op_kind op, bool is_signed, bool is_float)
{
    switch (op)
    {
        case WAVE_OP_ADD:
            return VKD3DSIH_WAVE_ACTIVE_OP_ADD;
        case WAVE_OP_MUL:
            return VKD3DSIH_WAVE_ACTIVE_OP_MUL;
        case WAVE_OP_MIN:
            if (is_float)
                return VKD3DSIH_WAVE_ACTIVE_OP_MIN;
            return is_signed ? VKD3DSIH_WAVE_ACTIVE_OP_IMIN : VKD3DSIH_WAVE_ACTIVE_OP_UMIN;
        case WAVE_OP_MAX:
            if (is_float)
                return VKD3DSIH_WAVE_ACTIVE_OP_MAX;
            return is_signed ? VKD3DSIH_WAVE_ACTIVE_OP_IMAX : VKD3DSIH_WAVE_ACTIVE_OP_UMAX;
        default:
            vkd3d_unreachable();
    }
}

static void sm6_parser_emit_dx_wave_active_op(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;
    enum dxil_wave_op_kind wave_op;
    bool is_signed;

    wave_op = sm6_value_get_constant_uint(operands[3]);
    is_signed = !sm6_value_get_constant_uint(operands[4]);
    shader_instruction_init(ins, sm6_dx_map_wave_op(wave_op, is_signed, sm6_type_is_floating_point(operands[2]->type)));

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_dx_wave_builtin(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    switch (op)
    {
        case DX_WAVE_GET_LANE_COUNT:
            sm6_parser_emit_dx_input_register_mov(sm6, ins, VKD3DSPR_WAVELANECOUNT, VKD3D_DATA_UINT);
            break;
        case DX_WAVE_GET_LANE_INDEX:
            sm6_parser_emit_dx_input_register_mov(sm6, ins, VKD3DSPR_WAVELANEINDEX, VKD3D_DATA_UINT);
            break;
        default:
            vkd3d_unreachable();
    }
}

static enum vkd3d_shader_opcode dx_map_wave_quad_op(enum dxil_wave_quad_op_kind op)
{
    switch (op)
    {
        case QUAD_OP_HORIZONTAL:
            return VKD3DSIH_WAVE_QUAD_OP_H;
        case QUAD_OP_VERTICAL:
            return VKD3DSIH_WAVE_QUAD_OP_V;
        case QUAD_OP_DIAGONAL:
            return VKD3DSIH_WAVE_QUAD_OP_D;
        default:
            FIXME("Unhandled quad op kind %u.\n", op);
            return VKD3DSIH_INVALID;
    }
}

static void sm6_parser_emit_dx_quad_op(struct sm6_parser *sm6, struct sm6_block *code_block,
        enum dx_intrinsic_opcode op, const struct sm6_value **operands, struct vkd3d_shader_instruction *ins)
{
    struct vkd3d_shader_src_param *src_param;
    enum dxil_wave_quad_op_kind quad_op;

    quad_op = sm6_value_get_constant_uint(operands[3]);
    shader_instruction_init(ins, dx_map_wave_quad_op(quad_op));
    if (ins->handler_idx == VKD3DSIH_INVALID)
        return;

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, operands[2]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

struct sm6_dx_opcode_info
{
    const char ret_type;
    const char *operand_info;
    void (*handler)(struct sm6_parser *, struct sm6_block *, enum dx_intrinsic_opcode,
            const struct sm6_value **, struct vkd3d_shader_instruction *);
};

/*
    1 -> int1
    8 -> int8
    i -> int32
    m -> int16/32/64
    n -> any numeric
    f -> float
    d -> double
    g -> float or double
    H -> handle
    D -> Dimensions
    S -> splitdouble
    V -> 4 x i32
    v -> void
    o -> overloaded
 */
static const struct sm6_dx_opcode_info sm6_dx_op_table[] =
{
    [DX_ACOS                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ASIN                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ATAN                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ATOMIC_BINOP                  ] = {'o', "Hiiiio", sm6_parser_emit_dx_atomic_binop},
    [DX_ATOMIC_CMP_XCHG               ] = {'o', "Hiiioo", sm6_parser_emit_dx_atomic_binop},
    [DX_BARRIER                       ] = {'v', "i", sm6_parser_emit_dx_barrier},
    [DX_BFREV                         ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_BUFFER_LOAD                   ] = {'o', "Hii", sm6_parser_emit_dx_buffer_load},
    [DX_BUFFER_STORE                  ] = {'v', "Hiioooo8", sm6_parser_emit_dx_buffer_store},
    [DX_BUFFER_UPDATE_COUNTER         ] = {'i', "H8", sm6_parser_emit_dx_buffer_update_counter},
    [DX_CALCULATE_LOD                 ] = {'f', "HHfff1", sm6_parser_emit_dx_calculate_lod},
    [DX_CBUFFER_LOAD_LEGACY           ] = {'o', "Hi", sm6_parser_emit_dx_cbuffer_load},
    [DX_COS                           ] = {'g', "g", sm6_parser_emit_dx_sincos},
    [DX_COUNT_BITS                    ] = {'i', "o", sm6_parser_emit_dx_unary},
    [DX_COVERAGE                      ] = {'i', "", sm6_parser_emit_dx_coverage},
    [DX_CREATE_HANDLE                 ] = {'H', "8ii1", sm6_parser_emit_dx_create_handle},
    [DX_CUT_STREAM                    ] = {'v', "8", sm6_parser_emit_dx_stream},
    [DX_DERIV_COARSEX                 ] = {'o', "o", sm6_parser_emit_dx_unary},
    [DX_DERIV_COARSEY                 ] = {'o', "o", sm6_parser_emit_dx_unary},
    [DX_DERIV_FINEX                   ] = {'o', "o", sm6_parser_emit_dx_unary},
    [DX_DERIV_FINEY                   ] = {'o', "o", sm6_parser_emit_dx_unary},
    [DX_DISCARD                       ] = {'v', "1", sm6_parser_emit_dx_discard},
    [DX_DOMAIN_LOCATION               ] = {'f', "8", sm6_parser_emit_dx_domain_location},
    [DX_DOT2                          ] = {'o', "oooo", sm6_parser_emit_dx_dot},
    [DX_DOT3                          ] = {'o', "oooooo", sm6_parser_emit_dx_dot},
    [DX_DOT4                          ] = {'o', "oooooooo", sm6_parser_emit_dx_dot},
    [DX_EMIT_STREAM                   ] = {'v', "8", sm6_parser_emit_dx_stream},
    [DX_EMIT_THEN_CUT_STREAM          ] = {'v', "8", sm6_parser_emit_dx_stream},
    [DX_EVAL_CENTROID                 ] = {'o', "ii8", sm6_parser_emit_dx_eval_attrib},
    [DX_EVAL_SAMPLE_INDEX             ] = {'o', "ii8i", sm6_parser_emit_dx_eval_attrib},
    [DX_EXP                           ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_FABS                          ] = {'g', "g", sm6_parser_emit_dx_fabs},
    [DX_FIRST_BIT_HI                  ] = {'i', "m", sm6_parser_emit_dx_unary},
    [DX_FIRST_BIT_LO                  ] = {'i', "m", sm6_parser_emit_dx_unary},
    [DX_FIRST_BIT_SHI                 ] = {'i', "m", sm6_parser_emit_dx_unary},
    [DX_FLATTENED_THREAD_ID_IN_GROUP  ] = {'i', "", sm6_parser_emit_dx_compute_builtin},
    [DX_FMA                           ] = {'g', "ggg", sm6_parser_emit_dx_ma},
    [DX_FMAD                          ] = {'g', "ggg", sm6_parser_emit_dx_ma},
    [DX_FMAX                          ] = {'g', "gg", sm6_parser_emit_dx_binary},
    [DX_FMIN                          ] = {'g', "gg", sm6_parser_emit_dx_binary},
    [DX_FRC                           ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_GET_DIMENSIONS                ] = {'D', "Hi", sm6_parser_emit_dx_get_dimensions},
    [DX_GROUP_ID                      ] = {'i', "i", sm6_parser_emit_dx_compute_builtin},
    [DX_HCOS                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_HSIN                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_HTAN                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_IMAD                          ] = {'m', "mmm", sm6_parser_emit_dx_ma},
    [DX_IMAX                          ] = {'m', "mm", sm6_parser_emit_dx_binary},
    [DX_IMIN                          ] = {'m', "mm", sm6_parser_emit_dx_binary},
    [DX_ISFINITE                      ] = {'1', "g", sm6_parser_emit_dx_unary},
    [DX_ISINF                         ] = {'1', "g", sm6_parser_emit_dx_unary},
    [DX_ISNAN                         ] = {'1', "g", sm6_parser_emit_dx_unary},
    [DX_LEGACY_F16TOF32               ] = {'f', "i", sm6_parser_emit_dx_unary},
    [DX_LEGACY_F32TOF16               ] = {'i', "f", sm6_parser_emit_dx_unary},
    [DX_LOAD_INPUT                    ] = {'o', "ii8i", sm6_parser_emit_dx_load_input},
    [DX_LOAD_OUTPUT_CONTROL_POINT     ] = {'o', "ii8i", sm6_parser_emit_dx_load_input},
    [DX_LOAD_PATCH_CONSTANT           ] = {'o', "ii8", sm6_parser_emit_dx_load_input},
    [DX_LOG                           ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_MAKE_DOUBLE                   ] = {'d', "ii", sm6_parser_emit_dx_make_double},
    [DX_OUTPUT_CONTROL_POINT_ID       ] = {'i', "", sm6_parser_emit_dx_output_control_point_id},
    [DX_PRIMITIVE_ID                  ] = {'i', "", sm6_parser_emit_dx_primitive_id},
    [DX_RAW_BUFFER_LOAD               ] = {'o', "Hii8i", sm6_parser_emit_dx_raw_buffer_load},
    [DX_RAW_BUFFER_STORE              ] = {'v', "Hiioooo8i", sm6_parser_emit_dx_raw_buffer_store},
    [DX_ROUND_NE                      ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ROUND_NI                      ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ROUND_PI                      ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_ROUND_Z                       ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_RSQRT                         ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_RT_GET_SAMPLE_POS             ] = {'o', "i", sm6_parser_emit_dx_get_sample_pos},
    [DX_RT_GET_SAMPLE_COUNT           ] = {'i', "", sm6_parser_emit_dx_get_sample_count},
    [DX_SAMPLE                        ] = {'o', "HHffffiiif", sm6_parser_emit_dx_sample},
    [DX_SAMPLE_B                      ] = {'o', "HHffffiiiff", sm6_parser_emit_dx_sample},
    [DX_SAMPLE_C                      ] = {'o', "HHffffiiiff", sm6_parser_emit_dx_sample},
    [DX_SAMPLE_C_LZ                   ] = {'o', "HHffffiiif", sm6_parser_emit_dx_sample},
    [DX_SAMPLE_GRAD                   ] = {'o', "HHffffiiifffffff", sm6_parser_emit_dx_sample},
    [DX_SAMPLE_INDEX                  ] = {'i', "", sm6_parser_emit_dx_sample_index},
    [DX_SAMPLE_LOD                    ] = {'o', "HHffffiiif", sm6_parser_emit_dx_sample},
    [DX_SATURATE                      ] = {'g', "g", sm6_parser_emit_dx_saturate},
    [DX_SIN                           ] = {'g', "g", sm6_parser_emit_dx_sincos},
    [DX_SPLIT_DOUBLE                  ] = {'S', "d", sm6_parser_emit_dx_split_double},
    [DX_SQRT                          ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_STORE_OUTPUT                  ] = {'v', "ii8o", sm6_parser_emit_dx_store_output},
    [DX_STORE_PATCH_CONSTANT          ] = {'v', "ii8o", sm6_parser_emit_dx_store_output},
    [DX_TAN                           ] = {'g', "g", sm6_parser_emit_dx_unary},
    [DX_TEXTURE_GATHER                ] = {'o', "HHffffiii", sm6_parser_emit_dx_texture_gather},
    [DX_TEXTURE_GATHER_CMP            ] = {'o', "HHffffiiif", sm6_parser_emit_dx_texture_gather},
    [DX_TEX2DMS_GET_SAMPLE_POS        ] = {'o', "Hi", sm6_parser_emit_dx_get_sample_pos},
    [DX_TEXTURE_LOAD                  ] = {'o', "Hiiiiiii", sm6_parser_emit_dx_texture_load},
    [DX_TEXTURE_STORE                 ] = {'v', "Hiiioooo8", sm6_parser_emit_dx_texture_store},
    [DX_THREAD_ID                     ] = {'i', "i", sm6_parser_emit_dx_compute_builtin},
    [DX_THREAD_ID_IN_GROUP            ] = {'i', "i", sm6_parser_emit_dx_compute_builtin},
    [DX_UMAD                          ] = {'m', "mmm", sm6_parser_emit_dx_ma},
    [DX_UMAX                          ] = {'m', "mm", sm6_parser_emit_dx_binary},
    [DX_UMIN                          ] = {'m', "mm", sm6_parser_emit_dx_binary},
    [DX_WAVE_ACTIVE_ALL_EQUAL         ] = {'1', "n", sm6_parser_emit_dx_unary},
    [DX_WAVE_ACTIVE_BALLOT            ] = {'V', "1", sm6_parser_emit_dx_wave_active_ballot},
    [DX_WAVE_ACTIVE_BIT               ] = {'o', "o8", sm6_parser_emit_dx_wave_active_bit},
    [DX_WAVE_ACTIVE_OP                ] = {'o', "o88", sm6_parser_emit_dx_wave_active_op},
    [DX_WAVE_ALL_BIT_COUNT            ] = {'i', "1", sm6_parser_emit_dx_unary},
    [DX_WAVE_ALL_TRUE                 ] = {'1', "1", sm6_parser_emit_dx_unary},
    [DX_WAVE_ANY_TRUE                 ] = {'1', "1", sm6_parser_emit_dx_unary},
    [DX_WAVE_GET_LANE_COUNT           ] = {'i', "", sm6_parser_emit_dx_wave_builtin},
    [DX_WAVE_GET_LANE_INDEX           ] = {'i', "", sm6_parser_emit_dx_wave_builtin},
    [DX_WAVE_IS_FIRST_LANE            ] = {'1', "", sm6_parser_emit_dx_void},
    [DX_WAVE_PREFIX_BIT_COUNT         ] = {'i', "1", sm6_parser_emit_dx_unary},
    [DX_WAVE_QUAD_OP                  ] = {'o', "o8", sm6_parser_emit_dx_quad_op},
    [DX_WAVE_READ_LANE_AT             ] = {'o', "oi", sm6_parser_emit_dx_binary},
    [DX_WAVE_READ_LANE_FIRST          ] = {'o', "o", sm6_parser_emit_dx_unary},
};

static bool sm6_parser_validate_operand_type(struct sm6_parser *sm6, const struct sm6_type *type, char info_type)
{
    switch (info_type)
    {
        case 0:
            FIXME("Invalid operand count.\n");
            return false;
        case '1':
            return sm6_type_is_bool(type);
        case '8':
            return sm6_type_is_i8(type);
        case 'i':
            return sm6_type_is_i32(type);
        case 'm':
            return sm6_type_is_wil(type);
        case 'n':
            return sm6_type_is_numeric(type);
        case 'f':
            return sm6_type_is_float(type);
        case 'd':
            return sm6_type_is_double(type);
        case 'g':
            return sm6_type_is_floating_point(type);
        case 'H':
            return type == sm6->handle_type;
        case 'D':
            return sm6_type_is_struct(type) && !strcmp(type->u.struc->name, "dx.types.Dimensions");
        case 'S':
            return sm6_type_is_struct(type) && !strcmp(type->u.struc->name, "dx.types.splitdouble");
        case 'V':
            return sm6_type_is_struct(type) && !strcmp(type->u.struc->name, "dx.types.fouri32");
        case 'v':
            return !type;
        case 'o':
            /* TODO: some type checking may be possible */
            return true;
        default:
            FIXME("Unhandled operand code '%c'.\n", info_type);
            return false;
    }
}

static void sm6_parser_validate_dx_op(struct sm6_parser *sm6, enum dx_intrinsic_opcode code,
        const struct sm6_value **operands, unsigned int operand_count, struct sm6_value *dst)
{
    const struct sm6_dx_opcode_info *info;
    unsigned int i;

    info = &sm6_dx_op_table[code];

    if (!sm6_parser_validate_operand_type(sm6, dst->type, info->ret_type))
        WARN("Failed to validate return type for code %u.\n", code);

    for (i = 2; i < operand_count; ++i)
    {
        const struct sm6_value *op = operands[i];
        if ((!sm6_value_is_register(op) && !sm6_value_is_handle(op))
                || !sm6_parser_validate_operand_type(sm6, op->type, info->operand_info[i - 2]))
            WARN("Failed to validate operand %u for dx intrinsic code %u.\n", i - 1, code);
    }
}

static void sm6_parser_emit_unhandled(struct sm6_parser *sm6, struct vkd3d_shader_instruction *ins,
        struct sm6_value *dst)
{
    const struct sm6_type *type;

    ins->handler_idx = VKD3DSIH_NOP;

    if (!dst->type)
        return;

    type = sm6_type_get_pseudo_vec4_element_type(dst->type);
    if (!type)
        type = sm6_type_get_scalar_type(dst->type, 0);

    register_init_with_id(&dst->u.reg, VKD3DSPR_UNDEF, vkd3d_data_type_from_sm6_type(type), sm6->value_count);
    /* dst->is_undefined is not set here because it flags unused parameters in dx intrinsic calls. */
}

static void sm6_parser_decode_dx_op(struct sm6_parser *sm6, struct sm6_block *code_block, enum dx_intrinsic_opcode op,
        const char *name, const struct sm6_value **operands, unsigned int operand_count,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    if (op >= ARRAY_SIZE(sm6_dx_op_table) || !sm6_dx_op_table[op].operand_info)
    {
        FIXME("Unhandled dx intrinsic function id %u, '%s'.\n", op, name);
        return sm6_parser_emit_unhandled(sm6, ins, dst);
    }
    sm6_parser_validate_dx_op(sm6, op, operands, operand_count, dst);
    sm6_dx_op_table[op].handler(sm6, code_block, op, operands, ins);
}

static void sm6_parser_emit_call(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_block *code_block, struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    const struct sm6_value *operands[DXIL_OP_MAX_OPERANDS];
    const struct sm6_value *fn_value, *op_value;
    unsigned int i = 1, j, operand_count;
    const struct sm6_type *type = NULL;
    uint64_t call_conv;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    /* TODO: load the 1-based attributes index from record->operands[0] and validate against attribute count. */

    if ((call_conv = record->operands[i++]) & (1u << 15))
    {
        type = sm6_parser_get_type(sm6, record->operands[i++]);
        DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(4);
    }

    if (!(fn_value = sm6_parser_get_value_by_ref(sm6, record, NULL, &i)))
        return;
    if (!sm6_value_is_function_dcl(fn_value))
    {
        WARN("Target function value is not a function declaration.\n");
        return;
    }

    if (!type)
    {
        type = fn_value->type->u.pointer.type;
    }
    else if (type != fn_value->type->u.pointer.type)
    {
        WARN("Explicit call type does not match function type.\n");
        type = fn_value->type->u.pointer.type;
    }

    if (!sm6_type_is_void(type->u.function->ret_type))
        dst->type = type->u.function->ret_type;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(i + type->u.function->param_count);

    for (j = 0; j < type->u.function->param_count; ++j)
    {
        if (!(operands[j + 1] = sm6_parser_get_value_by_ref(sm6, record, type->u.function->param_types[j], &i)))
            return;
    }
    operand_count = j + 1;
    assert(operand_count <= ARRAY_SIZE(operands));

    if (fn_value->u.function.is_prototype)
    {
        op_value = operands[1];
        if (!sm6_value_is_constant(op_value) || !sm6_type_is_integer(op_value->type))
        {
            WARN("dx intrinsic function id is not a constant int.\n");
            return;
        }
        if (!sm6_value_is_dx_intrinsic_dcl(fn_value))
            WARN("External function is not a dx intrinsic.\n");

        sm6_parser_decode_dx_op(sm6, code_block, register_get_uint_value(&op_value->u.reg),
                fn_value->u.function.name, operands, operand_count, ins, dst);
    }
    else
    {
        FIXME("Unhandled call to local function.\n");
    }
}

static enum vkd3d_shader_opcode sm6_map_cast_op(uint64_t code, const struct sm6_type *from,
        const struct sm6_type *to)
{
    enum vkd3d_shader_opcode op;
    bool from_int, to_int;
    bool is_valid = true;

    /* NOTE: DXIL currently doesn't use vectors here. */
    if (!sm6_type_is_scalar(from) || !sm6_type_is_scalar(to))
    {
        FIXME("Unhandled non-scalar type class %u and/or %u.\n", from->class, to->class);
        /* width is not available for non-scalar types so we can't attempt to compile it. */
        return VKD3DSIH_INVALID;
    }
    if (to->u.width == 8 || from->u.width == 8)
    {
        FIXME("Unhandled 8-bit value.\n");
        return VKD3DSIH_INVALID;
    }

    from_int = sm6_type_is_integer(from);
    to_int = sm6_type_is_integer(to);

    switch (code)
    {
        case CAST_TRUNC:
            if (to->u.width == 16 && from->u.width == 32)
                return VKD3DSIH_NOP;
            assert(to->u.width == 32 && from->u.width == 64);
            op = VKD3DSIH_U64TOU;
            break;
        case CAST_ZEXT:
            /* nop or min precision. TODO: native 16-bit */
            if (to->u.width == from->u.width || (to->u.width == 32 && from->u.width == 16))
                op = VKD3DSIH_NOP;
            else
                op = (from->u.width == 1) ? VKD3DSIH_BTOU : VKD3DSIH_UTOU64;
            is_valid = sm6_type_is_1wil(from) && to_int;
            break;
        case CAST_SEXT:
            /* nop or min precision. TODO: native 16-bit */
            if (to->u.width == from->u.width || (to->u.width == 32 && from->u.width == 16))
                op = VKD3DSIH_NOP;
            else
                op = (from->u.width == 1) ? VKD3DSIH_BTOI : VKD3DSIH_ITOI64;
            is_valid = sm6_type_is_1wil(from) && to_int;
            break;
        case CAST_FPTOUI:
            op = (from->u.width == 64) ? VKD3DSIH_DTOU : VKD3DSIH_FTOU;
            is_valid = !from_int && to_int;
            break;
        case CAST_FPTOSI:
            op = (from->u.width == 64) ? VKD3DSIH_DTOI : VKD3DSIH_FTOI;
            is_valid = !from_int && to_int;
            break;
        case CAST_UITOFP:
            op = (to->u.width == 64) ? VKD3DSIH_UTOD : VKD3DSIH_UTOF;
            is_valid = from_int && !to_int;
            break;
        case CAST_SITOFP:
            op = (to->u.width == 64) ? VKD3DSIH_ITOD : VKD3DSIH_ITOF;
            is_valid = from_int && !to_int;
            break;
        case CAST_FPTRUNC:
            op = (from->u.width == 64) ? VKD3DSIH_DTOF : VKD3DSIH_NOP;
            is_valid = !from_int && !to_int;
            break;
        case CAST_FPEXT:
            op = (to->u.width == 64) ? VKD3DSIH_FTOD : VKD3DSIH_NOP;
            is_valid = !from_int && !to_int;
            break;
        case CAST_BITCAST:
            op = VKD3DSIH_BITCAST;
            break;
        default:
            FIXME("Unhandled cast op %"PRIu64".\n", code);
            op = VKD3DSIH_INVALID;
            break;
    }

    if (!is_valid)
        FIXME("Invalid types %u and/or %u for op %"PRIu64".\n", from->class, to->class, code);

    return op;
}

static void sm6_parser_emit_cast(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_src_param *src_param;
    enum vkd3d_shader_opcode handler_idx;
    const struct sm6_value *value;
    const struct sm6_type *type;
    unsigned int idx = 0;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    if (!(value = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 2);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 2);

    if (!(type = sm6_parser_get_type(sm6, record->operands[idx++])))
        return;

    dst->type = type;

    if (sm6_type_is_pointer(type))
    {
        *dst = *value;
        dst->type = type;
        shader_instruction_init(ins, VKD3DSIH_NOP);
        return;
    }

    handler_idx = sm6_map_cast_op(record->operands[idx], value->type, type);

    shader_instruction_init(ins, handler_idx);

    if (handler_idx == VKD3DSIH_INVALID)
        return;
    if (handler_idx == VKD3DSIH_NOP)
    {
        dst->u.reg = value->u.reg;
        return;
    }

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_from_value(src_param, value);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

struct sm6_cmp_info
{
    enum vkd3d_shader_opcode handler_idx_32;
    enum vkd3d_shader_opcode handler_idx_64;
    bool src_swap;
    unsigned int flags;
};

static const struct sm6_cmp_info sm6_cmp_op_table[] =
{
    [FCMP_FALSE] = {VKD3DSIH_INVALID, VKD3DSIH_INVALID},
    [FCMP_OEQ]   = {VKD3DSIH_EQ, VKD3DSIH_DEQ},
    [FCMP_OGT]   = {VKD3DSIH_LT, VKD3DSIH_DLT, true},
    [FCMP_OGE]   = {VKD3DSIH_GE, VKD3DSIH_DGE},
    [FCMP_OLT]   = {VKD3DSIH_LT, VKD3DSIH_DLT},
    [FCMP_OLE]   = {VKD3DSIH_GE, VKD3DSIH_DGE, true},
    [FCMP_ONE]   = {VKD3DSIH_NE, VKD3DSIH_DNE, false, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_ORD]   = {VKD3DSIH_INVALID, VKD3DSIH_INVALID},
    [FCMP_UNO]   = {VKD3DSIH_INVALID, VKD3DSIH_INVALID},
    [FCMP_UEQ]   = {VKD3DSIH_EQ, VKD3DSIH_DEQ, false, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_UGT]   = {VKD3DSIH_LT, VKD3DSIH_DLT, true, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_UGE]   = {VKD3DSIH_GE, VKD3DSIH_DGE, false, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_ULT]   = {VKD3DSIH_LT, VKD3DSIH_DLT, false, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_ULE]   = {VKD3DSIH_GE, VKD3DSIH_DGE, true, VKD3DSI_FCMP_INVERT_ORDERED},
    [FCMP_UNE]   = {VKD3DSIH_NE, VKD3DSIH_DNE},
    [FCMP_TRUE]  = {VKD3DSIH_INVALID, VKD3DSIH_INVALID},

    [ICMP_EQ]    = {VKD3DSIH_IEQ, VKD3DSIH_INVALID},
    [ICMP_NE]    = {VKD3DSIH_INE, VKD3DSIH_INVALID},
    [ICMP_UGT]   = {VKD3DSIH_ULT, VKD3DSIH_INVALID, true},
    [ICMP_UGE]   = {VKD3DSIH_UGE, VKD3DSIH_INVALID},
    [ICMP_ULT]   = {VKD3DSIH_ULT, VKD3DSIH_INVALID},
    [ICMP_ULE]   = {VKD3DSIH_UGE, VKD3DSIH_INVALID, true},
    [ICMP_SGT]   = {VKD3DSIH_ILT, VKD3DSIH_INVALID, true},
    [ICMP_SGE]   = {VKD3DSIH_IGE, VKD3DSIH_INVALID},
    [ICMP_SLT]   = {VKD3DSIH_ILT, VKD3DSIH_INVALID},
    [ICMP_SLE]   = {VKD3DSIH_IGE, VKD3DSIH_INVALID, true},
};

static void sm6_parser_emit_cmp2(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_type *type_a, *type_b;
    enum vkd3d_shader_opcode handler_idx;
    const struct sm6_cmp_info *cmp;
    const struct sm6_value *a, *b;
    unsigned int idx = 0;
    uint64_t op;
    bool is_int;

    if (!(dst->type = sm6->bool_type))
    {
        FIXME("Bool type not found.\n");
        return;
    }

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    a = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx);
    b = sm6_parser_get_value_by_ref(sm6, record, a->type, &idx);
    if (!a || !b)
        return;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 1);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 2);

    type_a = a->type;
    type_b = b->type;
    is_int = sm6_type_is_1wil(type_a);

    if (!is_int && !sm6_type_is_floating_point(type_a))
    {
        FIXME("Argument type %u is not bool, int16/32/64 or floating point.\n", type_a->class);
        return;
    }

    op = record->operands[idx++];

    if (is_int != (op >= ICMP_EQ))
        WARN("Invalid type %u for operation %"PRIu64".\n", type_a->class, op);
    if (type_a != type_b)
        WARN("Type mismatch.\n");

    if (record->operand_count > idx)
    {
        if (sm6_type_is_floating_point(a->type))
            TRACE("Ignoring fast FP modifier %#"PRIx64".\n", record->operands[idx]);
        else
            FIXME("Unexpected extra value %#"PRIx64".\n", record->operands[idx]);
    }

    handler_idx = VKD3DSIH_INVALID;
    if (op < ARRAY_SIZE(sm6_cmp_op_table))
    {
        cmp = &sm6_cmp_op_table[op];
        handler_idx = (!is_int && type_a->u.width == 64) ? cmp->handler_idx_64 : cmp->handler_idx_32;
    }
    if (!handler_idx || handler_idx == VKD3DSIH_INVALID)
    {
        FIXME("Unhandled comparison op %"PRIu64".\n", op);
        return;
    }
    shader_instruction_init(ins, handler_idx);
    ins->flags = cmp->flags;

    src_params = instruction_src_params_alloc(ins, 2, sm6);
    if (cmp->src_swap)
    {
        src_param_init_from_value(&src_params[0], b);
        src_param_init_from_value(&src_params[1], a);
    }
    else
    {
        src_param_init_from_value(&src_params[0], a);
        src_param_init_from_value(&src_params[1], b);
    }

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_cmpxchg(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    uint64_t success_ordering, failure_ordering;
    struct vkd3d_shader_dst_param *dst_params;
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *ptr, *cmp, *new;
    const struct sm6_type *type;
    unsigned int idx = 0;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(6);

    if (!(ptr = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_POINTER_TO_I32(ptr);
    if (ptr->u.reg.type != VKD3DSPR_GROUPSHAREDMEM)
    {
        FIXME("Only groupshared variables are supported.\n");
        return;
    }

    if (!(dst->type = sm6_type_get_cmpxchg_result_struct(sm6)))
    {
        FIXME("Failed to find result struct.\n");
        return;
    }

    type = ptr->type->u.pointer.type;
    cmp = sm6_parser_get_value_by_ref(sm6, record, type, &idx);
    new = sm6_parser_get_value_by_ref(sm6, record, type, &idx);
    if (!cmp || !new)
        return;

    SM6_VALUE_VALIDATE_IS_I32(cmp);
    SM6_VALUE_VALIDATE_IS_I32(new);

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 3);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 5);

    if (record->operands[idx++])
        TRACE("Ignoring volatile modifier.\n");

    success_ordering = record->operands[idx++];

    if (record->operands[idx++] != 1)
        WARN("Ignoring synchronisation scope %"PRIu64".\n", record->operands[idx - 1]);

    failure_ordering = (record->operand_count > idx ) ? record->operands[idx++] : success_ordering;

    /* It's currently not possible to specify an atomic ordering in HLSL, and it defaults to seq_cst. */
    if (success_ordering != ORDERING_SEQCST)
        FIXME("Unhandled success ordering %"PRIu64".\n", success_ordering);
    if (success_ordering != failure_ordering)
        FIXME("Unhandled failure ordering %"PRIu64".\n", failure_ordering);

    if (record->operand_count > idx && record->operands[idx])
        FIXME("Ignoring weak cmpxchg.\n");

    shader_instruction_init(ins, VKD3DSIH_IMM_ATOMIC_CMP_EXCH);
    src_params = instruction_src_params_alloc(ins, 3, sm6);
    src_param_make_constant_uint(&src_params[0], 0);
    src_param_init_from_value(&src_params[1], cmp);
    src_param_init_from_value(&src_params[2], new);

    dst_params = instruction_dst_params_alloc(ins, 2, sm6);
    dst_param_init(&dst_params[0]);
    register_init_ssa_scalar(&dst_params[0].reg, dst->type, dst, sm6);
    dst_param_init(&dst_params[1]);
    dst_params[1].reg = ptr->u.reg;

    dst->u.reg = dst_params[0].reg;
}

static void sm6_parser_emit_extractval(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_src_param *src_param;
    unsigned int elem_idx, idx = 0;
    const struct sm6_type *type;
    const struct sm6_value *src;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(2);

    if (!(src = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 1);

    if (record->operand_count > idx + 1)
    {
        FIXME("Unhandled multiple indices.\n");
        return;
    }
    if (record->operands[idx] >= VKD3D_VEC4_SIZE)
    {
        FIXME("Unhandled element index %"PRIu64".\n", record->operands[idx]);
        return;
    }
    elem_idx = record->operands[idx];

    type = src->type;
    if (!sm6_type_is_aggregate(type))
    {
        FIXME("Invalid extraction from non-aggregate.\n");
        return;
    }
    if (!(type = sm6_type_get_element_type_at_index(type, elem_idx)))
    {
        FIXME("Invalid element index %u.\n", elem_idx);
        return;
    }
    if (!sm6_type_is_scalar(type))
    {
        FIXME("Value is not of scalar type.\n");
        return;
    }

    dst->type = type;

    shader_instruction_init(ins, VKD3DSIH_MOV);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init_scalar(src_param, elem_idx);
    src_param->reg = src->u.reg;

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_gep(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    const struct sm6_type *type, *pointee_type;
    unsigned int i, idx = 2, operand, value;
    enum bitcode_address_space addr_space;
    struct vkd3d_shader_register *reg;
    struct sm6_value *operands[2];
    struct sm6_value *elem_value;
    const struct sm6_value *src;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(4);

    if (record->operands[0])
        TRACE("Ignoring inbounds flag.\n");
    if (!(type = sm6_parser_get_type(sm6, record->operands[1])))
        return;

    if (!(src = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_REGISTER(src);
    SM6_VALUE_VALIDATE_IS_POINTER(src);

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 1);

    if ((pointee_type = src->type->u.pointer.type) != type)
        WARN("Type mismatch.\n");
    addr_space = src->type->u.pointer.addr_space;

    if (record->operand_count > 5)
    {
        FIXME("Multiple element indices are not implemented.\n");
        return;
    }

    for (type = src->type, i = 0; idx < record->operand_count; ++i)
    {
        if ((operand = sm6_parser_get_value_idx_by_ref(sm6, record, NULL, &idx)) == UINT_MAX)
            return;

        elem_value = &sm6->values[operand];
        value = sm6_value_is_constant(elem_value) ? sm6_value_get_constant_uint(elem_value) : UINT_MAX;
        if (!(type = sm6_type_get_element_type_at_index(type, value)))
            WARN("Failed to get element type at index %u.\n", value);
        if (!i && value)
            WARN("Pointer dereference offset is not zero.\n");

        operands[i] = elem_value;
    }

    if (!(dst->type = sm6_type_get_pointer_to_type(type, addr_space, sm6)))
    {
        FIXME("Failed to get pointer type for type %u.\n", type->class);
        return;
    }

    reg = &dst->u.reg;
    *reg = src->u.reg;
    reg->idx[1].offset = 0;
    register_address_init(reg, operands[1], 1, sm6);
    reg->idx_count = 2;
    dst->structure_stride = src->structure_stride;

    ins->handler_idx = VKD3DSIH_NOP;
}

static void sm6_parser_emit_load(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    const struct sm6_type *elem_type = NULL, *pointee_type;
    unsigned int alignment, operand_count, idx = 0;
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *ptr;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    if (!(ptr = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_POINTER(ptr);

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 2);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 3);

    if (record->operand_count > idx + 2 && !(elem_type = sm6_parser_get_type(sm6, record->operands[idx++])))
        return;

    if (!elem_type)
        elem_type = ptr->type->u.pointer.type;
    else if (elem_type != (pointee_type = ptr->type->u.pointer.type))
        WARN("Type mismatch.\n");

    dst->type = elem_type;

    SM6_VALUE_VALIDATE_IS_NUMERIC(dst);

    if (!bitcode_parse_alignment(record->operands[idx++], &alignment))
        WARN("Invalid alignment %"PRIu64".\n", record->operands[idx - 1]);
    if (!sm6_type_alignment_is_default(elem_type, alignment))
        WARN("Ignoring alignment %u.\n", alignment);

    if (record->operands[idx])
        WARN("Ignoring volatile modifier.\n");

    if (ptr->structure_stride)
    {
        assert(ptr->u.reg.type == VKD3DSPR_GROUPSHAREDMEM);
        shader_instruction_init(ins, VKD3DSIH_LD_STRUCTURED);

        src_params = instruction_src_params_alloc(ins, 3, sm6);
        if (ptr->u.reg.idx[1].rel_addr)
            src_params[0] = *ptr->u.reg.idx[1].rel_addr;
        else
            src_param_make_constant_uint(&src_params[0], ptr->u.reg.idx[1].offset);
        src_param_make_constant_uint(&src_params[1], 0);
        src_param_init_from_value(&src_params[2], ptr);
    }
    else
    {
        operand_count = 1 + (ptr->u.reg.type == VKD3DSPR_GROUPSHAREDMEM);
        shader_instruction_init(ins, operand_count > 1 ? VKD3DSIH_LD_RAW : VKD3DSIH_MOV);

        src_params = instruction_src_params_alloc(ins, operand_count, sm6);
        if (operand_count > 1)
            src_param_make_constant_uint(&src_params[0], 0);
        src_param_init_from_value(&src_params[operand_count - 1], ptr);
    }

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static void sm6_parser_emit_phi(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_function *function, struct sm6_block *code_block, struct vkd3d_shader_instruction *ins,
        struct sm6_value *dst)
{
    const struct sm6_type *type;
    const struct sm6_value *src;
    struct sm6_phi *phi;
    unsigned int i, j;
    uint64_t src_idx;

    if (record->operand_count < 3 || !(record->operand_count & 1))
    {
        WARN("Invalid operand count %u.\n", record->operand_count);
        return;
    }

    if (!(type = sm6_parser_get_type(sm6, record->operands[0])))
        return;
    if (!sm6_type_is_numeric(type))
    {
        /* dxc doesn't seem to use buffer/resource read return types here.
         * NOTE: Supporting vectors would also require support in emit_temps(). */
        FIXME("Vector types are not supported.\n");
        return;
    }

    dst->type = type;
    register_init_ssa_scalar(&dst->u.reg, type, dst, sm6);

    phi = sm6_block_phi_new(code_block, &dst->u.reg);
    phi->incoming_count = record->operand_count / 2u;

    if (!vkd3d_array_reserve((void **)&phi->incoming, &phi->incoming_capacity, phi->incoming_count,
            sizeof(phi->incoming[0])))
        return;

    for (i = 1; i < record->operand_count; i += 2)
    {
        src_idx = sm6->value_count - decode_rotated_signed_value(record->operands[i]);
        if (src_idx >= sm6->cur_max_value)
        {
            FIXME("Invalid value index %"PRIu64".\n", src_idx);
            return;
        }

        src = sm6_parser_get_value(sm6, src_idx);
        if (src->type && src->type != type)
            WARN("Type mismatch.\n");

        j = i / 2u;
        phi->incoming[j].reg.idx[0].offset = src_idx;
        phi->incoming[j].reg.idx_count = 1;

        if (record->operands[i + 1] >= function->block_count)
        {
            FIXME("Invalid block index %"PRIu64".\n", record->operands[i + 1]);
            return;
        }
        phi->incoming[j].block = sm6_function_get_block(function, record->operands[i + 1]);
    }

    ins->handler_idx = VKD3DSIH_NOP;
}

static void sm6_parser_emit_ret(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_block *code_block, struct vkd3d_shader_instruction *ins)
{
    DXIL_RECORD_VALIDATE_OPERAND_COUNT(0, 1);

    if (record->operand_count)
        FIXME("Non-void return is not implemented.\n");

    code_block->terminator.type = TERMINATOR_RET;

    ins->handler_idx = VKD3DSIH_NOP;
}

static void sm6_parser_emit_store(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    bool is_atomic = record->code == FUNC_CODE_INST_STOREATOMIC;
    unsigned int idx = 0, alignment, operand_count;
    struct vkd3d_shader_src_param *src_params;
    struct vkd3d_shader_dst_param *dst_param;
    const struct sm6_type *pointee_type;
    const struct sm6_value *ptr, *src;

    if (is_atomic)
    {
        FIXME("Atomic store is not implemented yet.\n");
        return;
    }

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(4);

    if (!(ptr = sm6_parser_get_value_by_ref(sm6, record, NULL, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_REGISTER(ptr);
    SM6_VALUE_VALIDATE_IS_POINTER(ptr);

    pointee_type = ptr->type->u.pointer.type;
    if (!(src = sm6_parser_get_value_by_ref(sm6, record, pointee_type, &idx)))
        return;
    SM6_VALUE_VALIDATE_IS_NUMERIC(src);

    if (pointee_type != src->type)
        WARN("Type mismatch.\n");

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(idx + 2);
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(idx + 2);

    if (!bitcode_parse_alignment(record->operands[idx++], &alignment))
        WARN("Invalid alignment %"PRIu64".\n", record->operands[idx - 1]);
    if (!sm6_type_alignment_is_default(pointee_type, alignment))
        WARN("Ignoring alignment %u.\n", alignment);

    if (record->operands[idx])
        WARN("Ignoring volatile modifier.\n");

    if (ptr->structure_stride)
    {
        assert(ptr->u.reg.type == VKD3DSPR_GROUPSHAREDMEM);
        shader_instruction_init(ins, VKD3DSIH_STORE_STRUCTURED);

        src_params = instruction_src_params_alloc(ins, 3, sm6);
        if (ptr->u.reg.idx[1].rel_addr)
            src_params[0] = *ptr->u.reg.idx[1].rel_addr;
        else
            src_param_make_constant_uint(&src_params[0], ptr->u.reg.idx[1].offset);
        src_param_make_constant_uint(&src_params[1], 0);
        src_param_init_from_value(&src_params[2], src);
    }
    else
    {
        operand_count = 1 + (ptr->u.reg.type == VKD3DSPR_GROUPSHAREDMEM);
        shader_instruction_init(ins, operand_count > 1 ? VKD3DSIH_STORE_RAW : VKD3DSIH_MOV);

        src_params = instruction_src_params_alloc(ins, operand_count, sm6);
        if (operand_count > 1)
            src_param_make_constant_uint(&src_params[0], 0);
        src_param_init_from_value(&src_params[operand_count - 1], src);
    }

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init(dst_param);
    dst_param->reg = ptr->u.reg;
    if (dst_param->reg.type != VKD3DSPR_IDXTEMP)
    {
        dst_param->reg.idx[1].offset = UINT_MAX;
        dst_param->reg.idx[1].rel_addr = NULL;
        dst_param->reg.idx_count = 1;
    }
}

static void sm6_parser_emit_switch(struct sm6_parser *sm6, const struct dxil_record *record,
        struct sm6_function *function, struct sm6_block *code_block, struct vkd3d_shader_instruction *ins)
{
    struct sm6_block_terminator *terminator = &code_block->terminator;
    const struct sm6_type *type;
    const struct sm6_value *src;
    unsigned int i = 1, j;
    uint64_t index;

    if (record->operand_count < 3 || !(record->operand_count & 1))
    {
        WARN("Invalid operand count %u.\n", record->operand_count);
        return;
    }

    if (!(type = sm6_parser_get_type(sm6, record->operands[0])))
        return;

    if (!(src = sm6_parser_get_value_by_ref(sm6, record, type, &i)))
        return;
    SM6_VALUE_VALIDATE_IS_REGISTER(src);
    terminator->conditional_reg = src->u.reg;
    terminator->type = TERMINATOR_SWITCH;

    terminator->case_count = record->operand_count / 2u;
    if (!(terminator->cases = vkd3d_calloc(terminator->case_count, sizeof(*terminator->cases))))
    {
        ERR("Failed to allocate case array.\n");
        return;
    }

    for (; i < record->operand_count; i += 2)
    {
        if ((index = record->operands[i]) >= function->block_count)
        {
            FIXME("Invalid block number %"PRIu64".\n", index);
            return;
        }
        j = i / 2u - 1;
        terminator->cases[j].node = function->blocks[index];
        sm6_block_init_succ(code_block, terminator->cases[j].node);
        terminator->cases[j].is_default = !j;
    }

    for (i = 3; i < record->operand_count; i += 2)
    {
        if (!(src = sm6_parser_get_value_unsafe(sm6, record->operands[i])))
        {
            FIXME("Invalid value index %"PRIu64".\n", record->operands[i]);
            return;
        }

        if (src->type != type)
            WARN("Type mismatch.\n");
        if (!sm6_value_is_constant(src))
            WARN("Case value is not a constant.\n");

        terminator->cases[i / 2u].value = sm6_value_get_constant_uint(src);
    }

    ins->handler_idx = VKD3DSIH_NOP;
}

static void sm6_parser_emit_vselect(struct sm6_parser *sm6, const struct dxil_record *record,
        struct vkd3d_shader_instruction *ins, struct sm6_value *dst)
{
    struct vkd3d_shader_src_param *src_params;
    const struct sm6_value *src[3];
    unsigned int i = 0;

    DXIL_RECORD_VALIDATE_OPERAND_MIN_COUNT(3);

    if (!(src[1] = sm6_parser_get_value_by_ref(sm6, record, NULL, &i))
            || !(src[2] = sm6_parser_get_value_by_ref(sm6, record, src[1]->type, &i))
            || !(src[0] = sm6_parser_get_value_by_ref(sm6, record, NULL, &i)))
        return;
    DXIL_RECORD_VALIDATE_OPERAND_MAX_COUNT(i);

    for (i = 0; i < 3; ++i)
        SM6_VALUE_VALIDATE_IS_REGISTER(src[i]);

    dst->type = src[1]->type;

    SM6_VALUE_VALIDATE_IS_BOOL(src[0]);

    shader_instruction_init(ins, VKD3DSIH_MOVC);

    src_params = instruction_src_params_alloc(ins, 3, sm6);
    for (i = 0; i < 3; ++i)
        src_param_init_from_value(&src_params[i], src[i]);

    instruction_dst_param_init_ssa_scalar(ins, sm6);
}

static const struct sm6_metadata_value *sm6_parser_find_metadata_kind(
        struct sm6_parser *sm6, uint64_t id)
{
    unsigned int i, j;

    for (i = 0; i < ARRAY_SIZE(sm6->metadata_tables); ++i)
    {
        for (j = 0; j < sm6->metadata_counts[i]; ++j)
            if (sm6->metadata_tables[i][j].type == VKD3D_METADATA_KIND
                    && sm6->metadata_tables[i][j].u.kind.id == id)
                return &sm6->metadata_tables[i][j];
    }

    return NULL;
}

static bool sm6_metadata_value_is_node(const struct sm6_metadata_value *m)
{
    return m && m->type == VKD3D_METADATA_NODE;
}

static bool sm6_metadata_value_is_value(const struct sm6_metadata_value *m)
{
    return m && m->type == VKD3D_METADATA_VALUE;
}

static bool sm6_metadata_value_is_string(const struct sm6_metadata_value *m)
{
    return m && m->type == VKD3D_METADATA_STRING;
}

static bool sm6_metadata_get_uint_value(const struct sm6_parser *sm6,
        const struct sm6_metadata_value *m, unsigned int *u)
{
    const struct sm6_value *value;

    if (!m || m->type != VKD3D_METADATA_VALUE)
        return false;

    value = m->u.value;
    if (!sm6_value_is_constant(value))
        return false;
    if (!sm6_type_is_integer(value->type))
        return false;

    *u = register_get_uint_value(&value->u.reg);

    return true;
}

static bool sm6_metadata_get_uint64_value(const struct sm6_parser *sm6,
        const struct sm6_metadata_value *m, uint64_t *u)
{
    const struct sm6_value *value;

    if (!m || m->type != VKD3D_METADATA_VALUE)
        return false;

    value = m->u.value;
    if (!sm6_value_is_constant(value))
        return false;
    if (!sm6_type_is_integer(value->type))
        return false;

    *u = register_get_uint64_value(&value->u.reg);
    return true;
}

static const struct sm6_metadata_value *shader_sm6_metadata_get(struct sm6_parser *sm6,
        uint64_t index)
{
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(sm6->metadata_tables); ++i)
    {
        if (sm6->metadata_counts[i] > index)
            break;
        index -= sm6->metadata_counts[i];
    }

    return (index >= sm6->metadata_counts[i]) ? NULL : &sm6->metadata_tables[i][index];
}

struct metadata_attachment
{
    uint64_t index;
    unsigned int operand_count;
    struct
    {
        const char *name;
        unsigned int operand;
    } operands[];
};

struct metadata_attachment_array
{
    struct metadata_attachment **elements;
    unsigned int count;
};

static void sm6_parser_metadata_attachment_block_init(struct sm6_parser *sm6, const struct dxil_block *block,
        struct metadata_attachment_array *attachments)
{
    const struct sm6_metadata_value *m;
    const struct dxil_record *record;
    struct metadata_attachment *node;
    unsigned int i, j, k;

    if (!block->record_count)
        return;
    if (attachments->elements)
    {
        FIXME("Skipping additional metadata attachment block.\n");
        return;
    }

    if (!(attachments->elements = vkd3d_calloc(block->record_count, sizeof(*attachments->elements))))
    {
        ERR("Failed to allocate metadata attachments.\n");
        return;
    }

    for (i = 0; i < block->record_count; ++i)
    {
        record = block->records[i];
        if (record->code != METADATA_ATTACHMENT)
        {
            WARN("Skiping record with code %u.\n", record->code);
            continue;
        }
        if (!(record->operand_count & 1))
        {
            WARN("Skipping function attachment.\n");
            continue;
        }

        if (!(node = vkd3d_malloc(offsetof(struct metadata_attachment, operands[record->operand_count / 2u]))))
        {
            ERR("Failed to allocate metadata attachments.\n");
            return;
        }

        attachments->elements[attachments->count++] = node;
        node->index = record->operands[0];
        node->operand_count = record->operand_count / 2u;
        for (j = 1; j < record->operand_count; j += 2)
        {
            if (!(m = sm6_parser_find_metadata_kind(sm6, record->operands[j])))
            {
                WARN("Failed to find metadata kind %"PRIx64".\n", record->operands[j]);
                continue;
            }

            k = j / 2u;
            node->operands[k].name = m->u.kind.name;
            node->operands[k].operand = 0;

            m = shader_sm6_metadata_get(sm6, record->operands[j + 1]);
            if (!m || !sm6_metadata_value_is_node(m))
            {
                WARN("Failed to retrieve metadata attachment node.\n");
                continue;
            }
            if (m->u.node->operand_count && sm6_metadata_value_is_value(m->u.node->operands[0])
                    && !sm6_metadata_get_uint_value(sm6, m->u.node->operands[0], &node->operands[k].operand))
                WARN("Failed to get operand value.\n");
        }
    }
}

static bool sm6_parser_metadata_attachments_init(struct sm6_parser *sm6, const struct dxil_block *block,
        const struct sm6_function *function, struct metadata_attachment_array *attachments)
{
    unsigned int i;

    attachments->elements = NULL;
    attachments->count = 0;

    for (i = 0; i < block->child_block_count; ++i)
    {
        if (block->child_blocks[i]->id == METADATA_ATTACHMENT_BLOCK)
            sm6_parser_metadata_attachment_block_init(sm6, block->child_blocks[i], attachments);
    }
    return true;
}

static const struct metadata_attachment *metadata_attachment_array_find(struct metadata_attachment_array *attachments,
        unsigned int index)
{
    unsigned int i;

    for (i = 0; i < attachments->count; ++i)
    {
        if (attachments->elements[i]->index == index)
            return attachments->elements[i];
    }

    return NULL;
}

static unsigned int metadata_attachment_array_apply(struct metadata_attachment_array *attachments,
        struct vkd3d_shader_instruction *ins, unsigned int ins_idx, struct sm6_value *dst)
{
    static const char *ignored_names[] = {"alias.scope", "dx.controlflow.hints", "llvm.loop", "noalias", "tbaa"};
    const struct metadata_attachment *attachment;
    unsigned int i, j;

    if (!(attachment = metadata_attachment_array_find(attachments, ins_idx)))
        return 0;

    for (i = 0; i < attachment->operand_count; ++i)
    {
        const char *name = attachment->operands[i].name;
        if (!strcmp(name, "dx.precise"))
        {
            ins->flags |= attachment->operands[i].operand ? VKD3DSI_PRECISE_XYZW : 0;
        }
        else if (!strcmp(name, "dx.nonuniform"))
        {
            if (!sm6_value_is_register(dst))
                WARN("Non-uniform result is not a register.\n");
            else
                dst->u.reg.non_uniform = !!attachment->operands[i].operand;
        }
        else
        {
            for (j = 0; j < ARRAY_SIZE(ignored_names); ++j)
                if (!strcmp(name, ignored_names[j]))
                    break;
            if (j == ARRAY_SIZE(ignored_names))
                WARN("Ignoring metadata modifier %s.\n", name);
        }
    }

    return 1;
}

static void metadata_attachment_array_cleanup(struct metadata_attachment_array *attachments)
{
    unsigned int i;

    for (i = 0; i < attachments->count; ++i)
        vkd3d_free(attachments->elements[i]);
    vkd3d_free(attachments->elements);
}

static bool sm6_function_blocks_reserve(struct sm6_function *function, unsigned int reserve)
{
    if (!vkd3d_array_reserve((void **)&function->blocks, &function->block_capacity,
            reserve, sizeof(*function->blocks)))
    {
        ERR("Failed to allocate code block array.\n");
        return false;
    }
    return true;
}

unsigned int sm6_function_alloc_block_id(struct sm6_function *function)
{
    return ++function->block_current_id;
}

struct sm6_block *sm6_function_blocks_new(struct sm6_function *function)
{
    struct sm6_block *block;

    if (!sm6_function_blocks_reserve(function, function->block_count + 1) || !(block = sm6_block_create()))
        return NULL;

    function->blocks[function->block_count++] = block;

    return block;
}

static enum vkd3d_result sm6_parser_function_init(struct sm6_parser *sm6, const struct dxil_block *block,
        struct sm6_function *function, struct metadata_attachment_array *attachments)
{
    unsigned int i, j, count, block_idx, ins_idx, attachment_count;
    struct vkd3d_shader_instruction *ins;
    const struct dxil_record *record;
    const struct sm6_type *fwd_type;
    struct sm6_block *code_block;
    const struct sm6_value *src;
    struct sm6_value *dst;
    bool ret_found;
    enum
    {
        RESULT_VOID,
        RESULT_VALUE,
        RESULT_TERMINATE,
    } result_type;

    if (!(function->declaration = sm6_parser_next_function_definition(sm6)))
    {
        WARN("Failed to find definition to match function body.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }

    if (block->record_count < 2)
    {
        /* It should contain at least a block count and a RET instruction. */
        WARN("Invalid function block record count %zu.\n", block->record_count);
        return VKD3D_ERROR_INVALID_SHADER;
    }
    if (block->records[0]->code != FUNC_CODE_DECLAREBLOCKS || !block->records[0]->operand_count
            || block->records[0]->operands[0] > UINT_MAX)
    {
        WARN("Block count declaration not found or invalid.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }

    if (!(count = block->records[0]->operands[0]))
    {
        WARN("Function contains no blocks.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }
    if (!sm6_function_blocks_reserve(function, count))
        return VKD3D_ERROR_OUT_OF_MEMORY;
    for (i = 0; i < count; ++i)
    {
        if (!sm6_function_blocks_new(function))
        {
            ERR("Failed to allocate code block.\n");
            return VKD3D_ERROR_OUT_OF_MEMORY;
        }
    }
    code_block = function->blocks[0];
    function->entry_block = code_block;

    sm6->cur_max_value = function->value_count;

    for (i = 1, block_idx = 0, ins_idx = 0, attachment_count = 0, ret_found = false; i < block->record_count; ++i)
    {
        /* some instructions can emit >1 IR instruction, so extra may be used. */
        if (!vkd3d_array_reserve((void **)&code_block->instructions, &code_block->instruction_capacity,
                max(code_block->instruction_count + VKD3D_SM6_MAX_IR_INSTRUCTIONS_PER_DXIL_INSTRUCTION, 16u),
                sizeof(*code_block->instructions)))
        {
            ERR("Failed to allocate instructions.\n");
            return VKD3D_ERROR_OUT_OF_MEMORY;
        }

        ins = &code_block->instructions[code_block->instruction_count];
        ins->handler_idx = VKD3DSIH_INVALID;

        dst = sm6_parser_get_current_value(sm6);
        fwd_type = dst->type;
        dst->type = NULL;
        dst->value_type = VALUE_TYPE_REG;
        result_type = RESULT_VALUE;

        record = block->records[i];
        switch (record->code)
        {
            case FUNC_CODE_INST_ALLOCA:
                sm6_parser_emit_alloca(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_ATOMICRMW:
                sm6_parser_emit_atomicrmw(sm6, record, code_block, ins, dst);
                break;
            case FUNC_CODE_INST_BINOP:
                sm6_parser_emit_binop(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_BR:
                sm6_parser_emit_br(sm6, record, function, code_block, ins);
                result_type = RESULT_TERMINATE;
                break;
            case FUNC_CODE_INST_CALL:
                sm6_parser_emit_call(sm6, record, code_block, ins, dst);
                result_type = dst->type ? RESULT_VALUE : RESULT_VOID;
                break;
            case FUNC_CODE_INST_CAST:
                sm6_parser_emit_cast(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_CMP2:
                sm6_parser_emit_cmp2(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_CMPXCHG:
                sm6_parser_emit_cmpxchg(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_EXTRACTVAL:
                sm6_parser_emit_extractval(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_GEP:
                sm6_parser_emit_gep(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_LOAD:
                sm6_parser_emit_load(sm6, record, ins, dst);
                break;
            case FUNC_CODE_INST_PHI:
                sm6_parser_emit_phi(sm6, record, function, code_block, ins, dst);
                break;
            case FUNC_CODE_INST_RET:
                sm6_parser_emit_ret(sm6, record, code_block, ins);
                result_type = RESULT_TERMINATE;
                ret_found = true;
                break;
            case FUNC_CODE_INST_STORE:
                sm6_parser_emit_store(sm6, record, ins, dst);
                result_type = RESULT_VOID;
                break;
            case FUNC_CODE_INST_SWITCH:
                sm6_parser_emit_switch(sm6, record, function, code_block, ins);
                result_type = RESULT_TERMINATE;
                break;
            case FUNC_CODE_INST_VSELECT:
                sm6_parser_emit_vselect(sm6, record, ins, dst);
                break;
            default:
                FIXME("Unhandled dxil instruction %u.\n", record->code);
                return VKD3D_ERROR_INVALID_SHADER;
        }

        attachment_count += metadata_attachment_array_apply(attachments, ins, ins_idx, dst);
        ++ins_idx;

        if (result_type == RESULT_TERMINATE)
        {
            ++block_idx;
            code_block = (block_idx < function->block_count) ? function->blocks[block_idx] : NULL;
        }
        if (ins->handler_idx == VKD3DSIH_INVALID)
        {
            sm6->p.failed = true;
            /* Parsing can continue if we have a result, otherwise the value references will be corrupted. */
            if (result_type == RESULT_VALUE && !dst->type)
                return VKD3D_ERROR_INVALID_SHADER;
        }
        if (code_block)
            code_block->instruction_count += ins->handler_idx != VKD3DSIH_NOP;
        else
            assert(ins->handler_idx == VKD3DSIH_NOP);
        if (dst->type && fwd_type && dst->type != fwd_type)
            FIXME("Type mismatch.\n");
        sm6->value_count += !!dst->type;
    }

    if (!ret_found)
    {
        WARN("Function contains no RET instruction.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }
    if (attachment_count < attachments->count)
        WARN("One or more metadata attachments did not match an instruction index.\n");

    for (block_idx = 0; block_idx < function->block_count; ++block_idx)
    {
        code_block = function->blocks[block_idx];

        if (block_idx)
        {
            char name[12];
            sprintf(name, "%u", block_idx);
            sm6_block_set_name1(code_block, name);
        }
        else
        {
            sm6_block_set_name1(code_block, ".entry");
        }

        for (i = 0; i < code_block->phi_count; ++i)
        {
            struct sm6_phi *phi = &code_block->phi[i];
            for (j = 0; j < phi->incoming_count; ++j)
            {
                src = sm6_parser_get_value(sm6, phi->incoming[j].reg.idx[0].offset);
                if (!sm6_value_is_constant(src) && !sm6_value_is_undef(src) && !sm6_value_is_ssa(src))
                {
                    FIXME("PHI incoming value is not an SSA register.\n");
                    return false;
                }
                phi->incoming[j].reg = src->u.reg;
            }
        }
    }

    return VKD3D_OK;
}

void sm6_function_reset_graphviz_ids(struct sm6_function *function)
{
    unsigned int i;
    for (i = 0; i < function->block_count; ++i)
        function->blocks[i]->graphviz_id = 0;
}

void sm6_function_reset_traversal(struct sm6_function *function)
{
    struct sm6_block *block;
    unsigned int i;

    for (i = 0; i < function->block_count; ++i)
    {
        block = function->blocks[i];
        block->visited = false;
        block->backward_visited = false;
        block->traversing = false;
        block->immediate_dominator = NULL;
        block->immediate_post_dominator = NULL;
        block->fake_pred.count = 0;
        block->fake_succ.count = 0;

        if (!block->freeze_structured_analysis)
        {
            block->headers.count = 0;
            block->merge_type = MERGE_TYPE_NONE;
            block->loop_merge_block = NULL;
            block->loop_ladder_block = NULL;
            block->selection_merge_block = NULL;
        }

        if (block->succ_back_edge)
            sm6_block_array_append(&block->succ, block->succ_back_edge);
        if (block->pred_back_edge)
            sm6_block_array_append(&block->pred, block->pred_back_edge);
        block->succ_back_edge = NULL;
        block->pred_back_edge = NULL;
    }
}

void sm6_function_reset_visited(struct sm6_function *function)
{
    unsigned int i;
    for (i = 0; i < function->block_count; ++i)
    {
        function->blocks[i]->delayed = false;
        function->blocks[i]->visited = false;
    }
}

void sm6_block_emit_copy_ssa_reg(struct sm6_block *block, const struct vkd3d_shader_register *dst,
        const struct vkd3d_shader_register *src, void *parser)
{
    struct vkd3d_shader_src_param *src_param;
    struct vkd3d_shader_dst_param *dst_param;
    struct vkd3d_shader_instruction *ins;
    struct sm6_parser *sm6 = parser;

    if (!vkd3d_array_reserve((void **)&block->instructions, &block->instruction_capacity,
            block->instruction_count + 1, sizeof(*block->instructions)))
    {
        ERR("Failed to allocate instructions.\n");
        return;
    }
    ins = &block->instructions[block->instruction_count++];
    shader_instruction_init(ins, VKD3DSIH_MOV);

    src_param = instruction_src_params_alloc(ins, 1, sm6);
    src_param_init(src_param);
    src_param->reg = *src;

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init(dst_param);
    dst_param->reg = *dst;
}

void sm6_block_emit_select_replace_phi(struct sm6_block *block, const struct vkd3d_shader_register *condition,
        struct vkd3d_shader_register *true_src, const struct vkd3d_shader_register *false_src, void *parser)
{
    struct vkd3d_shader_src_param *src_params;
    struct vkd3d_shader_dst_param *dst_param;
    struct vkd3d_shader_instruction *ins;
    struct sm6_parser *sm6 = parser;

    if (!vkd3d_array_reserve((void **)&block->instructions, &block->instruction_capacity,
            block->instruction_count + 1, sizeof(*block->instructions)))
    {
        ERR("Failed to allocate instructions.\n");
        return;
    }
    ins = &block->instructions[block->instruction_count++];
    shader_instruction_init(ins, VKD3DSIH_MOVC);

    src_params = instruction_src_params_alloc(ins, 3, sm6);
    src_param_init(&src_params[0]);
    src_params[0].reg = *condition;
    src_param_init(&src_params[1]);
    src_params[1].reg = *true_src;
    src_param_init(&src_params[2]);
    src_params[2].reg = *false_src;

    dst_param = instruction_dst_params_alloc(ins, 1, sm6);
    dst_param_init(dst_param);
    assert(register_is_ssa(true_src));
    true_src->idx[0].offset = sm6_parser_alloc_ssa_id(sm6);
    true_src->idx_count = 1;
    dst_param->reg = *true_src;
}

static void sm6_block_visit_resolve_id(struct sm6_function *function, struct sm6_block *block,
        unsigned int *next_index, unsigned int *id_table)
{
    struct sm6_block *continue_block = NULL;
    struct sm6_block *merge_block = NULL;
    unsigned int i;

    assert(block);
    if (block->visited || block->delayed)
        return;

    id_table[block->synthetic_id] = (*next_index)++;
    block->visited = true;

    if ((merge_block = block->merge_block))
    {
        merge_block->delayed = true;
        if (block->merge_type == MERGE_TYPE_LOOP && block->continue_block)
        {
            continue_block = block->continue_block;
            continue_block->delayed = true;
        }
    }

    for (i = 0; i < block->succ.count; ++i)
        sm6_block_visit_resolve_id(function, block->succ.blocks[i], next_index, id_table);

    if (block->fake_merge_block_id)
        id_table[block->fake_merge_block_id] = (*next_index)++;
    if (block->fake_loop_block_id)
        id_table[block->fake_loop_block_id] = (*next_index)++;

    if (continue_block)
    {
        continue_block->delayed = false;
        sm6_block_visit_resolve_id(function, continue_block, next_index, id_table);
    }
    if (merge_block)
    {
        merge_block->delayed = false;
        sm6_block_visit_resolve_id(function, merge_block, next_index, id_table);
    }
}

static void sm6_block_emit_phi(struct sm6_block *block, unsigned int *id_table, struct sm6_parser *sm6)
{
    struct vkd3d_shader_instruction *ins;
    struct vkd3d_shader_phi *dst_phi;
    const struct sm6_phi *src_phi;
    unsigned int i, j;

    sm6_parser_require_space(sm6, block->phi_count);

    for (i = 0; i < block->phi_count; ++i)
    {
        src_phi = &block->phi[i];
        if (!src_phi->incoming_count)
            continue;
        assert(src_phi->incoming_count > 1);

        ins = sm6_parser_add_instruction(sm6, VKD3DSIH_PHI);
        dst_phi = vkd3d_malloc(sizeof(*dst_phi) + src_phi->incoming_count * sizeof(dst_phi->incoming[0]));
        dst_phi->name = src_phi->name;
        dst_phi->reg = src_phi->reg;
        dst_phi->incoming_count = src_phi->incoming_count;
        for (j = 0; j < src_phi->incoming_count; ++j)
        {
            dst_phi->incoming[j].block = id_table[src_phi->incoming[j].block->synthetic_id];
            dst_phi->incoming[j].reg = src_phi->incoming[j].reg;
        }
        ins->declaration.phi = dst_phi;
        shader_instruction_array_add_opaque_param(&sm6->p.instructions, dst_phi);
    }
}

static bool sm6_block_emit_instructions(struct sm6_block *block, struct sm6_parser *sm6)
{
    /* Reserve extra instructions for a possible merge and the block terminator. */
    struct vkd3d_shader_instruction *ins = sm6_parser_require_space(sm6, block->instruction_count + 2);

    if (!ins)
        return false;

    memcpy(ins, block->instructions, block->instruction_count * sizeof(*block->instructions));
    sm6->p.instructions.count += block->instruction_count;
    return true;
}

static void sm6_block_emit_merge(struct sm6_block *block, unsigned int *id_table, struct sm6_parser *sm6)
{
    unsigned int merge_id = 0, continue_id = 0;
    struct vkd3d_shader_instruction *ins;

    if (block->merge_type == MERGE_TYPE_NONE)
        return;

    /* Emit structured merge information. Blocks required by SPIR-V can be missing from the DXIL
     * source. If a selection merge has no merge block, emit a fake merge block. A loop merge must
     * have at least a continue or a merge block, if not both. Emit a fake block for whichever
     * is missing. */
    switch (block->merge_type)
    {
        case MERGE_TYPE_SELECTION:
            if (block->merge_block)
                merge_id = block->merge_block->synthetic_id;
            else
                merge_id = block->fake_merge_block_id;
            break;

        case MERGE_TYPE_LOOP:
            if (block->merge_block && block->continue_block)
            {
                merge_id = block->merge_block->synthetic_id;
                continue_id = block->continue_block->synthetic_id;
            }
            else if (block->merge_block)
            {
                merge_id = block->merge_block->synthetic_id;
                continue_id = block->fake_loop_block_id;
            }
            else if (block->continue_block)
            {
                merge_id = block->fake_loop_block_id;
                continue_id = block->continue_block->synthetic_id;
            }
            break;

        default:
            vkd3d_unreachable();
    }

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_MERGE);
    ins->declaration.merge.type = block->merge_type;
    ins->declaration.merge.merge_block_idx = id_table[merge_id];
    ins->declaration.merge.continue_block_idx = id_table[continue_id];
}

static void sm6_block_emit_terminator(struct sm6_block *block, unsigned int *id_table, struct sm6_parser *sm6)
{
    struct vkd3d_shader_instruction *ins;
    unsigned int i, count;

    switch (block->terminator.type)
    {
        case TERMINATOR_UNREACHABLE:
        {
            sm6_parser_add_instruction(sm6, VKD3DSIH_UNREACHABLE);
            break;
        }

        case TERMINATOR_UNCOND_BR:
        {
            ins = sm6_parser_add_instruction(sm6, VKD3DSIH_BRANCH);
            ins->declaration.branch.condition.type = VKD3DSPR_INVALID;
            ins->declaration.branch.true_idx = id_table[block->terminator.direct_block->synthetic_id];
            break;
        }

        case TERMINATOR_COND_BR:
        {
            ins = sm6_parser_add_instruction(sm6, VKD3DSIH_BRANCH);
            ins->declaration.branch.condition = block->terminator.conditional_reg;
            ins->declaration.branch.true_idx = id_table[block->terminator.true_block->synthetic_id];
            ins->declaration.branch.false_idx = id_table[block->terminator.false_block->synthetic_id];
            break;
        }

        case TERMINATOR_SWITCH:
        {
            struct sm6_switch_case *cases;
            struct terminator_case *switch_case;

            for (i = 0; i < block->terminator.case_count; ++i)
                if (block->terminator.cases[i].is_default)
                    break;
            assert(i < block->terminator.case_count);
            ins = sm6_parser_add_instruction(sm6, VKD3DSIH_SWITCH);
            ins->declaration.sm6_switch.reg = block->terminator.conditional_reg;
            ins->declaration.sm6_switch.default_idx = id_table[block->terminator.cases[i].node->synthetic_id];
            cases = vkd3d_malloc((block->terminator.case_count - 1) * sizeof(*cases));
            for (i = 0, count = 0; i < block->terminator.case_count; ++i)
            {
                switch_case = &block->terminator.cases[i];
                if (switch_case->is_default)
                    continue;
                cases[count].value = switch_case->value;
                cases[count++].block_idx = id_table[switch_case->node->synthetic_id];
            }
            ins->declaration.sm6_switch.cases = cases;
            ins->declaration.sm6_switch.case_count = count;
            shader_instruction_array_add_opaque_param(&sm6->p.instructions, cases);
            break;
        }

        case TERMINATOR_RET:
        {
            sm6_parser_add_instruction(sm6, VKD3DSIH_RET);
            break;
        }

        default:
            break;
    }
}

static void sm6_parser_emit_unreachable(struct sm6_parser *sm6)
{
    sm6_parser_require_space(sm6, 1);
    sm6_parser_add_instruction(sm6, VKD3DSIH_UNREACHABLE);
}

static void sm6_parser_emit_branch(struct sm6_parser *sm6, unsigned int block_idx)
{
    struct vkd3d_shader_instruction *ins;

    sm6_parser_require_space(sm6, 1);
    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_BRANCH);
    ins->declaration.branch.condition.type = VKD3DSPR_INVALID;
    ins->declaration.branch.true_idx = block_idx;
}

static void sm6_block_visit_append_instructions(struct sm6_block *block, struct sm6_function *function,
        unsigned int *id_table, struct sm6_parser *sm6)
{
    struct sm6_block *continue_block = NULL;
    struct sm6_block *merge_block = NULL;
    unsigned int i;

    assert(block);
    if (block->visited || block->delayed)
        return;

    sm6_block_emit_phi(block, id_table, sm6);
    sm6_block_emit_instructions(block, sm6);
    sm6_block_emit_merge(block, id_table, sm6);
    sm6_block_emit_terminator(block, id_table, sm6);
    block->visited = true;

    if ((merge_block = block->merge_block))
    {
        merge_block->delayed = true;
        if (block->merge_type == MERGE_TYPE_LOOP && block->continue_block)
        {
            continue_block = block->continue_block;
            continue_block->delayed = true;
        }
    }

    for (i = 0; i < block->succ.count; ++i)
        sm6_block_visit_append_instructions(block->succ.blocks[i], function, id_table, sm6);

    if (block->fake_merge_block_id)
        sm6_parser_emit_unreachable(sm6);

    if (block->fake_loop_block_id)
    {
        if (block->merge_block)
            sm6_parser_emit_branch(sm6, id_table[block->synthetic_id]);
        else
            sm6_parser_emit_unreachable(sm6);
    }

    if (continue_block)
    {
        continue_block->delayed = false;
        sm6_block_visit_append_instructions(continue_block, function, id_table, sm6);
    }
    if (merge_block)
    {
        merge_block->delayed = false;
        sm6_block_visit_append_instructions(merge_block, function, id_table, sm6);
    }
}

static void sm6_parser_emit_block_count(struct sm6_parser *sm6, unsigned int block_count)
{
    struct vkd3d_shader_instruction *ins;

    sm6_parser_require_space(sm6, 1);
    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_MAX_BLOCK_COUNT);
    ins->declaration.count = block_count;
}

void sm6_function_emit_blocks(struct sm6_function *function, void *sm6)
{
    unsigned int i, next_index = 0;
    unsigned int *id_table;

    if (!(id_table = vkd3d_malloc((function->block_current_id + 1) * sizeof(*id_table))))
    {
        ERR("Failed to allocate id table.\n");
        abort();
    }
    sm6_block_visit_resolve_id(function, function->entry_block, &next_index, id_table);
    sm6_parser_emit_block_count(sm6, next_index);
    sm6_parser_emit_temps(sm6, function);

    for (i = 0; i < function->block_count; ++i)
    {
        function->blocks[i]->visited = false;
        function->blocks[i]->delayed = false;
    }

    sm6_block_visit_append_instructions(function->entry_block, function, id_table, sm6);

    vkd3d_free(id_table);
}

static enum vkd3d_result sm6_parser_module_init(struct sm6_parser *sm6, const struct dxil_block *block,
        unsigned int level)
{
    unsigned int i, old_value_count = sm6->value_count;
    struct metadata_attachment_array attachments;
    struct sm6_function *function;
    enum vkd3d_result ret;

    for (i = 0; i < block->child_block_count; ++i)
    {
        if ((ret = sm6_parser_module_init(sm6, block->child_blocks[i], level + 1)) < 0)
            return ret;
    }

    switch (block->id)
    {
        case CONSTANTS_BLOCK:
            if (level < 2)
                break;
            function = &sm6->functions[sm6->function_count];
            sm6->cur_max_value = function->value_count;
            return sm6_parser_constants_init(sm6, block);

        case FUNCTION_BLOCK:
            function = &sm6->functions[sm6->function_count];
            sm6_parser_metadata_attachments_init(sm6, block, function, &attachments);
            ret = sm6_parser_function_init(sm6, block, function, &attachments);
            metadata_attachment_array_cleanup(&attachments);
            if (ret < 0)
                return ret;
            /* The value index returns to its previous value after handling a function. It's usually nonzero
             * at the start because of global constants/variables/function declarations. Function constants
             * occur in a child block, so value_count is already saved before they are emitted. */
            memset(&sm6->values[old_value_count], 0, (sm6->value_count - old_value_count) * sizeof(*sm6->values));
            sm6->value_count = old_value_count;
            break;

        default:
            break;
    }

    return VKD3D_OK;
}

static bool sm6_parser_allocate_named_metadata(struct sm6_parser *sm6)
{
    struct dxil_block *block;
    unsigned int i, j, count;

    for (i = 0, count = 0; i < sm6->root_block.child_block_count; ++i)
    {
        block = sm6->root_block.child_blocks[i];
        if (block->id != METADATA_BLOCK)
            continue;
        for (j = 0; j < block->record_count; ++j)
            count += block->records[j]->code == METADATA_NAMED_NODE;
    }

    if (!count)
        return true;

    return !!(sm6->named_metadata = vkd3d_calloc(count, sizeof(*sm6->named_metadata)));
}

static enum vkd3d_result sm6_parser_metadata_init(struct sm6_parser *sm6, const struct dxil_block *block,
        unsigned int idx)
{
    unsigned int i, j, count, node_idx, offset;
    struct sm6_metadata_value *table, *m;
    const struct dxil_record *record;
    struct sm6_metadata_node *node;
    const struct sm6_value *value;
    char *name;

    for (i = 0, count = 0; i < block->record_count; ++i)
        count += block->records[i]->code != METADATA_NAME;

    if (!(sm6->metadata_tables[idx] = vkd3d_calloc(count, sizeof(*sm6->metadata_tables[0]))))
    {
        ERR("Failed to allocate metadata tables.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }
    table = sm6->metadata_tables[idx];

    for (i = 0, node_idx = 0, name = NULL; i < block->record_count; ++i)
    {
        record = block->records[i];
        m = &table[sm6->metadata_counts[idx]];
        m->index = sm6->metadata_counts[idx];
        offset = 1;

        switch (record->code)
        {
            case METADATA_NAMED_NODE:
                m = &sm6->named_metadata[sm6->named_metadata_count].value;
                m->index = UINT_MAX;
                offset = 0;
                /* fall through */
            case METADATA_DISTINCT_NODE:
            case METADATA_NODE:
                m->type = VKD3D_METADATA_NODE;
                if (!(m->value_type = sm6->metadata_type))
                {
                    WARN("Metadata type not found.\n");
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (!(node = vkd3d_malloc(offsetof(struct sm6_metadata_node, operands[record->operand_count]))))
                {
                    ERR("Failed to allocate metadata node with %u operands.\n", record->operand_count);
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }
                m->u.node = node;

                node->is_distinct = record->code == METADATA_DISTINCT_NODE;

                for (j = 0; j < record->operand_count; ++j)
                {
                    size_t ref;

                    if (record->operands[j] >= count + offset)
                    {
                        WARN("Invalid metadata index %"PRIu64".\n", record->operands[j]);
                        return VKD3D_ERROR_INVALID_SHADER;
                    }
                    ref = record->operands[j] - offset;
                    if (!node->is_distinct && ref == m->index)
                    {
                        WARN("Metadata self-reference at index %u.\n", m->index);
                        return VKD3D_ERROR_INVALID_SHADER;
                    }
                    node->operands[j] = (record->operands[j] >= offset) ? &table[ref] : NULL;
                }
                node->operand_count = record->operand_count;

                if (record->code == METADATA_NAMED_NODE)
                {
                    node->index = UINT_MAX;
                    if (!name)
                    {
                        WARN("Named node has no name.\n");
                        continue;
                    }
                    if (!node->operand_count || !sm6_metadata_value_is_node(node->operands[0]))
                    {
                        WARN("Named node does not contain a node.\n");
                        continue;
                    }
                    sm6->named_metadata[sm6->named_metadata_count++].name = name;
                    name = NULL;
                    continue;
                }

                node->index = node_idx++;
                break;

            case METADATA_KIND:
                m->type = VKD3D_METADATA_KIND;
                m->u.kind.id = record->operands[0];
                if (!(m->u.kind.name = dxil_record_to_string(record, 1)))
                {
                    ERR("Failed to allocate name of a kind.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }
                break;

            case METADATA_NAME:
                if (!(name = dxil_record_to_string(record, 0)))
                {
                    ERR("Failed to allocate name.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }
                continue;

            case METADATA_STRING:
                m->type = VKD3D_METADATA_STRING;
                if (!(m->u.string_value = dxil_record_to_string(record, 0)))
                {
                    ERR("Failed to allocate string.\n");
                    return VKD3D_ERROR_OUT_OF_MEMORY;
                }
                break;

            case METADATA_VALUE:
                m->type = VKD3D_METADATA_VALUE;
                if (record->operand_count < 2)
                {
                    WARN("Invalid operand count %u.\n", record->operand_count);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (record->operand_count > 2)
                    WARN("Ignoring %u extra operands.\n", record->operand_count - 2);

                if (!(m->value_type = sm6_parser_get_type(sm6, record->operands[0])))
                {
                    WARN("Invalid type id %"PRIu64".\n", record->operands[0]);
                    return VKD3D_ERROR_INVALID_SHADER;
                }

                if (record->operands[1] > UINT_MAX)
                    WARN("Truncating value index %"PRIu64".\n", record->operands[1]);
                if (!(value = sm6_parser_get_value(sm6, record->operands[1])))
                {
                    WARN("Invalid value index %"PRIu64".\n", record->operands[1]);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                if (!sm6_value_is_constant(value) && !sm6_value_is_undef(value) && !sm6_value_is_constant_array(value)
                        && !sm6_value_is_function_dcl(value))
                {
                    WARN("Value at index %"PRIu64" is not a constant or a function declaration.\n",
                            record->operands[1]);
                    return VKD3D_ERROR_INVALID_SHADER;
                }
                m->u.value = value;

                if (value->type != m->value_type)
                    WARN("Type mismatch.\n");

                break;

            default:
                FIXME("Unhandled metadata type %u.\n", record->code);
                return VKD3D_ERROR_INVALID_SHADER;
        }
        ++sm6->metadata_counts[idx];
    }

    if (name)
    {
        WARN("Unused metadata name '%s'.\n", name);
        vkd3d_free(name);
    }

    return VKD3D_OK;
}

static const struct sm6_metadata_value *sm6_parser_find_named_metadata(struct sm6_parser *sm6, const char *name)
{
    const struct sm6_metadata_node *node;
    unsigned int i;

    for (i = 0; i < sm6->named_metadata_count; ++i)
    {
        if (!strcmp(sm6->named_metadata[i].name, name))
        {
            node = sm6->named_metadata[i].value.u.node;
            if (node->operand_count > 1)
                FIXME("Ignoring %u additional operands for %s.\n", node->operand_count - 1, name);
            return node->operands[0];
        }
    }

    return NULL;
}

static bool sm6_parser_resources_load_register_range(const struct sm6_parser *sm6,
        const struct sm6_metadata_node *node, struct vkd3d_shader_register_range *range)
{
    unsigned int size;

    if (!sm6_metadata_value_is_value(node->operands[1]))
    {
        WARN("Resource data type is not a value.\n");
        return false;
    }
    if (!sm6_type_is_pointer(node->operands[1]->value_type))
        WARN("Resource type is not a pointer.\n");

    if (!sm6_metadata_get_uint_value(sm6, node->operands[3], &range->space))
    {
        WARN("Failed to load register space.\n");
        return false;
    }
    if (!sm6_metadata_get_uint_value(sm6, node->operands[4], &range->first))
    {
        WARN("Failed to load register first.\n");
        return false;
    }
    if (!sm6_metadata_get_uint_value(sm6, node->operands[5], &size))
    {
        WARN("Failed to load register range size.\n");
        return false;
    }
    if (!size || (size != UINT_MAX && !vkd3d_bound_range(range->first, size, UINT_MAX)))
    {
        WARN("Invalid register range, first %u, size %u.\n", range->first, size);
        return false;
    }
    range->last = (size == UINT_MAX) ? UINT_MAX : range->first + size - 1;

    return true;
}

static enum vkd3d_shader_resource_type shader_resource_type_from_dxil_resource_kind(enum dxil_resource_kind kind)
{
    if (kind >= RESOURCE_KIND_TEXTURE1D && kind <= RESOURCE_KIND_TEXTURECUBEARRAY)
        return kind + 1;

    switch (kind)
    {
        case RESOURCE_KIND_TYPEDBUFFER:
        case RESOURCE_KIND_RAWBUFFER:
        case RESOURCE_KIND_STRUCTUREDBUFFER:
            return VKD3D_SHADER_RESOURCE_BUFFER;
        default:
            return VKD3D_SHADER_RESOURCE_NONE;
    }
}

static const enum vkd3d_data_type data_type_table[] =
{
    [COMPONENT_TYPE_INVALID]     = VKD3D_DATA_UNUSED,
    [COMPONENT_TYPE_I1]          = VKD3D_DATA_UNUSED,
    [COMPONENT_TYPE_I16]         = VKD3D_DATA_INT,
    [COMPONENT_TYPE_U16]         = VKD3D_DATA_UINT,
    [COMPONENT_TYPE_I32]         = VKD3D_DATA_INT,
    [COMPONENT_TYPE_U32]         = VKD3D_DATA_UINT,
    [COMPONENT_TYPE_I64]         = VKD3D_DATA_UNUSED,
    [COMPONENT_TYPE_U64]         = VKD3D_DATA_UNUSED,
    [COMPONENT_TYPE_F16]         = VKD3D_DATA_FLOAT,
    [COMPONENT_TYPE_F32]         = VKD3D_DATA_FLOAT,
    [COMPONENT_TYPE_F64]         = VKD3D_DATA_DOUBLE,
    [COMPONENT_TYPE_SNORMF16]    = VKD3D_DATA_SNORM,
    [COMPONENT_TYPE_UNORMF16]    = VKD3D_DATA_UNORM,
    [COMPONENT_TYPE_SNORMF32]    = VKD3D_DATA_SNORM,
    [COMPONENT_TYPE_UNORMF32]    = VKD3D_DATA_UNORM,
    [COMPONENT_TYPE_SNORMF64]    = VKD3D_DATA_DOUBLE,
    [COMPONENT_TYPE_UNORMF64]    = VKD3D_DATA_DOUBLE,
    [COMPONENT_TYPE_PACKEDS8X32] = VKD3D_DATA_UNUSED,
    [COMPONENT_TYPE_PACKEDU8X32] = VKD3D_DATA_UNUSED,
};

static enum vkd3d_data_type vkd3d_data_type_from_dxil_component_type(enum dxil_component_type type)
{
    enum vkd3d_data_type data_type;

    if (type >= ARRAY_SIZE(data_type_table) || (data_type = data_type_table[type]) == VKD3D_DATA_UNUSED)
    {
        FIXME("Unhandled data type %u.\n", type);
        return VKD3D_DATA_FLOAT;
    }

    return data_type;
}

static struct vkd3d_shader_resource *sm6_parser_resources_load_common_info(const struct sm6_parser *sm6,
        const struct sm6_metadata_value *type_value, bool is_uav, enum dxil_resource_kind kind,
        const struct sm6_metadata_value *m, struct vkd3d_shader_instruction *ins)
{
    enum vkd3d_shader_resource_type resource_type;
    const struct sm6_metadata_node *node;
    const struct sm6_type *type = NULL;
    unsigned int i, count, values[2];
    enum vkd3d_data_type data_type;

    count = 0;
    /* Extract the component count from the LLVM type declaration stored in the resource metadata. */
    if (type_value->type == VKD3D_METADATA_VALUE && sm6_type_is_pointer(type = type_value->u.value->type))
    {
        type = type->u.pointer.type;
        if (type->class == TYPE_CLASS_STRUCT)
            type = type->u.struc->elem_types[0];
        if (sm6_type_is_vector(type))
            count = type->u.array.count;
        else if (sm6_type_is_numeric(type))
            count = 1;
    }

    if (!(resource_type = shader_resource_type_from_dxil_resource_kind(kind)))
    {
        WARN("Invalid resource kind %u.\n", kind);
        return NULL;
    }
    ins->resource_type = resource_type;

    if (!m)
    {
        ins->handler_idx = is_uav ? VKD3DSIH_DCL_UAV_RAW : VKD3DSIH_DCL_RESOURCE_RAW;
        ins->declaration.raw_resource.resource.reg.write_mask = 0;
        return &ins->declaration.raw_resource.resource;
    }

    if (!sm6_metadata_value_is_node(m))
    {
        WARN("Resource metadata list is not a node.\n");
        return NULL;
    }

    node = m->u.node;

    if (node->operand_count < 2)
    {
        WARN("Invalid operand count %u.\n", node->operand_count);
        return NULL;
    }
    if (node->operand_count > 2)
        FIXME("Ignoring %u extra operands.\n", node->operand_count - 2);

    for (i = 0; i < 2; ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &values[i]))
        {
            WARN("Failed to load uint value at index %u.\n", i);
            return NULL;
        }
    }

    if (values[0] == RESOURCE_TYPE_NON_RAW_STRUCTURED)
    {
        data_type = vkd3d_data_type_from_dxil_component_type(values[1]);
        ins->handler_idx = is_uav ? VKD3DSIH_DCL_UAV_TYPED : VKD3DSIH_DCL;
        for (i = 0; i < VKD3D_VEC4_SIZE; ++i)
            ins->declaration.semantic.resource_data_type[i] = data_type;
        ins->declaration.semantic.resource_type = resource_type;
        ins->declaration.semantic.resource.reg.write_mask = vkd3d_write_mask_from_component_count(count);

        return &ins->declaration.semantic.resource;
    }
    else if (values[0] == RESOURCE_TYPE_RAW_STRUCTURED)
    {
        if (kind == RESOURCE_KIND_RAWBUFFER)
        {
            ins->handler_idx = is_uav ? VKD3DSIH_DCL_UAV_RAW : VKD3DSIH_DCL_RESOURCE_RAW;
            ins->declaration.raw_resource.resource.reg.write_mask = 0;

            return &ins->declaration.raw_resource.resource;
        }

        if (kind != RESOURCE_KIND_STRUCTUREDBUFFER)
            FIXME("Unhandled resource kind %u.\n", kind);

        ins->handler_idx = is_uav ? VKD3DSIH_DCL_UAV_STRUCTURED : VKD3DSIH_DCL_RESOURCE_STRUCTURED;
        ins->declaration.structured_resource.byte_stride = values[1];
        ins->declaration.structured_resource.resource.reg.write_mask = 0;

        return &ins->declaration.structured_resource.resource;
    }
    else
    {
        FIXME("Unhandled resource type %u.\n", values[0]);
    }

    return NULL;
}

static void sm6_init_resource_dst_param(struct vkd3d_shader_dst_param *param,
        enum vkd3d_shader_register_type reg_type, enum vkd3d_data_type data_type, unsigned int index)
{
    param->modifiers = 0;
    param->shift = 0;
    register_init_with_id(&param->reg, reg_type, data_type, index);
}

static struct vkd3d_shader_register_range *sm6_parser_resources_load_srv(const struct sm6_parser *sm6,
        const struct sm6_metadata_node *node, struct vkd3d_shader_instruction *ins, struct sm6_resource *res)
{
    struct vkd3d_shader_resource *resource;
    unsigned int value;

    if (node->operand_count < 9)
    {
        WARN("Invalid operand count %u.\n", node->operand_count);
        return NULL;
    }
    if (node->operand_count > 9)
        FIXME("Ignoring %u extra operands.\n", node->operand_count - 9);

    if (!sm6_metadata_get_uint_value(sm6, node->operands[6], &value))
    {
        WARN("Failed to load resource type.\n");
        return NULL;
    }

    shader_instruction_init(ins, VKD3DSIH_INVALID);

    if (!(resource = sm6_parser_resources_load_common_info(sm6, node->operands[1], false, value,
            node->operands[8], ins)))
        return NULL;

    sm6_init_resource_dst_param(&resource->reg, VKD3DSPR_RESOURCE, ins->resource_type == VKD3D_SHADER_RESOURCE_BUFFER
            ? VKD3D_DATA_UINT : VKD3D_DATA_RESOURCE, res->id);

    res->type = ins->resource_type;
    res->kind = value;
    res->reg = &resource->reg.reg;
    res->data_type = ins->handler_idx == VKD3DSIH_DCL ? ins->declaration.semantic.resource_data_type[0]
            : VKD3D_DATA_UNUSED;

    return &resource->range;
}

static struct vkd3d_shader_register_range *sm6_parser_resources_load_uav(const struct sm6_parser *sm6,
        const struct sm6_metadata_node *node, struct vkd3d_shader_instruction *ins, struct sm6_resource *res)
{
    struct vkd3d_shader_resource *resource;
    unsigned int i, values[4];

    if (node->operand_count < 11)
    {
        WARN("Invalid operand count %u.\n", node->operand_count);
        return NULL;
    }
    if (node->operand_count > 11)
        FIXME("Ignoring %u extra operands.\n", node->operand_count - 11);

    for (i = 6; i < 10; ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &values[i - 6]))
        {
            WARN("Failed to load uint value at index %u.\n", i);
            return NULL;
        }
    }

    shader_instruction_init(ins, VKD3DSIH_INVALID);
    if (values[1])
        ins->flags = VKD3DSUF_GLOBALLY_COHERENT;
    if (values[2])
        ins->flags |= VKD3DSUF_ORDER_PRESERVING_COUNTER;
    if (values[3])
        FIXME("Ignoring rasterizer ordered view flag %#x.\n", values[3]);

    if (!(resource = sm6_parser_resources_load_common_info(sm6, node->operands[1], true, values[0],
            node->operands[10], ins)))
        return NULL;

    sm6_init_resource_dst_param(&resource->reg, VKD3DSPR_UAV, ins->resource_type == VKD3D_SHADER_RESOURCE_BUFFER
            ? VKD3D_DATA_UINT : VKD3D_DATA_UAV, res->id);

    res->type = ins->resource_type;
    res->kind = values[0];
    res->reg = &resource->reg.reg;
    res->data_type = ins->handler_idx == VKD3DSIH_DCL_UAV_TYPED
            ? ins->declaration.semantic.resource_data_type[0] : VKD3D_DATA_UNUSED;

    return &resource->range;
}

static struct vkd3d_shader_register_range *sm6_parser_resources_load_cbv(const struct sm6_parser *sm6,
        const struct sm6_metadata_node *node, struct vkd3d_shader_instruction *ins, struct sm6_resource *res)
{
    unsigned int aligned_size;

    if (node->operand_count < 7)
    {
        WARN("Invalid operand count %u.\n", node->operand_count);
        return NULL;
    }
    if (node->operand_count > 7 && node->operands[7])
        FIXME("Ignoring %u extra operands.\n", node->operand_count - 7);

    shader_instruction_init(ins, VKD3DSIH_DCL_CONSTANT_BUFFER);
    ins->resource_type = VKD3D_SHADER_RESOURCE_BUFFER;

    if (!sm6_metadata_get_uint_value(sm6, node->operands[6], &ins->declaration.cb.size))
    {
        WARN("Failed to load buffer size.\n");
        return NULL;
    }
    aligned_size = align(ins->declaration.cb.size, sizeof(float) * VKD3D_VEC4_SIZE);
    if (aligned_size != ins->declaration.cb.size)
        TRACE("Rounding up constant buffer byte size %u to %u.\n", ins->declaration.cb.size, aligned_size);
    ins->declaration.cb.size = aligned_size / (sizeof(float) * VKD3D_VEC4_SIZE);
    ins->declaration.cb.src.swizzle = VKD3D_SHADER_NO_SWIZZLE;
    ins->declaration.cb.src.modifiers = VKD3DSPSM_NONE;
    register_init_with_id(&ins->declaration.cb.src.reg, VKD3DSPR_CONSTBUFFER, VKD3D_DATA_FLOAT, res->id);

    res->type = VKD3D_SHADER_RESOURCE_BUFFER;
    res->kind = RESOURCE_KIND_CBUFFER;
    res->reg = &ins->declaration.cb.src.reg;

    return &ins->declaration.cb.range;
}

static struct vkd3d_shader_register_range *sm6_parser_resources_load_sampler(const struct sm6_parser *sm6,
        const struct sm6_metadata_node *node, struct vkd3d_shader_instruction *ins, struct sm6_resource *res)
{
    unsigned int kind;

    if (node->operand_count < 7)
    {
        WARN("Invalid operand count %u.\n", node->operand_count);
        return NULL;
    }
    if (node->operand_count > 7 && node->operands[7])
        FIXME("Ignoring %u extra operands.\n", node->operand_count - 7);

    shader_instruction_init(ins, VKD3DSIH_DCL_SAMPLER);
    ins->resource_type = VKD3D_SHADER_RESOURCE_NONE;

    if (!sm6_metadata_get_uint_value(sm6, node->operands[6], &kind))
    {
        WARN("Failed to load sampler mode.\n");
        return NULL;
    }
    switch (kind)
    {
        case SAMPLER_KIND_DEFAULT:
            ins->flags = 0;
            break;
        case SAMPLER_KIND_COMPARISON:
            ins->flags = VKD3DSI_SAMPLER_COMPARISON_MODE;
            break;
        default:
            FIXME("Ignoring sampler kind %u.\n", kind);
            break;
    }

    ins->declaration.sampler.src.swizzle = VKD3D_SHADER_NO_SWIZZLE;
    ins->declaration.sampler.src.modifiers = VKD3DSPSM_NONE;
    register_init_with_id(&ins->declaration.sampler.src.reg, VKD3DSPR_SAMPLER, VKD3D_DATA_UNUSED, res->id);

    res->type = VKD3D_SHADER_RESOURCE_NONE;
    res->kind = RESOURCE_KIND_SAMPLER;
    res->reg = &ins->declaration.sampler.src.reg;

    return &ins->declaration.sampler.range;
}

static enum vkd3d_result sm6_parser_resources_init(struct sm6_parser *sm6)
{
    const struct sm6_metadata_value *value, *m = sm6_parser_find_named_metadata(sm6, "dx.resources");
    const struct sm6_metadata_node *resources_node, *node;
    struct vkd3d_shader_register_range *range;
    struct vkd3d_shader_instruction *ins;
    struct sm6_resource *resource;
    unsigned int i, j, count, id;
    size_t capacity = 0;

    if (!m)
        return VKD3D_OK;

    resources_node = m->u.node;
    if (resources_node->operand_count != VKD3D_SHADER_DESCRIPTOR_TYPE_COUNT)
        WARN("Unexpected descriptor type count %u.\n", resources_node->operand_count);
    count = min(resources_node->operand_count, VKD3D_SHADER_DESCRIPTOR_TYPE_COUNT);

    for (i = 0; i < count; ++i)
    {
        if (!(m = resources_node->operands[i]))
            continue;

        if (!sm6_metadata_value_is_node(m))
        {
            WARN("Resource metadata list is not a node.\n");
            continue;
        }

        for (j = 0; j < m->u.node->operand_count; ++j)
        {
            value = m->u.node->operands[j];
            if (!sm6_metadata_value_is_node(value))
            {
                WARN("Resource metadata entry is not a node.\n");
                continue;
            }

            node = value->u.node;
            if (node->operand_count < 6)
            {
                WARN("Invalid operand count %u.\n", node->operand_count);
                continue;
            }

            if (!sm6_metadata_get_uint_value(sm6, node->operands[0], &id))
            {
                WARN("Failed to load uint id.\n");
                return VKD3D_ERROR_INVALID_SHADER;
            }

            if (!vkd3d_array_reserve((void **)&sm6->resources, &capacity,
                    sm6->resource_count + 1, sizeof(*sm6->resources)))
            {
                ERR("Failed to allocate resource array.\n");
                return VKD3D_ERROR_OUT_OF_MEMORY;
            }

            resource = &sm6->resources[sm6->resource_count];
            resource->class = i;
            resource->id = id;

            if (!(ins = sm6_parser_require_space(sm6, 1)))
            {
                ERR("Failed to allocate instruction.\n");
                return VKD3D_ERROR_OUT_OF_MEMORY;
            }

            switch (i)
            {
                case RESOURCE_CLASS_CBV:
                    range = sm6_parser_resources_load_cbv(sm6, node, ins, resource);
                    break;
                case RESOURCE_CLASS_SRV:
                    range = sm6_parser_resources_load_srv(sm6, node, ins, resource);
                    break;
                case RESOURCE_CLASS_UAV:
                    range = sm6_parser_resources_load_uav(sm6, node, ins, resource);
                    break;
                case RESOURCE_CLASS_SAMPLER:
                    range = sm6_parser_resources_load_sampler(sm6, node, ins, resource);
                    break;
                default:
                    vkd3d_unreachable();
            }
            if (!range)
                return VKD3D_ERROR_INVALID_SHADER;

            if (!sm6_parser_resources_load_register_range(sm6, node, range))
                return VKD3D_ERROR_INVALID_SHADER;
            resource->reg->idx[1].offset = range->first;
            resource->reg->idx[2].offset = range->last;
            resource->reg->idx_count = 3;

            ++sm6->resource_count;
            ++sm6->p.instructions.count;
        }
    }

    return VKD3D_OK;
}

static enum vkd3d_shader_component_type vkd3d_component_type_from_dxil_component_type(enum dxil_component_type type)
{
    switch (type)
    {
        case COMPONENT_TYPE_I1:
            return VKD3D_SHADER_COMPONENT_BOOL;
        case COMPONENT_TYPE_I16:
        case COMPONENT_TYPE_I32:
            return VKD3D_SHADER_COMPONENT_INT;
        case COMPONENT_TYPE_U16:
        case COMPONENT_TYPE_U32:
            return VKD3D_SHADER_COMPONENT_UINT;
        case COMPONENT_TYPE_F16:
        case COMPONENT_TYPE_F32:
        case COMPONENT_TYPE_SNORMF32:
        case COMPONENT_TYPE_UNORMF32:
            return VKD3D_SHADER_COMPONENT_FLOAT;
        case COMPONENT_TYPE_F64:
        case COMPONENT_TYPE_SNORMF64:
        case COMPONENT_TYPE_UNORMF64:
            return VKD3D_SHADER_COMPONENT_DOUBLE;
        default:
            FIXME("Unhandled component type %u.\n", type);
            return VKD3D_SHADER_COMPONENT_UINT;
    }
}

static const enum vkd3d_shader_sysval_semantic sysval_semantic_table[] =
{
    [SEMANTIC_KIND_ARBITRARY]            = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_VERTEXID]             = VKD3D_SHADER_SV_VERTEX_ID,
    [SEMANTIC_KIND_INSTANCEID]           = VKD3D_SHADER_SV_INSTANCE_ID,
    [SEMANTIC_KIND_POSITION]             = VKD3D_SHADER_SV_POSITION,
    [SEMANTIC_KIND_RTARRAYINDEX]         = VKD3D_SHADER_SV_RENDER_TARGET_ARRAY_INDEX,
    [SEMANTIC_KIND_VIEWPORTARRAYINDEX]   = VKD3D_SHADER_SV_VIEWPORT_ARRAY_INDEX,
    [SEMANTIC_KIND_CLIPDISTANCE]         = VKD3D_SHADER_SV_CLIP_DISTANCE,
    [SEMANTIC_KIND_CULLDISTANCE]         = VKD3D_SHADER_SV_CULL_DISTANCE,
    [SEMANTIC_KIND_OUTPUTCONTROLPOINTID] = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_DOMAINLOCATION]       = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_PRIMITIVEID]          = VKD3D_SHADER_SV_PRIMITIVE_ID,
    [SEMANTIC_KIND_GSINSTANCEID]         = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_SAMPLEINDEX]          = VKD3D_SHADER_SV_SAMPLE_INDEX,
    [SEMANTIC_KIND_ISFRONTFACE]          = VKD3D_SHADER_SV_IS_FRONT_FACE,
    [SEMANTIC_KIND_COVERAGE]             = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_INNERCOVERAGE]        = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_TARGET]               = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_DEPTH]                = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_DEPTHLESSEQUAL]       = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_DEPTHGREATEREQUAL]    = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_STENCILREF]           = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_DISPATCHTHREADID]     = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_GROUPID]              = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_GROUPINDEX]           = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_GROUPTHREADID]        = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_TESSFACTOR]           = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_INSIDETESSFACTOR]     = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_VIEWID]               = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_BARYCENTRICS]         = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_SHADINGRATE]          = VKD3D_SHADER_SV_NONE,
    [SEMANTIC_KIND_CULLPRIMITIVE]        = VKD3D_SHADER_SV_NONE,
};

static enum vkd3d_shader_sysval_semantic sysval_semantic_from_dxil_semantic_kind(enum dxil_semantic_kind kind,
        enum vkd3d_tessellator_domain domain)
{
    if (kind == SEMANTIC_KIND_TESSFACTOR)
    {
        switch (domain)
        {
            case VKD3D_TESSELLATOR_DOMAIN_LINE:
                return VKD3D_SHADER_SV_TESS_FACTOR_LINEDET;
            case VKD3D_TESSELLATOR_DOMAIN_TRIANGLE:
                return VKD3D_SHADER_SV_TESS_FACTOR_TRIEDGE;
            case VKD3D_TESSELLATOR_DOMAIN_QUAD:
                return VKD3D_SHADER_SV_TESS_FACTOR_QUADEDGE;
            default:
                FIXME("Unhandled domain %u.\n", domain);
                return VKD3D_SHADER_SV_TESS_FACTOR_TRIEDGE;
        }
    }
    else if (kind == SEMANTIC_KIND_INSIDETESSFACTOR)
    {
        switch (domain)
        {
            case VKD3D_TESSELLATOR_DOMAIN_LINE:
                return VKD3D_SHADER_SV_TESS_FACTOR_LINEDEN;
            case VKD3D_TESSELLATOR_DOMAIN_TRIANGLE:
                return VKD3D_SHADER_SV_TESS_FACTOR_TRIINT;
            case VKD3D_TESSELLATOR_DOMAIN_QUAD:
                return VKD3D_SHADER_SV_TESS_FACTOR_QUADINT;
            default:
                FIXME("Unhandled domain %u.\n", domain);
                return VKD3D_SHADER_SV_TESS_FACTOR_TRIEDGE;
        }
    }
    else if (kind < ARRAY_SIZE(sysval_semantic_table))
    {
        return sysval_semantic_table[kind];
    }
    else
    {
        FIXME("Unhandled kind %u.\n", kind);
        return VKD3D_SHADER_SV_NONE;
    }
}

static enum vkd3d_shader_register_type shader_io_register_type_from_dxil_semantic_kind(enum dxil_semantic_kind kind)
{
    if (kind == SEMANTIC_KIND_ARBITRARY || kind == SEMANTIC_KIND_TARGET)
        return VKD3DSPR_INVALID;

    switch (kind)
    {
        case SEMANTIC_KIND_COVERAGE:
        case SEMANTIC_KIND_INNERCOVERAGE:
            return VKD3DSPR_COVERAGE;
        case SEMANTIC_KIND_TARGET:
            return VKD3DSPR_COLOROUT;
        case SEMANTIC_KIND_DEPTH:
            return VKD3DSPR_DEPTHOUT;
        case SEMANTIC_KIND_DEPTHLESSEQUAL:
            return VKD3DSPR_DEPTHOUTLE;
        case SEMANTIC_KIND_DEPTHGREATEREQUAL:
            return VKD3DSPR_DEPTHOUTGE;
        case SEMANTIC_KIND_STENCILREF:
            return VKD3DSPR_OUTSTENCILREF;
        default:
            return VKD3DSPR_INVALID;
    }
}

static void metadata_node_read_additional_element_values(const struct sm6_metadata_node *node,
        struct signature_element *e, struct sm6_parser *sm6)
{
    unsigned int i, operand_count, tag, value;

    if (node->operand_count < 11 || !node->operands[10])
        return;

    if (!sm6_metadata_value_is_node(node->operands[10]))
    {
        WARN("Additional values array is not a node.\n");
        return;
    }

    node = node->operands[10]->u.node;
    if (node->operand_count & 1)
        WARN("Operand count is odd.\n");
    operand_count = node->operand_count & ~1;

    for (i = 0; i < operand_count; i += 2)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &tag)
                || !sm6_metadata_get_uint_value(sm6, node->operands[i + 1], &value))
        {
            WARN("Failed to extract tag/value pair.\n");
            continue;
        }
        switch (tag)
        {
            case 2:
                /* Something to do with relative addressing but doesn't seem useful. */
                break;
            case 3:
                e->used_mask = value;
                break;
            default:
                WARN("Unhandled tag %u, value %u.\n", tag, value);
                break;
        }
    }
}

static bool sm6_parser_read_signature(struct sm6_parser *sm6, const struct sm6_metadata_value *m,
        struct shader_signature *s, enum vkd3d_tessellator_domain tessellator_domain)
{
    const struct sm6_metadata_node *node, *element_node;
    enum vkd3d_shader_register_type reg_type;
    struct signature_element *elements, *e;
    unsigned int i, j, operand_count;
    unsigned int values[10];

    if (!m)
        return true;

    if (!sm6_metadata_value_is_node(m))
    {
        WARN("Signatures list is not a node.\n");
        return false;
    }

    node = m->u.node;
    operand_count = node->operand_count;

    if (operand_count > s->element_count)
    {
        FIXME("The DXBC signature has fewer elements than the DXIL signature.\n");
        return false;
    }
    if (!(elements = vkd3d_malloc(operand_count * sizeof(*elements))))
    {
        ERR("Failed to allocate signature.\n");
        return false;
    }

    for (i = 0; i < operand_count; ++i)
    {
        m = node->operands[i];
        if (!sm6_metadata_value_is_node(m))
        {
            WARN("Signature element is not a node.\n");
            return false;
        }

        element_node = m->u.node;
        if (element_node->operand_count < 10)
        {
            WARN("Invalid operand count %u.\n", element_node->operand_count);
            return false;
        }
        if (element_node->operand_count > 11)
            WARN("Ignoring %u additional operands.\n", element_node->operand_count);

        for (j = 0; j < 10; ++j)
        {
            /* 1 is the semantic name, 4 is semantic index metadata */
            if (j == 1 || j == 4)
                continue;
            if (!sm6_metadata_get_uint_value(sm6, element_node->operands[j], &values[j]))
            {
                WARN("Failed to load uint value at index %u.\n", j);
                return false;
            }
        }

        e = &elements[i];

        if (values[0] != i)
        {
            WARN("Element id is not an index.\n");
            return false;
        }

        if (!sm6_metadata_value_is_string(element_node->operands[1]))
        {
            WARN("Element name is not a string.\n");
            return false;
        }
        e->semantic_name = element_node->operands[1]->u.string_value;
        e->stream_index = 0;
        e->component_type = vkd3d_component_type_from_dxil_component_type(values[2]);

        e->interpolation_mode = values[5];

        if ((e->register_count = values[6]) > 4)
        {
            WARN("Invalid row count %u.\n", e->register_count);
            return false;
        }
        if ((j = values[7]) > 4)
        {
            WARN("Invalid column count %u.\n", j);
            return false;
        }
        e->mask = vkd3d_write_mask_from_component_count(j);
        e->used_mask = e->mask;
        metadata_node_read_additional_element_values(element_node, e, sm6);
        e->register_index = values[8];
        if ((j = values[9]) < VKD3D_VEC4_SIZE)
        {
            e->mask <<= j;
            e->used_mask <<= j;
        }
        if (e->mask > VKD3DSP_WRITEMASK_ALL)
        {
            WARN("Mask %#x has more than 4 components.\n", e->mask);
            e->mask &= VKD3DSP_WRITEMASK_ALL;
        }
        e->min_precision = VKD3D_SHADER_MINIMUM_PRECISION_NONE;
        if ((j = values[3]) >= SEMANTIC_KIND_INVALID)
        {
            WARN("Unknown semantic %u.\n", j);
            j = SEMANTIC_KIND_ARBITRARY;
        }
        e->sysval_semantic = sysval_semantic_from_dxil_semantic_kind(j, tessellator_domain);
        if ((reg_type = shader_io_register_type_from_dxil_semantic_kind(j)) != VKD3DSPR_INVALID)
            e->register_index = MAX_REG_OUTPUT + reg_type;
        if (!reg_type && j != SEMANTIC_KIND_ARBITRARY && e->sysval_semantic == VKD3D_SHADER_SV_NONE)
            FIXME("Unhandled kind %u.\n", j);

        /* bool I/O for non-sysvals is invalid in SPIR-V, and DXIL reads/writes uint. */
        if (e->component_type == VKD3D_SHADER_COMPONENT_BOOL && !e->sysval_semantic)
            e->component_type = VKD3D_SHADER_COMPONENT_UINT;

        m = element_node->operands[4];
        if (!sm6_metadata_value_is_node(m))
        {
            WARN("Semantic index list is not a node.\n");
            return false;
        }

        element_node = m->u.node;
        if (element_node->operand_count)
        {
            if (!sm6_metadata_get_uint_value(sm6, element_node->operands[0], &j))
            {
                WARN("Failed to get semantic index %u.\n", j);
                e->semantic_index = 0;
            }
            else
            {
                e->semantic_index = j;
            }
        }
    }

    vkd3d_free(s->elements);
    s->elements = elements;
    s->element_count = operand_count;

    return true;
}

static bool sm6_parser_signatures_init(struct sm6_parser *sm6, const struct sm6_metadata_value *m,
        enum vkd3d_tessellator_domain tessellator_domain)
{
    if (!sm6_metadata_value_is_node(m))
    {
        WARN("Signatures table is not a node.\n");
        return false;
    }

    if (m->u.node->operand_count && !sm6_parser_read_signature(sm6, m->u.node->operands[0],
            &sm6->p.shader_desc.input_signature, tessellator_domain))
        return false;
    if (m->u.node->operand_count > 1 && !sm6_parser_read_signature(sm6, m->u.node->operands[1],
            &sm6->p.shader_desc.output_signature, tessellator_domain))
        return false;
    if (m->u.node->operand_count > 2 && !sm6_parser_read_signature(sm6, m->u.node->operands[2],
            &sm6->p.shader_desc.patch_constant_signature, tessellator_domain))
        return false;

    sm6_parser_init_input_signature(sm6, &sm6->p.shader_desc.input_signature);
    sm6_parser_init_output_signature(sm6, &sm6->p.shader_desc.output_signature);
    sm6_parser_init_patch_constant_signature(sm6, &sm6->p.shader_desc.patch_constant_signature);

    return true;
}

static void sm6_parser_emit_global_flags(struct sm6_parser *sm6, const struct sm6_metadata_value *m)
{
    enum vkd3d_shader_global_flags global_flags;
    struct vkd3d_shader_instruction *ins;
    unsigned int mask, rotated_flags;

    if (!sm6_metadata_get_uint64_value(sm6, m, &global_flags))
    {
        WARN("Failed to load global flags.\n");
        return;
    }
    /* Rotate SKIP_OPTIMIZATION from bit 0 to bit 4. */
    mask = VKD3DSGF_ENABLE_MINIMUM_PRECISION - 1;
    rotated_flags = global_flags & mask;
    rotated_flags = (rotated_flags >> 1) | ((rotated_flags & 1) << 4);
    global_flags = (global_flags & ~mask) | rotated_flags;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_GLOBAL_FLAGS);
    ins->declaration.global_flags = global_flags;
}

static void sm6_parser_emit_thread_group(struct sm6_parser *sm6, const struct sm6_metadata_value *m)
{
    struct vkd3d_shader_thread_group_size thread_group_size = {1, 1, 1};
    const struct sm6_metadata_node *node;
    struct vkd3d_shader_instruction *ins;
    unsigned int group_sizes[3];
    unsigned int i;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_THREAD_GROUP);
    ins->declaration.thread_group_size = thread_group_size;

    if (sm6->p.shader_version.type != VKD3D_SHADER_TYPE_COMPUTE)
    {
        WARN("Non-compute shader has thread group dimensions.\n");
        return;
    }

    if (!m || !sm6_metadata_value_is_node(m))
    {
        WARN("Missing thread group dimension node.\n");
        return;
    }

    node = m->u.node;
    if (node->operand_count < 3)
        WARN("Unspecified dimensions are defaulting to 1.\n");

    for (i = 0; i < min(node->operand_count, 3); ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &group_sizes[i]))
        {
            WARN("Thread group dimension is not a uint value.\n");
            return;
        }
    }

    thread_group_size.x = group_sizes[0];
    thread_group_size.y = group_sizes[1];
    thread_group_size.z = group_sizes[2];
    ins->declaration.thread_group_size = thread_group_size;
}

static void sm6_parser_emit_dcl_count(struct sm6_parser *sm6, enum vkd3d_shader_opcode handler_idx, unsigned int count)
{
    struct vkd3d_shader_instruction *ins;

    ins = sm6_parser_add_instruction(sm6, handler_idx);
    ins->declaration.count = count;
}

static void sm6_parser_emit_dcl_primitive_topology(struct sm6_parser *sm6,
        enum vkd3d_shader_opcode handler_idx, enum vkd3d_primitive_type primitive_type,
        unsigned int patch_vertex_count)
{
    struct vkd3d_shader_instruction *ins;

    ins = sm6_parser_add_instruction(sm6, handler_idx);
    ins->declaration.primitive_type.type = primitive_type;
    ins->declaration.primitive_type.patch_vertex_count = patch_vertex_count;
}

static void sm6_parser_emit_dcl_tessellator_domain(struct sm6_parser *sm6,
        enum vkd3d_tessellator_domain tessellator_domain)
{
    struct vkd3d_shader_instruction *ins;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TESSELLATOR_DOMAIN);
    ins->declaration.tessellator_domain = tessellator_domain;
}

static void sm6_parser_emit_dcl_tessellator_partitioning(struct sm6_parser *sm6,
        enum vkd3d_shader_tessellator_partitioning tessellator_partitioning)
{
    struct vkd3d_shader_instruction *ins;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TESSELLATOR_PARTITIONING);
    ins->declaration.tessellator_partitioning = tessellator_partitioning;
}

static void sm6_parser_emit_dcl_tessellator_output_primitive(struct sm6_parser *sm6,
        enum vkd3d_shader_tessellator_output_primitive primitive)
{
    struct vkd3d_shader_instruction *ins;

    ins = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TESSELLATOR_OUTPUT_PRIMITIVE);
    ins->declaration.tessellator_output_primitive = primitive;
}

static void sm6_parser_gs_properties_init(struct sm6_parser *sm6, const struct sm6_metadata_value *m)
{
    unsigned int i, input_control_point_count = 1, patch_vertex_count = 0;
    enum vkd3d_primitive_type input_primitive;
    const struct sm6_metadata_node *node;
    unsigned int operands[5] = {0};

    if (!m || !sm6_metadata_value_is_node(m))
    {
        WARN("Missing GS properties.\n");
        return;
    }

    node = m->u.node;
    for (i = 0; i < min(node->operand_count, ARRAY_SIZE(operands)); ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &operands[i]))
            WARN("GS property at index %u is not a uint value.\n", i);
    }

    switch (operands[0])
    {
        case INPUT_PRIMITIVE_UNDEFINED:
            input_primitive = VKD3D_PT_UNDEFINED;
            break;
        case INPUT_PRIMITIVE_POINT:
            input_primitive = VKD3D_PT_POINTLIST;
            input_control_point_count = 1;
            break;
        case INPUT_PRIMITIVE_LINE:
            input_primitive = VKD3D_PT_LINELIST;
            input_control_point_count = 2;
            break;
        case INPUT_PRIMITIVE_TRIANGLE:
            input_primitive = VKD3D_PT_TRIANGLELIST;
            input_control_point_count = 3;
            break;
        case INPUT_PRIMITIVE_LINEWITHADJACENCY:
            input_primitive = VKD3D_PT_LINELIST_ADJ;
            input_control_point_count = 4;
            break;
        case INPUT_PRIMITIVE_TRIANGLEWITHADJACENY:
            input_primitive = VKD3D_PT_TRIANGLELIST_ADJ;
            input_control_point_count = 6;
            break;
        default:
            if (operands[0] >= INPUT_PRIMITIVE_PATCH1 && operands[0] <= INPUT_PRIMITIVE_PATCH32)
            {
                input_primitive = VKD3D_PT_PATCH;
                patch_vertex_count = operands[0] - INPUT_PRIMITIVE_PATCH1 + 1;
                break;
            }
            FIXME("Unhandled input primitive %u.\n", operands[0]);
            input_primitive = VKD3D_PT_UNDEFINED;
            break;
    }
    sm6_parser_emit_dcl_primitive_topology(sm6, VKD3DSIH_DCL_INPUT_PRIMITIVE, input_primitive, patch_vertex_count);
    sm6->input_control_point_count = input_control_point_count;
    sm6_parser_emit_dcl_count(sm6, VKD3DSIH_DCL_VERTICES_OUT, operands[1]);
    sm6_parser_emit_dcl_primitive_topology(sm6, VKD3DSIH_DCL_OUTPUT_TOPOLOGY, operands[3], 0);
    sm6_parser_emit_dcl_count(sm6, VKD3DSIH_DCL_GS_INSTANCES,
            (node->operand_count > 4 && operands[4]) ? operands[4] : 1);
}

static enum vkd3d_tessellator_domain sm6_parser_ds_properties_init(struct sm6_parser *sm6,
        const struct sm6_metadata_value *m)
{
    const struct sm6_metadata_node *node;
    unsigned int operands[2] = {0};
    unsigned int i;

    if (!m || !sm6_metadata_value_is_node(m))
    {
        WARN("Missing DS properties.\n");
        return 0;
    }

    node = m->u.node;
    for (i = 0; i < min(node->operand_count, ARRAY_SIZE(operands)); ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &operands[i]))
            WARN("DS property at index %u is not a uint value.\n", i);
    }

    sm6_parser_emit_dcl_tessellator_domain(sm6, operands[0]);
    sm6_parser_emit_dcl_count(sm6, VKD3DSIH_DCL_INPUT_CONTROL_POINT_COUNT, operands[1]);
    sm6->input_control_point_count = operands[1];

    return operands[0];
}

static enum vkd3d_tessellator_domain sm6_parser_hs_properties_init(struct sm6_parser *sm6,
        const struct sm6_metadata_value *m)
{
    const struct sm6_metadata_node *node;
    unsigned int operands[6] = {0};
    unsigned int i;

    if (!m || !sm6_metadata_value_is_node(m))
    {
        WARN("Missing HS properties.\n");
        return 0;
    }

    node = m->u.node;

    if (!sm6_metadata_value_is_value(node->operands[0]))
        WARN("Patch constant function node is not a value.\n");
    else
        sm6->patch_constant_function = node->operands[0]->u.value->u.function.name;

    for (i = 1; i < min(node->operand_count, ARRAY_SIZE(operands)); ++i)
    {
        if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &operands[i]))
            WARN("HS property at index %u is not a uint value.\n", i);
    }

    sm6_parser_emit_dcl_count(sm6, VKD3DSIH_DCL_INPUT_CONTROL_POINT_COUNT, operands[1]);
    sm6->input_control_point_count = operands[1];
    sm6_parser_emit_dcl_count(sm6, VKD3DSIH_DCL_OUTPUT_CONTROL_POINT_COUNT, operands[2]);
    sm6->output_control_point_count = operands[2];
    sm6_parser_emit_dcl_tessellator_domain(sm6, operands[3]);
    sm6_parser_emit_dcl_tessellator_partitioning(sm6, operands[4]);
    sm6_parser_emit_dcl_tessellator_output_primitive(sm6, operands[5]);

    return operands[3];
}

static enum vkd3d_result sm6_parser_entry_point_init(struct sm6_parser *sm6)
{
    const struct sm6_metadata_value *m = sm6_parser_find_named_metadata(sm6, "dx.entryPoints");
    enum vkd3d_tessellator_domain tessellator_domain = 0;
    const struct sm6_metadata_node *node, *entry_node;
    const struct sm6_value *value;
    unsigned int i, tag;

    if (!m)
        return VKD3D_OK;

    if (!sm6_parser_require_space(sm6, 6))
    {
        ERR("Failed to allocate instructions.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    entry_node = m->u.node;
    if (!entry_node->operand_count)
    {
        WARN("No entry point definition found.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }

    if (!(m = entry_node->operands[0]))
        return VKD3D_OK;

    if (m->type != VKD3D_METADATA_VALUE)
    {
        WARN("Entry point definition is not a value.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }

    value = m->u.value;
    if (!sm6_value_is_function_dcl(value))
    {
        WARN("Entry point value is not a function definition.\n");
        return VKD3D_ERROR_INVALID_SHADER;
    }

    sm6->entry_point = value->u.function.name;

    if (entry_node->operand_count >= 5 && (m = entry_node->operands[4]))
    {
        if (!sm6_metadata_value_is_node(m))
        {
            WARN("Compute thread info is not a node.\n");
            return VKD3D_ERROR_INVALID_SHADER;
        }

        node = m->u.node;
        for (i = 0; i < node->operand_count - 1; i += 2)
        {
            if (!sm6_metadata_get_uint_value(sm6, node->operands[i], &tag))
            {
                WARN("Tag is not a uint value.\n");
                return VKD3D_ERROR_INVALID_SHADER;
            }

            switch (tag)
            {
                case SHADER_PROPERTIES_FLAGS:
                    sm6_parser_emit_global_flags(sm6, node->operands[i + 1]);
                    break;
                case SHADER_PROPERTIES_GEOMETRY:
                    sm6_parser_gs_properties_init(sm6, node->operands[i + 1]);
                    break;
                case SHADER_PROPERTIES_DOMAIN:
                    tessellator_domain = sm6_parser_ds_properties_init(sm6, node->operands[i + 1]);
                    break;
                case SHADER_PROPERTIES_HULL:
                    tessellator_domain = sm6_parser_hs_properties_init(sm6, node->operands[i + 1]);
                    break;
                case SHADER_PROPERTIES_COMPUTE:
                    sm6_parser_emit_thread_group(sm6, node->operands[i + 1]);
                    break;
                default:
                    FIXME("Unhandled tag %#x.\n", tag);
                    break;
            }
        }
    }

    if (entry_node->operand_count >= 3 && (m = entry_node->operands[2]))
        sm6_parser_signatures_init(sm6, m, tessellator_domain);

    return VKD3D_OK;
}

static void sm6_metadata_value_destroy(struct sm6_metadata_value *m)
{
    switch (m->type)
    {
        case VKD3D_METADATA_NODE:
            vkd3d_free(m->u.node);
            break;
        case VKD3D_METADATA_KIND:
            vkd3d_free(m->u.kind.name);
            break;
        case VKD3D_METADATA_STRING:
            vkd3d_free(m->u.string_value);
            break;
        default:
            break;
    }
}

static void sm6_parser_metadata_cleanup(struct sm6_parser *sm6)
{
    unsigned int i, j;

    for (i = 0; i < ARRAY_SIZE(sm6->metadata_tables); ++i)
    {
        for (j = 0; j < sm6->metadata_counts[i]; ++j)
            sm6_metadata_value_destroy(&sm6->metadata_tables[i][j]);
        vkd3d_free(sm6->metadata_tables[i]);
    }
    for (i = 0; i < sm6->named_metadata_count; ++i)
    {
        sm6_metadata_value_destroy(&sm6->named_metadata[i].value);
        vkd3d_free(sm6->named_metadata[i].name);
    }
    vkd3d_free(sm6->named_metadata);
}

static void sm6_type_table_cleanup(struct sm6_type *types, unsigned int count)
{
    unsigned int i;

    if (!types)
        return;

    for (i = 0; i < count; ++i)
    {
        switch (types[i].class)
        {
            case TYPE_CLASS_STRUCT:
                vkd3d_free((void *)types[i].u.struc->name);
                vkd3d_free(types[i].u.struc);
                break;
            case TYPE_CLASS_FUNCTION:
                vkd3d_free(types[i].u.function);
                break;
            default:
                break;
        }
    }

    vkd3d_free(types);
}

static void sm6_symtab_cleanup(struct sm6_symbol *symbols, unsigned int count)
{
    unsigned int i;

    for (i = 0; i < count; ++i)
        vkd3d_free((void *)symbols[i].name);
    vkd3d_free(symbols);
}

void sm6_phi_destroy(struct sm6_phi *phi)
{
    vkd3d_free(phi->name);
    vkd3d_free(phi->incoming);
}

static void sm6_block_destroy(struct sm6_block *block)
{
    unsigned int i;

    vkd3d_free(block->instructions);
    vkd3d_free(block->name);
    for (i = 0; i < block->phi_count; ++i)
        sm6_phi_destroy(&block->phi[i]);
    vkd3d_free(block->phi);
    sm6_block_array_destroy(&block->headers);
    sm6_block_array_destroy(&block->pred);
    sm6_block_array_destroy(&block->succ);
    sm6_block_array_destroy(&block->fake_succ);
    sm6_block_array_destroy(&block->fake_pred);
    sm6_block_array_destroy(&block->dominance_frontier);
    sm6_block_array_destroy(&block->post_dominance_frontier);
    vkd3d_free(block->terminator.cases);
    vkd3d_free(block);
}

static void sm6_functions_cleanup(struct sm6_function *functions, unsigned int count)
{
    unsigned int i, j;

    for (i = 0; i < count; ++i)
    {
        for (j = 0; j < functions[i].block_count; ++j)
            sm6_block_destroy(functions[i].blocks[j]);
        vkd3d_free(functions[i].blocks);
    }
    vkd3d_free(functions);
}

static void sm6_parser_destroy(struct vkd3d_shader_parser *parser)
{
    struct sm6_parser *sm6 = sm6_parser(parser);

    dxil_block_destroy(&sm6->root_block);
    dxil_global_abbrevs_cleanup(sm6->abbrevs, sm6->abbrev_count);
    shader_instruction_array_destroy(&parser->instructions);
    sm6_type_table_cleanup(sm6->types, sm6->type_count);
    sm6_symtab_cleanup(sm6->global_symbols, sm6->global_symbol_count);
    vkd3d_free(sm6->output_params);
    vkd3d_free(sm6->input_params);
    vkd3d_free(sm6->patch_constant_params);
    sm6_functions_cleanup(sm6->functions, sm6->function_count);
    sm6_parser_metadata_cleanup(sm6);
    vkd3d_free(sm6->resources);
    vkd3d_free(sm6->values);
    free_shader_desc(&parser->shader_desc);
    vkd3d_free(sm6);
}

static const struct vkd3d_shader_parser_ops sm6_parser_ops =
{
    .parser_destroy = sm6_parser_destroy,
};

static struct sm6_function *sm6_parser_get_function(const struct sm6_parser *sm6, const char *name)
{
    unsigned int i;
    for (i = 0; i < sm6->function_count; ++i)
        if (!ascii_strcasecmp(sm6->functions[i].declaration->u.function.name, name))
            return &sm6->functions[i];
    return NULL;
}

static enum vkd3d_result sm6_parser_init(struct sm6_parser *sm6, const uint32_t *byte_code, size_t byte_code_size,
        const char *source_name, struct vkd3d_shader_message_context *message_context)
{
    const struct shader_signature *patch_constant_signature = &sm6->p.shader_desc.patch_constant_signature;
    const struct shader_signature *output_signature = &sm6->p.shader_desc.output_signature;
    const struct shader_signature *input_signature = &sm6->p.shader_desc.input_signature;
    const struct vkd3d_shader_location location = {.source_name = source_name};
    unsigned int count, length, chunk_offset, chunk_size, function_count;
    uint32_t version_token, dxil_version, token_count, magic;
    enum bitcode_block_abbreviation abbr;
    struct vkd3d_shader_instruction *ins;
    struct vkd3d_shader_version version;
    struct dxil_block *block;
    enum vkd3d_result ret;
    unsigned int i, j;

    count = byte_code_size / sizeof(*byte_code);
    if (count < 6)
    {
        WARN("Invalid data size %zu.\n", byte_code_size);
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_SIZE,
                "DXIL chunk size %zu is smaller than the DXIL header size.", byte_code_size);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    version_token = byte_code[0];
    TRACE("Compiler version: 0x%08x.\n", version_token);
    token_count = byte_code[1];
    TRACE("Token count: %u.\n", token_count);

    if (token_count < 6 || count < token_count)
    {
        WARN("Invalid token count %u (word count %u).\n", token_count, count);
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_CHUNK_SIZE,
                "DXIL chunk token count %#x is invalid (word count %u).", token_count, count);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    if (byte_code[2] != TAG_DXIL)
        WARN("Unknown magic number 0x%08x.\n", byte_code[2]);

    dxil_version = byte_code[3];
    if (dxil_version > 0x102)
        WARN("Unknown DXIL version: 0x%08x.\n", dxil_version);
    else
        TRACE("DXIL version: 0x%08x.\n", dxil_version);

    chunk_offset = byte_code[4];
    if (chunk_offset < 16 || chunk_offset >= byte_code_size)
    {
        WARN("Invalid bitcode chunk offset %#x (data size %zu).\n", chunk_offset, byte_code_size);
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_CHUNK_OFFSET,
                "DXIL bitcode chunk has invalid offset %#x (data size %#zx).", chunk_offset, byte_code_size);
        return VKD3D_ERROR_INVALID_SHADER;
    }
    chunk_size = byte_code[5];
    if (chunk_size > byte_code_size - chunk_offset)
    {
        WARN("Invalid bitcode chunk size %#x (data size %zu, chunk offset %#x).\n",
                chunk_size, byte_code_size, chunk_offset);
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_CHUNK_SIZE,
                "DXIL bitcode chunk has invalid size %#x (data size %#zx, chunk offset %#x).",
                chunk_size, byte_code_size, chunk_offset);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    sm6->start = (const uint32_t *)((const char*)&byte_code[2] + chunk_offset);
    if ((magic = sm6->start[0]) != BITCODE_MAGIC)
    {
        WARN("Unknown magic number 0x%08x.\n", magic);
        vkd3d_shader_parser_warning(&sm6->p, VKD3D_SHADER_WARNING_DXIL_UNKNOWN_MAGIC_NUMBER,
                "DXIL bitcode chunk magic number 0x%08x is not the expected 0x%08x.", magic, BITCODE_MAGIC);
    }

    sm6->end = &sm6->start[(chunk_size + sizeof(*sm6->start) - 1) / sizeof(*sm6->start)];

    if ((version.type = version_token >> 16) >= VKD3D_SHADER_TYPE_COUNT)
    {
        FIXME("Unknown shader type %#x.\n", version.type);
        vkd3d_shader_parser_warning(&sm6->p, VKD3D_SHADER_WARNING_DXIL_UNKNOWN_SHADER_TYPE,
                "Unknown shader type %#x.", version.type);
    }

    version.major = VKD3D_SM6_VERSION_MAJOR(version_token);
    version.minor = VKD3D_SM6_VERSION_MINOR(version_token);

    if ((abbr = sm6->start[1] & 3) != ENTER_SUBBLOCK)
    {
        WARN("Initial block abbreviation %u is not ENTER_SUBBLOCK.\n", abbr);
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_BITCODE,
                "DXIL bitcode chunk has invalid initial block abbreviation %u.", abbr);
        return VKD3D_ERROR_INVALID_SHADER;
    }

    /* Estimate instruction count to avoid reallocation in most shaders. */
    count = max(token_count, 400) - 400;
    vkd3d_shader_parser_init(&sm6->p, message_context, source_name, &version, &sm6_parser_ops,
            (count + (count >> 2)) / 2u + 10);
    sm6->ptr = &sm6->start[1];
    sm6->bitpos = 2;

    block = &sm6->root_block;
    if ((ret = dxil_block_init(block, NULL, sm6)) < 0)
    {
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL bitcode chunk.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_BITCODE,
                    "DXIL bitcode chunk has invalid bitcode.");
        else
            vkd3d_unreachable();
        return ret;
    }

    dxil_global_abbrevs_cleanup(sm6->abbrevs, sm6->abbrev_count);
    sm6->abbrevs = NULL;
    sm6->abbrev_count = 0;

    length = sm6->ptr - sm6->start - block->start;
    if (length != block->length)
    {
        WARN("Invalid block length %u; expected %u.\n", length, block->length);
        vkd3d_shader_parser_warning(&sm6->p, VKD3D_SHADER_WARNING_DXIL_INVALID_BLOCK_LENGTH,
                "Root block ends with length %u but indicated length is %u.", length, block->length);
    }
    if (sm6->ptr != sm6->end)
    {
        unsigned int expected_length = sm6->end - sm6->start;
        length = sm6->ptr - sm6->start;
        WARN("Invalid module length %u; expected %u.\n", length, expected_length);
        vkd3d_shader_parser_warning(&sm6->p, VKD3D_SHADER_WARNING_DXIL_INVALID_MODULE_LENGTH,
                "Module ends with length %u but indicated length is %u.", length, expected_length);
    }

    if ((ret = sm6_parser_type_table_init(sm6)) < 0)
    {
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL type table.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_TYPE_TABLE,
                    "DXIL type table is invalid.");
        else
            vkd3d_unreachable();
        return ret;
    }

    if ((ret = sm6_parser_symtab_init(sm6)) < 0)
    {
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL value symbol table.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_VALUE_SYMTAB,
                    "DXIL value symbol table is invalid.");
        else
            vkd3d_unreachable();
        return ret;
    }

    if (!(sm6->output_params = vkd3d_malloc(output_signature->element_count * sizeof(*sm6->output_params)))
            || !(sm6->input_params = vkd3d_malloc(input_signature->element_count * sizeof(*sm6->input_params)))
            || !(sm6->patch_constant_params = vkd3d_malloc(patch_constant_signature->element_count
            * sizeof(*sm6->patch_constant_params))))
    {
        ERR("Failed to allocate input/output registers.\n");
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                "Out of memory allocating output parameters.");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    function_count = dxil_block_compute_function_count(&sm6->root_block);
    if (!(sm6->functions = vkd3d_calloc(function_count, sizeof(*sm6->functions))))
    {
        ERR("Failed to allocate function array.\n");
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                "Out of memory allocating DXIL function array.");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    sm6_parser_compute_max_value_count(sm6, &sm6->root_block, 0);
    if (!(sm6->values = vkd3d_calloc(sm6->value_capacity, sizeof(*sm6->values))))
    {
        ERR("Failed to allocate value array.\n");
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                "Out of memory allocating DXIL value array.");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }
    sm6->function_count = 0;
    sm6->ssa_next_id = 1;

    if ((ret = sm6_parser_globals_init(sm6)) < 0)
    {
        WARN("Failed to load global declarations.\n");
        return ret;
    }

    if (!sm6_parser_allocate_named_metadata(sm6))
    {
        ERR("Failed to allocate named metadata array.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    for (i = 0, j = 0; i < sm6->root_block.child_block_count; ++i)
    {
        block = sm6->root_block.child_blocks[i];
        if (block->id != METADATA_BLOCK)
            continue;

        if (j == ARRAY_SIZE(sm6->metadata_tables))
        {
            FIXME("Skipping metadata table.\n");
            ret = VKD3D_ERROR_INVALID_SHADER;
        }
        else
        {
            ret = sm6_parser_metadata_init(sm6, block, j++);
        }
        if (ret < 0)
        {
            WARN("Failed to load metadata.\n");
            if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
                vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                        "Out of memory parsing DXIL metadata tables.");
            else if (ret == VKD3D_ERROR_INVALID_SHADER)
                vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_METADATA,
                        "A DXIL metadata table is invalid.");
            else
                vkd3d_unreachable();
            return ret;
        }
    }

    if ((ret = sm6_parser_entry_point_init(sm6)) < 0)
    {
        WARN("Failed to parse shader properties.\n");
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL shader properties.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_PROPERTIES,
                    "The DXIL shader properties are invalid.");
        else
            vkd3d_unreachable();
        return ret;
    }

    if ((ret = sm6_parser_resources_init(sm6)) < 0)
    {
        WARN("Failed to init resources.\n");
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL resources.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_RESOURCE,
                    "A DXIL resource description is invalid.");
        else
            vkd3d_unreachable();
        return ret;
    }

    if ((ret = sm6_parser_module_init(sm6, &sm6->root_block, 0)) < 0)
    {
        if (ret == VKD3D_ERROR_OUT_OF_MEMORY)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                    "Out of memory parsing DXIL module.");
        else if (ret == VKD3D_ERROR_INVALID_SHADER)
            vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_INVALID_MODULE,
                    "DXIL module is invalid.");
        else
            vkd3d_unreachable();
        return ret;
    }

    if (!sm6_parser_require_space(sm6, output_signature->element_count + input_signature->element_count
            + patch_constant_signature->element_count + 3))
    {
        vkd3d_shader_error(message_context, &location, VKD3D_SHADER_ERROR_DXIL_OUT_OF_MEMORY,
                "Out of memory emitting shader signature declarations.");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    j = sm6_parser_add_instruction(sm6, VKD3DSIH_DCL_TEMPS) - sm6->p.instructions.elements;

    if (version.type == VKD3D_SHADER_TYPE_HULL)
    {
        struct sm6_function *fn;
        if (!(fn = sm6_parser_get_function(sm6, sm6->entry_point)))
        {
            FIXME("Failed to find entry point %s.\n", sm6->entry_point);
            return VKD3D_ERROR_INVALID_SHADER;
        }
        sm6_parser_add_instruction(sm6, VKD3DSIH_HS_CONTROL_POINT_PHASE);
        sm6_parser_emit_input_signature(sm6, input_signature);
        sm6_parser_emit_output_signature(sm6, output_signature);
        cfg_structuriser_emit_blocks(sm6, fn);
        if (!(fn = sm6_parser_get_function(sm6, sm6->patch_constant_function)))
        {
            FIXME("Failed to find patch constant function %s.\n", sm6->patch_constant_function);
            return VKD3D_ERROR_INVALID_SHADER;
        }
        sm6_parser_add_instruction(sm6, VKD3DSIH_HS_FORK_PHASE);
        sm6_parser_emit_patch_constant_signature(sm6, patch_constant_signature);
        cfg_structuriser_emit_blocks(sm6, fn);
        if (sm6->function_count > 2)
            FIXME("%u unhandled functions.\n", sm6->function_count - 2);
    }
    else
    {
        sm6_parser_emit_input_signature(sm6, input_signature);
        sm6_parser_emit_output_signature(sm6, output_signature);
        sm6_parser_emit_patch_constant_signature(sm6, patch_constant_signature);
        for (i = 0; i < sm6->function_count; ++i)
            cfg_structuriser_emit_blocks(sm6, &sm6->functions[i]);
    }

    ins = &sm6->p.instructions.elements[j];
    ins->flags = VKD3DSDTF_SSA;
    ins->declaration.count = sm6->ssa_next_id;

    dxil_block_destroy(&sm6->root_block);

    return VKD3D_OK;
}

int vkd3d_shader_sm6_parser_create(const struct vkd3d_shader_compile_info *compile_info,
        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_parser **parser)
{
    struct vkd3d_shader_desc *shader_desc;
    uint32_t *byte_code = NULL;
    struct sm6_parser *sm6;
    int ret;

    if (!(sm6 = vkd3d_calloc(1, sizeof(*sm6))))
    {
        ERR("Failed to allocate parser.\n");
        return VKD3D_ERROR_OUT_OF_MEMORY;
    }

    shader_desc = &sm6->p.shader_desc;
    shader_desc->is_dxil = true;
    if ((ret = shader_extract_from_dxbc(&compile_info->source, message_context, compile_info->source_name,
            shader_desc)) < 0)
    {
        WARN("Failed to extract shader, vkd3d result %d.\n", ret);
        vkd3d_free(sm6);
        return ret;
    }

    sm6->p.shader_desc = *shader_desc;
    shader_desc = &sm6->p.shader_desc;

    if (((uintptr_t)shader_desc->byte_code & (VKD3D_DXBC_CHUNK_ALIGNMENT - 1)))
    {
        /* LLVM bitcode should be 32-bit aligned, but before dxc v1.7.2207 this was not always the case in the DXBC
         * container due to missing padding after signature names. Get an aligned copy to prevent unaligned access. */
        if (!(byte_code = vkd3d_malloc(align(shader_desc->byte_code_size, VKD3D_DXBC_CHUNK_ALIGNMENT))))
            ERR("Failed to allocate aligned chunk. Unaligned access will occur.\n");
        else
            memcpy(byte_code, shader_desc->byte_code, shader_desc->byte_code_size);
    }

    ret = sm6_parser_init(sm6, byte_code ? byte_code : shader_desc->byte_code, shader_desc->byte_code_size,
            compile_info->source_name, message_context);
    vkd3d_free(byte_code);

    if (ret < 0)
    {
        WARN("Failed to initialise shader parser.\n");
        sm6_parser_destroy(&sm6->p);
        return ret;
    }

    *parser = &sm6->p;

    return ret;
}
