/*
 * Copyright 2022 Conor McCarthy for CodeWeavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

/*
 * This library contains code derived from the dxil-spirv subproject of
 * vkd3d-proton, the license for which follows:
 *
 * Copyright (c) 2019-2022 Hans-Kristian Arntzen for Valve Corporation
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "vkd3d_shader_private.h"

#ifdef HAVE_SPIRV_UNIFIED1_SPIRV_H
# include "spirv/unified1/spirv.h"
#else
# include "vulkan/spirv.h"
#endif  /* HAVE_SPIRV_UNIFIED1_SPIRV_H */

#define vkd3d_swap(a, b, type) do { type tmp = a; a = b; b = tmp; } while(0)

static bool sm6_block_add_unique_pred(struct sm6_block *block, struct sm6_block *pred)
{
    return sm6_block_array_append_unique(&block->pred, pred);
}

static bool sm6_block_add_unique_fake_pred(struct sm6_block *block, struct sm6_block *fake_pred)
{
    return sm6_block_array_append_unique(&block->fake_pred, fake_pred);
}

static bool sm6_block_add_unique_header(struct sm6_block *block, struct sm6_block *header)
{
    return sm6_block_array_append_unique(&block->headers, header);
}

static bool sm6_block_array_append_array(struct sm6_block_array *dst, const struct sm6_block_array *src)
{
    unsigned int i;

    if (!sm6_block_array_reserve(dst, dst->count + src->count))
        return false;
    for (i = 0; i < src->count; ++i)
        dst->blocks[dst->count++] = src->blocks[i];
    return true;
};

static inline bool sm6_block_array_contains(const struct sm6_block_array *array,
        const struct sm6_block *block)
{
    unsigned int i = sm6_block_array_find(array, block);
    return i < array->count;
}

static unsigned int sm6_block_array_contains_succ_back_edge(const struct sm6_block_array *array)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (array->blocks[i]->succ_back_edge)
            return true;
    return false;
}

static void sm6_block_array_erase(struct sm6_block_array *array, unsigned int i)
{
    memmove(&array->blocks[i], &array->blocks[i + 1], (--array->count - i) * sizeof(*array->blocks));
};

static void sm6_block_array_delete(struct sm6_block_array *array, const struct sm6_block *block)
{
    unsigned int i = sm6_block_array_find(array, block);
    if (i != array->count)
        sm6_block_array_erase(array, i);
}

/* For sorted arrays only. */
static void sm6_block_array_deduplicate(struct sm6_block_array *array)
{
    unsigned int i, j;
    for (i = 1, j = 1; i < array->count; ++i)
        if (array->blocks[i] != array->blocks[i - 1])
            array->blocks[j++] = array->blocks[i];
    array->count = j;
}

static bool sm6_block_array_clone(struct sm6_block_array *dst, const struct sm6_block_array *src)
{
    if (!sm6_block_array_reserve(dst, src->count))
        return false;
    memcpy(dst->blocks, src->blocks, src->count * sizeof(*src->blocks));
    dst->count = src->count;
    return true;
}

static void sm6_block_array_retarget(struct sm6_block_array *array, const struct sm6_block *from,
        struct sm6_block *to)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (array->blocks[i] == from)
            array->blocks[i] = to;
}

static unsigned int block_terminator_case_array_contains_block(struct sm6_block_terminator *terminator,
        const struct sm6_block *block)
{
    unsigned int i;
    for (i = 0; i < terminator->case_count; ++i)
        if (terminator->cases[i].node == block)
            return true;
    return false;
};

static bool sm6_block_dominates(const struct sm6_block *block, const struct sm6_block *other)
{
    /* Follow immediate dominator graph. Either we end up at this, or entry block. */
    while (block != other)
    {
        if (!other->immediate_dominator || other == other->immediate_dominator)
            break;
        other = other->immediate_dominator;
    }

    return block == other;
}

static bool sm6_block_post_dominates(const struct sm6_block *block, const struct sm6_block *start_node)
{
    while (start_node != block)
    {
        /* Reached exit node. */
        if (!start_node->immediate_post_dominator || start_node == start_node->immediate_post_dominator)
            break;
        start_node = start_node->immediate_post_dominator;
    }

    return block == start_node;
}

static bool sm6_block_post_dominates_perfect_structured_construct(const struct sm6_block *block)
{
    unsigned int i;

    if (!sm6_block_post_dominates(block, block->immediate_dominator))
        return false;

    for (i = 0; i < block->pred.count; ++i)
        if (!sm6_block_post_dominates(block, block->pred.blocks[i]))
            return false;
    return true;
}

static struct sm6_block *sm6_block_find_common_post_dominator(struct sm6_block *a,
        struct sm6_block *b);

static void sm6_block_recompute_immediate_post_dominator(struct sm6_block *block)
{
    unsigned int i;

    if (!block->succ.count && !block->fake_succ.count)
        return;

    /* For non-leaf blocks only. The immediate post dominator is already set up to be the exit node in leaf nodes. */
    block->immediate_post_dominator = NULL;
    for (i = 0; i < block->succ.count; ++i)
    {
        struct sm6_block *edge = block->succ.blocks[i];

        if (block->immediate_post_dominator)
            block->immediate_post_dominator = sm6_block_find_common_post_dominator(block->immediate_post_dominator, edge);
        else
            block->immediate_post_dominator = edge;
    }

    for (i = 0; i < block->fake_succ.count; ++i)
    {
        struct sm6_block *edge = block->fake_succ.blocks[i];

        if (block->immediate_post_dominator)
            block->immediate_post_dominator = sm6_block_find_common_post_dominator(block->immediate_post_dominator, edge);
        else
            block->immediate_post_dominator = edge;
    }
}

static struct sm6_block *sm6_block_find_common_post_dominator(struct sm6_block *a,
        struct sm6_block *b)
{
    unsigned int i;

    assert(a);
    assert(b);

    while (a != b)
    {
        if (!a->immediate_post_dominator)
        {
            for (i = 0; i < a->succ.count; ++i)
                sm6_block_recompute_immediate_post_dominator(a->succ.blocks[i]);
            sm6_block_recompute_immediate_post_dominator(a);
        }

        if (!b->immediate_post_dominator)
        {
            for (i = 0; i < b->succ.count; ++i)
                sm6_block_recompute_immediate_post_dominator(b->succ.blocks[i]);
            sm6_block_recompute_immediate_post_dominator(b);
        }

        if (a->backward_post_visit_order == b->backward_post_visit_order)
        {
            /* Should not normally happen, but when we insert ladder blocks, we might have assigned
             * temporary visit orders which can alias with other nodes in some cases. Fixing this
             * up requires a full traversal of the entire CFG, so as a fallback we can do direct
             * reachability and domination analysis. */
            if (sm6_block_post_dominates(b, a))
                return b;
            else if (sm6_block_post_dominates(a, b))
                return a;

            /* If there is no clear domination relationship, then we need to iterate both a and b.
             * This is fine as we now know that neither a nor b can be the common node. */
            assert(a->immediate_post_dominator);
            assert(b->immediate_post_dominator);
            a = a->immediate_post_dominator;
            b = b->immediate_post_dominator;
        }
        else if (a->backward_post_visit_order < b->backward_post_visit_order)
        {
            assert(a->immediate_post_dominator);
            a = a->immediate_post_dominator;
        }
        else
        {
            assert(b->immediate_post_dominator);
            b = b->immediate_post_dominator;
        }
    }
    return a;
}

static struct sm6_block *sm6_block_array_find_common_post_dominator(const struct sm6_block_array *candidates)
{
    struct sm6_block *common_post;
    unsigned int i;

    if (!candidates->count)
        return NULL;
    else if (candidates->count == 1)
        return candidates->blocks[0];

    common_post = sm6_block_find_common_post_dominator(candidates->blocks[0], candidates->blocks[1]);
    for (i = 2; i < candidates->count; ++i)
        common_post = sm6_block_find_common_post_dominator(common_post, candidates->blocks[i]);

    return common_post != common_post->immediate_post_dominator ? common_post : NULL;
}

static bool sm6_block_array_contains_dominated_block(const struct sm6_block_array *array,
        const struct sm6_block *dom)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (sm6_block_dominates(dom, array->blocks[i]))
            return true;
    return false;
};

static bool sm6_block_array_contains_post_dominated_block(const struct sm6_block_array *array,
        const struct sm6_block *post_dom)
{
    unsigned int i;
    for (i = 0; i < array->count; ++i)
        if (sm6_block_post_dominates(post_dom, array->blocks[i]))
            return true;
    return false;
};

static struct sm6_block *sm6_block_find_common_dominator(struct sm6_block *a, struct sm6_block *b);

static void sm6_block_recompute_immediate_dominator(struct sm6_block *block)
{
    unsigned int i;

    if (!block->pred.count)
    {
        /* For entry block only. */
        block->immediate_dominator = block;
    }
    else
    {
        block->immediate_dominator = NULL;

        for (i = 0; i < block->pred.count; ++i)
        {
            struct sm6_block *edge = block->pred.blocks[i];

            if (block->immediate_dominator)
                block->immediate_dominator = sm6_block_find_common_dominator(block->immediate_dominator, edge);
            else
                block->immediate_dominator = edge;
        }
    }
}

static struct sm6_block *sm6_block_find_common_dominator(struct sm6_block *a, struct sm6_block *b)
{
    unsigned int i;

    assert(a);
    assert(b);

    while (a != b)
    {
        if (!a->immediate_dominator)
        {
            for (i = 0; i < a->pred.count; ++i)
                sm6_block_recompute_immediate_dominator(a->pred.blocks[i]);
            sm6_block_recompute_immediate_dominator(a);
        }

        if (!b->immediate_dominator)
        {
            for (i = 0; i < b->pred.count; ++i)
                sm6_block_recompute_immediate_dominator(b->pred.blocks[i]);
            sm6_block_recompute_immediate_dominator(b);
        }

        if (a->forward_post_visit_order == b->forward_post_visit_order)
        {
            /* Should not normally happen, but when we insert ladder blocks, we might have assigned
             * temporary visit orders which can alias with other nodes in some cases. Fixing this
             * up requires a full traversal of the entire CFG, so as a fallback we can do direct
             * reachability and domination analysis. */
            if (sm6_block_dominates(b, a))
                return b;
            else if (sm6_block_dominates(a, b))
                return a;

            /* If there is no clear domination relationship, then we need to iterate both a and b.
             * This is fine as we now know that neither a nor b can be the common node. */
            assert(a->immediate_dominator);
            assert(b->immediate_dominator);
            if (a == a->immediate_dominator)
                return b;
            else if (b == b->immediate_dominator)
                return a;
            a = a->immediate_dominator;
            b = b->immediate_dominator;
        }
        else if (a->forward_post_visit_order < b->forward_post_visit_order)
        {
            /* Awkward case which can happen when nodes are unreachable in the CFG.
             * Can occur with the dummy blocks we create. */
            if (a == a->immediate_dominator)
                return b;
            a = a->immediate_dominator;
        }
        else
        {
            /* Awkward case which can happen when nodes are unreachable in the CFG.
             * Can occur with the dummy blocks we create. */
            if (b == b->immediate_dominator)
                return a;

            b = b->immediate_dominator;
        }
    }
    return a;
}

static unsigned int sm6_block_find_phi_by_reg(struct sm6_block *block, const struct vkd3d_shader_register *reg)
{
    unsigned int i, id;

    if (!register_is_ssa(reg))
        return block->phi_count;

    for (i = 0, id = vkd3d_shader_register_ssa_id(reg); i < block->phi_count; ++i)
        if (sm6_phi_ssa_id(&block->phi[i]) == id)
            break;
    return i;
}

static void sm6_block_erase_phi(struct sm6_block *block, unsigned int index)
{
    unsigned int i, j;

    assert(index < block->phi_count);
    vkd3d_free(block->phi[index].incoming);

    for (i = 0, j = 0; i < block->phi_count; ++i)
        if (i != index)
            block->phi[j++] = block->phi[i];
    block->phi_count = j;
}

static void sm6_block_add_branch(struct sm6_block *block, struct sm6_block *to)
{
    sm6_block_array_append_unique(&block->succ, to);
    sm6_block_array_append_unique(&to->pred, block);
}

static void sm6_block_add_fake_branch(struct sm6_block *block, struct sm6_block *to)
{
    sm6_block_array_append_unique(&block->fake_succ, to);
    sm6_block_array_append_unique(&to->fake_pred, block);
}

static void sm6_terminator_retarget(struct sm6_block_terminator *terminator, const struct sm6_block *from,
        struct sm6_block *to)
{
    unsigned int i;

    if (terminator->direct_block == from)
        terminator->direct_block = to;
    if (terminator->true_block == from)
        terminator->true_block = to;
    if (terminator->false_block == from)
        terminator->false_block = to;
    for (i = 0; i < terminator->case_count; ++i)
        if (terminator->cases[i].node == from)
            terminator->cases[i].node = to;
}

static void sm6_block_retarget_branch(struct sm6_block *block, struct sm6_block *to_prev,
        struct sm6_block *to_next)
{
    unsigned int i;

    assert(sm6_block_array_contains(&block->succ, to_prev));
    assert(sm6_block_array_contains(&to_prev->pred, block));
    assert(!sm6_block_array_contains(&block->succ, to_next));
    assert(!sm6_block_array_contains(&to_next->pred, block));

    sm6_block_array_delete(&to_prev->pred, block);

    sm6_block_array_retarget(&block->succ, to_prev, to_next);

    i = sm6_block_array_find(&to_next->pred, to_prev);
    /* If to_prev now becomes unreachable, just replace pred in-place to avoid a stale pred.
     * The stale pred will be cleaned up later when recomputing CFG. */
    if (!to_prev->pred.count && !to_prev->pred_back_edge && i < to_next->pred.count)
        to_next->pred.blocks[i] = block;
    else
        sm6_block_add_unique_pred(to_next, block);

    /* Branch targets have changed, so recompute immediate dominators. */
    if (to_prev->forward_post_visit_order > to_next->forward_post_visit_order)
    {
        sm6_block_recompute_immediate_dominator(to_prev);
        sm6_block_recompute_immediate_dominator(to_next);
    }
    else
    {
        sm6_block_recompute_immediate_dominator(to_next);
        sm6_block_recompute_immediate_dominator(to_prev);
    }

    /* ... and post dominator for ourself.
     * I am not sure if it's technically possible that we have to recompute the entire post domination graph now? */
    sm6_block_recompute_immediate_post_dominator(block);

    sm6_terminator_retarget(&block->terminator, to_prev, to_next);
}

static void sm6_block_retarget_branch_with_intermediate_node(struct sm6_block *block,
        struct sm6_block *to_prev, struct sm6_block *intermediate, struct sm6_block *to_next)
{
    sm6_block_set_name3(intermediate, block->name, ".intermediate.", to_next->name);
    intermediate->terminator.type = TERMINATOR_UNCOND_BR;
    intermediate->terminator.direct_block = to_next;
    sm6_block_add_branch(intermediate, to_next);
    intermediate->immediate_post_dominator = to_next;
    intermediate->immediate_dominator = block;
    intermediate->forward_post_visit_order = block->forward_post_visit_order;
    intermediate->backward_post_visit_order = block->backward_post_visit_order;
    sm6_block_retarget_branch(block, to_prev, intermediate);
    sm6_block_recompute_immediate_dominator(to_next);
}

static void sm6_block_retarget_fake_succ(struct sm6_block *block, struct sm6_block *to_prev,
        struct sm6_block *to_next)
{
    assert(sm6_block_array_contains(&block->fake_succ, to_prev));
    assert(sm6_block_array_contains(&to_prev->fake_pred, block));
    assert(!sm6_block_array_contains(&block->fake_succ, to_next));
    assert(!sm6_block_array_contains(&to_next->fake_pred, block));

    sm6_block_array_retarget(&block->fake_succ, to_prev, to_next);
    sm6_block_add_unique_fake_pred(to_next, block);

    sm6_block_recompute_immediate_post_dominator(block);
}

struct vkd3d_cfg_structuriser
{
    void *parser;

    struct sm6_function *function;

    struct sm6_block *entry_block;
    struct sm6_block *exit_block;

    /* For dominance analysis. */
    struct sm6_block_array forward_post_visit_order;
    /* For post-dominance analysis. */
    struct sm6_block_array backward_post_visit_order;

    uint32_t *reachability_bitset;
    size_t reachability_capacity;
    unsigned int reachability_stride;

    unsigned int true_value;
    unsigned int loop_merge_target_true_value;

    unsigned int reachable_true_value;
};

static inline unsigned int cfg_structuriser_next_true_value(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;
    /* Virtually impossible outside an infinite loop. */
    if (!(i = ++structuriser->true_value))
        FIXME("True value has wrapped.\n");
    return i;
}

static void cfg_structuriser_build_immediate_dominators(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;

    for (i = structuriser->forward_post_visit_order.count; i; --i)
    {
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[i - 1];
        sm6_block_recompute_immediate_dominator(node);
    }
}

static void cfg_structuriser_visit_reachability(struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node)
{
    uint32_t *dst_reachability = &structuriser->reachability_bitset[node->forward_post_visit_order * structuriser->reachability_stride];
    const uint32_t *src_reachability;
    struct sm6_block *succ;
    unsigned int i, j;

    for (i = 0; i < node->succ.count; ++i)
    {
        succ = node->succ.blocks[i];
        /* Inherit reachability from all successors. */
        src_reachability = &structuriser->reachability_bitset[succ->forward_post_visit_order
                * structuriser->reachability_stride];
        for (j = 0; j < structuriser->reachability_stride; ++j)
            dst_reachability[j] |= src_reachability[j];
    }

    /* We can reach ourselves. */
    dst_reachability[node->forward_post_visit_order / 32u] |= 1u << (node->forward_post_visit_order & 31u);
}

static void cfg_structuriser_build_reachability(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;

    structuriser->reachability_stride = align(structuriser->forward_post_visit_order.count, 32);
    i = structuriser->reachability_stride * structuriser->forward_post_visit_order.count;
    if (!vkd3d_array_reserve((void **)&structuriser->reachability_bitset, &structuriser->reachability_capacity,
            i, sizeof(*structuriser->reachability_bitset)))
    {
        ERR("Failed to allocate reachability bitset.\n");
        abort();
    }
    memset(structuriser->reachability_bitset, 0, i * sizeof(*structuriser->reachability_bitset));
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
        cfg_structuriser_visit_reachability(structuriser, structuriser->forward_post_visit_order.blocks[i]);
}

static bool cfg_structuriser_query_reachability(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *from, const struct sm6_block *to)
{
    const uint32_t *src_reachability;

    if (from == to)
        return true;

    src_reachability = &structuriser->reachability_bitset[
            from->forward_post_visit_order * structuriser->reachability_stride];
    return !!(src_reachability[to->forward_post_visit_order / 32u] & (1u << (to->forward_post_visit_order & 31u)));
}

static void sm6_block_array_delete_unreachable(struct sm6_block_array *array,
        const struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *must_reach_frontier)
{
    unsigned int i, j;
    for (i = 0, j = 0; i < array->count; ++i)
        if (cfg_structuriser_query_reachability(structuriser, must_reach_frontier, array->blocks[i]))
            array->blocks[j++] = array->blocks[i];
    array->count = j;
}

static void cfg_structuriser_reset_traversal(struct vkd3d_cfg_structuriser *structuriser)
{
    structuriser->reachable_true_value = cfg_structuriser_next_true_value(structuriser);
    structuriser->forward_post_visit_order.count = 0;
    structuriser->backward_post_visit_order.count = 0;

    sm6_function_reset_traversal(structuriser->function);
}

static void cfg_structuriser_visit(struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *entry)
{
    unsigned int i;

    entry->visited = true;
    entry->traversing = true;
    entry->reachable = structuriser->reachable_true_value;

    for (i = 0; i < entry->succ.count; ++i)
    {
        struct sm6_block *succ = entry->succ.blocks[i];
        if (succ->traversing)
        {
            /* For now, only support one back edge.
             * DXIL seems to obey this. */
            assert(!entry->succ_back_edge || entry->succ_back_edge == succ);
            entry->succ_back_edge = succ;

            /* For now, only support one back edge.
             * DXIL seems to obey this. */
            assert(!succ->pred_back_edge || succ->pred_back_edge == entry);
            succ->pred_back_edge = entry;
        }
        else if (!succ->visited)
            cfg_structuriser_visit(structuriser, succ);
    }

    /* Any back edges need to be handled specifically, only keep forward edges in succ/pred lists.
     * This avoids any infinite loop scenarios and needing to special case a lot of checks. */
    if (entry->succ_back_edge)
        sm6_block_array_delete(&entry->succ, entry->succ_back_edge);
    if (entry->pred_back_edge)
        sm6_block_array_delete(&entry->pred, entry->pred_back_edge);

    entry->traversing = false;
    entry->forward_post_visit_order = structuriser->forward_post_visit_order.count;
    sm6_block_array_append(&structuriser->forward_post_visit_order, entry);
}

static void cfg_structuriser_prune_dead_preds(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block *node;
    unsigned int i, j, k;

    /* We do not want to see unreachable preds.
     * Having a pred means we need to map it to an incoming value when dealing with PHI. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        for (j = 0, k = 0; j < node->pred.count; ++j)
        {
            if (node->pred.blocks[j]->reachable == structuriser->reachable_true_value)
                node->pred.blocks[k++] = node->pred.blocks[j];
        }
        node->pred.count = k;
    }
}

struct loop_merge_tracer
{
    struct sm6_block_array loop_exits;
    unsigned int true_value;
};

static void loop_merge_tracer_init(struct loop_merge_tracer *tracer, unsigned int true_value)
{
    sm6_block_array_init(&tracer->loop_exits);
    tracer->true_value = true_value;
}

static void loop_merge_tracer_destroy(struct loop_merge_tracer *tracer)
{
    sm6_block_array_destroy(&tracer->loop_exits);
}

static void loop_merge_tracer_trace_to_parent(struct loop_merge_tracer *tracer, struct sm6_block *header,
        struct sm6_block *node)
{
    unsigned int i, tracer_true_value = tracer->true_value;

    if (node == header)
    {
        node->back_traced = tracer_true_value;
        return;
    }

    if (node->back_traced != tracer_true_value)
    {
        node->back_traced = tracer_true_value;
        for (i = 0; i < node->pred.count; ++i)
            loop_merge_tracer_trace_to_parent(tracer, header, node->pred.blocks[i]);

        /* A backtrace will not pick up continue blocks which only branch back to header,
         * and thus they will be considered loop exits by mistake.
         * Start traversing from the continue block to catch these nodes as well.
         * If a loop header is part of an outer loop construct, the loop body must
         * also be part of the loop construct. */
        if (node->pred_back_edge)
            loop_merge_tracer_trace_to_parent(tracer, header, node->pred_back_edge);
    }
}

static void loop_merge_tracer_trace_from_parent(struct loop_merge_tracer *tracer, struct sm6_block *header)
{
    struct sm6_block *succ;
    unsigned int i;

    if (header->back_traced != tracer->true_value)
    {
        sm6_block_array_append(&tracer->loop_exits, header);
        return;
    }

    for (i = 0; i < header->succ.count; ++i)
    {
        succ = header->succ.blocks[i];
        if (succ->forward_traced != tracer->true_value)
        {
            loop_merge_tracer_trace_from_parent(tracer, succ);
            succ->forward_traced = tracer->true_value;
        }
    }
}

static void cfg_structuriser_backwards_visit(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *entry)
{
    unsigned int i;

    entry->backward_visited = true;

    for (i = 0; i < entry->pred.count; ++i)
        if (!entry->pred.blocks[i]->backward_visited)
            cfg_structuriser_backwards_visit(structuriser, entry->pred.blocks[i]);

    for (i = 0; i < entry->fake_pred.count; ++i)
        if (!entry->fake_pred.blocks[i]->backward_visited)
            cfg_structuriser_backwards_visit(structuriser, entry->fake_pred.blocks[i]);

    entry->backward_post_visit_order = structuriser->backward_post_visit_order.count;
    sm6_block_array_append(&structuriser->backward_post_visit_order, entry);
}

static int forward_post_visit_order_compare(const void *a, const void *b)
{
    const struct sm6_block * const *block_a = a, * const *block_b = b;
    return vkd3d_u32_compare((*block_a)->forward_post_visit_order, (*block_b)->forward_post_visit_order);
}

static int forward_post_visit_order_compare_desc(const void *a, const void *b)
{
    const struct sm6_block * const *block_a = a, * const *block_b = b;
    return vkd3d_u32_compare((*block_b)->forward_post_visit_order, (*block_a)->forward_post_visit_order);
}

static int backward_post_visit_order_compare(const void *a, const void *b)
{
    const struct sm6_block * const *block_a = a, * const *block_b = b;
    return vkd3d_u32_compare((*block_a)->backward_post_visit_order, (*block_b)->backward_post_visit_order);
}

static bool sm6_block_dominates_all_reachable_exits(struct sm6_block *node, const struct sm6_block *header,
        unsigned int true_value)
{
    unsigned int i;

    if (node->dominates_exits_completed != true_value)
    {
        if (node->succ_back_edge)
            return false;

        for (i = 0; i < node->succ.count; ++i)
            if (!sm6_block_dominates(header, node->succ.blocks[i])
                    || !sm6_block_dominates_all_reachable_exits(node->succ.blocks[i], header, true_value))
                return false;

        node->dominates_exits_completed = true_value;
    }

    return true;
}

static inline bool cfg_structuriser_node_dominates_all_reachable_exits(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node)
{
    return sm6_block_dominates_all_reachable_exits(node, node, cfg_structuriser_next_true_value(structuriser));
}

static const struct sm6_block *cfg_structuriser_get_innermost_loop_header_for(
        const struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *header,
        const struct sm6_block *other)
{
    const struct sm6_block *node = other;

    while (header != other)
    {
        /* Entry block case. */
        if (!other->pred.count)
            break;

        /* Found a loop header. This better be the one.
         * Detect false positive if back-edge can reach the node, this means we just skip over
         * the loop. We want to detect loops in a structured sense.
         * Breaking constructs should still detect the loop header as we'd expect. */
        if (other->pred_back_edge && !cfg_structuriser_query_reachability(structuriser, other->pred_back_edge, node))
            break;

        assert(other->immediate_dominator);
        other = other->immediate_dominator;
    }

    return other;
}

static void cfg_structuriser_backwards_visit_all(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block_array leaf_nodes = {0};
    struct sm6_block *node, *exit_block;
    bool need_revisit = false;
    unsigned int i, j;

    /* Traverse from leaf nodes, back through their preds instead.
     * Clear out some state set by forward visit earlier. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];

        node->backward_visited = false;
        node->traversing = false;

        /* For loops which can only exit from their header block,
         * certain loops will be unreachable when doing a backwards traversal.
         * We'll visit them explicitly later. */
        if (!node->succ.count && !node->succ_back_edge)
            sm6_block_array_append(&leaf_nodes, node);
    }

    for (i = 0; i < leaf_nodes.count; ++i)
        cfg_structuriser_backwards_visit(structuriser, leaf_nodes.blocks[i]);

    /* It might be case that some continue blocks are not reachable through backwards traversal.
     * This effectively means that our flipped CFG is not reducible, which is rather annoying.
     * To work around this, we fake some branches from the continue block out to other blocks.
     * This way, we ensure that every forward-reachable block is reachable in a backwards traversal as well.
     * The algorithm works where given the innermost loop header A, a block B (A dom B) and continue block C,
     * For successors of B, we will observe some successors which can reach C ({E}), and some successors which can not reach C.
     * C will add fake successor edges to {E}. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->pred_back_edge)
        {
            if (!node->pred_back_edge->backward_visited)
            {
                struct sm6_block_array exits = {0};
                struct loop_merge_tracer merge_tracer;
                bool transpose_loop_exit = false;

                loop_merge_tracer_init(&merge_tracer, cfg_structuriser_next_true_value(structuriser));

                loop_merge_tracer_trace_to_parent(&merge_tracer, node, node->pred_back_edge);
                loop_merge_tracer_trace_from_parent(&merge_tracer, node);

                /* If we have an infinite loop, the continue block will not be reachable with backwards traversal.
                 * Also, the only way to exit the loop construct could be through a single return block.
                 * In this case, the return block should be moved and considered to be the merge block.
                 * We add true branches from the continue block to return block instead of fake branches. */

                /* Ensure stable codegen order. */
                sm6_block_array_clone(&exits, &merge_tracer.loop_exits);
                qsort(exits.blocks, exits.count, sizeof(*exits.blocks), forward_post_visit_order_compare_desc);

                if (exits.count == 1)
                {
                    struct sm6_block *exit_node = exits.blocks[0];
                    /* If this is true, we never really leave the loop, which is problematic. */
                    transpose_loop_exit = cfg_structuriser_node_dominates_all_reachable_exits(structuriser, exit_node);

                    /* Only transpose if we're the innermost header, otherwise, inner loops which try to branch
                     * to the return will be considered a multi-break which is very awkward. */
                    if (transpose_loop_exit)
                    {
                        const struct sm6_block *innermost_header = cfg_structuriser_get_innermost_loop_header_for(
                                structuriser, node, exit_node);
                        transpose_loop_exit = innermost_header == node;
                    }
                }

                if (transpose_loop_exit)
                {
                    for (j = 0; j < exits.count; ++j)
                        sm6_block_add_branch(node->pred_back_edge, exits.blocks[j]);
                }
                else
                {
                    for (j = 0; j < exits.count; ++j)
                        if (exits.blocks[j]->backward_visited)
                            sm6_block_add_fake_branch(node->pred_back_edge, exits.blocks[j]);
                }

                loop_merge_tracer_destroy(&merge_tracer);
                sm6_block_array_destroy(&exits);

                need_revisit = true;
            }
        }
    }

    if (need_revisit)
    {
        for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
        {
            node = structuriser->forward_post_visit_order.blocks[i];
            node->backward_visited = false;
            node->traversing = false;
            node->backward_post_visit_order = 0;
        }

        for (i = 0; i < leaf_nodes.count; ++i)
            cfg_structuriser_backwards_visit(structuriser, leaf_nodes.blocks[i]);
    }

    exit_block = structuriser->exit_block;
    exit_block->backward_post_visit_order = structuriser->backward_post_visit_order.count;
    exit_block->immediate_post_dominator = exit_block;
    exit_block->backward_visited = true;
    for (i = 0; i < leaf_nodes.count; ++i)
        leaf_nodes.blocks[i]->immediate_post_dominator = exit_block;

    sm6_block_array_destroy(&leaf_nodes);
}

static void cfg_structuriser_build_immediate_post_dominators(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block *node;
    unsigned int i;

    for (i = structuriser->backward_post_visit_order.count; i; --i)
    {
        node = structuriser->backward_post_visit_order.blocks[i - 1];
        sm6_block_recompute_immediate_post_dominator(node);
    }
}

static void sm6_block_recompute_dominance_frontier(struct sm6_block *node)
{
    struct sm6_block *idom;
    unsigned int i, j;

    for (i = 0; i < node->succ.count; ++i)
    {
        struct sm6_block *succ = node->succ.blocks[i];
        if (succ->immediate_dominator != node)
            sm6_block_array_append_unique(&node->dominance_frontier, succ);

        if ((idom = node->immediate_dominator))
        {
            for (j = 0; j < node->dominance_frontier.count; ++j)
            {
                struct sm6_block *frontier_node = node->dominance_frontier.blocks[j];
                if (!sm6_block_dominates(idom, frontier_node))
                    sm6_block_array_append_unique(&idom->dominance_frontier, frontier_node);
            }
        }
    }
}

static void cfg_structuriser_compute_dominance_frontier(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
        structuriser->forward_post_visit_order.blocks[i]->dominance_frontier.count = 0;
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
        sm6_block_recompute_dominance_frontier(structuriser->forward_post_visit_order.blocks[i]);
}

static void sm6_block_recompute_post_dominance_frontier(struct sm6_block *node)
{
    struct sm6_block *pd;
    unsigned int i, j;

    for (i = 0; i < node->pred.count; ++i)
    {
        struct sm6_block *pred = node->pred.blocks[i];
        if (pred->immediate_post_dominator != node)
            sm6_block_array_append_unique(&node->post_dominance_frontier, pred);

        if ((pd = node->immediate_post_dominator))
        {
            for (j = 0; j < node->post_dominance_frontier.count; ++j)
            {
                struct sm6_block *frontier_node = node->post_dominance_frontier.blocks[j];
                if (!sm6_block_post_dominates(pd, frontier_node))
                    sm6_block_array_append_unique(&pd->post_dominance_frontier, frontier_node);
            }
        }
    }
}

static void cfg_structuriser_compute_post_dominance_frontier(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;

    for (i = 0; i < structuriser->backward_post_visit_order.count; ++i)
        structuriser->backward_post_visit_order.blocks[i]->post_dominance_frontier.count = 0;
    for (i = 0; i < structuriser->backward_post_visit_order.count; ++i)
        sm6_block_recompute_post_dominance_frontier(structuriser->backward_post_visit_order.blocks[i]);
}

static void cfg_structuriser_recompute_cfg(struct vkd3d_cfg_structuriser *structuriser)
{
    cfg_structuriser_reset_traversal(structuriser);
    cfg_structuriser_visit(structuriser, structuriser->entry_block);
    /* Need to prune dead preds before computing dominance. */
    cfg_structuriser_prune_dead_preds(structuriser);
    cfg_structuriser_build_immediate_dominators(structuriser);
    cfg_structuriser_build_reachability(structuriser);

    cfg_structuriser_backwards_visit_all(structuriser);
    cfg_structuriser_build_immediate_post_dominators(structuriser);

    cfg_structuriser_compute_dominance_frontier(structuriser);
    cfg_structuriser_compute_post_dominance_frontier(structuriser);
}

static void cfg_structuriser_init(struct vkd3d_cfg_structuriser *structuriser, void *parser,
        struct sm6_function *function)
{
    memset(structuriser, 0, sizeof(*structuriser));
    structuriser->parser = parser;
    structuriser->function = function;
    structuriser->entry_block = function->entry_block;
    structuriser->exit_block = sm6_function_blocks_new(function);
    sm6_block_set_name1(structuriser->exit_block, "EXIT");
    structuriser->true_value = 1;
    structuriser->loop_merge_target_true_value = 1;
    structuriser->reachable_true_value = 1;
}

static unsigned int sm6_block_graphviz_id(struct sm6_block *block, unsigned int *accum_id, FILE *file)
{
    const char *shape = NULL;

    if (block->graphviz_id)
        return block->graphviz_id;

    if (block->merge_type == MERGE_TYPE_LOOP)
        shape = "circle";
    else if (block->merge_type == MERGE_TYPE_SELECTION)
        shape = "triangle";
    else
        shape = "box";

    block->graphviz_id = ++(*accum_id);
    fprintf(file, "%u [label=\"%s\", shape=\"%s\"];\n", *accum_id, block->name, shape);
    return *accum_id;
}

static void cfg_structuriser_log_cfg_graphviz(struct vkd3d_cfg_structuriser *structuriser,
        const char *log_path, const char *suffix)
{
    unsigned int i, j, accum_id = 0;
    struct sm6_block *node;
    char path[PATH_MAX];
    FILE *file;

    if (!log_path)
        return;

    sprintf(path, "%s%s", log_path, suffix);
    file = fopen(path, "w");
    if (!file)
    {
        WARN("Failed to open graphviz dump path: %s\n", path);
        return;
    }

    sm6_function_reset_graphviz_ids(structuriser->function);

    fprintf(file, "digraph {\n");
    for (i = structuriser->forward_post_visit_order.count; i; i--)
    {
        node = structuriser->forward_post_visit_order.blocks[i - 1];
        switch (node->terminator.type)
        {
        case TERMINATOR_UNCOND_BR:
            fprintf(file, "%u -> %u;\n", sm6_block_graphviz_id(node, &accum_id, file), sm6_block_graphviz_id(node->terminator.direct_block, &accum_id, file));
            break;

        case TERMINATOR_COND_BR:
            fprintf(file, "%u -> %u;\n", sm6_block_graphviz_id(node, &accum_id, file), sm6_block_graphviz_id(node->terminator.true_block, &accum_id, file));
            fprintf(file, "%u -> %u;\n", sm6_block_graphviz_id(node, &accum_id, file), sm6_block_graphviz_id(node->terminator.false_block, &accum_id, file));
            break;

        case TERMINATOR_SWITCH:
            for (j = 0; j < node->terminator.case_count; ++j)
                fprintf(file, "%u -> %u;\n", sm6_block_graphviz_id(node, &accum_id, file), sm6_block_graphviz_id(node->terminator.cases[j].node, &accum_id, file));
            break;

        default:
            break;
        }

        if (node->merge_type == MERGE_TYPE_LOOP)
        {
            if (node->pred_back_edge)
                fprintf(file, "%u -> %u [style=\"dotted\"];\n", sm6_block_graphviz_id(node, &accum_id, file),
                        sm6_block_graphviz_id(node->pred_back_edge, &accum_id, file));
            if (node->loop_merge_block)
                fprintf(file, "%u -> %u [style=\"dashed\"];\n", sm6_block_graphviz_id(node, &accum_id, file),
                        sm6_block_graphviz_id(node->loop_merge_block, &accum_id, file));
        }
        else if (node->merge_type == MERGE_TYPE_SELECTION)
        {
            if (node->selection_merge_block)
                fprintf(file, "%u -> %u [style=\"dashed\"];\n", sm6_block_graphviz_id(node, &accum_id, file),
                        sm6_block_graphviz_id(node->selection_merge_block, &accum_id, file));
        }
    }

    fprintf(file, "}\n");
    fclose(file);
}

static bool sm6_block_ladder_chain_has_phi_dependencies(const struct sm6_block *succ,
        const struct sm6_block *block)
{
    unsigned int i, j;

    while (succ)
    {
        for (i = 0; i < succ->phi_count; ++i)
        {
            struct sm6_phi *phi = &succ->phi[i];
            for (j = 0; j < phi->incoming_count; ++j)
                if (phi->incoming[j].block == block)
                    return true;
        }

        if (succ->succ.count == 1)
            succ = succ->succ.blocks[0];
        else
            break;
    }

    return false;
}

#ifdef PHI_DEBUG
static int incoming_value_compare(const void *a, const void *b)
{
    const struct incoming_value *incoming_a = a;
    const struct incoming_value *incoming_b = b;
    return (incoming_a->block > incoming_b->block) - (incoming_a->block < incoming_b->block);
}

static void validate_phi(struct sm6_phi *phi)
{
    struct incoming_value *incomings = phi->incoming;
    unsigned int i;

    qsort(incomings, phi->incoming_count, sizeof(*incomings), incoming_value_compare);
    for (i = 1; i < phi->incoming_count; ++i)
    {
        if (incomings[i].block == incomings[i - 1].block)
        {
            FIXME("Duplicate phi found.\n");
            abort();
        };
    }
}

static void validate_phis(struct sm6_block *block)
{
    unsigned int i;
    for (i = 0; i < block->phi_count; ++i)
        validate_phi(&block->phi[i]);
}
#else
#define validate_phis(block) ((void)0)
#endif

static unsigned int sm6_phi_find_incoming_by_block(const struct sm6_phi *phi, struct sm6_block *block)
{
    unsigned int i;
    for (i = 0; i < phi->incoming_count; ++i)
        if (phi->incoming[i].block == block)
            break;
    return i;
}

static bool sm6_phi_incoming_append_array(struct sm6_phi *dst, const struct sm6_phi *src)
{
    if (!vkd3d_array_reserve((void **)&dst->incoming, &dst->incoming_capacity,
            dst->incoming_count + src->incoming_count, sizeof(*dst->incoming)))
    {
        ERR("Failed to allocate incoming value array.\n");
        abort();
    }
    memcpy(&dst->incoming[dst->incoming_count], src->incoming, src->incoming_count * sizeof(src->incoming[0]));
    dst->incoming_count += src->incoming_count;
    return true;
}

static void sm6_phi_incoming_erase(struct sm6_phi *phi, unsigned int index)
{
    /* In case the order is important later, it's safer to do exactly what the C++ version
     * does instead of copying the last element over the erased one. */
    memmove(&phi->incoming[index], &phi->incoming[index + 1], (--phi->incoming_count - index) * sizeof(*phi->incoming));
}

static void cfg_structuriser_eliminate_node_link_preds_to_succ(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node)
{
    struct sm6_block_array pred_copy = {0}, break_nodes = {0};
    struct sm6_block *succ = node->succ.blocks[0];
    struct sm6_block *pred, *break_node;
    unsigned int i, j, k;

    assert(node->succ.count == 1);

    validate_phis(succ);

    sm6_block_array_clone(&pred_copy, &node->pred);
    for (i = 0; i < pred_copy.count; ++i)
    {
        pred = pred_copy.blocks[i];
        break_node = sm6_function_blocks_new(structuriser->function);
        sm6_block_set_name3(break_node, node->name, ".break.", pred->name);
        break_node->terminator.type = TERMINATOR_UNCOND_BR;
        break_node->terminator.direct_block = succ;
        sm6_block_add_branch(break_node, succ);
        break_node->immediate_post_dominator = succ;
        break_node->immediate_dominator = pred;
        sm6_block_retarget_branch(pred, node, break_node);

        sm6_block_array_append(&break_nodes, break_node);

        for (j = 0; j < node->phi_count; ++j)
        {
            struct sm6_phi *phi = &node->phi[j];
            for (k = 0; k < phi->incoming_count; ++k)
                if (phi->incoming[k].block == pred)
                    phi->incoming[k].block = break_node;
        }
    }
    assert(!node->pred.count);

    for (i = 0; i < succ->phi_count; ++i)
    {
        struct sm6_phi *phi = &succ->phi[i];
        struct vkd3d_shader_register incoming_from_node;

        /* Find incoming ID from the block we're splitting up. */
        j = sm6_phi_find_incoming_by_block(phi, node);
        assert(j != phi->incoming_count);
        incoming_from_node = phi->incoming[j].reg;
        sm6_phi_incoming_erase(phi, j);

        /* Try to see if the ID is a PHI that was generated by this block. */
        j = sm6_block_find_phi_by_reg(node, &incoming_from_node);

        if (j < node->phi_count)
        {
            /* If it was then we need to split up the PHI node. The break block will serve as a proxy
             * incoming block instead. */
            sm6_phi_incoming_append_array(phi, &node->phi[j]);
            sm6_block_erase_phi(node, j);
            validate_phis(succ);
        }
        else
        {
            /* We know this block does not generate this ID, so it must be either a value generated at global scope
             * (constant), or a value created by a block which dominates this node,
             * which also means it dominates all preds to this node. */
            if (!vkd3d_array_reserve((void **)&phi->incoming, &phi->incoming_capacity,
                    phi->incoming_count + break_nodes.count, sizeof(phi->incoming[0])))
            {
                ERR("Failed to allocate incoming value array.\n");
                abort();
            }
            for (j = 0; j < break_nodes.count; ++j)
            {
                phi->incoming[phi->incoming_count].block = break_nodes.blocks[j];
                phi->incoming[phi->incoming_count++].reg = incoming_from_node;
            }
            validate_phis(succ);
        }
    }
    assert(!node->phi_count);

    sm6_block_array_destroy(&break_nodes);
    sm6_block_array_destroy(&pred_copy);
}

static void cfg_structuriser_cleanup_breaking_phi_constructs(struct vkd3d_cfg_structuriser *structuriser)
{
    bool is_merge_block_candidate, did_work = false;
    unsigned int i;

    /* There might be cases where we have a common break block from different scopes which only serves to PHI together some values
     * before actually breaking, and passing that PHI node on to the actual break block.
     * This causes problems because this looks very much like a merge, but it is actually not and forces validation errors.
     * Another case is where the succ block takes PHI nodes from the breaking block only,
     * which is relevant if only constants are somehow used in the PHI construct. */

    for (i = structuriser->forward_post_visit_order.count; i; --i)
    {
        struct sm6_block *succ, *node = structuriser->forward_post_visit_order.blocks[i - 1];

        /* Only bother with blocks which don't do any useful work.
         * The only opcodes they should have are PHI nodes and a direct branch. */
        if (node->instruction_count)
            continue;
        if (node->pred.count <= 1)
            continue;
        if (node->succ.count != 1)
            continue;
        if (node->terminator.type != TERMINATOR_UNCOND_BR)
            continue;

        succ = node->succ.blocks[0];

        /* Checks if either the merge block or successor is sensitive to PHI somehow. */
        if (!sm6_block_ladder_chain_has_phi_dependencies(succ, node))
            continue;

        if (sm6_block_dominates(node, succ))
            continue;

        /* Anything related to loop/continue blocks, we don't bother with. */
        if (node->succ_back_edge || node->pred_back_edge)
            continue;

        /* This is a merge block candidate for a loop, don't split.
         * It will only confuse things where we'll need to re-merge the split blocks anyways. */
        for (i = 0, is_merge_block_candidate = false; i < node->pred.count; ++i)
        {
            struct sm6_block *pred = node->pred.blocks[i];
            if (pred->succ_back_edge)
            {
                is_merge_block_candidate = true;
                break;
            }
        }

        if (is_merge_block_candidate)
            continue;

        cfg_structuriser_eliminate_node_link_preds_to_succ(structuriser, node);
        did_work = true;
    }

    if (did_work)
        cfg_structuriser_recompute_cfg(structuriser);
}

static inline bool sm6_block_is_plain_continue(const struct sm6_block *block)
{
    return block->succ_back_edge && block != block->succ_back_edge;
}

static bool sm6_continue_block_can_merge(struct sm6_block *node)
{
    const struct sm6_block *header = node->succ_back_edge;
    const struct sm6_block *pred_candidate = NULL;
    bool found_another_escape_edge = false;
    unsigned int i;

    /* This algorithm is very arbitrary and should be seen as a nasty heuristic which solves real shaders
     * we see in the wild. It's probably safe to block continue merge in far more cases than this, but we
     * want to be maximally convergent as often as we can. */

    for (i = 0; i < node->pred.count; ++i)
    {
        struct sm6_block *pred = node->pred.blocks[i];
        /* If we have a situation where a continue block has a pred which is itself a selection merge target, that
         * node is the merge target where we follow maximum convergence.
         * The candidate must be inside loop body and not the header itself.
         * Neither continue block nor merge target have any dominance relationship. */

        if (pred->pred.count >= 2 && pred->succ.count >= 2 && header != pred
                && !sm6_block_dominates(pred, node) && !sm6_block_post_dominates(node, pred))
        {
            /* If execution does not merge up right at the natural break block, things will get very complicated.
             * In practice, we can handle merges as long as the candidate just breaks out normally.
             * If not, we have to introduce ladder breaking and this is (almost) impossible to get right. */
            struct sm6_block *common_post_dominator = sm6_block_find_common_post_dominator(node, pred);
            if (common_post_dominator && !sm6_block_array_contains(&node->succ, common_post_dominator))
            {
                pred_candidate = pred;
                break;
            }
        }
    }

    /* No obviously nasty case to handle, probably safe to let the algorithm do its thing ... */
    if (!pred_candidate)
        return true;

    /* Need to find another escape edge which is neither header nor the candidate. */
    for (i = 0; i < node->pred.count; ++i)
    {
        struct sm6_block *pred = node->pred.blocks[i];
        if (pred != header && pred != pred_candidate && !sm6_block_dominates(pred, node))
        {
            found_another_escape_edge = true;
            break;
        }
    }

    /* If we have yet another escape edge, we probably cannot merge to continue ... */
    return !found_another_escape_edge;
}

static void sm6_block_fixup_merge_info_after_branch_rewrite(struct sm6_block *node,
        struct sm6_block *from, struct sm6_block *to)
{
    /* If we end up re-seating merge sites, make sure we add it to headers in the target block, since we might have
     * to keep splitting merge scopes in innermost scopes. */
    if (sm6_block_array_contains(&from->headers, node))
    {
        sm6_block_array_append_unique(&to->headers, node);
        if (node->selection_merge_block == from)
            node->selection_merge_block = to;
        if (node->loop_merge_block == from)
            node->loop_merge_block = to;
        if (node->loop_ladder_block == from)
            node->loop_ladder_block = to;
    }
}

static void cfg_structuriser_retarget_pred_from(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *new_node, struct sm6_block *old_succ)
{
    struct sm6_block *node;
    unsigned int i;

    for (i = 0; i < new_node->pred.count; ++i)
    {
        node = new_node->pred.blocks[i];
        sm6_block_array_retarget(&node->succ, old_succ, new_node);
        sm6_terminator_retarget(&node->terminator, old_succ, new_node);
    }

    /* Do not swap back edges. */

    /* Retarget immediate post dominators. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->immediate_post_dominator == old_succ)
            node->immediate_post_dominator = new_node;
    }
}

static struct sm6_block *cfg_structuriser_create_helper_pred_block(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node)
{
    struct sm6_block *pred_node;
    unsigned int i;

    pred_node = sm6_function_blocks_new(structuriser->function);
    sm6_block_set_name2(pred_node, node->name, ".pred");

    /* Fixup visit order later. */
    pred_node->forward_post_visit_order = node->forward_post_visit_order;
    pred_node->backward_post_visit_order = node->backward_post_visit_order;

    vkd3d_swap(pred_node->pred, node->pred, struct sm6_block_array);
    for (i = 0; i < node->headers.count; ++i)
        sm6_block_fixup_merge_info_after_branch_rewrite(node->headers.blocks[i], node, pred_node);
    node->headers.count = 0;

    /* We're replacing entry block. */
    if (node == node->immediate_dominator)
        pred_node->immediate_dominator = pred_node;
    else
        pred_node->immediate_dominator = node->immediate_dominator;

    pred_node->immediate_post_dominator = node;
    node->immediate_dominator = pred_node;

    cfg_structuriser_retarget_pred_from(structuriser, pred_node, node);

    sm6_block_add_branch(pred_node, node);

    if (node == structuriser->entry_block)
    {
        structuriser->entry_block = pred_node;
    }

    pred_node->terminator.type = TERMINATOR_UNCOND_BR;
    pred_node->terminator.direct_block = node;

    return pred_node;
}

static void cfg_structuriser_create_continue_block_ladders(struct vkd3d_cfg_structuriser *structuriser)
{
    bool need_recompute_cfg = false;
    struct sm6_block *node;
    unsigned int i;

    /* It does not seem to be legal to merge directly to continue blocks.
     * To make it possible to merge execution, we need to create a ladder block which we can merge to.
     * There are certain scenarios where it is impossible to merge to a continue block.
     * In this case, we will abandom maximum convergence and use the continue block as a "break"-like target. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (sm6_block_is_plain_continue(node) && sm6_continue_block_can_merge(node))
        {
            cfg_structuriser_create_helper_pred_block(structuriser, node);
            need_recompute_cfg = true;
        }
    }

    if (need_recompute_cfg)
        cfg_structuriser_recompute_cfg(structuriser);
}

static const struct sm6_block *cfg_structuriser_get_post_dominance_frontier_with_cfg_subset_that_reaches(
        const struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node,
        const struct sm6_block *must_reach, const struct sm6_block *must_reach_frontier)
{
    struct sm6_block_array promoted_post_dominators = {0};
    const struct sm6_block *succ, *pdom, *result;
    struct sm6_block_array frontiers = {0};
    struct sm6_block *frontier;
    bool all_succs_must_go_via_node;
    unsigned int i, j;

    assert(cfg_structuriser_query_reachability(structuriser, node, must_reach));

    if (!node->post_dominance_frontier.count)
        return NULL;

    sm6_block_array_append(&promoted_post_dominators, node);

    sm6_block_array_clone(&frontiers, &node->post_dominance_frontier);
    while (frontiers.count)
    {
        /* We might not be interested in post-domination-frontiers that we cannot reach.
         * Filter our search based on this. */
        if (must_reach_frontier)
            sm6_block_array_delete_unreachable(&frontiers, structuriser, must_reach_frontier);

        if (frontiers.count > 1)
        {
            qsort(frontiers.blocks, frontiers.count, sizeof(*frontiers.blocks), backward_post_visit_order_compare);
            sm6_block_array_deduplicate(&frontiers);
        }
        else if (!frontiers.count)
            break;

        frontier = frontiers.blocks[frontiers.count - 1];

        /* For a frontier to be discounted, we look at all successors and check if there no node in
         * promoted_post_dominators that post-dominate the successor, that path cannot reach must_reach.
         * If a post dominance frontier satisfies this rule, it is promoted to be considered an alias of node. */

        all_succs_must_go_via_node = true;
        for (i = 0; i < frontier->succ.count; ++i)
        {
            bool promote = true;
            succ = frontier->succ.blocks[i];
            if (cfg_structuriser_query_reachability(structuriser, succ, must_reach))
            {
                promote = false;
                for (j = 0; j < promoted_post_dominators.count; ++j)
                {
                    pdom = promoted_post_dominators.blocks[j];
                    if (sm6_block_post_dominates(pdom, succ))
                    {
                        promote = true;
                        break;
                    }
                }
            }

            if (!promote)
            {
                all_succs_must_go_via_node = false;
                break;
            }
        }

        if (!all_succs_must_go_via_node)
        {
            result = frontier;
            goto done;
        }
        else
        {
            sm6_block_array_append(&promoted_post_dominators, frontier);
            --frontiers.count;
            sm6_block_array_append_array(&frontiers, &frontier->post_dominance_frontier);
        }
    }

    result = !frontiers.count ? NULL : frontiers.blocks[0];

done:
    sm6_block_array_destroy(&frontiers);
    sm6_block_array_destroy(&promoted_post_dominators);
    return result;
}

static bool cfg_structuriser_exists_path_in_cfg_without_intermediate_node(
        const struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *start_block,
        const struct sm6_block *end_block, struct sm6_block *stop_block)
{
    if (cfg_structuriser_query_reachability(structuriser, start_block, end_block)
            && cfg_structuriser_query_reachability(structuriser, start_block, stop_block)
            && cfg_structuriser_query_reachability(structuriser, stop_block, end_block))
    {
        const struct sm6_block *frontier = cfg_structuriser_get_post_dominance_frontier_with_cfg_subset_that_reaches(
                structuriser, stop_block, end_block, start_block);
        /* We already know start_block reaches the frontier. */
        return !!frontier;
    }
    else
    {
        return cfg_structuriser_query_reachability(structuriser, start_block, end_block);
    }
}

typedef bool (*traverse_op)(const struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node,
        unsigned int true_value, void *opaque);

static void cfg_structuriser_traverse_dominated_blocks(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *block, const struct sm6_block *header,
        unsigned int true_value, traverse_op op, void *opaque)
{
    unsigned int i;
    for (i = 0; i < block->succ.count; ++i)
    {
        struct sm6_block *node = block->succ.blocks[i];
        if (sm6_block_dominates(header, node) && op(structuriser, node, true_value, opaque))
            cfg_structuriser_traverse_dominated_blocks(structuriser, node, header, true_value, op, opaque);
    }
}

struct inner_merge_target_search_info
{
    const struct sm6_block *header;
    const struct sm6_block *merge;
    const struct sm6_block *potential_inner_merge_target;
    bool found_inner_merge_target;
};

struct natural_break_search_info
{
    const struct sm6_block *merge;
    const struct sm6_block *potential_inner_merge_target;
    const struct sm6_block *first_natural_breaks_to_outer;
    const struct sm6_block *first_natural_breaks_to_inner;
    const struct sm6_block *last_natural_breaks_to_outer;
    const struct sm6_block *last_natural_breaks_to_inner;
};

static bool sm6_block_is_earlier(const struct sm6_block *candidate, const struct sm6_block *existing)
{
    return !existing || (candidate->forward_post_visit_order > existing->forward_post_visit_order);
}

static bool sm6_block_is_later(const struct sm6_block *candidate, const struct sm6_block *existing)
{
    return !existing || (candidate->forward_post_visit_order < existing->forward_post_visit_order);
}

static bool find_inner_merge_target(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, unsigned int true_value, void *opaque)
{
    struct inner_merge_target_search_info *info = opaque;
    struct sm6_block *idom;

    if (node == info->merge)
        return false;
    if (node->traversed == true_value)
        return false;
    node->traversed = true_value;

    /* Don't analyze loops, this path is mostly for selections only. */
    if (node->pred_back_edge)
        return false;

    if (node->pred.count <= 1)
        return true;

    idom = node->immediate_dominator;
    if (idom == info->header)
    {
        info->found_inner_merge_target = true;
        return false;
    }
    else if (sm6_block_is_later(node, info->potential_inner_merge_target)
            && idom->immediate_post_dominator == info->merge
            && !cfg_structuriser_exists_path_in_cfg_without_intermediate_node(structuriser, info->header, node, idom))
    {
        /* Need to analyze this further to determine if it's one of those insane crossing merge cases ...
         * Find the lowest post visit order if there are multiple candidates. */
        info->potential_inner_merge_target = node;
    }

    return true;
}

static bool find_natural_breaks(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, unsigned int true_value, void *opaque)
{
    struct natural_break_search_info *info = opaque;
    bool breaks_to_outer, breaks_to_inner;

    if (node == info->merge || node == info->potential_inner_merge_target)
        return false;
    if (!cfg_structuriser_query_reachability(structuriser, node, info->merge)
            || !cfg_structuriser_query_reachability(structuriser, node, info->potential_inner_merge_target))
        return false;
    if (node->traversed == true_value)
        return false;
    node->traversed = true_value;

    if (node->succ.count < 2)
        return true;

    breaks_to_inner = sm6_block_array_contains_post_dominated_block(&node->succ, info->potential_inner_merge_target);
    breaks_to_outer = breaks_to_inner ? false : sm6_block_array_contains_post_dominated_block(&node->succ,
            info->merge);

    if (breaks_to_outer)
    {
        if (sm6_block_is_earlier(node, info->first_natural_breaks_to_outer))
            info->first_natural_breaks_to_outer = node;
        if (sm6_block_is_later(node, info->last_natural_breaks_to_outer))
            info->last_natural_breaks_to_outer = node;
    }

    if (breaks_to_inner)
    {
        if (sm6_block_is_earlier(node, info->first_natural_breaks_to_inner))
            info->first_natural_breaks_to_inner = node;
        if (sm6_block_is_later(node, info->last_natural_breaks_to_inner))
            info->last_natural_breaks_to_inner = node;
    }

    return true;
}

static bool sm6_block_is_ordered(const struct sm6_block *a, const struct sm6_block *b,
        const struct sm6_block *c)
{
    return a != b && b != c && sm6_block_dominates(a, b) && sm6_block_dominates(b, c);
}

static bool cfg_structuriser_header_and_merge_block_have_entry_exit_relationship(
        struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *header,
        const struct sm6_block *merge)
{
    struct inner_merge_target_search_info merge_info;
    struct natural_break_search_info break_info;
    unsigned int true_value;

    if (!sm6_block_post_dominates(merge, header))
        return false;

    merge_info.header = header;
    merge_info.merge = merge;
    /* If there are other blocks which need merging, and that idom is the header,
     * then header is some kind of exit block. */
    merge_info.potential_inner_merge_target = NULL;
    merge_info.found_inner_merge_target = false;
    true_value = cfg_structuriser_next_true_value(structuriser);
    cfg_structuriser_traverse_dominated_blocks(structuriser, header, header, true_value,
            find_inner_merge_target, &merge_info);

    if (merge_info.found_inner_merge_target)
        return true;
    if (!merge_info.potential_inner_merge_target)
        return false;

    /* Alternatively, try to find a situation where the natural merge is difficult to determine.
     * In this scenario, selection constructs appear to be "breaking" in different directions.
     * Any attempt to split scopes here will fail spectacularly. */

    break_info.merge = merge;
    break_info.potential_inner_merge_target = merge_info.potential_inner_merge_target;
    break_info.first_natural_breaks_to_outer = NULL;
    break_info.first_natural_breaks_to_inner = NULL;
    break_info.last_natural_breaks_to_outer = NULL;
    break_info.last_natural_breaks_to_inner = NULL;
    true_value = cfg_structuriser_next_true_value(structuriser);

    cfg_structuriser_traverse_dominated_blocks(structuriser, header, header, true_value,
            find_natural_breaks, &break_info);

    if (!break_info.first_natural_breaks_to_outer || !break_info.first_natural_breaks_to_inner
            || !break_info.last_natural_breaks_to_outer || !break_info.last_natural_breaks_to_inner)
    {
        return false;
    }

    /* Crossing break scenario. */
    if (sm6_block_is_ordered(break_info.first_natural_breaks_to_inner, break_info.first_natural_breaks_to_outer,
            break_info.last_natural_breaks_to_inner))
        return true;
    if (sm6_block_is_ordered(break_info.first_natural_breaks_to_outer, break_info.first_natural_breaks_to_inner,
            break_info.last_natural_breaks_to_outer))
        return true;

    return false;
}

static bool sm6_block_can_backtrace_to(struct sm6_block *node, const struct sm6_block *parent,
        unsigned int true_value)
{
    unsigned int i;

    if (node->can_backtrace_visited == true_value)
        return false;
    node->can_backtrace_visited = true_value;

    for (i = 0; i < node->pred.count; ++i)
        if (node->pred.blocks[i] == parent || sm6_block_can_backtrace_to(node->pred.blocks[i], parent, true_value))
            return true;

    return false;
}

static bool cfg_structuriser_can_backtrace_to(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, const struct sm6_block *parent)
{
    /* If parent can branch to this, then post_order(parent) must be greater than post_order(this). */
    if (parent->forward_post_visit_order < node->forward_post_visit_order)
        return false;

    return sm6_block_can_backtrace_to(node, parent, cfg_structuriser_next_true_value(structuriser));
}

static void cfg_structuriser_recurse_traverse_dominated_blocks_and_rewrite_branch(struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *dominator, struct sm6_block *candidate, struct sm6_block *from,
        struct sm6_block *to, const struct sm6_block *skip_block, unsigned int visitation_true_value)
{
    struct sm6_block *node;
    unsigned int i;

    candidate->rewrite_visited = visitation_true_value;

    for (i = 0; i < candidate->succ.count; ++i)
    {
        node = candidate->succ.blocks[i];
        if (node == skip_block)
            continue;

        if (node == from)
        {
            /* Don't introduce a cycle.
             * We only retarget branches when we have "escape-like" edges. */
            bool introduces_cycle;

            if (to->forward_post_visit_order == candidate->forward_post_visit_order && to != candidate)
            {
                /* Can happen when resolving ladders. We cannot use reachability query, do it slow way. */
                introduces_cycle = cfg_structuriser_can_backtrace_to(structuriser, candidate, to);
            }
            else
            {
                introduces_cycle = cfg_structuriser_query_reachability(structuriser, to, candidate);
            }

            if (!introduces_cycle)
            {
                /* If we already have a branch to "to", need to branch there via an intermediate node.
                 * This way, we can distinguish between a normal branch and a rewritten branch. */
                if (sm6_block_array_contains(&candidate->succ, to))
                    sm6_block_retarget_branch_with_intermediate_node(candidate, from,
                            sm6_function_blocks_new(structuriser->function), to);
                else
                    sm6_block_retarget_branch(candidate, from, to);
            }
        }
        else if (sm6_block_dominates(dominator, node) && node != to) /* Do not traverse beyond the new branch target. */
        {
            if (node->rewrite_visited != visitation_true_value)
                cfg_structuriser_recurse_traverse_dominated_blocks_and_rewrite_branch(structuriser,
                        dominator, node, from, to, skip_block, visitation_true_value);
        }
    }

    /* In case we are rewriting branches to a new merge block, we might change the immediate post dominator
     * for continue blocks inside this loop construct. When analysing post dominance in these cases, we
     * need to make sure that we merge to the new merge block, and not the old one. This avoids some
     * redundant awkward loop constructs. */
    for (i = 0; i < candidate->fake_succ.count; ++i)
    {
        if (candidate->fake_succ.blocks[i] == from)
        {
            sm6_block_retarget_fake_succ(candidate, from, to);
            break;
        }
    }
}

static void cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *dominator, struct sm6_block *from, struct sm6_block *to,
        const struct sm6_block *skip_block)
{
    if (from == to)
        return;

    cfg_structuriser_recurse_traverse_dominated_blocks_and_rewrite_branch(structuriser, dominator, dominator,
            from, to, skip_block, cfg_structuriser_next_true_value(structuriser));
    sm6_block_fixup_merge_info_after_branch_rewrite(dominator, from, to);
}

struct selection_break_search_info
{
    const struct sm6_block *header;
    const struct sm6_block *ladder_to;
    const struct sm6_block *inner_continue_block;
    bool ladder_to_dominates_continue;
    bool break_post_dominates_ladder_to;
    struct sm6_block_array sorted_construct;
};

static bool find_selection_breaks(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, unsigned int true_value, void *opaque)
{
    struct selection_break_search_info *info = opaque;
    bool branch_is_loop_or_switch;

    /* Inner loop headers are not candidates for a rewrite. They are split in split_merge_blocks.
     * Similar with switch blocks.
     * Also, we need to stop traversing when we hit the target block ladder_to. */
    if (node == info->ladder_to || node->traversed == true_value)
        return false;

    node->traversed = true_value;
    if (!cfg_structuriser_query_reachability(structuriser, node, info->ladder_to))
        return false;

    branch_is_loop_or_switch = node->pred_back_edge || node->terminator.type == TERMINATOR_SWITCH;

    /* If our candidate scope splits a loop scope in half, ignore this candidate. */
    if (info->break_post_dominates_ladder_to && !info->ladder_to_dominates_continue
            && sm6_block_dominates(node, info->inner_continue_block))
    {
        return false;
    }

    if (node->succ.count >= 2 && !branch_is_loop_or_switch)
    {
        const struct sm6_block *outer_header = cfg_structuriser_get_post_dominance_frontier_with_cfg_subset_that_reaches(
                structuriser, node, info->ladder_to, NULL);
        if (outer_header == info->header && node->rewrite_visited != true_value)
        {
            node->rewrite_visited = true_value;
            sm6_block_array_append(&info->sorted_construct, node);
        }
    }

    return true;
}

static void cfg_structuriser_rewrite_selection_breaks(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *header, struct sm6_block *ladder_to)
{
    const struct sm6_block *innermost_loop_header, *inner_continue_succ;
    struct selection_break_search_info info = {0};
    unsigned int i, true_value;

    /* Don't rewrite loops here (since this is likely a loop merge block),
     * unless we're rewriting header -> inner construct scenario.
     * Check if the ladder_to block has a path to continue block.
     * If it does, it is part of the loop construct, and cannot be a loop merge block. */
    if (header->pred_back_edge && !cfg_structuriser_can_backtrace_to(structuriser, header->pred_back_edge, ladder_to))
        return;

    /* Don't rewrite switch blocks either. */
    if (header->terminator.type == TERMINATOR_SWITCH)
        return;

    true_value = cfg_structuriser_next_true_value(structuriser);

    info.header = header;
    info.ladder_to = ladder_to;
    /* Be careful about rewriting branches in continuing constructs. */
    inner_continue_succ = NULL;
    innermost_loop_header = cfg_structuriser_get_innermost_loop_header_for(structuriser,
            structuriser->entry_block, header);
    if (innermost_loop_header && innermost_loop_header->pred_back_edge)
        info.inner_continue_block = innermost_loop_header->pred_back_edge;
    if (info.inner_continue_block && info.inner_continue_block->succ.count == 1)
    {
        inner_continue_succ = info.inner_continue_block->succ.blocks[0];
        info.break_post_dominates_ladder_to = sm6_block_post_dominates(inner_continue_succ, ladder_to);
        info.ladder_to_dominates_continue = sm6_block_dominates(ladder_to, info.inner_continue_block);
    }

    cfg_structuriser_traverse_dominated_blocks(structuriser, header, header, true_value,
            find_selection_breaks, &info);

    /* Emit inner constructs before outer constructs.
     * This way we get natural nesting in case of certain if/else if ladders. */
    qsort(info.sorted_construct.blocks, info.sorted_construct.count, sizeof(*info.sorted_construct.blocks),
            forward_post_visit_order_compare);

    for (i = 0; i < info.sorted_construct.count; ++i)
    {
        struct sm6_block *inner_block = info.sorted_construct.blocks[i];
        struct sm6_block *ladder = sm6_function_blocks_new(structuriser->function);
        sm6_block_set_name4(ladder, ladder_to->name, ".", inner_block->name, ".ladder");

        sm6_block_add_branch(ladder, ladder_to);
        ladder->terminator.type = TERMINATOR_UNCOND_BR;
        ladder->terminator.direct_block = ladder_to;
        ladder->immediate_post_dominator = ladder_to;
        sm6_block_array_append(&ladder->dominance_frontier, ladder_to);
        ladder->forward_post_visit_order = ladder_to->forward_post_visit_order;
        ladder->backward_post_visit_order = ladder_to->backward_post_visit_order;

        /* Stop rewriting once we hit a merge block. */
        cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, inner_block,
                ladder_to, ladder, inner_block->selection_merge_block);

        sm6_block_recompute_immediate_dominator(ladder);
        cfg_structuriser_rewrite_selection_breaks(structuriser, inner_block, ladder);
    }

    sm6_block_array_destroy(&info.sorted_construct);
}

struct sm6_block_pair
{
    struct sm6_block *first;
    struct sm6_block *second;
};

static bool cfg_structuriser_serialise_interleaved_merge_scopes(struct vkd3d_cfg_structuriser *structuriser)
{
    /* Try to fixup scenarios which arise from unrolled loops with multiple break blocks.
     * DXC will emit maximal convergence and force all dynamic instances of a given break to branch to the same
     * block, which then breaks, e.g.:
     * for (int i = 0; i < CONSTANT; i++) { cond_break_construct1(); cond_break_construct2(); cond_break_construct3(); }
     * When this unrolls we can end up with merge blocks which are entangled. Only sane way to make this work
     * is to serialize the breaks to after the merge block. */
    unsigned int i, j, k, count, pdf_range_count, cutoff_index;
    struct sm6_block_array potential_merge_nodes = {0};
    struct sm6_block *node, *idom, *dispatcher;
    struct sm6_block_terminator *terminator;
    struct sm6_block_pair *pdf_ranges;
    struct incoming_value *incoming;
    struct sm6_phi *phi;

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->pred.count >= 2 && !sm6_block_is_plain_continue(node))
            sm6_block_array_append(&potential_merge_nodes, node);
    }

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        struct sm6_block_array inner_constructs = {0};
        struct sm6_block_array valid_constructs = {0};
        bool need_deinterleave, need_default_case;

        node = structuriser->forward_post_visit_order.blocks[i];

        if (node->pred.count <= 1)
            continue;
        if (sm6_block_is_plain_continue(node))
            continue;

        idom = node->immediate_dominator;

        /* Find merge block candidates that are strictly dominated by idom and immediately post-dominated by node.
         * They also must not be good merge candidates on their own.
         * Also, we're not interested in any loop merge candidates. */
        for (j = 0; j < potential_merge_nodes.count; ++j)
        {
            struct sm6_block *candidate = potential_merge_nodes.blocks[j];

            if (candidate != idom && candidate->immediate_post_dominator == node
                && sm6_block_dominates(idom, candidate)
                && !sm6_block_post_dominates_perfect_structured_construct(candidate)
                && cfg_structuriser_get_innermost_loop_header_for(structuriser, idom, node) == idom)
            {
                bool direct_dominance_frontier = candidate->dominance_frontier.count == 1
                        && candidate->dominance_frontier.blocks[0] == node;
                /* The candidate must not try to merge to other code since we might end up introducing loops that way.
                 * All code reachable by candidate must cleanly break to node. */
                if (direct_dominance_frontier)
                    sm6_block_array_append(&inner_constructs, candidate);
            }
        }

        /* Ensure stable order. */
        qsort(inner_constructs.blocks, inner_constructs.count, sizeof(*inner_constructs.blocks),
                forward_post_visit_order_compare);

        /* Prune any candidate that can reach another candidate. The sort ensures the
         * candidate to be removed comes last. */
        count = inner_constructs.count;
        for (j = 0; j < count; j++)
        {
            bool valid = true;
            for (k = 0; k < j; k++)
            {
                if (cfg_structuriser_query_reachability(structuriser, inner_constructs.blocks[k],
                        inner_constructs.blocks[j]))
                {
                    valid = false;
                    break;
                }
            }

            if (valid)
                sm6_block_array_append(&valid_constructs, inner_constructs.blocks[j]);
        }

        sm6_block_array_destroy(&inner_constructs);

        if (valid_constructs.count < 2)
        {
            sm6_block_array_destroy(&valid_constructs);
            continue;
        }

        if (!(pdf_ranges = vkd3d_malloc(valid_constructs.count * sizeof(*pdf_ranges))))
        {
            ERR("Failed to allocate range buffer.\n");
            abort();
        }

        /* If breaking merge constructs are entangled, their PDFs will overlap. */
        for (j = 0, pdf_range_count = 0; j < valid_constructs.count; ++j)
        {
            struct sm6_block *first, *last, *candidate = valid_constructs.blocks[j];
            const struct sm6_block_array *pdf = &candidate->post_dominance_frontier;

            assert(pdf->count);
            first = pdf->blocks[0];
            last = first;

            for (k = 0; k < pdf->count; ++k)
            {
                struct sm6_block *n = pdf->blocks[k];
                if (n->forward_post_visit_order > first->forward_post_visit_order)
                    first = n;
                if (n->forward_post_visit_order < last->forward_post_visit_order)
                    last = n;
            }

            pdf_ranges[pdf_range_count].first = first;
            pdf_ranges[pdf_range_count++].second = last;
        }

        count = valid_constructs.count;
        for (j = 0, need_deinterleave = false; j < count && !need_deinterleave; j++)
            for (k = 0; k < count && !need_deinterleave; k++)
                if (j != k)
                    need_deinterleave = sm6_block_is_ordered(pdf_ranges[j].first, pdf_ranges[k].first, pdf_ranges[j].second);

        vkd3d_free(pdf_ranges);

        if (!need_deinterleave)
        {
            sm6_block_array_destroy(&valid_constructs);
            continue;
        }

        WARN("De-interleaving merge scopes.\n");

        /* Rewrite the control flow to serialize execution of the candidate blocks. */

        dispatcher = cfg_structuriser_create_helper_pred_block(structuriser, node);
        shader_register_make_scalar_ssa(&dispatcher->terminator.conditional_reg, VKD3D_DATA_UINT,
                structuriser->parser);

        for (j = 0; j < valid_constructs.count; ++j)
        {
            struct sm6_block *candidate = valid_constructs.blocks[j];
            cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser,
                    candidate, dispatcher, node, NULL);
        }

        cutoff_index = dispatcher->pred.count;

        /* If there is no direct branch intended for node, the default case label will never be reached,
         * so just pilfer one of the cases as default. */
        need_default_case = !!dispatcher->pred.count;

        phi = sm6_block_phi_new(dispatcher, &dispatcher->terminator.conditional_reg);

        for (j = 0; j < cutoff_index; j++)
        {
            incoming = sm6_phi_incoming_new(phi);
            incoming->block = dispatcher->pred.blocks[j];
            register_make_constant_uint(&incoming->reg, ~0u);
        }

        for (j = 0; j < count; j++)
        {
            struct sm6_block *candidate = valid_constructs.blocks[j];
            unsigned int next_cutoff_index;

            cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser,
                    idom, candidate, dispatcher, NULL);
            next_cutoff_index = dispatcher->pred.count;
            for (k = cutoff_index; k < next_cutoff_index; k++)
            {
                incoming = sm6_phi_incoming_new(phi);
                incoming->block = dispatcher->pred.blocks[k];
                register_make_constant_uint(&incoming->reg, j);
            }
            cutoff_index = next_cutoff_index;
        }

        idom->freeze_structured_analysis = true;
        idom->merge_type = MERGE_TYPE_LOOP;
        idom->loop_merge_block = dispatcher;

        terminator = &dispatcher->terminator;
        if (!(terminator->cases = vkd3d_malloc((count + 1) * sizeof(*terminator->cases))))
        {
            ERR("Failed to allocate case array.\n");
            abort();
        }

        terminator->type = TERMINATOR_SWITCH;
        terminator->cases[0].node = need_default_case ? node : valid_constructs.blocks[0];
        terminator->cases[0].is_default = true;
        terminator->case_count = 1;

        for (j = 0; j < count; j++)
        {
            struct sm6_block *candidate = valid_constructs.blocks[j];
            assert(!candidate->pred.count);
            sm6_block_add_branch(dispatcher, candidate);

            if (need_default_case || j)
            {
                terminator->cases[terminator->case_count].node = candidate;
                terminator->cases[terminator->case_count].value = j;
                terminator->cases[terminator->case_count++].is_default = false;
            }
        }

        sm6_block_array_destroy(&valid_constructs);
        sm6_block_array_destroy(&potential_merge_nodes);

        /* This completely transposes the CFG, so need to recompute CFG to keep going. */
        cfg_structuriser_recompute_cfg(structuriser);
        return true;
    }

    sm6_block_array_destroy(&potential_merge_nodes);
    return false;
}

static void cfg_structuriser_split_merge_scopes(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block *node, *idom;
    unsigned int i;

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        /* Setup a preliminary merge scope so we know when to stop traversal.
         * We don't care about traversing inner scopes, out starting from merge block as well. */
        if (node->pred.count <= 1)
            continue;

        if (sm6_block_is_plain_continue(node))
            continue;

        /* The immediate dominator is the natural header block. */
        idom = node->immediate_dominator;
        assert(idom->succ.count >= 2);

        if (idom->merge_type == MERGE_TYPE_NONE)
        {
            idom->merge_type = MERGE_TYPE_SELECTION;
            idom->selection_merge_block = node;
        }
        sm6_block_array_append(&node->headers, idom);
    }

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->pred.count <= 1)
            continue;

        /* Continue blocks can always be branched to, from any scope, so don't rewrite anything here. */
        if (node->succ_back_edge)
            continue;

        /* The immediate dominator is the natural header block. */
        idom = node->immediate_dominator;
        assert(idom->succ.count >= 2);

        /* We already rewrote this selection construct in serialize_interleaved_merge_scopes.
         * Don't try to introduce unnecessary ladders. */
        if (idom->merge_type == MERGE_TYPE_LOOP && idom->loop_merge_block == node)
            continue;

        /* If we find a construct which is a typical entry <-> exit scenario, do not attempt to rewrite
         * any branches. The real merge block might be contained inside this construct, and this block merely
         * serves as the exit merge point. It should generally turn into a loop merge later. */
        if (cfg_structuriser_header_and_merge_block_have_entry_exit_relationship(structuriser, idom, node))
            continue;

        /* Now we want to deal with cases where we are using this selection merge block as "goto" target
         * for inner selection constructs. Using a loop header might be possible, but we will need to
         * split up blocks to make sure that we don't end up with headers where the only branches are
         * either merges or breaks. */

        /* This case is relevant when we have something like:
         * A -> B -> C -> D -> M
         * A -> M
         * B -> M
         * C -> M
         * D -> M
         * We'll need intermediate blocks which merge each layer of the selection "onion". */
        cfg_structuriser_rewrite_selection_breaks(structuriser, idom, node);
    }
}

static bool cfg_structuriser_block_is_load_bearing(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, const struct sm6_block *merge)
{
    return node->pred.count >= 2 && !cfg_structuriser_exists_path_in_cfg_without_intermediate_node(structuriser,
            node->immediate_dominator, merge, node);
}

static bool cfg_structuriser_control_flow_is_escaping_from_loop(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *node, const struct sm6_block *merge)
{
    const struct sm6_block *innermost_loop_header;
    bool escaping_path = false;
    unsigned int i;

    if (node == merge)
        return escaping_path;

    assert(sm6_block_post_dominates(merge, node));

    /* First, test the loop scenario.
     * If we're inside a loop, we're a break construct if we can prove that:
     * - node has a loop header which dominates it.
     * - node cannot reach the continue block.
     * - Continue block cannot reach node.
     * - All post-domination frontiers can reach the continue block, meaning that at some point control flow
     *   decided to break out of the loop construct. */
    innermost_loop_header = cfg_structuriser_get_innermost_loop_header_for(structuriser, structuriser->entry_block, node);
    if (innermost_loop_header && innermost_loop_header->pred_back_edge)
    {
        bool dominates_merge = sm6_block_dominates(node, merge);
        bool can_reach_continue = cfg_structuriser_query_reachability(structuriser, node, innermost_loop_header->pred_back_edge);
        bool continue_can_reach = cfg_structuriser_query_reachability(structuriser, innermost_loop_header->pred_back_edge, node);
        bool pdf_can_reach_continue = true;

        for (i = 0; i < node->post_dominance_frontier.count; ++i)
        {
            const struct sm6_block *frontier = node->post_dominance_frontier.blocks[i];
            bool header_dominates_frontier = sm6_block_dominates(innermost_loop_header, frontier);
            bool frontier_is_inside_loop_construct = cfg_structuriser_query_reachability(structuriser,
                    frontier, innermost_loop_header->pred_back_edge);
            if (!header_dominates_frontier || !frontier_is_inside_loop_construct)
            {
                pdf_can_reach_continue = false;
                break;
            }
        }

        if (!dominates_merge && !continue_can_reach && !can_reach_continue && pdf_can_reach_continue)
            escaping_path = true;
    }

    return escaping_path;
}

static bool sm6_block_reaches_domination_frontier_before_merge(const struct sm6_block *node,
        const struct sm6_block *merge)
{
    const struct sm6_block *frontier;
    unsigned int i;

    for (i = 0; i < node->dominance_frontier.count; ++i)
    {
        frontier = node->dominance_frontier.blocks[i];
        if (merge != frontier && sm6_block_post_dominates(merge, frontier))
            return true;
    }
    return false;
}

static bool sm6_block_pred_post_dominates_any_work(const struct sm6_block *start_node,
        struct sm6_block *parent, unsigned int true_value)
{
    unsigned int i;

    /* If we reached this node before and didn't terminate, it must have returned false. */
    if (parent->pdom_any_work_visited == true_value)
        return false;
    parent->pdom_any_work_visited = true_value;

    /* This is not a dummy block, we have an answer. */
    if (parent->instruction_count || parent->phi_count)
        return sm6_block_post_dominates(start_node, parent);

    for (i = 0; i < parent->pred.count; ++i)
        if (sm6_block_pred_post_dominates_any_work(start_node, parent->pred.blocks[i], true_value))
            return true;

    return false;
}

static bool sm6_block_post_dominates_any_work(const struct sm6_block *start_node, unsigned int true_value)
{
    unsigned int i;

    /* Trivial back-trace as far as we can go. */
    while (start_node->pred.count == 1
            && !start_node->instruction_count && !start_node->phi_count
            && sm6_block_post_dominates(start_node, start_node->pred.blocks[0]))
    {
        start_node = start_node->pred.blocks[0];
    }

    if (start_node->instruction_count || start_node->phi_count)
        return true;

    for (i = 0; i < start_node->pred.count; ++i)
        if (sm6_block_pred_post_dominates_any_work(start_node, start_node->pred.blocks[i], true_value))
            break;

    return i < start_node->pred.count;
}

static bool cfg_structuriser_control_flow_is_escaping(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, const struct sm6_block *merge)
{
    const struct sm6_block *frontier, *idom, *pred;
    bool found_false_positive;
    bool escaping_path;
    unsigned int i;

    if (node == merge)
        return false;

    if (cfg_structuriser_control_flow_is_escaping_from_loop(structuriser, node, merge))
        return true;

    /* Try to test if our block is load bearing, in which case it cannot be considered a break block.
     * If the only path from idom to merge goes through node, it must be considered load bearing,
     * since removing break paths must not change reachability. */
    if (cfg_structuriser_block_is_load_bearing(structuriser, node, merge))
        return false;

    /* If we cannot prove the escape through loop analysis, we might be able to deduce it from domination frontiers.
     * If control flow is not escaping, then there must exist a dominance frontier node A,
     * where merge strictly post-dominates A.
     * This means that control flow can merge somewhere before we hit the merge block, and we consider that
     * normal structured control flow. */

    escaping_path = !sm6_block_reaches_domination_frontier_before_merge(node, merge);

    /* This is a strong check.
     * If node directly branches to merge, but PDF does not, we have detected a control flow pattern
     * which is clearly a break. The PDF candidate must dominate node for this check to be meaningful. */
    if (escaping_path)
    {
        for (i = 0; i < node->post_dominance_frontier.count; ++i)
        {
            frontier = node->post_dominance_frontier.blocks[i];
            if (sm6_block_dominates(frontier, node) && sm6_block_reaches_domination_frontier_before_merge(frontier,
                    merge))
                return true;
        }

        /* Strong check as well.
         * If branching directly to continue block like this, this is a non-merging continue,
         * which we should always consider an escape. */
        if (node->succ_back_edge)
            return true;
    }

    if (escaping_path && !node->instruction_count && !node->phi_count)
    {
        /* If we post-dominate nothing useful or do nothing useful ourselves,
         * this is a good indication we're a common escape edge ladder block.
         * This can happen if we have a graph of:
         * A -> B
         * A -> C
         * B -> merge
         * C -> merge
         * B -> block
         * C -> block
         * block -> merge
         * This super jank diamond pattern will break the heuristics. */
        if (!sm6_block_post_dominates_any_work(node, cfg_structuriser_next_true_value(structuriser)))
            return true;
    }

    if (escaping_path && node->pred.count >= 2)
    {
        /* We also need to consider false positives here, which are mostly only relevant for merge candidates. */

        /* One case would be selection construct A, which terminates in block B. B then branches to C.
         * Earlier in the A -> B construct, there might be a break block D which also branches to B.
         * This means that C will be a "false" domination frontier of B and our analysis above is wrong. */

        /* The algorithm here:
         * - Get idom of block, which represents the header. For this analysis, we're only interested
         *   in code paths which are dominated by idom.
         * - Find all preds of merge which are dominated by idom(block).
         * - Backtrace every pred P until they can reach B, or B can reach P.
         * - If B has strictly lowest post-visit order, we are not escaping. P was. */

        idom = node->immediate_dominator;
        found_false_positive = false;

        for (i = 0; i < merge->pred.count; ++i)
        {
            pred = merge->pred.blocks[i];

            /* Don't care about these. */
            if (!sm6_block_dominates(idom, pred))
                continue;

            while (pred != node && !cfg_structuriser_query_reachability(structuriser, pred, node)
                    && !cfg_structuriser_query_reachability(structuriser, node, pred))
                pred = pred->immediate_dominator;

            /* Ignore these. */
            if (pred == node)
                continue;

            if (cfg_structuriser_query_reachability(structuriser, pred, node))
            {
                /* Seems good. Keep going. If we don't find a counter example, we'll accept this as a false positive. */
                found_false_positive = true;
            }
            else
            {
                /* Indeed, this is an escape. */
                found_false_positive = false;
                break;
            }
        }

        escaping_path = !found_false_positive;
    }

    return escaping_path;
}

static bool cfg_structuriser_merge_candidate_is_on_breaking_path(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node)
{
    return node->pred.count >= 2 && node->succ.count == 1
            && !sm6_block_dominates(node, node->succ.blocks[0])
            && sm6_block_post_dominates(node->succ.blocks[0], node)
            && cfg_structuriser_control_flow_is_escaping(structuriser, node, node->succ.blocks[0])
            && !sm6_block_post_dominates_perfect_structured_construct(node);
}

static void cfg_structuriser_eliminate_degenerate_blocks(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block_array tmp_pred = {0};
    struct sm6_block *node, *pred, *succ;
    unsigned int i, j;
    bool did_work;

    /* After we create ladder blocks, we will likely end up with a lot of blocks which don't do much.
     * We might also have created merge scenarios which should *not* merge, i.e. cleanup_breaking_phi_constructs(),
     * except we caused it ourselves. */

    /* Eliminate bottom-up. First eliminate B, in A -> B -> C, where B contributes nothing. */
    for (i = 0, did_work = false; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (!node->instruction_count
                && !node->phi_count
                && !node->pred_back_edge
                && !node->succ_back_edge
                && node->succ.count == 1
                && node->terminator.type == TERMINATOR_UNCOND_BR
                && node->merge_type == MERGE_TYPE_NONE
                /* Loop merge targets are sacred, and must not be removed. */
                && node->is_loop_merge_target != structuriser->loop_merge_target_true_value
                && !sm6_block_ladder_chain_has_phi_dependencies(node->succ.blocks[0], node))
        {
            /* If any pred is a continue block, this block is also load-bearing, since it can be used as a merge block. */
            if (sm6_block_array_contains_succ_back_edge(&node->pred))
                continue;

            /* If any succ is a continue block, this block is also load-bearing, since it can be used as a merge block
             * (merge-to-continue ladder). */
            if (sm6_block_array_contains_succ_back_edge(&node->succ))
                continue;

            pred = node->pred.blocks[0];
            succ = node->succ.blocks[0];

            if (node->pred.count == 1 && sm6_block_post_dominates(node, pred))
            {
                /* Trivial case. */
                did_work = true;
                sm6_block_retarget_branch(pred, node, succ);
            }
            else if (cfg_structuriser_merge_candidate_is_on_breaking_path(structuriser, node))
            {
                /* If we have two or more preds, we have to be really careful.
                 * If this node is on a breaking path, without being important for merging control flow,
                 * it is fine to eliminate the block. */
                did_work = true;
                sm6_block_array_clone(&tmp_pred, &node->pred);
                for (j = 0; j < tmp_pred.count; ++j)
                    sm6_block_retarget_branch(tmp_pred.blocks[j], node, node->succ.blocks[0]);

                /* Iteratively, we need to recompute the dominance frontier for all preds.
                 * When we eliminate nodes like this, we might cause the pred blocks to become degenerate in
                 * future iterations in this loop. */
                qsort(tmp_pred.blocks, tmp_pred.count, sizeof(*tmp_pred.blocks), forward_post_visit_order_compare);

                /* Need to compute dominance frontiers from inside out. */
                for (j = 0; j < tmp_pred.count; ++j)
                {
                    pred = tmp_pred.blocks[j];
                    pred->dominance_frontier.count = 0;
                    sm6_block_recompute_dominance_frontier(pred);
                }

                tmp_pred.count = 0;
            }
        }
    }

    sm6_block_array_destroy(&tmp_pred);

    if (did_work)
        cfg_structuriser_recompute_cfg(structuriser);
}

static bool opcode_is_control_dependent(enum vkd3d_shader_opcode opcode)
{
    /* An opcode is considered control dependent if it is affected by other invocations in the subgroup. */
    switch (opcode)
    {
        /* Anything derivatives */
        case VKD3DSIH_DSX:
        case VKD3DSIH_DSX_COARSE:
        case VKD3DSIH_DSX_FINE:
        case VKD3DSIH_DSY:
        case VKD3DSIH_DSY_COARSE:
        case VKD3DSIH_DSY_FINE:

        /* Anything implicit LOD */
        case VKD3DSIH_SAMPLE:
        case VKD3DSIH_SAMPLE_B:
        case VKD3DSIH_SAMPLE_C:
        case VKD3DSIH_LOD:
        case VKD3DSIH_GATHER4:
        case VKD3DSIH_GATHER4_C:
        case VKD3DSIH_GATHER4_PO:
        case VKD3DSIH_GATHER4_PO_C:

        /* Control barriers */
        case VKD3DSIH_SYNC:

            return true;

        default:
            return false;
    }
}

static bool sm6_block_is_control_dependent(const struct sm6_block *node)
{
    unsigned int i;
    for (i = 0; i < node->instruction_count; ++i)
        if (opcode_is_control_dependent(node->instructions[i].handler_idx))
            return true;
    return false;
}

static void sm6_phi_build_name(struct sm6_phi *phi, const char *str1, const char *str2)
{
    int len = snprintf(NULL, 0, "%s%s", str1, str2);
    if (!(phi->name = vkd3d_realloc(phi->name, len + 1)))
        return;
    sprintf(phi->name, "%s%s", str1, str2);
}

static const struct incoming_value *sm6_phi_find_incoming_value(const struct sm6_phi *phi,
        const struct sm6_block *frontier_pred)
{
    const struct incoming_value *candidate = NULL;
    unsigned int i;

    /* Find the incoming block which dominates frontier_pred and has the lowest post visit order.
     * There are cases where two or more blocks dominate, but we want the most immediate dominator. */
    for (i = 0; i < phi->incoming_count; ++i)
    {
        struct sm6_block *block = phi->incoming[i].block;
        if (sm6_block_dominates(block, frontier_pred))
        {
            if (!candidate || (block->forward_post_visit_order < candidate->block->forward_post_visit_order))
                candidate = &phi->incoming[i];
        }
    }

    return candidate;
}
#if 0
static unsigned int sm6_phi_find_incoming_from_block(const struct sm6_phi *phi,
        const struct sm6_block *block)
{
    unsigned int i;

    for (i = 0; i < phi->incoming_count; ++i)
    {
        if (phi->incoming[i].block == block)
            break;
    }

    return i;
}

static bool cfg_structuriser_can_duplicate_phis(struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *node)
{
    unsigned int i, j;

    /* If we want to duplicate nodes, we cannot do so in complicated scenarios where
     * we need to resolve PHIs. For example, if a node is split, the split nodes might have to
     * insert PHI nodes covering the subset of nodes which can reach each split.
     * This get very hairy, very quickly.
     * To check this, ensure that the node we want to split does not require any complex PHI handling. */

    /* First, validate that we can even find incoming values properly. */
    for (i = 0; i < node->pred.count; ++i)
    {
        const struct sm6_block *pred = node->pred.blocks[i];
        for (j = 0; j < node->phi_count; ++j)
        {
            if (!sm6_phi_find_incoming_value(&node->phi[j], pred))
                return false;
        }
    }

    /* Then, make sure that every incoming value dominates at least pred of node.
     * This way, we know that we don't need complicated PHI frontier merges along the way. */
    for (i = 0; i < node->phi_count; ++i)
    {
        struct sm6_phi *phi = &node->phi[i];
        for (j = 0; j < phi->incoming_count; ++j)
        {
            struct incoming_value *incoming = &phi->incoming[j];
            if (!sm6_block_array_contains_dominated_block(&node->pred, incoming->block))
                return false;
        }
    }

    return true;
}

static void cfg_structuriser_duplicate_node(struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node)
{
    struct spirv_compiler *compiler = structuriser->compiler;
    struct sm6_block_array tmp_pred;
    struct sm6_block **break_blocks;
    struct vkd3d_spirv_builder *builder;
    struct sm6_block *succ;
    unsigned int i, j, map_size;
    uint32_t *rewritten_ids;

    assert(node->succ.count == 1);
    assert(node->pred.count >= 2);
    assert(!sm6_block_dominates(node, node->succ.blocks[0]));

    builder = dxbc_compiler_get_builder(compiler);
    map_size = vkd3d_spirv_peek_next_id(builder);
    if (!(break_blocks = vkd3d_malloc(node->pred.count * sizeof(*break_blocks)))
            || !(rewritten_ids = vkd3d_calloc(map_size * node->pred.count, sizeof(*rewritten_ids)))
            || !sm6_block_array_clone(&tmp_pred, &node->pred))
    {
        ERR("Failed to allocate temporary objects.\n");
        abort();
    }
    succ = node->succ.blocks[0];

    for (i = 0; i < tmp_pred.count; ++i)
    {
        struct sm6_block *pred = tmp_pred.blocks[i];
        uint32_t *remap = &rewritten_ids[i * map_size];

        /* First, rewrite PHI inputs.
         * Since we only have one pred now, we can resolve PHIs directly. */
        struct sm6_block *block = sm6_function_blocks_new(structuriser->function);
        sm6_block_set_name3(block, node->name, ".dup.", pred->name);
        block->terminator.type = TERMINATOR_UNCOND_BR;
        block->terminator.direct_block = succ;
        block->immediate_post_dominator = succ;
        block->immediate_dominator = pred;
        sm6_block_retarget_branch(pred, node, block);
        sm6_block_add_branch(block, succ);

        for (j = 0; j < node->phi_count; ++j)
        {
            struct sm6_phi *phi = &node->phi[j];
            const struct incoming_value *incoming = sm6_phi_find_incoming_value(phi, pred);
            assert(incoming);
            assert(phi->id < map_size);
            remap[sm6_phi_ssa_id(phi)] = sm6_phi_incoming_ssa_id(incoming);
        }

        vkd3d_spirv_sm6_stream_duplicate(&block->stream, &node->stream, remap, builder);

        break_blocks[i] = block;
    }

    assert(!node->pred.count);

    /* Finally, look at succ. If it takes PHI inputs from node, we'll have to rewrite the PHIs.
     * We know that node does not dominate succ, so succ cannot use any SSA variables node generated
     * directly without using PHI nodes. */

    /* We might have placed ladders in between so that we need to fixup PHI later than just plain succ.
     * Chase down the chain and replace all PHIs. */

    do
    {
        bool done = false;
        for (i = 0; i < succ->phi_count; ++i)
        {
            struct sm6_phi *phi = &succ->phi[i];
            uint32_t incoming_from_node;

            /* Find incoming ID from the block we're splitting up. */
            j = sm6_phi_find_incoming_from_block(phi, node);

            if (j == phi->incoming_count)
                continue;

            incoming_from_node = sm6_phi_incoming_ssa_id(&phi->incoming[j]);
            sm6_phi_incoming_erase(phi, j);

            for (j = 0; j < tmp_pred.count; ++j)
            {
                uint32_t *remap = &rewritten_ids[j * map_size];
                struct incoming_value *incoming = sm6_phi_incoming_new(phi);
                incoming->block = break_blocks[j];
                incoming->reg.idx[0].offset = remap[incoming_from_node] ? remap[incoming_from_node] : incoming_from_node;
                incoming->reg.idx_count = 1;
            }

            /* We've found the block we wanted to rewrite, terminate loop now. */
            done = true;
        }

        succ = (!done && succ->succ.count == 1) ? succ->succ.blocks[0] : NULL;
    }
    while (succ);

    sm6_block_array_destroy(&tmp_pred);
    vkd3d_free(rewritten_ids);
    vkd3d_free(break_blocks);
}
#endif
static void cfg_structuriser_duplicate_impossible_merge_constructs(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i;

    for (i = structuriser->forward_post_visit_order.count; i; --i)
    {
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[i - 1];

        /* Check for breaking merge blocks which were not considered degenerate.
         * This can happen if we actually have code in the breaking construct ... (scary!)
         * We'll have to split this block somehow.
         * If the candidate has control dependent effects like barriers and such, this will likely
         * break completely, but I don't see how that would work on native drivers either ... */

        /* WARNING: This check is EXTREMELY sensitive and microscopic changes to the implementation
         * will dramatically affect codegen. */
        bool breaking = cfg_structuriser_merge_candidate_is_on_breaking_path(structuriser, node);

        if (breaking && node->instruction_count && !sm6_block_is_control_dependent(node))
        {
            FIXME("Probably require block duplication.\n");
            return;
#if 0
            sm6_block_array_append(&duplicate_queue, node);
#endif
        }
    }
#if 0
    if (!duplicate_queue.count)
        return;

    for (i = 0; i < duplicate_queue.count; ++i)
    {
        struct sm6_block *node = duplicate_queue.blocks[i];
        if (!cfg_structuriser_can_duplicate_phis(structuriser, node))
        {
            /* A block could be subtly load bearing, in that if we split the node, it becomes impossible to resolve
             * PHIs and we hit assertions in cfg_structuriser_duplicate_node().
             * This means the block is probably load bearing after all, and we should not split it.
             * Normally, we only want to break up blocks which have fairly trivial PHI resolves. */
            WARN("Was asked to duplicate node %s, but cannot split phis without crashing ...\n", node->name);
            continue;
        }

        cfg_structuriser_duplicate_node(structuriser, node);
    }

    sm6_block_array_destroy(&duplicate_queue);
    cfg_structuriser_recompute_cfg(structuriser);
#endif
}

static struct sm6_block *cfg_structuriser_create_ladder_block(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *header, struct sm6_block *node, const char *tag)
{
    struct sm6_block *ladder = sm6_function_blocks_new(structuriser->function);
    sm6_block_set_name2(ladder, node->name, tag);
    sm6_block_add_branch(ladder, node);
    ladder->terminator.type = TERMINATOR_UNCOND_BR;
    ladder->terminator.direct_block = node;
    ladder->immediate_post_dominator = node;
    ladder->forward_post_visit_order = node->forward_post_visit_order;
    ladder->backward_post_visit_order = node->backward_post_visit_order;

    cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, header, node, ladder, NULL);
    sm6_block_recompute_immediate_dominator(ladder);

    return ladder;
}

struct sm6_loop_analysis
{
    struct sm6_block_array direct_exits;
    struct sm6_block_array inner_direct_exits;
    struct sm6_block_array dominated_exit;
    struct sm6_block_array inner_dominated_exit;
    struct sm6_block_array non_dominated_exit;
    struct sm6_block_array dominated_continue_exit;
};

static void sm6_loop_analysis_destroy(struct sm6_loop_analysis *result)
{
    sm6_block_array_destroy(&result->direct_exits);
    sm6_block_array_destroy(&result->inner_direct_exits);
    sm6_block_array_destroy(&result->dominated_exit);
    sm6_block_array_destroy(&result->inner_dominated_exit);
    sm6_block_array_destroy(&result->non_dominated_exit);
    sm6_block_array_destroy(&result->dominated_continue_exit);
}

enum loop_exit_type
{
    LOOP_EXIT_TYPE_EXIT,
    LOOP_EXIT_TYPE_MERGE,
    LOOP_EXIT_TYPE_ESCAPE,
    LOOP_EXIT_TYPE_MERGETOINFINITELOOP,
    LOOP_EXIT_TYPE_INNERLOOPEXIT,
    LOOP_EXIT_TYPE_INNERLOOPMERGE,
    LOOP_EXIT_TYPE_INNERLOOPFALSEPOSITIVE,
};

static bool sm6_block_can_loop_merge_to(const struct sm6_block *node, const struct sm6_block *other)
{
    const struct sm6_block *c;

    if (!sm6_block_dominates(node, other))
        return false;

    c = node->pred_back_edge;

    if (c && c->succ.count)
    {
        /* If the continue block branches to something which is not the loop header, it must be the
         * merge block we're after, i.e., it must be a clean break (or we are kind of screwed).
         * Detect a "fake" merge branch here.
         * E.g., we have a fake merge branch if an escaping edge is branching to one block beyond the real merge block.
         * This can happen after split_merge_scopes() transform where inner loop
         * tries to break out of multiple loops and multiple selection scopes at the same time.
         * We can still dominate this escape target, but it's still an escape which must be resolved some other way with ladders. */
        if (!sm6_block_array_contains(&c->succ, other))
            return false;
    }

    return true;
}

static enum loop_exit_type cfg_structuriser_get_loop_exit_type(struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *header, struct sm6_block *node)
{
    /* If there exists an inner loop which dominates this exit, we treat it as an inner loop exit. */
    const struct sm6_block *innermost_loop_header, *outer_infinite_loop, *post;
    bool is_innermost_loop_header;

    innermost_loop_header = cfg_structuriser_get_innermost_loop_header_for(structuriser, header, node);
    is_innermost_loop_header = header == innermost_loop_header;

    if (sm6_block_dominates(header, node) && cfg_structuriser_node_dominates_all_reachable_exits(structuriser, node))
    {
        if (is_innermost_loop_header)
            return LOOP_EXIT_TYPE_EXIT;
        else
            return LOOP_EXIT_TYPE_INNERLOOPEXIT;
    }

    if (sm6_block_dominates(header, node))
    {
        if (is_innermost_loop_header)
        {
            /* Even if we dominate node, we might not be able to merge to it. */
            if (!sm6_block_can_loop_merge_to(header, node))
            {
                /* This is an escape we dominate, but this could also be a case where we break
                 * to a continue construct in the outer loop which is not reachable through back traversal.
                 * This will confuse loop analysis, since this kind of double continue will not resolve properly.
                 * In this case we need to rendezvous at this block with a ladder to avoid double-continue. */

                outer_infinite_loop = cfg_structuriser_get_innermost_loop_header_for(structuriser,
                        structuriser->entry_block, innermost_loop_header->immediate_dominator);
                if (outer_infinite_loop && outer_infinite_loop->pred_back_edge
                        && !outer_infinite_loop->pred_back_edge->succ.count
                        && sm6_block_post_dominates(outer_infinite_loop->pred_back_edge, node))
                {
                    return LOOP_EXIT_TYPE_MERGETOINFINITELOOP;
                }
                else
                {
                    return LOOP_EXIT_TYPE_ESCAPE;
                }
            }

            return LOOP_EXIT_TYPE_MERGE;
        }
        else
        {
            /* Try to detect if this is a degenerate inner loop merge.
             * If the inner loop header is the only way to exit the loop construct, the loop exit block
             * is a false exit. This is the case if the candidate must pass through the back edge, and
             * the back edge can only branch to header. In this case, the loop will not be visible
             * through back-propagation, but it is definitely part of the loop construct. */

            if (!innermost_loop_header->pred_back_edge
                    || innermost_loop_header->pred_back_edge->terminator.type != TERMINATOR_UNCOND_BR)
                return LOOP_EXIT_TYPE_INNERLOOPMERGE;

            post = sm6_block_find_common_post_dominator(node, innermost_loop_header->pred_back_edge);
            if (post == innermost_loop_header->pred_back_edge)
                return LOOP_EXIT_TYPE_INNERLOOPFALSEPOSITIVE;
            else
                return LOOP_EXIT_TYPE_INNERLOOPMERGE;
        }
    }
    else
    {
        return LOOP_EXIT_TYPE_ESCAPE;
    }
}

static void cfg_structuriser_analyse_loop(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, struct sm6_loop_analysis *result)
{
    struct sm6_block *pred, *loop_exit, *common;
    struct loop_merge_tracer merge_tracer;
    unsigned int i;

    /* Re-use existing result allocations. */
    result->direct_exits.count = 0;
    result->inner_direct_exits.count = 0;
    result->dominated_exit.count = 0;
    result->inner_dominated_exit.count = 0;
    result->non_dominated_exit.count = 0;
    result->dominated_continue_exit.count = 0;

    /* Now, we need to figure out which blocks belong in the loop construct.
     * The way to figure out a natural loop is any block which is dominated by loop header
     * and control flow passes to one of the back edges. */

    /* Unfortunately, it can be ambiguous which block is the merge block for a loop.
     * Ideally, there is a unique block which is the loop exit block, but if there are multiple breaks
     * there are multiple blocks which are not part of the loop construct. */

    pred = node->pred_back_edge;

    /* Back-trace from here.
     * The CFG is reducible, so node must dominate pred.
     * Since node dominates pred, there is no pred chain we can follow without
     * eventually hitting node, and we'll stop traversal there. */

    /* All nodes which are touched during this traversal must be part of the loop construct. */

    loop_merge_tracer_init(&merge_tracer, cfg_structuriser_next_true_value(structuriser));
    loop_merge_tracer_trace_to_parent(&merge_tracer, node, pred);
    loop_merge_tracer_trace_from_parent(&merge_tracer, node);

    for (i = 0; i < merge_tracer.loop_exits.count; ++i)
    {
        struct sm6_block_array *dst_array;
        enum loop_exit_type exit_type;

        loop_exit = merge_tracer.loop_exits.blocks[i];
        exit_type = cfg_structuriser_get_loop_exit_type(structuriser, node, loop_exit);
        switch (exit_type)
        {
            case LOOP_EXIT_TYPE_EXIT:
                dst_array = &result->direct_exits;
                break;

            case LOOP_EXIT_TYPE_INNERLOOPEXIT:
                /* It's not an exit for us, but the inner loop. */
                dst_array = &result->inner_direct_exits;
                break;

            case LOOP_EXIT_TYPE_MERGE:
                dst_array = &result->dominated_exit;
                break;

            case LOOP_EXIT_TYPE_INNERLOOPMERGE:
                dst_array = &result->inner_dominated_exit;
                break;

            case LOOP_EXIT_TYPE_INNERLOOPFALSEPOSITIVE:
                /* In this case, the inner loop can only exit at the loop header,
                 * and thus post-dominance analysis will always fail.
                 * Ignore this case as it's a false exit. */
                continue;

            case LOOP_EXIT_TYPE_ESCAPE:
                dst_array = &result->non_dominated_exit;
                break;

            case LOOP_EXIT_TYPE_MERGETOINFINITELOOP:
                dst_array = &result->dominated_continue_exit;
                break;

            default:
                FIXME("Unhandled loop exit type %u.\n", exit_type);
                continue;
        }
        sm6_block_array_append(dst_array, loop_exit);
    }

    if (result->dominated_continue_exit.count > 1)
    {
        /* If we have multiple continue exit candidates, they better merge into a single clean candidate that we
         * still dominate, otherwise, ignore this case and treat them all as normal Escape nodes. */
        common = sm6_block_array_find_common_post_dominator(&result->dominated_continue_exit);
        if (common && sm6_block_dominates(node, common))
        {
            result->dominated_continue_exit.blocks[0] = common;
            result->dominated_continue_exit.count = 1;
        }
        else
        {
            sm6_block_array_append_array(&result->non_dominated_exit, &result->dominated_continue_exit);
            result->dominated_continue_exit.count = 0;
        }
    }

    /* If the only merge candidates we have are inner dominated, treat them as true dominated exits. */
    if (!result->dominated_exit.count && result->inner_dominated_exit.count)
        vkd3d_swap(result->dominated_exit, result->inner_dominated_exit, struct sm6_block_array);

    /* If there are no direct exits, treat inner direct exits as direct exits. */
    if (!result->direct_exits.count)
        vkd3d_swap(result->direct_exits, result->inner_direct_exits, struct sm6_block_array);

    /* A direct exit can be considered a dominated exit if there are no better candidates. */
    if (!result->dominated_exit.count && result->direct_exits.count)
        vkd3d_swap(result->dominated_exit, result->direct_exits, struct sm6_block_array);

    /* If we only have one direct exit, consider it our merge block.
     * Pick either Merge or Escape. */
    if (result->direct_exits.count == 1 && !result->dominated_exit.count && !result->non_dominated_exit.count)
    {
        if (sm6_block_dominates(node, result->direct_exits.blocks[0]))
            vkd3d_swap(result->dominated_exit, result->direct_exits, struct sm6_block_array);
        else
            vkd3d_swap(result->non_dominated_exit, result->direct_exits, struct sm6_block_array);
    }

    if (result->dominated_exit.count >= 2)
    {
        /* Try to see if we can reduce the number of merge blocks to just 1.
         * This is relevant if we have various "clean" break blocks. */
        struct sm6_block *post_dominator = sm6_block_array_find_common_post_dominator(&result->dominated_exit);
        if (sm6_block_array_contains(&result->dominated_exit, post_dominator))
        {
            result->dominated_exit.blocks[0] = post_dominator;
            result->dominated_exit.count = 1;
        }
    }

    loop_merge_tracer_destroy(&merge_tracer);
}

static bool cfg_structuriser_loop_exit_supports_infinite_loop(struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *header, struct sm6_block *loop_exit)
{
    const struct sm6_block *inner_header = cfg_structuriser_get_innermost_loop_header_for(structuriser,
            header, loop_exit);
    /* A loop exit can exit out to an outer scope such that inner_header dominates the header.
     * If there is no inner loop we can transform the loop exit into a merge block quite easily
     * and avoid the infinite loop. */
    if (sm6_block_dominates(inner_header, header))
        return false;

    /* We have a candidate. If the candidates dominates all reachable exits, there is never a need to merge later. */
    return cfg_structuriser_node_dominates_all_reachable_exits(structuriser, loop_exit);
}

static bool cfg_structuriser_loop_exits_support_infinite_loop(struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *header, const struct sm6_block_array *loop_exits)
{
    unsigned int i;

    for (i = 0; i < loop_exits->count; ++i)
        if (!cfg_structuriser_loop_exit_supports_infinite_loop(structuriser, header, loop_exits->blocks[i]))
            return false;
    return true;
}

struct sm6_loop_merge_analysis
{
    struct sm6_block *merge;
    struct sm6_block *dominated_merge;
    struct sm6_block *infinite_continue_ladder;
};

static void cfg_structuriser_analyse_loop_merge(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, const struct sm6_loop_analysis *analysis,
        struct sm6_loop_merge_analysis *merge_result)
{
    struct sm6_block *merge, *dominated_merge;
    struct sm6_block_array merges = {0};
    unsigned int i;

    /* We have multiple blocks which are merge candidates. We need to figure out where execution reconvenes. */

    sm6_block_array_reserve(&merges, analysis->inner_dominated_exit.count + analysis->dominated_exit.count
            + analysis->non_dominated_exit.count);
    sm6_block_array_append_array(&merges, &analysis->inner_dominated_exit);
    sm6_block_array_append_array(&merges, &analysis->dominated_exit);
    sm6_block_array_append_array(&merges, &analysis->non_dominated_exit);
    merge = sm6_block_array_find_common_post_dominator(&merges);
    sm6_block_array_destroy(&merges);

    /* Try to find the sensible target first.
     * If one of our merge blocks is the successor of the continue block,
     * this is a prime candidate for a ladder block. */
    if (node->pred_back_edge && node->pred_back_edge->succ.count == 1
            && sm6_block_array_contains(&analysis->dominated_exit, node->pred_back_edge->succ.blocks[0]))
    {
        dominated_merge = node->pred_back_edge->succ.blocks[0];
    }
    else if (merge && !sm6_block_dominates(node, merge) && analysis->dominated_exit.count > 1)
    {
        /* Now, we might have Merge blocks which end up escaping out of the loop construct.
         * We might have to remove candidates which end up being break blocks after all. */
        struct sm6_block_array non_breaking_exits = {0};

        sm6_block_array_reserve(&non_breaking_exits, analysis->dominated_exit.count);
        for (i = 0; i < analysis->dominated_exit.count; ++i)
            if (!cfg_structuriser_control_flow_is_escaping(structuriser, analysis->dominated_exit.blocks[i], merge))
                sm6_block_array_append(&non_breaking_exits, analysis->dominated_exit.blocks[i]);

        dominated_merge = sm6_block_array_find_common_post_dominator(&non_breaking_exits);

        sm6_block_array_destroy(&non_breaking_exits);
    }
    else
    {
        dominated_merge = sm6_block_array_find_common_post_dominator(&analysis->dominated_exit);
    }

    if (!dominated_merge)
    {
        WARN("There is no candidate for ladder merging.\n");
    }

    if (dominated_merge && !sm6_block_dominates(node, dominated_merge))
    {
        WARN("Node does not dominate the merge target.\n");
        dominated_merge = NULL;
    }

    merge_result->merge = merge;
    merge_result->dominated_merge = dominated_merge;
    merge_result->infinite_continue_ladder = NULL;

    if (analysis->dominated_continue_exit.count)
    {
        assert(analysis->dominated_continue_exit.count == 1);
        merge_result->infinite_continue_ladder = analysis->dominated_continue_exit.blocks[0];
    }
}

static void cfg_structuriser_rewrite_transposed_loop_outer(struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *node, struct sm6_block *impossible_merge_target,
        const struct sm6_loop_merge_analysis *analysis)
{
    struct sm6_block_array *impossible_preds = &impossible_merge_target->pred;
    struct sm6_block *replaced_merge_block;
    struct sm6_phi *phi;
    unsigned int i;

    replaced_merge_block = cfg_structuriser_create_helper_pred_block(structuriser, analysis->dominated_merge);
    sm6_block_set_name2(replaced_merge_block, analysis->dominated_merge->name, ".transposed-merge-outer");

    for (i = 0; i < impossible_preds->count; ++i)
    {
        struct sm6_block *pred = impossible_preds->blocks[i];
        if (!cfg_structuriser_query_reachability(structuriser, analysis->dominated_merge, pred))
            sm6_block_retarget_branch(pred, impossible_merge_target, replaced_merge_block);
    }

    sm6_block_add_branch(replaced_merge_block, impossible_merge_target);
    replaced_merge_block->terminator.true_block = impossible_merge_target;
    replaced_merge_block->terminator.false_block = analysis->dominated_merge;
    replaced_merge_block->terminator.type = TERMINATOR_COND_BR;
    shader_register_make_scalar_ssa(&replaced_merge_block->terminator.conditional_reg, VKD3D_DATA_BOOL,
            structuriser->parser);

    phi = sm6_block_phi_new(replaced_merge_block, &replaced_merge_block->terminator.conditional_reg);
    sm6_phi_build_name(phi, "transposed_selector_", node->name);

    for (i = 0; i < replaced_merge_block->pred.count; ++i)
    {
        struct sm6_block *ladder_pred = replaced_merge_block->pred.blocks[i];
        bool branches_to_impossible = sm6_block_array_contains(impossible_preds, ladder_pred);
        struct incoming_value *incoming = sm6_phi_incoming_new(phi);
        incoming->block = ladder_pred;
        register_make_constant_bool(&incoming->reg, branches_to_impossible);
    }
}

static void cfg_structuriser_rewrite_transposed_loop_inner(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, struct sm6_block *impossible_merge_target,
        const struct sm6_loop_merge_analysis *analysis)
{
    /* Rewrite the control flow from the inside out through a transposition.
     * The common break target will become the merge block instead.
     * The continue will break out to the transposed merge instead.
     * In the ladder, we will enter a breaking path which branches out to loop_ladder. */

    /* We just arbitrary call this "inner", since I don't think it has a formal name.
     * In this case, dominated merge cannot reach impossible merge target. */

    struct sm6_block *merge, *dominated_merge, *ladder_break, *ladder_selection;
    struct sm6_block_array *ladder_preds;
    struct sm6_phi *phi;
    unsigned int i;

    merge = analysis->merge;
    dominated_merge = analysis->dominated_merge;

    ladder_break = sm6_function_blocks_new(structuriser->function);
    sm6_block_set_name2(ladder_break, node->name, ".transposed-merge-inner.break");
    ladder_break->terminator.type = TERMINATOR_UNCOND_BR;
    ladder_break->terminator.direct_block = impossible_merge_target;
    ladder_break->immediate_post_dominator = impossible_merge_target;
    ladder_break->forward_post_visit_order = impossible_merge_target->forward_post_visit_order;
    ladder_break->backward_post_visit_order = impossible_merge_target->backward_post_visit_order;

    ladder_selection = sm6_function_blocks_new(structuriser->function);
    sm6_block_set_name2(ladder_selection, node->name, ".transposed-merge-inner");
    ladder_selection->forward_post_visit_order = impossible_merge_target->forward_post_visit_order;
    ladder_selection->backward_post_visit_order = impossible_merge_target->backward_post_visit_order;
    ladder_selection->immediate_post_dominator = merge;
    ladder_break->immediate_dominator = ladder_selection;

    ladder_preds = &dominated_merge->pred;

    sm6_block_add_branch(ladder_selection, ladder_break);
    sm6_block_add_branch(ladder_selection, dominated_merge);
    cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, node, impossible_merge_target,
            ladder_selection, NULL);
    sm6_block_recompute_immediate_dominator(ladder_selection);

    sm6_block_add_branch(ladder_break, impossible_merge_target);

    /* Branches from these blocks should be rewritten to target transposed-merge. */
    for (i = 0; i < ladder_preds->count; ++i)
        sm6_block_retarget_branch(ladder_preds->blocks[i], dominated_merge, ladder_selection);

    ladder_selection->terminator.type = TERMINATOR_COND_BR;
    ladder_selection->terminator.true_block = dominated_merge;
    ladder_selection->terminator.false_block = ladder_break;
    shader_register_make_scalar_ssa(&ladder_selection->terminator.conditional_reg, VKD3D_DATA_BOOL,
            structuriser->parser);

    phi = sm6_block_phi_new(ladder_selection, &ladder_selection->terminator.conditional_reg);
    sm6_phi_build_name(phi, "transposed_selector_", node->name);
    for (i = 0; i < ladder_selection->pred.count; ++i)
    {
        struct sm6_block *ladder_pred = ladder_selection->pred.blocks[i];
        bool branches_to_dominated_merge = sm6_block_array_contains(ladder_preds, ladder_pred);
        struct incoming_value *incoming = sm6_phi_incoming_new(phi);
        incoming->block = ladder_pred;
        register_make_constant_bool(&incoming->reg, branches_to_dominated_merge);
    }
}

static bool cfg_structuriser_rewrite_transposed_loops(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_loop_analysis result = {0};
    bool did_rewrite = false;
    unsigned int i, index;

    for (index = structuriser->forward_post_visit_order.count; index && !did_rewrite; --index)
    {
        /* Visit in reverse order so we resolve outer loops first,
         * this lets us detect ladder-breaking loops. */
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[index - 1];
        struct sm6_block *merge, *dominated_merge, *impossible_merge_target;
        struct sm6_loop_merge_analysis merge_result;

        if (node->freeze_structured_analysis && node->merge_type == MERGE_TYPE_LOOP)
            continue;
        if (!node->pred_back_edge)
            continue;

        cfg_structuriser_analyse_loop(structuriser, node, &result);
        cfg_structuriser_analyse_loop_merge(structuriser, node, &result, &merge_result);

        merge = merge_result.merge;
        dominated_merge = merge_result.dominated_merge;

        if (!merge || !dominated_merge)
            continue;

        /* We might have a horribly complex scenario where a loop breaks, but it breaks to an outer scope
         * which is not consistent with the merge block, i.e. we need structured control flow to resolve properly
         * before we can break. This is ... problematic. */

        /* We call this an "inner" transposed loop here since merge block cannot reach this block. */

        /* Always resolve infinite continue ladders. This is where we break to
         * an outer infinite loop. We must resolve the scopes by making this ladder the
         * merge point, then we can break further. */
        impossible_merge_target = merge_result.infinite_continue_ladder;

        if (!impossible_merge_target && result.non_dominated_exit.count)
        {
            struct sm6_block *common_break_target = sm6_block_array_find_common_post_dominator(&result.non_dominated_exit);
            if (common_break_target && common_break_target != merge
                    && sm6_block_reaches_domination_frontier_before_merge(common_break_target, merge)
                    && !cfg_structuriser_query_reachability(structuriser, dominated_merge, common_break_target)
                    && !cfg_structuriser_query_reachability(structuriser, common_break_target, dominated_merge))
            {
                impossible_merge_target = common_break_target;
            }
        }

        if (!impossible_merge_target)
        {
            /* We might have a different scenario where there are multiple breaks, but they break out to different
             * scopes. One of these might require a similar impossible merge.
             * Common post dominator analysis would not catch this.
             * What we're looking for is a node which:
             * - Is dominated by loop header (or is in the domination frontier of loop header)
             * - Is reachable, but not dominated by dominated_merge.
             * - Post dominates one of the non_dominated_exits.
             * This means the node is in a twilight zone where the node is kinda in the loop construct, but kinda not. */

            /* Structured rules for a loop state that a node is in the construct if:
             * - It is dominated by loop header
             * - Not dominated by merge block.
             * In a sense, the merge block ends up branching back into its own loop, which is irreducible, kinda ... */

            /* We call this an "outer" transposed loop here since merge block *can* reach this block. */

            for (i = 0; i < result.non_dominated_exit.count && !impossible_merge_target; ++i)
            {
                struct sm6_block *candidate = result.non_dominated_exit.blocks[i];

                while (candidate != merge && candidate != dominated_merge)
                {
                    if (cfg_structuriser_query_reachability(structuriser, dominated_merge, candidate)
                            && !sm6_block_dominates(dominated_merge, candidate))
                    {
                        /* Merge block attempts to branch back into its own loop construct (yikes). */
                        impossible_merge_target = candidate;

                        /* If we don't dominate the merge target, i.e. we're in the domination frontier,
                         * we have to synthesize a fake impossible merge target first since the rewrite
                         * algorithm depends on node dominating the merge target. */
                        if (!sm6_block_dominates(node, impossible_merge_target))
                            impossible_merge_target = cfg_structuriser_create_ladder_block(structuriser, node,
                                    impossible_merge_target, ".impossible-ladder");
                        break;
                    }
                    else if (sm6_block_dominates(node, candidate) && candidate != candidate->immediate_post_dominator)
                    {
                        candidate = candidate->immediate_post_dominator;
                    }
                    else
                    {
                        /* We will be able to select a candidate in the domination frontier once.
                         * If we failed to find a candidate in the domination frontier, we're done checking. */
                        break;
                    }
                }
            }
        }

        if (impossible_merge_target)
        {
            if (cfg_structuriser_query_reachability(structuriser, dominated_merge, impossible_merge_target))
                cfg_structuriser_rewrite_transposed_loop_outer(structuriser, node, impossible_merge_target, &merge_result);
            else
                cfg_structuriser_rewrite_transposed_loop_inner(structuriser, node, impossible_merge_target, &merge_result);

            /* We have obliterated the existing control flow through transposition,
             * and any domination or post-domination analysis will break.
             * Re-traverse the CFG and try again.
             * Continue until we have eliminated all impossible loops (should be extremely rare). */
            did_rewrite = true;
        }
    }

    sm6_loop_analysis_destroy(&result);

    if (did_rewrite)
        cfg_structuriser_recompute_cfg(structuriser);
    return did_rewrite;
}

static int switch_case_forward_post_visit_order_compare_desc(const void *a, const void *b)
{
    const struct terminator_case *ca = a, *cb = b;
    return vkd3d_u32_compare(cb->node->forward_post_visit_order, ca->node->forward_post_visit_order);
}

static int switch_case_global_order_compare(const void *a, const void *b)
{
    const struct terminator_case *ca = a, *cb = b;

    if (ca->global_order != cb->global_order)
        return (ca->global_order > cb->global_order) - (ca->global_order < cb->global_order);
    /* stable sort */
    return vkd3d_u32_compare(ca->index, cb->index);
}

static struct sm6_block *cfg_structuriser_find_natural_switch_merge_block(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, struct sm6_block *post_dominator)
{
    bool has_impossible_fallthrough;
    uint64_t target_order;
    unsigned int i, j, n;
    uint64_t order = 1;

    /* Maintain the original switch block order if possible to avoid awkward churn in reference output. */
    for (i = 0; i < node->terminator.case_count; ++i)
    {
        struct terminator_case *c = &node->terminator.cases[i];
        /* We'll need to decrement global order up to N times in the worst case.
         * Use 64-bit here as a safeguard in case the module is using a ridiculous amount of case labels. */
        c->global_order = order++ * node->terminator.case_count;
    }

    /* First, sort so that any fallthrough parent comes before fallthrough target. */
    qsort(node->terminator.cases, node->terminator.case_count, sizeof(*node->terminator.cases),
            switch_case_forward_post_visit_order_compare_desc);

    /* Look at all potential fallthrough candidates and reassign global order. */
    for (size_t i = 0, n = node->terminator.case_count; i < n; ++i)
    {
        for (size_t j = i + 1; j < n; ++j)
        {
            struct terminator_case *parent = &node->terminator.cases[i];
            struct terminator_case *child = &node->terminator.cases[j];

            /* A case label might be the merge block candidate of the switch.
             * Don't consider case fallthrough if b post-dominates the entire switch statement. */
            if (child->node != post_dominator && parent->node != child->node
                    && cfg_structuriser_query_reachability(structuriser, parent->node, child->node))
            {
                parent->global_order = child->global_order - 1;
                break;
            }
        }
    }

    for (i = 0; i < node->terminator.case_count; ++i)
        node->terminator.cases[i].index = i;
    /* Sort again, but this time, stably by global order. */
    qsort(node->terminator.cases, node->terminator.case_count, sizeof(*node->terminator.cases),
              switch_case_global_order_compare);

    /* Detect impossible fallthrough scenarios. We can have A -> B -> C fallthrough, but not
     * A -> C and B -> C. In this situation, we should see C as the actual switch merge block,
     * and rewrite the switch to loop + switch.
     * Detect this by having two entries with identical global order. */

    has_impossible_fallthrough = false;
    target_order = 0;

    for (i = 1, n = node->terminator.case_count; i < n; ++i)
    {
        if (node->terminator.cases[i].global_order == node->terminator.cases[i - 1].global_order)
        {
            target_order = node->terminator.cases[i].global_order + 1;
            has_impossible_fallthrough = true;
            break;
        }
    }

    if (has_impossible_fallthrough)
    {
        for (i = 0; i < node->terminator.case_count; ++i)
        {
            struct terminator_case *c = &node->terminator.cases[i];
            if (c->global_order == target_order)
                return c->node;
        }
    }

    /* If two case labels merge execution before the candidate merge, we should consider that the natural merge,
     * since it is not possible to express this without a switch merge. */
    for (i = 0; i < node->terminator.case_count; ++i)
    {
        const struct terminator_case *c = &node->terminator.cases[i];
        for (j = 0; j < c->node->dominance_frontier.count; ++j)
        {
            struct sm6_block *front = c->node->dominance_frontier.blocks[j];
            /* Ignore frontiers that are other case labels.
             * We allow simple fallthrough, and if we found an impossible case we would have handled it already. */
            if (block_terminator_case_array_contains_block(&node->terminator, front))
                continue;

            if (front->forward_post_visit_order != post_dominator->forward_post_visit_order
                    && cfg_structuriser_query_reachability(structuriser, front, post_dominator))
            {
                /* If this is reachable by a different case label, we have a winner. This must be a fake fallthrough
                 * that we should promote to switch merge. */
                for (n = 0; n < node->terminator.case_count; ++n)
                {
                    const struct sm6_block *ic_block = node->terminator.cases[n].node;
                    if (ic_block != c->node && cfg_structuriser_query_reachability(structuriser, ic_block, front))
                        return front;
                }
            }
        }
    }

    return post_dominator;
}

static struct sm6_block *sm6_block_array_find_common_post_dominator_with_ignored_break(
        const struct sm6_block_array *src_candidates, const struct sm6_block *ignored_node)
{
    struct sm6_block_array candidates = {0}, next_nodes = {0};
    const struct sm6_block *front;
    struct sm6_block *pdom;
    unsigned int i;

    if (!src_candidates->count)
        return NULL;

    sm6_block_array_clone(&candidates, src_candidates);
    while (candidates.count != 1)
    {
        qsort(candidates.blocks, candidates.count, sizeof(*candidates.blocks), forward_post_visit_order_compare_desc);
        front = candidates.blocks[0];

        /* We reached exit without merging execution, there is no common post dominator. */
        if (!front->succ.count)
            return NULL;

        for (i = 0; i < front->succ.count; ++i)
            if (front->succ.blocks[i] != ignored_node)
                sm6_block_array_append_unique(&next_nodes, front->succ.blocks[i]);
        for (i = 1; i < candidates.count; ++i)
            if (candidates.blocks[i] != ignored_node)
                sm6_block_array_append_unique(&next_nodes, candidates.blocks[i]);

        candidates.count = 0;
        vkd3d_swap(candidates, next_nodes, struct sm6_block_array);
    }

    pdom = candidates.blocks[0];
    sm6_block_array_destroy(&next_nodes);
    sm6_block_array_destroy(&candidates);

    return pdom;
}

static void cfg_structuriser_retarget_succ_from(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *new_node, struct sm6_block *old_pred)
{
    struct sm6_block *s, *node;
    unsigned int i;

    for (i = 0; i < new_node->succ.count; ++i)
    {
        s = new_node->succ.blocks[i];
        sm6_block_array_retarget(&s->pred, old_pred, new_node);
    }

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node != old_pred)
        {
            /* Don't override immediate dominator for entry block. */
            if (node->immediate_dominator == old_pred)
                node->immediate_dominator = new_node;
        }
    }
    new_node->immediate_dominator = old_pred;

    /* Do not swap back edges. */
}

static void block_terminator_clone(struct sm6_block_terminator *dst, const struct sm6_block_terminator *src)
{
    *dst = *src;
    dst->cases = vkd3d_malloc(dst->case_count * sizeof(*dst->cases));
    memcpy(dst->cases, src->cases, dst->case_count * sizeof(*dst->cases));
}

static struct sm6_block *cfg_structuriser_create_helper_succ_block(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node)
{
    struct sm6_block *succ_node = sm6_function_blocks_new(structuriser->function);

    sm6_block_set_name2(succ_node, node->name, ".succ");

    /* Fixup visit order later. */
    succ_node->forward_post_visit_order = node->forward_post_visit_order;
    succ_node->backward_post_visit_order = node->backward_post_visit_order;

    vkd3d_swap(succ_node->succ, node->succ, struct sm6_block_array);

    /* Do not swap back edges, only forward edges. */
    succ_node->immediate_post_dominator = node->immediate_post_dominator;
    node->immediate_post_dominator = succ_node;

    block_terminator_clone(&succ_node->terminator, &node->terminator);
    node->terminator.type = TERMINATOR_UNCOND_BR;
    node->terminator.direct_block = succ_node;

    cfg_structuriser_retarget_succ_from(structuriser, succ_node, node);

    sm6_block_add_branch(node, succ_node);
    return succ_node;
}

static struct sm6_block *cfg_structuriser_create_switch_merge_ladder(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *header, struct sm6_block *merge)
{
    /* We did not rewrite switch blocks w.r.t. selection breaks.
     * We might be in a situation where the switch block is trying to merge to a block which is already being merged to.
     * Create a ladder which the switch block could merge to. */
    return cfg_structuriser_create_ladder_block(structuriser, header, merge, ".switch-merge");
}

static bool cfg_structuriser_find_switch_blocks(struct vkd3d_cfg_structuriser *structuriser, unsigned int pass)
{
    struct sm6_function *function = structuriser->function;
    struct sm6_block *node, *merge, *natural_merge;
    bool modified_cfg = false;
    unsigned int index;

    for (index = structuriser->forward_post_visit_order.count; index; --index)
    {
        node = structuriser->forward_post_visit_order.blocks[index - 1];
        if (node->terminator.type != TERMINATOR_SWITCH)
            continue;

        merge = sm6_block_array_find_common_post_dominator(&node->succ);
        natural_merge = cfg_structuriser_find_natural_switch_merge_block(structuriser, node, merge);

        if (node->freeze_structured_analysis && node->merge_type == MERGE_TYPE_SELECTION)
        {
            natural_merge = node->selection_merge_block;
        }
        else if (pass == 0)
        {
            bool can_merge_to_post_dominator = merge && sm6_block_dominates(node, merge) && !merge->headers.count;

            /* Need to guarantee that we can merge somewhere.
             * If possible we want to make it so that by creating a ladder,
             * we change the post-dominator to something we dominate.
             * For this to work, the dominance frontier of node must only contain the merge node. */
            if (merge != natural_merge && !can_merge_to_post_dominator &&
                node->dominance_frontier.count == 1 && node->dominance_frontier.blocks[0] == merge)
            {
                merge = cfg_structuriser_create_switch_merge_ladder(structuriser, node, merge);
                assert(sm6_block_dominates(node, merge));
                modified_cfg = true;
                can_merge_to_post_dominator = true;
            }

            /* Need to rewrite the switch. */
            if (merge != natural_merge && can_merge_to_post_dominator)
            {
                struct sm6_block *switch_outer = cfg_structuriser_create_helper_pred_block(structuriser, node);
                switch_outer->merge_type = MERGE_TYPE_LOOP;
                switch_outer->loop_merge_block = merge;
                switch_outer->freeze_structured_analysis = true;
                sm6_block_array_append(&merge->headers, switch_outer);

                /* Shouldn't be needed (I believe), but spirv-val is a bit temperamental when double breaking
                 * straight out of a switch block in some situations, so try not to ruffle too many feathers. */
                if (sm6_block_array_contains(&node->succ, natural_merge))
                {
                    struct sm6_block *dummy_case = sm6_function_blocks_new(function);
                    sm6_block_set_name2(dummy_case, natural_merge->name, ".pred");
                    dummy_case->immediate_dominator = node;
                    dummy_case->immediate_post_dominator = natural_merge;
                    dummy_case->forward_post_visit_order = node->forward_post_visit_order;
                    dummy_case->backward_post_visit_order = node->backward_post_visit_order;
                    dummy_case->terminator.type = TERMINATOR_UNCOND_BR;
                    dummy_case->terminator.direct_block = natural_merge;
                    sm6_block_add_branch(dummy_case, natural_merge);
                    sm6_block_retarget_branch(node, natural_merge, dummy_case);
                }

                if (sm6_block_array_contains(&node->succ, merge))
                {
                    struct sm6_block *dummy_break = sm6_function_blocks_new(function);
                    sm6_block_set_name2(dummy_break, node->name, ".break");
                    dummy_break->immediate_dominator = node;
                    dummy_break->immediate_post_dominator = merge;
                    dummy_break->forward_post_visit_order = node->forward_post_visit_order;
                    dummy_break->backward_post_visit_order = node->backward_post_visit_order;
                    dummy_break->terminator.type = TERMINATOR_UNCOND_BR;
                    dummy_break->terminator.direct_block = merge;
                    sm6_block_add_branch(dummy_break, merge);
                    sm6_block_retarget_branch(node, merge, dummy_break);
                }

                node->freeze_structured_analysis = true;
            }
        }

        merge = natural_merge;

        /* We cannot rewrite the CFG in pass 1 safely, this should have happened in pass 0. */
        if (pass == 0 && (!sm6_block_dominates(node, merge) || sm6_block_is_plain_continue(merge)))
        {
            cfg_structuriser_create_switch_merge_ladder(structuriser, node, merge);
            merge = sm6_block_array_find_common_post_dominator(&node->succ);
            modified_cfg = true;
        }

        if (sm6_block_dominates(node, merge))
        {
            node->merge_type = MERGE_TYPE_SELECTION;
            node->selection_merge_block = merge;
            sm6_block_add_unique_header(merge, node);
        }
        else
        {
            /* We got a switch block where someone is escaping. Similar idea as for loop analysis.
             * Find a post-dominator where we ignore branches which are "escaping". */
            struct sm6_block *dominated_merge_target = sm6_block_array_find_common_post_dominator_with_ignored_break(
                    &node->succ, merge);
            assert(dominated_merge_target);
            if (sm6_block_dominates(node, dominated_merge_target))
            {
                node->merge_type = MERGE_TYPE_SELECTION;
                node->selection_merge_block = merge;
                sm6_block_add_unique_header(dominated_merge_target, node);
                sm6_block_add_unique_header(merge, node);
            }
        }

        /* A switch header might also be a loop header. Create a helper succ block for this case. */
        if (pass == 0 && node->pred_back_edge)
        {
            node = cfg_structuriser_create_helper_succ_block(structuriser, node);
            modified_cfg = true;
        }
    }

    return modified_cfg;
}

static void cfg_structuriser_find_loops(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_loop_analysis result = {0};
    bool force_infinite_loop;
    unsigned int index;

    for (index = structuriser->forward_post_visit_order.count; index; --index)
    {
        /* Visit in reverse order so we resolve outer loops first,
         * this lets us detect ladder-breaking loops. */
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[index - 1];

        if (node->freeze_structured_analysis)
        {
            /* If we have a pre-created dummy loop for ladding breaking,
             * just propagate the header information and be done with it. */
            if (node->merge_type == MERGE_TYPE_LOOP)
            {
                sm6_block_array_append(&node->loop_merge_block->headers, node);
                continue;
            }
        }

        if (!node->pred_back_edge)
            continue;

        /* There are back-edges here, this must be a loop header. */
        node->merge_type = MERGE_TYPE_LOOP;

        cfg_structuriser_analyse_loop(structuriser, node, &result);

        /* This should not come up here, and must be handled in transpose loops. */
        assert(!result.dominated_continue_exit.count);

        /* Detect infinite loop with an exit which is only in inner loop construct.
         * It is impossible to construct a merge block in this case since the merge targets,
         * so just merge to unreachable. */
        force_infinite_loop = false;
        if (!node->pred_back_edge->succ.count)
        {
            force_infinite_loop = cfg_structuriser_loop_exits_support_infinite_loop(structuriser, node, &result.dominated_exit)
                    && cfg_structuriser_loop_exits_support_infinite_loop(structuriser, node, &result.non_dominated_exit)
                    && cfg_structuriser_loop_exits_support_infinite_loop(structuriser, node, &result.inner_dominated_exit)
                    && cfg_structuriser_loop_exits_support_infinite_loop(structuriser, node, &result.direct_exits)
                    && cfg_structuriser_loop_exits_support_infinite_loop(structuriser, node, &result.inner_direct_exits);
        }

        if (force_infinite_loop || (!result.dominated_exit.count && !result.inner_dominated_exit.count
                && !result.non_dominated_exit.count))
        {
            /* There can be zero loop exits, i.e. infinite loop. This means we have no merge block.
             * We will invent a merge block to satisfy SPIR-V validator, and declare it as unreachable. */
            node->loop_merge_block = NULL;
        }
        else if (result.dominated_exit.count == 1 && !result.non_dominated_exit.count && !result.inner_dominated_exit.count)
        {
            /* Clean merge.
             * This is a unique merge block. There can be no other merge candidate. */
            node->loop_merge_block = result.dominated_exit.blocks[0];

            sm6_block_add_unique_header(node->loop_merge_block, node);
        }
        else if (!result.dominated_exit.count && !result.inner_dominated_exit.count && result.non_dominated_exit.count == 1)
        {
            struct sm6_block *merge_block, *ladder;

            /* Single-escape merge.
             * It is unique, but we need workarounds later. */
            merge_block = result.non_dominated_exit.blocks[0];

            /* We can make the non-dominated exit dominated by
             * adding a ladder block in-between. This allows us to merge the loop cleanly
             * before breaking out. */

            ladder = sm6_function_blocks_new(structuriser->function);
            sm6_block_set_name2(ladder, node->name, ".merge");
            sm6_block_add_branch(ladder, merge_block);
            ladder->terminator.type = TERMINATOR_UNCOND_BR;
            ladder->terminator.direct_block = merge_block;
            ladder->immediate_post_dominator = merge_block;
            ladder->forward_post_visit_order = merge_block->forward_post_visit_order;
            ladder->backward_post_visit_order = merge_block->backward_post_visit_order;

            cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, node, merge_block, ladder, NULL);
            node->loop_ladder_block = NULL;
            node->loop_merge_block = ladder;
            sm6_block_recompute_immediate_dominator(ladder);

            sm6_block_add_unique_header(node->loop_merge_block, node);
        }
        else
        {
            struct sm6_loop_merge_analysis merge_result;
            struct sm6_block *dominated_merge;
            struct sm6_block *merge;

            cfg_structuriser_analyse_loop_merge(structuriser, node, &result, &merge_result);
            merge = merge_result.merge;
            dominated_merge = merge_result.dominated_merge;

            if (!merge)
            {
                WARN("Failed to find a common merge point.\n");
            }
            else if (sm6_block_can_loop_merge_to(node, merge))
            {
                /* Clean merge.
                 * This is a unique merge block. There can be no other merge candidate. */

                node->loop_merge_block = merge;
                sm6_block_add_unique_header(node->loop_merge_block, node);
            }
            else
            {
                /* Single-escape merge.
                 * It is unique, but we need workarounds later. */

                /* We will use this block as a ladder. */
                node->loop_ladder_block = dominated_merge;
                node->loop_merge_block = merge;

                sm6_block_add_unique_header(node->loop_merge_block, node);
            }
        }
    }

    sm6_loop_analysis_destroy(&result);
}

static const struct sm6_block *sm6_block_get_outer_selection_dominator(const struct sm6_block *node)
{
    const struct sm6_block *idom = node->immediate_dominator;

    assert(idom);

    /* We need to find an immediate dominator which we do not post-dominate.
     * That first idom is considered the outer selection header. */
    while (idom->terminator.type != TERMINATOR_SWITCH && sm6_block_post_dominates(node, idom))
    {
        if (!idom->pred.count)
            break;

        /* Skip from merge block to header. */
        while (sm6_block_array_contains(&idom->headers, idom->immediate_dominator))
            idom = idom->immediate_dominator;

        if (sm6_block_post_dominates(node, idom))
        {
            assert(idom->immediate_dominator);
            idom = idom->immediate_dominator;
        }
    }

    return idom;
}

static struct sm6_block *cfg_structuriser_find_break_target_for_selection_construct(
        struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *idom, struct sm6_block *merge)
{
    struct sm6_block_array new_visit_queue = {0};
    struct sm6_block_array visit_queue = {0};
    struct sm6_block_array candidates = {0};
    unsigned int i, true_value;
    struct sm6_block *n;

    true_value = cfg_structuriser_next_true_value(structuriser);
    sm6_block_array_append(&visit_queue, idom);
    do
    {
        for (i = 0; i < visit_queue.count; ++i)
        {
            n = visit_queue.blocks[i];

            if (n->break_target_visited == true_value)
                continue;
            n->break_target_visited = true_value;

            if (cfg_structuriser_query_reachability(structuriser, n, merge))
                sm6_block_array_append_array(&new_visit_queue, &n->succ);
            else
                sm6_block_array_append(&candidates, n);
        }

        vkd3d_swap(visit_queue, new_visit_queue, struct sm6_block_array);
        new_visit_queue.count = 0;
    } while (visit_queue.count);

    sm6_block_array_destroy(&visit_queue);
    sm6_block_array_destroy(&new_visit_queue);
    n = sm6_block_array_find_common_post_dominator(&candidates);
    sm6_block_array_destroy(&candidates);

    return n;
}

static void cfg_structuriser_find_selection_merges(struct vkd3d_cfg_structuriser *structuriser, unsigned int pass)
{
    const struct sm6_block *inner_header;
    struct sm6_block *node, *idom;
    unsigned int i, j;

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->pred.count <= 1)
            continue;

        /* Never merge to continue block.
         * We should never hit this path unless we explicitly
         * avoided creating a continue ladder block earlier. */
        if (sm6_block_is_plain_continue(node))
            continue;

        /* If there are 2 or more pred edges, try to merge execution. */

        /* The idom is the natural header block. */
        idom = node->immediate_dominator;
        assert(idom->succ.count >= 2);

        /* Check for case fallthrough here. In this case, we do not have a merge scenario, just ignore. */
        inner_header = sm6_block_get_outer_selection_dominator(node);
        if (inner_header && inner_header->terminator.type == TERMINATOR_SWITCH)
        {
            if (inner_header->selection_merge_block == node)
            {
                /* We just found a switch block which we have already handled. */
                continue;
            }

            if (sm6_block_array_contains(&inner_header->succ, node))
            {
                /* Fallthrough. */
                continue;
            }
        }

        for (j = 0; j < node->headers.count; ++j)
        {
            /* If we have a loop header already associated with this block, treat that as our idom. */
            if (node->headers.blocks[j]->forward_post_visit_order > idom->forward_post_visit_order)
                idom = node->headers.blocks[j];
        }

        /* Similar, but also check if we have associated ladder blocks with the idom. */
        if (!idom->pred_back_edge)
        {
            const struct sm6_block *inner_loop_header = cfg_structuriser_get_innermost_loop_header_for(structuriser,
                    structuriser->entry_block, idom);
            if (inner_loop_header && inner_loop_header->loop_ladder_block == node)
                idom = (struct sm6_block *)inner_loop_header;
        }

        if (pass == 0)
        {
            /* Check that we're not merging ourselves into the aether.
             * This is a scenario that can happen if we attempt to merge to a block which terminates the CFG
             * (return or unreachable), but does not post-dominate the idom candidate,
             * i.e. the selection construct needs to break to some other scope.
             * If this happens, we won't be able to register a typical breaking scenario
             * (since post-domination analysis won't help us),
             * and we need to do some magic fixups. */

            struct sm6_block *merge_candidate = sm6_block_find_common_post_dominator(idom, node);
            bool post_dominator_is_exit_node =
                    merge_candidate && merge_candidate->immediate_post_dominator == merge_candidate;
            bool merged_into_terminating_path = post_dominator_is_exit_node
                    && cfg_structuriser_node_dominates_all_reachable_exits(structuriser, node);

            /* If our candidate idom post dominates the entry block, we consider this the main path of execution. */
            if (merged_into_terminating_path)
                merged_into_terminating_path = !sm6_block_post_dominates(idom, structuriser->entry_block);

            if (merged_into_terminating_path)
            {
                /* Similar to loops, find the break target for this construct. */
                struct sm6_block *break_target = cfg_structuriser_find_break_target_for_selection_construct(
                        structuriser, idom, node);

                /* Have not observed any scenario where we won't have a dominated break target we can use. */
                if (break_target && sm6_block_dominates(idom, break_target) && !break_target->headers.count)
                {
                    /* Enclose this scope in a loop. */
                    struct sm6_block *helper_pred = cfg_structuriser_create_helper_pred_block(structuriser, idom);
                    helper_pred->merge_type = MERGE_TYPE_LOOP;
                    helper_pred->loop_merge_block = break_target;
                    helper_pred->freeze_structured_analysis = true;
                    sm6_block_array_append(&break_target->headers, helper_pred);
                }
            }
        }

        if (idom->merge_type == MERGE_TYPE_NONE || idom->merge_type == MERGE_TYPE_SELECTION)
        {
            /* We just found a switch block which we have already handled. */
            if (idom->terminator.type == TERMINATOR_SWITCH)
                continue;

            /* If the idom is already a selection construct, this must mean
             * we have some form of breaking construct inside this inner construct.
             * This fooled find_selection_merges() to think we had a selection merge target at the break target.
             * Fix this up here, where we rewrite the outer construct as a fixed loop instead. */
            if (idom->merge_type == MERGE_TYPE_SELECTION && pass == 0)
            {
                struct sm6_block *merge_candidate;

                assert(idom->selection_merge_block);

                /* If we turn the outer selection construct into a loop,
                 * we remove the possibility to break further out (without adding ladders like we do for loops).
                 * To make this work, we must ensure that the new merge block post-dominates the loop and selection merge. */
                merge_candidate = sm6_block_find_common_post_dominator(idom->selection_merge_block, idom);

                if (!merge_candidate || merge_candidate == idom->selection_merge_block)
                {
                    idom->loop_merge_block = idom->selection_merge_block;
                }
                else
                {
                    /* Make sure we split merge scopes. Pretend we have a true loop. */
                    idom->loop_ladder_block = idom->selection_merge_block;
                    idom->loop_merge_block = merge_candidate;
                }

                sm6_block_add_unique_header(idom->loop_merge_block, idom);

                idom->merge_type = MERGE_TYPE_LOOP;
                idom->selection_merge_block = NULL;
                idom->freeze_structured_analysis = true;
                idom = cfg_structuriser_create_helper_succ_block(structuriser, idom);
            }
            else if (idom->merge_type == MERGE_TYPE_SELECTION)
            {
                WARN("Headers mismatch in pass 1.\n");
            }

            idom->merge_type = MERGE_TYPE_SELECTION;
            sm6_block_add_unique_header(node, idom);
            idom->selection_merge_block = node;
        }
        else if (idom->merge_type == MERGE_TYPE_LOOP)
        {
            if (idom->loop_merge_block == node && idom->loop_ladder_block)
            {
                /* We need to create an outer shell for this header since we need to ladder break to this node. */
                struct sm6_block *loop = cfg_structuriser_create_helper_pred_block(structuriser, idom);
                loop->merge_type = MERGE_TYPE_LOOP;
                loop->loop_merge_block = node;
                loop->freeze_structured_analysis = true;
                sm6_block_add_unique_header(node, loop);
            }
            else if (idom->loop_merge_block != node && idom->loop_ladder_block != node)
            {
                struct sm6_block *selection_idom = cfg_structuriser_create_helper_succ_block(structuriser, idom);
                /* If we split the loop header into the loop header -> selection merge header,
                 * then we can merge into a continue block for example. */
                selection_idom->merge_type = MERGE_TYPE_SELECTION;
                selection_idom->selection_merge_block = node;
                sm6_block_add_unique_header(node, selection_idom);
            }
        }
        else
        {
            /* We are hosed. There is no obvious way to merge execution here.
             * This might be okay. */
            WARN("Cannot merge execution for node %p (%s).\n", node, node->name);
        }
    }
}

static void cfg_structuriser_merge_to_succ(struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node,
        unsigned int index)
{
    sm6_block_array_append(&node->succ.blocks[index]->headers, node);
    node->selection_merge_block = node->succ.blocks[index];
    node->merge_type = MERGE_TYPE_SELECTION;
}

static struct sm6_block *sm6_block_get_outer_header_dominator(struct sm6_block *block)
{
    struct sm6_block *node = block->immediate_dominator;

    assert(node);

    while (node->succ.count == 1 && node->terminator.type != TERMINATOR_SWITCH && !node->pred_back_edge)
    {
        if (!node->pred.count)
            break;

        assert(node->immediate_dominator);
        node = node->immediate_dominator;
    }

    return node;
}

static void cfg_structuriser_fixup_broken_selection_merges(struct vkd3d_cfg_structuriser *structuriser,
        unsigned int pass)
{
    bool dominates_a, dominates_b, merge_a_has_header, merge_b_has_header;
    struct sm6_block *node;
    int trivial_merge_index;
    unsigned int i;

    /* Here we deal with selection branches where one path breaks and one path merges.
     * This is common case for ladder blocks where we need to merge to the "true" merge block.
     * The selection header has two succs, but the merge block might only have one pred block,
     * which means it was not considered a merge candidate earlier in find_selection_merges(). */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->succ.count != 2)
            continue;
        if (node->merge_type != MERGE_TYPE_NONE)
            continue;

        /* A continue block will never need to merge execution, but it shouldn't have succ.count == 2,
         * but rather succ.count == 1 and a back edge. */
        if (node->succ_back_edge)
            continue;

        dominates_a = sm6_block_dominates(node, node->succ.blocks[0]);
        dominates_b = sm6_block_dominates(node, node->succ.blocks[1]);

        /* Continue blocks should also be considered to have a header already. Makes sure we don't merge to them. */
        merge_a_has_header = node->succ.blocks[0]->headers.count || sm6_block_is_plain_continue(node->succ.blocks[0]);
        merge_b_has_header = node->succ.blocks[1]->headers.count || sm6_block_is_plain_continue(node->succ.blocks[1]);

        trivial_merge_index = -1;

        /* Only allow the obvious merge candidates in pass 1.
         * In pass 0, we might have a clear merge candidate,
         * but the other path might be an escaping edge, which needs to be considered. */
        if (dominates_a && !dominates_b && !merge_a_has_header)
        {
            /* A is obvious candidate. B is a direct break/continue construct target most likely. */
            cfg_structuriser_merge_to_succ(structuriser, node, 0);
            trivial_merge_index = 0;
        }
        else if (dominates_b && !dominates_a && !merge_b_has_header)
        {
            /* B is obvious candidate. A is a direct break/continue construct target most likely. */
            cfg_structuriser_merge_to_succ(structuriser, node, 1);
            trivial_merge_index = 1;
        }
        else if (dominates_a && dominates_b && !merge_a_has_header && merge_b_has_header)
        {
            /* Not as obvious of a candidate, but this can happen if one path hits continue block,
             * and other path hits a ladder merge block.
             * For do/while(false) style loop, the loop body may dominate the merge block. */
            cfg_structuriser_merge_to_succ(structuriser, node, 0);
            trivial_merge_index = 0;
        }
        else if (dominates_a && dominates_b && !merge_b_has_header && merge_a_has_header)
        {
            /* Not as obvious of a candidate, but this can happen if one path hits continue block,
             * and other path hits a ladder merge block.
             * For do/while style loop, the loop body may dominate the merge block. */
            cfg_structuriser_merge_to_succ(structuriser, node, 1);
            trivial_merge_index = 1;
        }
        else if (dominates_a && dominates_b && !merge_a_has_header && !merge_b_has_header)
        {
            /* We could merge to both, no obvious merge point.
             * Figure out where execution reconvenes.
             * If we have a "break"-like construct inside a selection construct, we will not end up dominating the merge block.
             * This will be fixed up with ladder constructs later in first pass. */

            /* In second pass, we will have redirected any branches which escape through a ladder block.
             * If we find that one path of the selection construct must go through that ladder block, we know we have a break construct. */
            struct sm6_block *merge = sm6_block_array_find_common_post_dominator(&node->succ);
            if (merge)
            {
                bool tie_break_merge;

                bool dominates_merge = sm6_block_dominates(node, merge);
                bool merges_to_continue = sm6_block_is_plain_continue(merge);

                /* Here we have a likely case where one block is doing a clean "break" out of a loop, and
                 * the other path continues as normal, and then conditionally breaks in a continue block or something similar. */
                bool ambiguous_merge_case = !merges_to_continue && dominates_merge && merge->headers.count;

                /* Happens first iteration. We'll have to split blocks, so register a merge target where we want it.
                 * Otherwise, this is the easy case if we observe it in pass 1.
                 * This shouldn't really happen though, as we'd normally resolve this earlier in find_selection_merges. */
                bool mark_merge_block_case = !merges_to_continue && (!merge->headers.count || pass == 0);

                /* Another scenario is that we don't dominate the merge block in pass 1. We cannot split blocks now.
                 * Check to see which paths can actually reach the merge target without going through a ladder block.
                 * If we don't go through ladder it means an outer scope will actually reach the merge node.
                 * If we reach a ladder it means a block we dominate will make the escape. */

                /* If we're in pass 1 and we still don't dominate our merge target, consider it ambiguous. */
                if (pass == 1 && !dominates_merge)
                    ambiguous_merge_case = true;

                /* Another case is when one path is "breaking" out to a continue block which we don't dominate.
                 * We should not attempt to do ladder breaking here in pass 0 since it's unnecessary. */
                tie_break_merge = ambiguous_merge_case || !mark_merge_block_case;

                if (tie_break_merge)
                {
                    bool a_path_is_break = cfg_structuriser_control_flow_is_escaping(structuriser,
                            node->succ.blocks[0], merge);
                    bool b_path_is_break = cfg_structuriser_control_flow_is_escaping(structuriser,
                            node->succ.blocks[1], merge);

                    if (a_path_is_break && b_path_is_break)
                    {
                        /* Both paths break, so we don't need to merge anything. Use Unreachable merge target. */
                        node->merge_type = MERGE_TYPE_SELECTION;
                        node->selection_merge_block = NULL;
                    }
                    else if (b_path_is_break)
                        cfg_structuriser_merge_to_succ(structuriser, node, 0);
                    else if (a_path_is_break)
                        cfg_structuriser_merge_to_succ(structuriser, node, 1);
                    else
                    {
                        /* Need more interesting tie-breaking.
                         * We can deduce which path is breaking or not based on the dominance frontier.
                         * If a dominance frontier for A can reach B, then we assume that B is breaking further than A
                         * is, so we should merge to A.
                         * The breaking path for B will likely need to ensure that the selection header can
                         * support such a break.
                         * If we hit this path, the common post-dominator will not find the intended merge
                         * target for B, so we never get to perform the necessary fixup. */
                        struct sm6_block *a_front = node->succ.blocks[0]->dominance_frontier.count == 1 ?
                                        node->succ.blocks[0]->dominance_frontier.blocks[0] : NULL;
                        struct sm6_block *b_front = node->succ.blocks[1]->dominance_frontier.count == 1 ?
                                        node->succ.blocks[1]->dominance_frontier.blocks[0] : NULL;
                        bool found_candidate = false;

                        struct sm6_block *inner_merge_candidate = NULL;

                        /* If there is no unique dominance frontier for one path, pick the one that has a unique frontier
                         * as that in considered a merge. */
                        if ((a_front || b_front) && a_front != b_front)
                        {
                            if (!b_front || (a_front && cfg_structuriser_query_reachability(structuriser, a_front, b_front)))
                            {
                                cfg_structuriser_merge_to_succ(structuriser, node, 0);
                                inner_merge_candidate = b_front;
                                found_candidate = true;
                            }
                            else if (!a_front || (b_front && cfg_structuriser_query_reachability(structuriser, b_front, a_front)))
                            {
                                cfg_structuriser_merge_to_succ(structuriser, node, 1);
                                inner_merge_candidate = a_front;
                                found_candidate = true;
                            }
                        }

                        if (!found_candidate)
                        {
                            node->merge_type = MERGE_TYPE_SELECTION;
                            node->selection_merge_block = NULL;

                            if (a_front && b_front && a_front->headers.count == 1 && b_front->headers.count == 1)
                            {
                                /* Extremely ambiguous merge where the selection construct can merge to two different paths.
                                 * Our only option at this point is to pick an arbitrary winner
                                 * and consider one path the breaking one arbitrarily. */
                                struct sm6_block *a_header = a_front->headers.blocks[0];
                                struct sm6_block *b_header = b_front->headers.blocks[0];

                                /* Pick the largest enclosing header as a heuristic. */
                                inner_merge_candidate =
                                        a_header->forward_post_visit_order > b_header->forward_post_visit_order ?
                                        a_front : b_front;
                            }
                        }

                        if (inner_merge_candidate && inner_merge_candidate->headers.count == 1)
                        {
                            /* The breaking path tries to break to this node.
                             * This will only trigger in pass 1. */
                            struct sm6_block *header = inner_merge_candidate->headers.blocks[0];
                            if (header->merge_type == MERGE_TYPE_SELECTION)
                            {
                                /* Promote to loop header instead.
                                 * We might have to enter the loop ladder fixup stages later
                                 * to insert ladders as required. */
                                header->merge_type = MERGE_TYPE_LOOP;
                                header->loop_merge_block = header->selection_merge_block;
                                header->selection_merge_block = NULL;
                                header->freeze_structured_analysis = true;
                            }
                        }
                    }
                }
                else
                {
                    assert(merge);
                    node->selection_merge_block = merge;
                    node->merge_type = MERGE_TYPE_SELECTION;
                    sm6_block_array_append(&merge->headers, node);
                }
            }
            else
            {
                /* We likely had one side of the branch take an "exit", in which case there is no common post-dominator. */
                bool a_dominates_exit = cfg_structuriser_node_dominates_all_reachable_exits(structuriser, node->succ.blocks[0]);
                bool b_dominates_exit = cfg_structuriser_node_dominates_all_reachable_exits(structuriser, node->succ.blocks[1]);
                if (!a_dominates_exit && b_dominates_exit)
                    cfg_structuriser_merge_to_succ(structuriser, node, 0);
                else if (!b_dominates_exit && a_dominates_exit)
                    cfg_structuriser_merge_to_succ(structuriser, node, 1);
                else
                {
                    /* Both paths lead to exit. Do we even need to merge here?
                     * In worst case we can always merge to an unreachable node in the CFG. */
                    node->merge_type = MERGE_TYPE_SELECTION;
                    node->selection_merge_block = NULL;
                }
            }
        }
        else if (pass == 0)
        {
            /* No possible merge target. Just need to pick whatever node is the merge block here.
             * Only do this in first pass, so that we can get a proper ladder breaking mechanism in place if we are escaping. */
            struct sm6_block *merge = sm6_block_array_find_common_post_dominator(&node->succ);

            if (merge)
            {
                /* Don't try to merge to our switch block. */
                struct sm6_block *inner_header = sm6_block_get_outer_header_dominator(node);
                bool conditional_switch_break = inner_header && inner_header->merge_type == MERGE_TYPE_SELECTION
                        && inner_header->selection_merge_block == merge;

                if (!conditional_switch_break)
                {
                    node->selection_merge_block = merge;
                    node->merge_type = MERGE_TYPE_SELECTION;
                    sm6_block_array_append(&merge->headers, node);
                }
            }
        }

        if (trivial_merge_index >= 0 && pass == 0)
        {
            struct sm6_block *merge = sm6_block_array_find_common_post_dominator(&node->succ);
            if (merge && !sm6_block_dominates(node, merge) && !sm6_block_is_plain_continue(merge))
            {
                struct sm6_block *current_candidate, *other_candidate;
                bool current_escapes, other_escapes;

                if (merge->headers.count)
                {
                    /* We might have a trivial merge, yet the other branch direction
                     * is a breaking construct. We will have to split some blocks. */
                    sm6_block_array_append(&merge->headers, node);
                }

                current_candidate = node->succ.blocks[trivial_merge_index];
                other_candidate = node->succ.blocks[1 - trivial_merge_index];

                current_escapes = current_candidate == merge || cfg_structuriser_control_flow_is_escaping(structuriser,
                        current_candidate, merge);
                other_escapes = other_candidate == merge || cfg_structuriser_control_flow_is_escaping(structuriser,
                        other_candidate, merge);

                /* If we tried to merge in a direction which is a breaking construct,
                 * this means that the other path is actual desired break path. */
                if (current_escapes && !other_escapes)
                {
                    struct sm6_block *target_block = node->succ.blocks[1 - trivial_merge_index];
                    /* We kinda want to merge the other way, but to do that, we need an interim block. */
                    struct sm6_block *ladder = sm6_function_blocks_new(structuriser->function);
                    sm6_block_set_name4(ladder, node->name, ".", target_block->name, ".interim");
                    sm6_block_add_branch(ladder, target_block);
                    ladder->terminator.type = TERMINATOR_UNCOND_BR;
                    ladder->terminator.direct_block = target_block;
                    ladder->immediate_dominator = node;
                    ladder->immediate_post_dominator = target_block;
                    sm6_block_array_append(&ladder->dominance_frontier, target_block);
                    ladder->forward_post_visit_order = node->forward_post_visit_order;
                    ladder->backward_post_visit_order = node->backward_post_visit_order;
                    sm6_block_retarget_branch(node, target_block, ladder);
                    node->selection_merge_block = ladder;
                }
            }
        }
    }
}

static struct sm6_block *cfg_structuriser_get_target_break_block_for_inner_header(
        struct vkd3d_cfg_structuriser *structuriser, const struct sm6_block *node, unsigned int header_index)
{
    struct sm6_block *inner_header = node->headers.blocks[header_index];
    struct sm6_block *target_header = NULL;
    unsigned int i;

    for (i = header_index; i && !target_header; --i)
    {
        struct sm6_block *candidate_header = node->headers.blocks[i - 1];

        if (candidate_header->merge_type == MERGE_TYPE_LOOP)
        {
            /* We might have two loops, each at equal scopes.
             * In order to break out to an outer loop, we must verify that the loops actually nest.
             * We must not introduce any backwards branches here. */
            struct sm6_block *candidate_merge = NULL;
            if (candidate_header->loop_ladder_block)
                candidate_merge = candidate_header->loop_ladder_block;
            else if (candidate_header->loop_merge_block)
                candidate_merge = candidate_header->loop_merge_block;

            if (!candidate_merge)
                continue;

            /* Check for backwards branch. */
            if (cfg_structuriser_query_reachability(structuriser, candidate_merge, inner_header))
                continue;

            /* An outer header is expected to dominate the inner header. Otherwise, they live in
             * separate scopes, and we should look for a header that is further out. */
            if (!sm6_block_dominates(candidate_header, inner_header))
                continue;

            target_header = candidate_header;
        }
    }

    return target_header;
}

static struct sm6_block *cfg_structuriser_get_or_create_ladder_block(struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *node, unsigned int header_index)
{
    struct sm6_block *header = node->headers.blocks[header_index];
    struct sm6_block *loop_ladder = header->loop_ladder_block;

    if (!loop_ladder)
    {
        /* We don't have a ladder, because the loop merged to an outer scope, so we need to fake a ladder.
         * If we hit this case, we did not hit the simpler case in find_loops(). */
        header->loop_ladder_block = cfg_structuriser_create_ladder_block(structuriser, header, node, ".merge");

        /* If this is the second outermost scope, we don't need to deal with ladders.
         * ladder is a dummy branch straight out to the outer merge point. */
        if (header_index > 1)
            loop_ladder = header->loop_ladder_block;
    }

    return loop_ladder;
}

static struct sm6_block *cfg_structuriser_build_enclosing_break_target_for_loop_ladder(
        struct vkd3d_cfg_structuriser *structuriser, struct sm6_block **node_ref, struct sm6_block *loop_ladder)
{
    struct sm6_block *node = *node_ref;

    /* A loop ladder needs to break out somewhere. If we don't have a candidate
     * place to break out to, we will need to create one for the outer scope.
     * This is the purpose of the full_break_target fallback. */

    bool ladder_to_merge_is_trivial = loop_ladder->succ.count == 1 && loop_ladder->succ.blocks[0] == node;
    struct sm6_block *full_break_target = NULL;

    /* We have to break somewhere, turn the outer selection construct into a loop. */
    if (!ladder_to_merge_is_trivial)
    {
        /* Selection merge to this dummy instead. */
        struct sm6_block *new_selection_merge = cfg_structuriser_create_helper_pred_block(structuriser, node);

        struct sm6_block *loop = cfg_structuriser_create_helper_pred_block(structuriser,
                new_selection_merge->headers.blocks[0]);

        /* This is now our fallback loop break target. */
        full_break_target = node;

        /* Reassign header node. */
        assert(new_selection_merge->headers.blocks[0]->merge_type == MERGE_TYPE_SELECTION);
        new_selection_merge->headers.blocks[0]->selection_merge_block = new_selection_merge;
        new_selection_merge->headers.blocks[0] = loop;

        loop->merge_type = MERGE_TYPE_LOOP;
        loop->loop_merge_block = node;
        loop->freeze_structured_analysis = true;

        /* After the loop ladder, make sure we always branch to the break target. */
        cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, loop_ladder,
                new_selection_merge, node, NULL);

        *node_ref = new_selection_merge;
    }

    return full_break_target;
}

static struct sm6_block *cfg_structuriser_build_ladder_block_for_escaping_edge_handling(
        struct vkd3d_cfg_structuriser *structuriser, struct sm6_block *node, struct sm6_block *header,
        struct sm6_block *loop_ladder, struct sm6_block *target_header,
        struct sm6_block *full_break_target, const struct sm6_block_array *normal_preds)
{
    struct sm6_block *new_ladder_block = NULL;
    unsigned int i;

    if (target_header || full_break_target)
    {
        struct sm6_phi *phi;

        /* If we have a ladder block, there exists a merge candidate which the loop header dominates.
         * We create a ladder block before the merge block, which becomes the true merge block.
         * In this ladder block, we can detect with Phi nodes whether the break was "clean",
         * or if we had an escape edge.
         * If we have an escape edge, we can break to outer level, and continue the ladder that way.
         * Otherwise we branch to the existing merge block and continue as normal.
         * We'll also need to rewrite a lot of Phi nodes this way as well. */
        struct sm6_block *ladder = cfg_structuriser_create_helper_pred_block(structuriser, loop_ladder);
        new_ladder_block = ladder;

        /* Merge to ladder instead. */
        cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, header, node, ladder, NULL);

        ladder->terminator.type = TERMINATOR_COND_BR;
        shader_register_make_scalar_ssa(&ladder->terminator.conditional_reg, VKD3D_DATA_BOOL, structuriser->parser);
        ladder->terminator.false_block = loop_ladder;

        phi = sm6_block_phi_new(ladder, &ladder->terminator.conditional_reg);
        sm6_phi_build_name(phi, "ladder_phi_", loop_ladder->name);

        for (i = 0; i < ladder->pred.count; ++i)
        {
            struct sm6_block *pred = ladder->pred.blocks[i];
            struct incoming_value *incoming = sm6_phi_incoming_new(phi);
            bool is_breaking_pred = !sm6_block_array_contains(normal_preds, pred);
            incoming->block = pred;
            register_make_constant_bool(&incoming->reg, is_breaking_pred);
        }

        /* Ladder breaks out to outer scope. */
        if (target_header && target_header->loop_ladder_block)
        {
            ladder->terminator.true_block = target_header->loop_ladder_block;
            sm6_block_add_branch(ladder, target_header->loop_ladder_block);
        }
        else if (target_header && target_header->loop_merge_block)
        {
            ladder->terminator.true_block = target_header->loop_merge_block;
            sm6_block_add_branch(ladder, target_header->loop_merge_block);
        }
        else if (full_break_target)
        {
            ladder->terminator.true_block = full_break_target;
            sm6_block_add_branch(ladder, full_break_target);
        }
        else
            WARN("No loop merge block?\n");

        /* This can happen in some scenarios, fixup the branch to be a direct one instead. */
        if (ladder->terminator.true_block == ladder->terminator.false_block)
        {
            ladder->terminator.direct_block = ladder->terminator.true_block;
            ladder->terminator.type = TERMINATOR_UNCOND_BR;
        }
    }
    else
    {
        /* Here, loop_ladder -> final merge is a trivial, direct branch. */

        if (!loop_ladder->instruction_count)
        {
            /* Simplest common case.
             * If the loop ladder just branches to outer scope, and this block does not perform
             * any operations we can avoid messing around with ladder PHI variables and just execute the branch.
             * This block will likely become a frontier node when merging PHI instead.
             * This is a common case when breaking out of a simple for loop. */
            cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, header, node, loop_ladder, NULL);
        }
        else
        {
            struct sm6_phi *phi;

            /* We have a case where we're trivially breaking out of a selection construct,
             * but the loop ladder block contains operations which we must not execute,
             * since we were supposed to branch directly out to node.
             * We cannot directly break out of a selection construct, so our ladder must be a bit more sophisticated.
             * ladder-pre -> merge -> ladder-post -> selection merge
             *      \-------------------/ */
            struct sm6_block *ladder_pre = cfg_structuriser_create_helper_pred_block(structuriser, loop_ladder);
            struct sm6_block *ladder_post = cfg_structuriser_create_helper_succ_block(structuriser, loop_ladder);
            sm6_block_add_branch(ladder_pre, ladder_post);

            ladder_pre->terminator.type = TERMINATOR_COND_BR;
            shader_register_make_scalar_ssa(&ladder_pre->terminator.conditional_reg, VKD3D_DATA_BOOL,
                    structuriser->parser);
            ladder_pre->terminator.true_block = ladder_post;
            ladder_pre->terminator.false_block = loop_ladder;

            /* Merge to ladder instead. */
            cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, header, node, ladder_pre, NULL);
            new_ladder_block = ladder_pre;

            phi = sm6_block_phi_new(ladder_pre, &ladder_pre->terminator.conditional_reg);
            sm6_phi_build_name(phi, "ladder_phi_", loop_ladder->name);
            for (i = 0; i < ladder_pre->pred.count; ++i)
            {
                struct sm6_block *pred = ladder_pre->pred.blocks[i];
                struct incoming_value *incoming = sm6_phi_incoming_new(phi);
                bool is_breaking_pred = !sm6_block_array_contains(normal_preds, pred);
                incoming->block = pred;
                register_make_constant_bool(&incoming->reg, is_breaking_pred);
            }
        }
    }

    return new_ladder_block;
}

static int split_merge_block_compare(const void *a, const void *b)
{
    const struct sm6_block *block_a = *(const struct sm6_block * const *)a;
    const struct sm6_block *block_b = *(const struct sm6_block * const *)b;

    if (sm6_block_dominates(block_a, block_b))
        return -1;
    if (sm6_block_dominates(block_b, block_a))
        return 1;
    return vkd3d_u32_compare(block_b->forward_post_visit_order, block_a->forward_post_visit_order);
}

static void cfg_structuriser_split_merge_blocks(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block_array *normal_preds;
    unsigned int i, j, max_headers;
    struct sm6_block *node;

    for (i = 0, max_headers = 0; i < structuriser->forward_post_visit_order.count; ++i)
        max_headers = max(max_headers, structuriser->forward_post_visit_order.blocks[i]->headers.count);
    if (!(normal_preds = vkd3d_calloc(max_headers, sizeof(*normal_preds))))
    {
        ERR("Failed to allocate predecessor buffer.\n");
        abort();
    }

    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        struct sm6_block *full_break_target = NULL;

        node = structuriser->forward_post_visit_order.blocks[i];
        if (node->headers.count <= 1)
            continue;

        assert(!sm6_block_is_plain_continue(node));

        /* If this block was the merge target for more than one construct, we will need to split the block. In
         * SPIR-V, a merge block can only be the merge target for one construct. However, we can set up a chain
         * of merges where inner scope breaks to outer scope with a dummy basic block.
         * The outer scope comes before the inner scope merge. */

        /* We cannot fully trust a sort on post-visit order, since if we have two split blocks here,
         * they will have the same post-visit order until we recompute them.
         * FIXME: Should probably be smarter about this ... */
        qsort(node->headers.blocks, node->headers.count, sizeof(*node->headers.blocks), split_merge_block_compare);

        /* Before we start splitting and rewriting branches, we need to know which preds are considered "normal",
         * and which branches are considered ladder breaking branches (rewritten branches).
         * This will influence if a pred block gets false or true when emitting ladder breaking blocks later. */
        assert(node->headers.count <= max_headers);
        for (j = 0; j < node->headers.count; ++j)
            if (node->headers.blocks[j]->loop_ladder_block)
                sm6_block_array_clone(&normal_preds[j], &node->headers.blocks[j]->loop_ladder_block->pred);

        /* Start from innermost scope, and rewrite all escape branches to a merge block which is dominated by the loop header in question.
         * The merge block for the loop must have a ladder block before the old merge block.
         * This ladder block will break to outer scope, or keep executing the old merge block. */
        for (j = node->headers.count - 1; j; --j)
        {
            struct sm6_block *current_node = node->headers.blocks[j];

            /* Find innermost loop header scope we can break to when resolving ladders. */
            struct sm6_block *target_header = cfg_structuriser_get_target_break_block_for_inner_header(
                    structuriser, node, j);

            if (current_node->merge_type == MERGE_TYPE_LOOP)
            {
                struct sm6_block *loop_ladder = cfg_structuriser_get_or_create_ladder_block(structuriser, node, j);
                struct sm6_block *new_ladder_block = NULL;

                /* The loop ladder needs to break to somewhere.
                 * Either this is an outer loop scope, or we need to create a fake loop we can break out of if
                 * the break is non-trivial. */
                if (loop_ladder && !target_header && !full_break_target)
                    full_break_target = cfg_structuriser_build_enclosing_break_target_for_loop_ladder(structuriser,
                            &node, loop_ladder);

                if (loop_ladder)
                {
                    new_ladder_block = cfg_structuriser_build_ladder_block_for_escaping_edge_handling(structuriser,
                            node, current_node, loop_ladder, target_header, full_break_target, &normal_preds[j]);
                }

                /* We won't analyze this again, so make sure header knows about the new merge block. */
                if (current_node->freeze_structured_analysis)
                {
                    if (new_ladder_block)
                        current_node->loop_ladder_block = new_ladder_block;
                    current_node->loop_merge_block = current_node->loop_ladder_block;
                    current_node->loop_ladder_block = NULL;
                }
            }
            else if (current_node->merge_type == MERGE_TYPE_SELECTION)
            {
                if (target_header)
                {
                    /* Breaks out to outer available scope. */
                    struct sm6_block *rewrite_to = NULL;
                    if (target_header->loop_ladder_block)
                        rewrite_to = target_header->loop_ladder_block;
                    else if (target_header->loop_merge_block)
                        rewrite_to = target_header->loop_merge_block;

                    if (rewrite_to)
                        cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser,
                                current_node, node, rewrite_to, NULL);
                    else
                        WARN("No loop merge block?\n");
                }
                else if (full_break_target)
                {
                    cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, current_node,
                            node, full_break_target, NULL);
                }
                else
                {
                    /* The outer scope *must* now become a loop, no matter what.
                     * We cannot rely on a traversal to rewrite breaking constructs in the entire loop,
                     * so "everything" must essentially become a break instead. */
                    full_break_target = node;
                    assert(node->headers.blocks[0]->merge_type == MERGE_TYPE_SELECTION);
                    node->headers.blocks[0]->merge_type = MERGE_TYPE_LOOP;
                    node->headers.blocks[0]->freeze_structured_analysis = true;

                    assert(node->headers.blocks[0]->selection_merge_block == node);
                    node->headers.blocks[0]->loop_merge_block = node->headers.blocks[0]->selection_merge_block;
                    node->headers.blocks[0]->selection_merge_block = NULL;
                }
            }
            else
            {
                FIXME("Invalid merge type.\n");
            }
        }
    }

    for (i = 0; i < max_headers; ++i)
        sm6_block_array_destroy(&normal_preds[i]);
    vkd3d_free(normal_preds);
}

static void cfg_structuriser_structurise(struct vkd3d_cfg_structuriser *structuriser, unsigned int pass)
{
    if (cfg_structuriser_find_switch_blocks(structuriser, pass))
    {
        cfg_structuriser_recompute_cfg(structuriser);
        if (cfg_structuriser_find_switch_blocks(structuriser, pass))
        {
            FIXME("Fatal, detected infinite loop.\n");
            abort();
        }
    }

    cfg_structuriser_find_loops(structuriser);
    cfg_structuriser_find_selection_merges(structuriser, pass);
    cfg_structuriser_fixup_broken_selection_merges(structuriser, pass);
    if (pass == 0)
        cfg_structuriser_split_merge_blocks(structuriser);
}

static void cfg_structuriser_update_structured_loop_merge_targets(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int true_value = ++structuriser->loop_merge_target_true_value;
    unsigned int i;

    /* First, we need to do this before recomputing the CFG, since we lose
     * normal loop merge targets when recomputing. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[i];
        if (node->loop_merge_block)
            node->loop_merge_block->is_loop_merge_target = true_value;
        if (node->loop_ladder_block)
            node->loop_ladder_block->is_loop_merge_target = true_value;
    }

    cfg_structuriser_recompute_cfg(structuriser);

    /* Make sure we include merge blocks which are frozen merge targets in ladder blocks, which
     * were not included in the post visit order yet. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        struct sm6_block *node = structuriser->forward_post_visit_order.blocks[i];
        if (node->loop_merge_block)
            node->loop_merge_block->is_loop_merge_target = true_value;
        if (node->loop_ladder_block)
            node->loop_ladder_block->is_loop_merge_target = true_value;
    }
}

struct invalid_loop_break_search_info
{
    struct sm6_block *merge;
    struct sm6_block *invalid_target;
};

static bool find_invalid_loop_break(const struct vkd3d_cfg_structuriser *structuriser,
        struct sm6_block *candidate, unsigned int true_value, void *opaque)
{
    struct invalid_loop_break_search_info *info = opaque;
    unsigned int i;

    if (candidate == info->merge || info->invalid_target)
        return false;

    /* If the succ can reach outside the loop construct, we have an error condition. */
    for (i = 0; i < candidate->succ.count; ++i)
    {
        struct sm6_block *succ = candidate->succ.blocks[i];
        if (!cfg_structuriser_query_reachability(structuriser, succ, info->merge))
        {
            /* Determine if we're an inner terminate/return, or a loop exit.
             * If the common post-dominator is EXIT node, this is a return-like relationship,
             * and we skip any fixup. */
            struct sm6_block *pdom = sm6_block_find_common_post_dominator(succ, info->merge);
            if (pdom && pdom->pred.count)
                info->invalid_target = succ;
        }
    }

    return true;
}

static bool cfg_structuriser_rewrite_invalid_loop_breaks(struct vkd3d_cfg_structuriser *structuriser)
{
    /* Keep iterating here until we have validated a clean CFG w.r.t. block-like loops.
     * This should pass through first time without issue with extremely high probability,
     * so hitting the slow path isn't a real concern until proven otherwise. */
    struct invalid_loop_break_search_info break_info;
    struct sm6_block *rewrite_header = NULL;
    unsigned int i;

    break_info.invalid_target = NULL;

    /* Process from inside out. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        struct sm6_block *merge, *node = structuriser->forward_post_visit_order.blocks[i];

        /* Structured loop constructs can end up with problematic merge scenarios where we missed
         * some cases where blocks branch outside our construct.
         * At some point, we were considered mere selection constructs and breaking out of it is fine,
         * but if the selection is promoted to a loop at some point after this analysis, we are a bit screwed.
         * This can happen in complex ladder resolve scenarios.
         * The fix-up means introducing multiple levels of ladder blocks. */
        if (node->merge_type != MERGE_TYPE_LOOP || !node->freeze_structured_analysis)
            continue;

        if (!(merge = node->loop_merge_block) || sm6_block_post_dominates(merge, node))
            continue;

        break_info.merge = merge;
        cfg_structuriser_traverse_dominated_blocks(structuriser, node, node, 0, find_invalid_loop_break,
                &break_info);

        if (break_info.invalid_target)
        {
            rewrite_header = node;
            break;
        }
    }

    if (break_info.invalid_target)
    {
        struct sm6_block *merge = rewrite_header->loop_merge_block;
        struct sm6_block *dispatcher = cfg_structuriser_create_helper_pred_block(structuriser, merge);
        unsigned int i, natural_preds;
        struct sm6_phi *phi;

        rewrite_header->loop_merge_block = dispatcher;

        natural_preds = dispatcher->pred.count;
        cfg_structuriser_traverse_dominated_blocks_and_rewrite_branch(structuriser, rewrite_header,
                break_info.invalid_target, dispatcher, NULL);

        shader_register_make_scalar_ssa(&dispatcher->terminator.conditional_reg, VKD3D_DATA_BOOL, structuriser->parser);
        phi = sm6_block_phi_new(dispatcher, &dispatcher->terminator.conditional_reg);
        sm6_phi_build_name(phi, "break_selector_", merge->name);

        for (i = 0; i < natural_preds; ++i)
        {
            struct incoming_value *incoming = sm6_phi_incoming_new(phi);
            incoming->block = dispatcher->pred.blocks[i];
            register_make_constant_bool(&incoming->reg, true);
        }

        for (; i < dispatcher->pred.count; ++i)
        {
            struct incoming_value *incoming = sm6_phi_incoming_new(phi);
            incoming->block = dispatcher->pred.blocks[i];
            register_make_constant_bool(&incoming->reg, false);
        }

        dispatcher->terminator.type = TERMINATOR_COND_BR;
        dispatcher->terminator.true_block = merge;
        dispatcher->terminator.false_block = break_info.invalid_target;
        dispatcher->terminator.direct_block = NULL;

        sm6_block_add_branch(dispatcher, break_info.invalid_target);

        cfg_structuriser_recompute_cfg(structuriser);
        return true;
    }

    return false;
}

struct phi_node
{
    struct sm6_block *block;
    unsigned int phi_index;
};

static void cfg_phi_node_fixup(struct phi_node *node,
        struct sm6_block **value_id_to_block, unsigned int map_size)
{
    struct sm6_phi *phi = &node->block->phi[node->phi_index];
    unsigned int i;

    /* We want to move any incoming block to where the ID was created.
     * This avoids some problematic cases of crossing edges when using ladders. */

    for (i = 0; i < phi->incoming_count; ++i)
    {
        struct incoming_value *incoming = &phi->incoming[i];
        struct sm6_block *block;
        unsigned int id;

        if (!sm6_phi_incoming_is_ssa(incoming))
            continue;
        id = sm6_phi_incoming_ssa_id(incoming);
        /* Out of bounds means no value. May never occur in practice but handle it instead of proving that. */
        if (id >= map_size || !(block = value_id_to_block[id]))
        {
            /* This is a global. */
            continue;
        }

        if (!sm6_block_dominates(block, incoming->block))
        {
#ifdef PHI_DEBUG
            TRACE("For node %s, move incoming node %s to %s.\n", node->block->name, incoming->block->name,
                 block->name);
#endif
            incoming->block = block;
        }
    }
}

static bool sm6_phi_can_complete_phi_insertion(const struct sm6_phi *phi, const struct sm6_block *block)
{
    unsigned int i;

    /* If all incoming values have at least one pred block they dominate, we can merge the final PHI. */
    for (i = 0; i < phi->incoming_count; ++i)
    {
        struct incoming_value *incoming = &phi->incoming[i];
        if (!sm6_block_array_contains_dominated_block(&block->pred, incoming->block)
                && (!block->pred_back_edge || !sm6_block_dominates(incoming->block, block->pred_back_edge)))
        {
            return false;
        }
    }

    return true;
}

static bool cfg_structuriser_query_reachability_through_back_edges(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *from, const struct sm6_block *to)
{
    if (sm6_block_dominates(to, from))
    {
        /* If we're dominated by end node, only way we can reach is through a back edge. */
        return to->pred_back_edge && cfg_structuriser_query_reachability(structuriser, from, to->pred_back_edge);
    }
    else
    {
        return cfg_structuriser_query_reachability(structuriser, from, to);
    }
}

static bool cfg_structuriser_query_reachability_split_loop_header(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_block *from, const struct sm6_block *to, const struct sm6_block *end_node)
{
    bool from_reaches_header, to_reaches_header;

    /* A special query where from and to must lie on the same side of a loop header to be considered reachable. */
    if (!end_node->pred_back_edge)
        return cfg_structuriser_query_reachability(structuriser, from, to);

    from_reaches_header = cfg_structuriser_query_reachability(structuriser, from, end_node);
    to_reaches_header = cfg_structuriser_query_reachability(structuriser, to, end_node);
    if (from_reaches_header != to_reaches_header)
        return false;

    return cfg_structuriser_query_reachability(structuriser, from, to);
}

static bool cfg_structuriser_phi_frontier_makes_forward_progress(const struct vkd3d_cfg_structuriser *structuriser,
        const struct sm6_phi *phi, struct sm6_block *frontier, const struct sm6_block *end_node)
{
    unsigned int i;

    /* Not all PHI frontiers are nodes we need to care about.
     * There are two conditions we must meet to disregard a placement.
     * - We do not remove any inputs as a result.
     * - The frontier can reach another incoming value.
     * In this situation, a frontier is completely meaningless. */

    for (i = 0; i < phi->incoming_count; ++i)
    {
        struct sm6_block *incoming_block = phi->incoming[i].block;
        /* We will remove an input, this is forward progress.
         * Avoid checking the edge case where frontier candidate == incoming block.
         * Removing an input only to place a new frontier there is nonsensical. */
        if (frontier != incoming_block && !cfg_structuriser_exists_path_in_cfg_without_intermediate_node(structuriser,
                incoming_block, end_node, frontier))
            return true;
    }

    /* Nothing is removed as a result, so check if the frontier can reach another incoming value.
     * If end_node is a loop header, makes sure we only consider a node visible if they are both on the
     * correct side of the loop header. */
    for (i = 0; i < phi->incoming_count; ++i)
        if (cfg_structuriser_query_reachability_split_loop_header(structuriser,
                frontier, phi->incoming[i].block, end_node))
            return false;

    /* Assume we make forward progress. Either way, we will never look at a frontier twice,
     * so this should be safe. The only real risk is that we add some redundant PHI nodes. */
    return true;
}

static void sm6_phi_copy_incoming_for_block(struct sm6_phi *dst, const struct sm6_phi *src,
        struct sm6_block *block)
{
    struct incoming_value *value = sm6_phi_incoming_new(dst);
    const struct incoming_value *incoming;

    if ((incoming = sm6_phi_find_incoming_value(src, block)))
        value->reg = incoming->reg;
    else
        register_make_undef(&value->reg, src->reg.data_type);

    value->block = block;
}

static void cfg_structuriser_insert_phi(struct vkd3d_cfg_structuriser *structuriser, struct phi_node *node)
{
    unsigned int i, j, placed_frontier_true_value = cfg_structuriser_next_true_value(structuriser);

    /* We start off with N values defined in N blocks.
     * These N blocks *used* to branch to the PHI node, but due to our structurizer,
     * there might not be branch targets here anymore, primary example here is ladders.
     * In order to fix this we need to follow control flow from these values and insert phi nodes as
     * necessary to link up a set of values where dominance frontiers are shared. */

    struct sm6_phi * const phi = &node->block->phi[node->phi_index];
    struct incoming_value *incoming_values = phi->incoming;

#ifdef PHI_DEBUG
    TRACE("\n=== INSERT PHI FOR %s ===\n", node->block->name);
#endif

    for (;;)
    {
        struct sm6_block *frontier, *candidate_frontier;
        struct incoming_value *dominated_incoming = NULL;
        unsigned int num_alive_incoming_values;
        struct sm6_phi *frontier_phi;

#ifdef PHI_DEBUG
        TRACE("\n=== PHI iteration ===\n");

        for (i = 0; i < phi->incoming_count; ++i)
            TRACE("  Incoming value from %s\n", incoming_values[i].block->name);
#endif

        /* Inside the CFG subset, find dominance frontiers where we merge PHIs this iteration. */
        frontier = node->block;
        if (!sm6_phi_can_complete_phi_insertion(phi, node->block))
        {
            frontier = NULL;

            /* We need some intermediate merge, so find a frontier node to work on. */
            for (i = 0; i < phi->incoming_count; ++i)
            {
                struct incoming_value *incoming = &incoming_values[i];
                for (j = 0; j < incoming->block->dominance_frontier.count; ++j)
                {
                    candidate_frontier = incoming->block->dominance_frontier.blocks[j];
                    if (candidate_frontier->placed_frontier == placed_frontier_true_value)
                        continue;

                    if (!cfg_structuriser_phi_frontier_makes_forward_progress(structuriser,
                            phi, candidate_frontier, node->block))
                    {
                        /* Makes sure we don't redundantly test this again. */
                        candidate_frontier->placed_frontier = placed_frontier_true_value;
                        continue;
                    }

                    /* Only consider a frontier if we can reach node->block or its back edge from it. */
                    if (cfg_structuriser_query_reachability_through_back_edges(structuriser, candidate_frontier, node->block))
                    {
                        if (!frontier || candidate_frontier->forward_post_visit_order > frontier->forward_post_visit_order)
                        {
                            /* Pick the earliest frontier in the CFG.
                             * We want to merge top to bottom. */
                            frontier = candidate_frontier;
                        }
                    }
                }
            }

            if (frontier)
                frontier->placed_frontier = placed_frontier_true_value;
        }

        if (!frontier)
            FIXME("Frontier not found.\n");
        assert(frontier);

        if (frontier == node->block)
        {
            if (frontier->pred.count == 1 && !frontier->pred_back_edge)
            {
                /* The PHI node has already been merged.
                 * This can happen if a ladder pred block merged all inputs, and we would
                 * end up with a single-pred PHI, which makes no sense (even if it should work).
                 * Just copy the ID for the frontier node which made the final merge. */
                const struct incoming_value *incoming = sm6_phi_find_incoming_value(phi, frontier->pred.blocks[0]);
                assert(incoming);

                sm6_block_emit_copy_ssa_reg(frontier->pred.blocks[0], &phi->reg, &incoming->reg,
                        structuriser->parser);

                /* Ignore this one when emitting PHIs later. */
                phi->incoming_count = 0;
            }
            else
            {
                struct sm6_phi final_incoming = {0};

                /* Final merge. */
                for (i = 0; i < frontier->pred.count; ++i)
                    sm6_phi_copy_incoming_for_block(&final_incoming, phi, frontier->pred.blocks[i]);

                if (frontier->pred_back_edge)
                    sm6_phi_copy_incoming_for_block(&final_incoming, phi, frontier->pred_back_edge);

                vkd3d_free(phi->incoming);
                phi->incoming = final_incoming.incoming;
                phi->incoming_capacity = final_incoming.incoming_capacity;
                phi->incoming_count = final_incoming.incoming_count;
            }
            return;
        }

        /* A candidate dominance frontier is a place where we might want to place a PHI node in order to merge values.
         * For a successful iteration, we need to find at least one candidate where we can merge PHI. */

#ifdef PHI_DEBUG
        TRACE("Testing dominance frontier %s ...\n", frontier->name);
#endif

        /* Remove old inputs. */
        frontier_phi = sm6_block_phi_new(frontier, NULL);
        shader_register_make_scalar_ssa(&frontier_phi->reg, phi->reg.data_type, structuriser->parser);
        sm6_phi_build_name(frontier_phi, "frontier_phi_", frontier->name);

        assert(!frontier->pred_back_edge);
        for (i = 0; i < frontier->pred.count; ++i)
        {
            struct incoming_value *value = sm6_phi_incoming_new(frontier_phi);
            struct sm6_block *input = frontier->pred.blocks[i];
            const struct incoming_value *incoming;

            if ((incoming = sm6_phi_find_incoming_value(phi, input)))
            {
#ifdef PHI_DEBUG
                struct sm6_block *incoming_block = incoming->block;
                TRACE("   ... found incoming block %s for input %s.\n", incoming_block->name, input->name);
                TRACE(" ... For pred %s (%p), found incoming value from %s (%p)\n", input->name,
                        input, incoming_block->name, incoming_block);
#endif

                value->reg = incoming->reg;
                value->block = input;
            }
            else
            {
#ifdef PHI_DEBUG
                TRACE("   ... creating undefined input for %s\n", input->name);
#endif
                /* If there is no incoming value, we need to hallucinate an undefined value. */
                register_make_undef(&value->reg, phi->reg.data_type);
                value->block = input;
            }
        }

        /* Do we remove the incoming value now or not?
         * If all paths from incoming value must go through frontier, we can remove it,
         * otherwise, we might still need to use the incoming value somewhere else. */
        num_alive_incoming_values = phi->incoming_count;
        for (i = 0; i < num_alive_incoming_values; )
        {
            struct sm6_block *incoming_block = incoming_values[i].block;
            if (!cfg_structuriser_exists_path_in_cfg_without_intermediate_node(structuriser,
                    incoming_block, node->block, frontier))
            {
#ifdef PHI_DEBUG
                TRACE("     ... removing input in %s\n", incoming_block->name);
#endif
                if (i != num_alive_incoming_values - 1)
                    vkd3d_swap(incoming_values[num_alive_incoming_values - 1], incoming_values[i], struct incoming_value);
                --num_alive_incoming_values;
            }
            else
            {
#ifdef PHI_DEBUG
                TRACE("     ... keeping input in %s\n", incoming_block->name);
#endif
                ++i;
            }
        }

        /* Need to clean up exhausted incoming values after the loop,
         * since an incoming value can be used multiple times before a frontier PHI is resolved. */
        phi->incoming_count = num_alive_incoming_values;

        for (i = 0; i < phi->incoming_count; ++i)
        {
            struct incoming_value *incoming = &incoming_values[i];
            if (sm6_block_dominates(frontier, incoming->block)
                    && !cfg_structuriser_exists_path_in_cfg_without_intermediate_node(structuriser,
                    frontier, node->block, incoming->block))
            {
                /* There should be only one block the frontier can dominate.
                 * The candidate block must also post-dominate the frontier on the CFG subset which
                 * terminates at node->block, otherwise we will get a proper merge later anyways. */
                assert(!dominated_incoming);
                dominated_incoming = incoming;
            }
        }

        if (dominated_incoming)
        {
            /* If our frontier dominates another incoming block, we need to merge two incoming values
             * using an auxillary phi node as well as an OpSelect to resolve two conflicting values into one. */

            /* For every pred edge of the frontier where pred did not dominate, we are now suddenly dominating.
             * If we came from such a block, we should replace the incoming value of dominating_incoming
             * rather than adding a new incoming value. */
            struct sm6_phi merge_phi = {0};

            /* Here we need to figure out if we have a cross branch which functions as a ladder.
             * If we have such a special edge, the PHI value we find here will override any other value on this path.
             * However, if we only have expected branches, there is nothing to override, and any PHI values
             * we created along this path turned out to be irrelevant after all. */

            unsigned normal_branch_count = 0;
            for (i = 0; i < frontier->pred.count; ++i)
            {
                struct incoming_value *value = sm6_phi_incoming_new(&merge_phi);
                struct sm6_block *input = frontier->pred.blocks[i];

                if (sm6_phi_find_incoming_value(phi, input))
                {
                    /* If the input does not dominate the frontier, this might be a case of cross-edge PHI merge.
                     * However, if we still have an incoming value which dominates the input block, ignore.
                     * This is considered a normal path and we will merge the actual result in a later iteration, because
                     * the frontier is not a post-dominator of the input value. */
                    bool input_is_normal_edge = true;
                    if (!sm6_block_dominates(input, frontier))
                    {
                        input_is_normal_edge = false;
                        for (j = 0; j < phi->incoming_count; ++j)
                        {
                            struct incoming_value *incoming = &incoming_values[j];
                            if (sm6_block_dominates(incoming->block, input))
                            {
                                input_is_normal_edge = true;
                                break;
                            }
                        }
                    }

                    normal_branch_count += input_is_normal_edge;

                    register_make_constant_bool(&value->reg, input_is_normal_edge);
                }
                else
                {
                    /* The input is undefined, so we don't really care. Just treat this as a normal edge. */
                    ++normal_branch_count;
                    register_make_constant_bool(&value->reg, true);
                }

                value->block = input;
            }

            if (normal_branch_count != frontier->pred.count)
            {
                struct sm6_phi *new_phi = sm6_block_phi_new(frontier, NULL);
                shader_register_make_scalar_ssa(&merge_phi.reg, VKD3D_DATA_BOOL, structuriser->parser);
                *new_phi = merge_phi;

                assert(sm6_phi_incoming_is_ssa(dominated_incoming));
                /* On the bool in merge_phi.reg, select between dominated_incoming and frontier_phi, and make
                 * the result the source reg of dominated_incoming. */
                sm6_block_emit_select_replace_phi(dominated_incoming->block, &merge_phi.reg,
                        &dominated_incoming->reg, &frontier_phi->reg, structuriser->parser);

                sm6_phi_build_name(&merge_phi, "merged_phi_", dominated_incoming->block->name);
            }
            else
            {
                sm6_phi_destroy(&merge_phi);
            }
        }
        else
        {
            /* Replace with merged value. */
            struct incoming_value *new_incoming = sm6_phi_incoming_new(phi);
            new_incoming->reg = frontier_phi->reg;
            new_incoming->block = frontier;
        }

#ifdef PHI_DEBUG
        TRACE("=========================\n");
#endif
    }
}

struct id_rewrite
{
    unsigned int id;
    const struct sm6_block_array *consumers;
};

static int rewrite_compare(const void *a, const void *b)
{
    const struct id_rewrite *ra = a, *rb = b;
    return vkd3d_u32_compare(ra->id, rb->id);
}

struct id_origin
{
    struct sm6_block *node;
    enum vkd3d_data_type data_type;
};

static unsigned int mark_node_value_access(const struct id_origin *origin, struct sm6_block *node,
        const struct vkd3d_shader_register *reg, struct sm6_block_array *id_to_non_local_consumers)
{
    unsigned int id, count = 0;

    if (!register_is_ssa(reg))
        return 0;
    id = vkd3d_shader_register_ssa_id(reg);

    if (!origin[id].node)
        return 0;

    if (!sm6_block_dominates(origin[id].node, node))
    {
        count = !id_to_non_local_consumers[id].count;
        /* We have a problem. Mark that we need to rewrite a certain variable. */
        sm6_block_array_append(&id_to_non_local_consumers[id], node);
    }
    return count;
}

static void sm6_block_rewrite_consumed_ids(struct sm6_block *node, unsigned int src_ssa_id,
        const struct vkd3d_shader_register *rewrite_reg)
{
    unsigned int i, j;

    for (i = 0; i < node->instruction_count; ++i)
    {
        struct vkd3d_shader_instruction *ins = &node->instructions[i];

        if (!ins->src)
            continue;

        for (j = 0; j < ins->src_count; ++j)
        {
            struct vkd3d_shader_register *reg = (struct vkd3d_shader_register *)&ins->src[j].reg;
            if (register_is_ssa(reg) && vkd3d_shader_register_ssa_id(reg) == src_ssa_id)
                *reg = *rewrite_reg;
        }
    }
}

static void cfg_structuriser_fixup_broken_value_dominance(struct vkd3d_cfg_structuriser *structuriser)
{
    unsigned int i, j, k, count, rewrite_count, id, table_count;
    struct sm6_block_array *id_to_non_local_consumers;
    struct sm6_block *node;
    struct id_rewrite *rewrites;
    struct id_origin *origin;

    table_count = sm6_get_ssa_register_count(structuriser->parser);
    if (!(origin = vkd3d_calloc(table_count, sizeof(*origin)))
            || !(id_to_non_local_consumers = vkd3d_calloc(table_count, sizeof(*id_to_non_local_consumers))))
    {
        ERR("Failed to allocate id lookup tables.\n");
        abort();
    }
    /* First, scan through all blocks and figure out which block creates an ID. */
    for (i = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        for (j = 0; j < node->instruction_count; ++j)
        {
            for (k = 0; k < node->instructions[j].dst_count; ++k)
            {
                const struct vkd3d_shader_register *reg = &node->instructions[j].dst[k].reg;
                if (!register_is_ssa(reg))
                    continue;
                id = vkd3d_shader_register_ssa_id(reg);
                origin[id].node = node;
                origin[id].data_type = reg->data_type;
            }
        }
        for (j = 0; j < node->phi_count; ++j)
        {
            const struct vkd3d_shader_register *reg = &node->phi[j].reg;
            uint32_t id = vkd3d_shader_register_ssa_id(reg);
            origin[id].node = node;
            origin[id].data_type = reg->data_type;
        }
    }

    /* Now, scan through all blocks and figure out which values are consumed in different blocks. */
    for (i = 0, count = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        for (j = 0; j < node->instruction_count; ++j)
        {
            struct vkd3d_shader_instruction *ins = &node->instructions[j];

            if (!ins->src)
                continue;

            for (k = 0; k < ins->src_count; ++k)
            {
                const struct vkd3d_shader_register *reg = &ins->src[k].reg;
                count += mark_node_value_access(origin, node, reg, id_to_non_local_consumers);
            }
        }

        if (node->terminator.type == TERMINATOR_COND_BR)
            count += mark_node_value_access(origin, node, &node->terminator.conditional_reg, id_to_non_local_consumers);
#if 0
        if (node->terminator.return_value)
            count += mark_node_value_access(origin, node, &node->terminator.return_value, id_to_non_local_consumers);
#endif

        /* Incoming PHI values are handled elsewhere by modifying the incoming block to the creating block.
         * Ignore these kinds of usage here. */
    }

    if (!count)
        goto done;

    WARN("Resolving %u phis using variables.\n", count);

    /* Resolve these broken PHIs by using OpVariable. It is the simplest solution, and this is a very rare case to begin with. */
    if (!(rewrites = vkd3d_malloc(count * sizeof(*rewrites))))
    {
        ERR("Failed to allocate rewrite array.\n");
        abort();
    }

    for (id = 0, rewrite_count = 0; id < table_count; ++id)
    {
        struct sm6_block_array *nodes = &id_to_non_local_consumers[id];
        if (!nodes->count)
            continue;
        /* Fixup nodes in order. */
        qsort(nodes->blocks, nodes->count, sizeof(*nodes->blocks), forward_post_visit_order_compare_desc);
        sm6_block_array_deduplicate(nodes);
        rewrites[rewrite_count].id = id;
        rewrites[rewrite_count++].consumers = nodes;
    }
    assert(rewrite_count <= count);

    /* Ensure ordering so that output remains stable. */
    qsort(rewrites, rewrite_count, sizeof(*rewrites), rewrite_compare);

    for (i = 0; i < rewrite_count; ++i)
    {
        struct vkd3d_shader_register reg;
        struct id_rewrite *rewrite;
        struct id_origin *orig;

        rewrite = &rewrites[i];
        orig = &origin[rewrite->id];
        sm6_function_declare_temp(structuriser->function, &reg, orig->data_type, orig->node, rewrite->id);

        /* For every non-local node which consumes ID, we load from the alloca'd variable instead.
         * Rewrite all matching src registers to point to the loaded value. */
        for (j = 0; j < rewrite->consumers->count; ++j)
            sm6_block_rewrite_consumed_ids(rewrite->consumers->blocks[j], rewrite->id, &reg);
    }

    vkd3d_free(rewrites);
    for (i = 0; i < table_count; ++i)
        sm6_block_array_destroy(&id_to_non_local_consumers[i]);
done:
    vkd3d_free(id_to_non_local_consumers);
    vkd3d_free(origin);
}

static int phi_node_compare(const void *a, const void *b)
{
    const struct phi_node *pna = a, *pnb = b;
    return vkd3d_u32_compare(pnb->block->forward_post_visit_order, pna->block->forward_post_visit_order);
}

static void cfg_structuriser_insert_phis(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block **value_id_to_block;
    unsigned int i, j, dst_idx, map_size, phi_count;
    struct sm6_block *node;
    struct phi_node *phi_nodes;

    cfg_structuriser_prune_dead_preds(structuriser);

    /* It is possible that an SSA value was created in a block, and consumed in another.
     * With CFG rewriting branches, it is possible that dominance relationship no longer holds
     * and we must insert new dummy IDs to resolve this. */
    cfg_structuriser_fixup_broken_value_dominance(structuriser);

    for (i = 0, phi_count = 0; i < structuriser->forward_post_visit_order.count; ++i)
        phi_count += structuriser->forward_post_visit_order.blocks[i]->phi_count;
    if (!(phi_nodes = vkd3d_malloc(phi_count * sizeof(*phi_nodes))))
    {
        ERR("Failed to allocate phi nodes.\n");
        abort();
    }
    map_size = sm6_get_ssa_register_count(structuriser->parser);
    if (!(value_id_to_block = vkd3d_calloc(map_size, sizeof(*value_id_to_block))))
    {
        ERR("Failed to allocate block table.\n");
        abort();
    }

    /* Build a map of value ID -> creating block.
     * This allows us to detect if a value is consumed in a situation where the declaration does not dominate use.
     * This can happen when introducing ladder blocks or similar. */
    for (i = 0, phi_count = 0; i < structuriser->forward_post_visit_order.count; ++i)
    {
        node = structuriser->forward_post_visit_order.blocks[i];
        for (j = 0; j < node->phi_count; ++j)
        {
            phi_nodes[phi_count].phi_index = j;
            phi_nodes[phi_count++].block = node;
            if (node->phi[j].incoming_count)
                value_id_to_block[sm6_phi_ssa_id(&node->phi[j])] = node;
        }

        for (j = 0; j < node->instruction_count; ++j)
        {
            for (dst_idx = 0; dst_idx < node->instructions[j].dst_count; ++dst_idx)
            {
                const struct vkd3d_shader_register *reg = &node->instructions[j].dst[dst_idx].reg;
                if (!register_is_ssa(reg))
                    continue;
                value_id_to_block[vkd3d_shader_register_ssa_id(reg)] = node;
            }
        }

        /* If we inserted dummy branches from back-edge to rewrite infinite loops, we must prune these branches
         * now, so we don't end up creating a wrong amount of PHI incoming values.
         * We don't have to recompute the CFG since we don't really care about post-visit orders at this stage. */
        if (node->pred_back_edge && node->pred_back_edge->terminator.type == TERMINATOR_UNCOND_BR &&
            node->pred_back_edge->succ_back_edge == node->pred_back_edge->terminator.direct_block &&
            node->pred_back_edge->succ.count == 1)
        {
            struct sm6_block *back_edge = node->pred_back_edge;
            struct sm6_block *succ = back_edge->succ.blocks[0];
            back_edge->succ.count = 0;
            sm6_block_array_delete(&succ->pred, back_edge);
        }
    }

    /* Resolve phi-nodes top-down since PHI nodes may depend on other PHI nodes. */
    qsort(phi_nodes, phi_count, sizeof(*phi_nodes), phi_node_compare);

    for (i = 0; i < phi_count; ++i)
    {
        cfg_phi_node_fixup(&phi_nodes[i], value_id_to_block, map_size);
        cfg_structuriser_insert_phi(structuriser, &phi_nodes[i]);
    }

    vkd3d_free(value_id_to_block);
    vkd3d_free(phi_nodes);
}

static void cfg_structuriser_run(struct vkd3d_cfg_structuriser *structuriser)
{
    const char *graphviz_path = getenv("DXIL_SPIRV_GRAPHVIZ_PATH");
    bool need_restructure = false;

    if (graphviz_path)
    {
        cfg_structuriser_reset_traversal(structuriser);
        cfg_structuriser_visit(structuriser, structuriser->entry_block);
        cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".input");
    }

    cfg_structuriser_recompute_cfg(structuriser);

    cfg_structuriser_cleanup_breaking_phi_constructs(structuriser);
    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".phi-split");

    cfg_structuriser_create_continue_block_ladders(structuriser);

    while (cfg_structuriser_serialise_interleaved_merge_scopes(structuriser))
        cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".serialize");

    cfg_structuriser_split_merge_scopes(structuriser);
    cfg_structuriser_recompute_cfg(structuriser);

    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".split");

    cfg_structuriser_eliminate_degenerate_blocks(structuriser);
    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".eliminate0");

    /* Similar to cleanup_breaking_phi_constructs() in spirit,
     * but here we are forced to duplicate code blocks to make it work. */
    cfg_structuriser_duplicate_impossible_merge_constructs(structuriser);
    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".duplicate");

    while (cfg_structuriser_rewrite_transposed_loops(structuriser))
        cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".transpose-loop-rewrite");

    cfg_structuriser_structurise(structuriser, 0);
    cfg_structuriser_update_structured_loop_merge_targets(structuriser);

    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".struct0");

    cfg_structuriser_eliminate_degenerate_blocks(structuriser);
    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".eliminate1");

    cfg_structuriser_structurise(structuriser, 1);
    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".struct1");

    while (cfg_structuriser_rewrite_invalid_loop_breaks(structuriser))
    {
        cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".loop-break-rewrite");

        need_restructure = true;
    }

    if (need_restructure)
    {
        /* Need to redo the final structurization pass if we end up here. */
        cfg_structuriser_structurise(structuriser, 1);
    }

    cfg_structuriser_log_cfg_graphviz(structuriser, graphviz_path, ".final");

    cfg_structuriser_insert_phis(structuriser);
}

static void sm6_block_register(struct sm6_block *block, struct sm6_function *function)
{
    if (!block->synthetic_id)
        block->synthetic_id = sm6_function_alloc_block_id(function);
}

static void cfg_structuriser_traverse(struct vkd3d_cfg_structuriser *structuriser)
{
    struct sm6_block *block;
    unsigned int index;

    /* Make sure all blocks are known to the backend before we emit code.
     * Prefer that IDs grow the further down the function we go. */
    for (index = structuriser->forward_post_visit_order.count; index > 0; --index)
    {
        block = structuriser->forward_post_visit_order.blocks[index - 1];
        block->synthetic_id = 0;
        sm6_block_register(block, structuriser->function);
    }

    index = structuriser->forward_post_visit_order.count;
    assert(index && structuriser->forward_post_visit_order.blocks[index - 1] == structuriser->entry_block);
    for (; index > 0; --index)
    {
        block = structuriser->forward_post_visit_order.blocks[index - 1];

        switch (block->merge_type)
        {
            case MERGE_TYPE_SELECTION:
                if ((block->merge_block = block->selection_merge_block))
                    sm6_block_register(block->merge_block, structuriser->function);
                else
                    block->fake_merge_block_id = sm6_function_alloc_block_id(structuriser->function);
                break;

            case MERGE_TYPE_LOOP:
                block->merge_block = block->loop_merge_block;
                block->continue_block = block->pred_back_edge;
                if (block->merge_block)
                    sm6_block_register(block->merge_block, structuriser->function);
                if (block->continue_block)
                    sm6_block_register(block->continue_block, structuriser->function);

                if ((!!block->merge_block) ^ (!!block->continue_block))
                    block->fake_loop_block_id = sm6_function_alloc_block_id(structuriser->function);
                break;

            default:
                break;
        }
    }
}

static void cfg_structuriser_free(struct vkd3d_cfg_structuriser *structuriser)
{
    sm6_block_array_destroy(&structuriser->forward_post_visit_order);
    sm6_block_array_destroy(&structuriser->backward_post_visit_order);
    vkd3d_free(structuriser->reachability_bitset);
}

void cfg_structuriser_emit_blocks(void *parser, struct sm6_function *function)
{
    struct vkd3d_cfg_structuriser structuriser;

    cfg_structuriser_init(&structuriser, parser, function);
    cfg_structuriser_run(&structuriser);
    cfg_structuriser_traverse(&structuriser);
    sm6_function_reset_visited(function);
    function->entry_block = structuriser.entry_block;
    sm6_function_emit_blocks(function, parser);
    cfg_structuriser_free(&structuriser);
}
